package jp.co.sra.jun.collections.linkedlist;

import java.io.IOException;
import java.io.Writer;

import jp.co.sra.smalltalk.SmalltalkException;
import jp.co.sra.smalltalk.StBlockClosure;
import jp.co.sra.smalltalk.StBlockClosure1;

import jp.co.sra.jun.collections.sequences.JunSequence;

/**
 * JunAbstractLinkedList class
 * 
 *  @author    nisinaka
 *  @created   2008/04/24 (by nisinaka)
 *  @updated   N/A
 *  @version   758 (with StPL9.7) based on JunXXX for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunAbstractLinkedList.java,v 9.9 2010/03/09 09:16:43 nisinaka Exp $
 */
public abstract class JunAbstractLinkedList<E, L extends JunAbstractLink<E, L>> extends JunSequence<E> {

	/** The first link of the list. */
	protected L firstLink = null;

	/** The last link of the list. */
	protected L lastLink = null;

	/**
	 * Answer the object of the first entity.
	 * 
	 * @return java.lang.Object
	 * @category accessing
	 */
	public E first() {
		this.emptyCheck();
		return firstLink.entityObject();
	}

	/**
	 * Answer the object of the last entity.
	 * 
	 * @return java.lang.Object
	 * @category accessing
	 */
	public E last() {
		this.emptyCheck();
		return lastLink.entityObject;
	}

	/**
	 * Answer the size of the list.
	 * 
	 * @return int
	 * @category accessing
	 */
	public int size() {
		int size = 0;
		L aLink = firstLink;
		while (aLink != null) {
			size++;
			aLink = aLink.nextLink();
		}
		return size;
	}

	/**
	 * Add the object as an entity of the receiver.
	 * 
	 * @param anObject java.lang.Object
	 * @category adding
	 */
	public void add_(E anObject) {
		this.addLast_(anObject);
	}

	/**
	 * Add the object as the first entity of the receiver.
	 * 
	 * @param anObject java.lang.Object
	 * @category adding
	 */
	public void addFirst_(E anObject) {
		L aLink = this._createLink(anObject);

		if (this.isEmpty()) {
			lastLink = aLink;
		} else {
			aLink.nextLink = firstLink;
		}

		firstLink = aLink;
	}

	/**
	 * Add the object as the last entity of the receiver.
	 * 
	 * @param anObject java.lang.Object
	 * @category adding
	 */
	public void addLast_(E anObject) {
		L aLink = this._createLink(anObject);

		if (this.isEmpty()) {
			firstLink = aLink;
		} else {
			lastLink.nextLink_(aLink);
		}

		lastLink = aLink;
	}

	/**
	 * Remove the first entity from the receiver.
	 * 
	 * @return java.lang.Object
	 * @category removing
	 */
	public Object removeFirst() {
		this.emptyCheck();

		L oldLink = firstLink;

		if (firstLink == lastLink) {
			firstLink = null;
			lastLink = null;
		} else {
			firstLink = oldLink.nextLink;
		}

		oldLink.nextLink = null;

		return oldLink.entityObject;
	}

	/**
	 * Enumerate the objects and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @see jp.co.sra.jun.collections.sequences.JunSequence#do_(jp.co.sra.smalltalk.StBlockClosure1)
	 * @category enumerating
	 */
	@Override
	public <R> R do_(final StBlockClosure1<R, E> aBlock) {
		return this.linksDo_(new StBlockClosure1<R, L>() {
			public R value(L aLink) {
				return aBlock.value(aLink.entityObject());
			}
		});
	}

	/**
	 * Enumerate the objects and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category enumerating
	 * @deprecated since Jun701, use Generics.
	 */
	public Object do_(final StBlockClosure<?, ?, ?, ?> aBlock) {
		return this.linksDo_(new StBlockClosure1<Object, L>() {
			public Object value(L aLink) {
				return aBlock.value_(aLink.entityObject());
			}
		});
	}

	/**
	 * Print my string representation on the Writer.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException
	 * @category printing
	 */
	public void printOn_(final Writer aWriter) throws IOException {
		aWriter.write('(');

		if (this.isEmpty() == false) {
			this.do_(new StBlockClosure1<Object, E>() {
				public E value(E each) {
					try {
						aWriter.write(each.toString());
						aWriter.write(' ');
					} catch (IOException e) {
					}
					return null;
				}
			});
		}

		aWriter.write(')');
	}

	/**
	 * Answer true if the receiver includes the Object, otherwise false.
	 * 
	 * @param anObject java.lang.Object
	 * @return boolean
	 * @category testing
	 */
	public boolean includes_(E anObject) {
		L aLink = firstLink;
		while (aLink != null) {
			if (aLink.entityObject == anObject) {
				return true;
			}
			aLink = aLink.nextLink;
		}
		return false;
	}

	/**
	 * Answer true if the receiver is empty, otherwise false.
	 * 
	 * @return boolean
	 * @category testing
	 */
	public boolean isEmpty() {
		return firstLink == null;
	}

	/**
	 * Enumerate all links and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category private
	 */
	protected <R> R linksDo_(StBlockClosure1<R, L> aBlock) {
		L aLink = firstLink;
		while (aLink != null) {
			R result = aBlock.value(aLink);
			if (result != null) {
				return result;
			}
			aLink = aLink.nextLink();
		}
		return null;
	}

	/**
	 * If the receiver is empty, throw an exception.
	 * 
	 * @category private
	 */
	protected void emptyCheck() {
		if (this.isEmpty()) {
			throw SmalltalkException.Error("this collection is empty");
		}
	}

	/**
	 * Create a link.
	 * 
	 * @param anObject java.lang.Object
	 * @return jp.co.sra.jun.collections.linkedlist.JunAbstractLink
	 * @category private
	 */
	protected abstract L _createLink(E anObject);

}
