package jp.co.sra.jun.collections.sequences;

import java.awt.Point;
import java.util.ArrayList;

import jp.co.sra.smalltalk.StBlockClosure0;
import jp.co.sra.smalltalk.StBlockClosure1;

import jp.co.sra.jun.goodies.lisp.JunLispCons;
import jp.co.sra.jun.goodies.lisp.JunLispList;

/**
 * JunLinearEquations class
 * 
 *  @author    nisinaka
 *  @created   1998/10/30 (by nisinaka)
 *  @updated   2008/04/24 (by nisinaka)
 *  @updated   2009/03/18 (by m-asada)
 *  @version   758 (with StPL9.7) based on Jun721 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2005 Information-technology Promotion Agency, Japan (IPA)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunLinearEquations.java,v 9.12 2010/03/09 09:16:43 nisinaka Exp $
 */
public class JunLinearEquations extends JunDoubleMatrix {

	/** The array of row index for the solution. */
	protected int[] solutionLocations = null;

	/**
	 * Create a new instance of <code>JunLinearEquations</code> and initialize it with collection.
	 * 
	 * @param rowSize int
	 * @param columnSize int
	 * @param collection double[]
	 * @category Instance creation
	 */
	public JunLinearEquations(int rowSize, int columnSize, double[] collection) {
		super(rowSize, columnSize, collection);
	}

	/**
	 * Create a new instance of <code>JunLinearEquations</code> and initialize it.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category Instance creation
	 */
	public JunLinearEquations(JunLispList aList) {
		super(aList);
	}

	/**
	 * Answer the solution of the receiver.
	 * 
	 * @return double[]
	 * @category accessing
	 */
	public double[] solution() {
		int size = this.unknownSize();
		double[] solution = new double[size];
		for (int i = 0; i < size; i++) {
			solution[i] = this.solutionAt_(i);
		}
		return solution;
	}

	/**
	 * Answer the unknown number at the place specified with the index. Note:
	 * The index starts with 0, while the Smalltalk version starts with 1.
	 * 
	 * @param index int
	 * @return double
	 * @category accessing
	 */
	public double solutionAt_(int index) {
		this.solve();
		return -this._at(solutionLocations[index], columnSize - 1);
	}

	/**
	 * Answer the number of the unknowns.
	 * 
	 * @return int
	 * @category accessing
	 */
	public int unknownSize() {
		return columnSize - 1;
	}

	/**
	 * Get the receiver's attributes from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.collections.sequences.JunSequence#fromLispList(jp.co.sra.jun.goodies.lisp.JunLispList)
	 * @category lisp support
	 */
	@Override
	public void fromLispList(JunLispList aList) {
		super.fromLispList(aList);
		this.solutionLocationsFromLispList(aList);
	}

	/**
	 * Convert the receiver's attributes to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#toLispList()
	 * @category lisp support
	 */
	@Override
	public JunLispCons toLispList() {
		JunLispCons aList = super.toLispList();
		if (solutionLocations != null) {
			aList.add_(this.solutionLocationsToLispList());
		}
		return aList;
	}

	/**
	 * Get the receiver's solution locations from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected void solutionLocationsFromLispList(JunLispList aList) {
		JunLispList list = (JunLispList) aList.detect_ifNone_(new StBlockClosure1<Boolean, Object>() {
			public Boolean value(Object car) {
				return new Boolean(car instanceof JunLispCons && ((JunLispCons) car).head() == $("solutionLocations"));
			}
		}, new StBlockClosure0<Object>());
		if (list == null) {
			return;
		}

		Object[] values = ((JunLispList) list.tail()).asArray();
		int[] asArray = new int[values.length];
		for (int i = 0; i < values.length; i++) {
			asArray[i] = ((Number) values[i]).intValue();
		}
		solutionLocations = asArray;
	}

	/**
	 * Convert the receiver's solution locations to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected JunLispCons solutionLocationsToLispList() {
		JunLispCons list = this.lispCons();
		list.head_($("solutionLocations"));
		int[] asArray = solutionLocations;
		Integer[] values = new Integer[asArray.length];
		for (int i = 0; i < values.length; i++) {
			values[i] = new Integer(asArray[i]);
		}
		list.tail_(JunLispList.List_(values));
		return list;
	}

	/**
	 * Solve the receiver equation.
	 * 
	 * @category solving
	 */
	public void solve() {
		if (this.isSolved()) {
			return;
		}

		int unknownSize = this.unknownSize();
		final ArrayList<Integer> rowsToBeProcessed = new ArrayList<Integer>(rowSize);
		for (int i = 0; i < rowSize; i++) {
			rowsToBeProcessed.add(new Integer(i));
		}
		final ArrayList<Integer> columnsToBeProcessed = new ArrayList<Integer>(unknownSize);
		for (int i = 0; i < unknownSize; i++) {
			columnsToBeProcessed.add(new Integer(i));
		}

		StBlockClosure0<Point> nextPivotBlock = new StBlockClosure0<Point>() {
			public Point value() {
				Point nextPivot = null;
				double nextPivotValue = -1.0d;
				for (int i = 0; i < rowsToBeProcessed.size(); i++) {
					int row = rowsToBeProcessed.get(i);
					for (int j = 0; j < columnsToBeProcessed.size(); j++) {
						int column = columnsToBeProcessed.get(j);
						double cellValue = Math.abs(_at(row, column));
						if (cellValue >= nextPivotValue) {
							nextPivot = new Point(row, column);
							nextPivotValue = cellValue;
						}
					}
				}
				if (nextPivot != null) {
					rowsToBeProcessed.remove(new Integer(nextPivot.x));
					columnsToBeProcessed.remove(new Integer(nextPivot.y));
				}
				return nextPivot;
			}
		};

		StBlockClosure1<Void, Point> sweepBlock = new StBlockClosure1<Void, Point>() {
			public Void value(Point pivot) {
				double pivotRatio;
				for (int row = 0; row < rowSize(); row++) {
					pivotRatio = _at(row, pivot.y) / _at(pivot.x, pivot.y);
					if ((row != pivot.x) && (pivotRatio != 0)) {
						for (int column = 0; column < pivot.y; column++) {
							_put(row, column, _at(row, column) - (_at(pivot.x, column) * pivotRatio));
						}
						_put(row, pivot.y, 0.0);
						for (int column = pivot.y + 1; column < columnSize(); column++) {
							_put(row, column, _at(row, column) - (_at(pivot.x, column) * pivotRatio));
						}
					}
				}
				return null;
			}
		};

		// Start calculations from here.
		solutionLocations = new int[unknownSize];
		for (int i = 0; i < unknownSize; i++) {
			solutionLocations[i] = -1;
		}
		for (int i = 0; i < unknownSize; i++) {
			Point pivot = nextPivotBlock.value();
			if (pivot != null) {
				if (this._at(pivot.x, pivot.y) != 0) {
					sweepBlock.value(pivot);
					solutionLocations[pivot.y] = pivot.x;
				}
			}
		}
		for (int pivotColumn = 0; pivotColumn < unknownSize; pivotColumn++) {
			int pivotRow = solutionLocations[pivotColumn];
			if (pivotRow >= 0) {
				this._put(pivotRow, columnSize - 1, (this._at(pivotRow, columnSize - 1) / this._at(pivotRow, pivotColumn)));
				this._put(pivotRow, pivotColumn, 1.0);
			}
		}
	}

	/**
	 * Answer true if the receiver equation is already solved, otherwise false.
	 * 
	 * @return boolean
	 * @category testing
	 */
	public boolean isSolved() {
		return solutionLocations != null;
	}

}
