package jp.co.sra.jun.delaunay.twoD;

import java.awt.Graphics;
import java.io.IOException;
import java.io.Writer;

import jp.co.sra.smalltalk.StBlockClosure;
import jp.co.sra.smalltalk.StBlockClosure1;
import jp.co.sra.smalltalk.StValueHolder;

import jp.co.sra.jun.geometry.basic.JunPoint;

/**
 * Jun2dDelaunayLoop class
 * 
 *  @author    Ryouichi Matsuda
 *  @created   2002/01/14 (by Ryouichi Matsuda)
 *  @updated   2008/04/28 (by nisinaka)
 *  @version   758 (with StPL9.7) based on JunXXX for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2005 Information-technology Promotion Agency, Japan (IPA)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: Jun2dDelaunayLoop.java,v 9.10 2010/03/09 09:16:48 nisinaka Exp $
 */
public class Jun2dDelaunayLoop extends Jun2dDelaunayElement<Jun2dDelaunayLoop> {

	protected Jun2dDelaunayHalfEdge halfEdge;

	/**
	 * Answer my current half edge.
	 * 
	 * @return jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayHalfEdge
	 * @category accessing
	 */
	public Jun2dDelaunayHalfEdge halfEdge() {
		return halfEdge;
	}

	/**
	 * Set my new half edge.
	 * 
	 * @param aJun2dDelaunayHalfEdge jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayHalfEdge
	 * @category accessing
	 */
	public void halfEdge_(Jun2dDelaunayHalfEdge aJun2dDelaunayHalfEdge) {
		halfEdge = aJun2dDelaunayHalfEdge;
	}

	/**
	 * Answer my area.
	 * 
	 * @return double
	 * @category accessing
	 */
	public double area() {
		Jun2dDelaunayVertex vertex1;
		Jun2dDelaunayVertex vertex2;
		Jun2dDelaunayHalfEdge hEdge;
		double area;

		vertex1 = halfEdge.vertex();
		vertex2 = halfEdge.next().vertex();
		hEdge = halfEdge.next().next();
		area = 0.0d;

		do {
			area = area + (hEdge.vertex().areaWith_with_(vertex1, vertex2));
			hEdge = hEdge.next();
		} while (hEdge != halfEdge);

		return area;
	}

	/**
	 * Answer my center.
	 * 
	 * @return jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @category accessing
	 */
	public Jun2dDelaunayVertex center() {
		final StValueHolder<Double> x = new StValueHolder<Double>(); // double
		final StValueHolder<Double> y = new StValueHolder<Double>(); // double
		final StValueHolder<Integer> num = new StValueHolder<Integer>(); // int

		x.value_(0.0d);
		y.value_(0.0d);
		num.value_(0);
		this.verticesDo_(new StBlockClosure1<Void, Jun2dDelaunayVertex>() {
			public Void value(Jun2dDelaunayVertex vertex) {
				x.value_(x.value() + vertex.x());
				y.value_(y.value() + vertex.y());
				num.value_(num.value() + 1);

				return null;
			}
		});

		return new Jun2dDelaunayVertex(x.value() / num.value(), y.value() / num.value());
	}

	/**
	 * Answer my first point.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.JunPoint
	 * @category accessing
	 */
	public JunPoint point1() {
		return halfEdge.vertex().point();
	}

	/**
	 * Answer my second point.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.JunPoint
	 * @category accessing
	 */
	public JunPoint point2() {
		return halfEdge.next().vertex().point();
	}

	/**
	 * Answer my third point.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.JunPoint
	 * @category accessing
	 */
	public JunPoint point3() {
		return halfEdge.next().next().vertex().point();
	}

	/**
	 * Answer the farest half edge.
	 * 
	 * @param aJun2dDelaunayVertex jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @return jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayHalfEdge
	 * @category functions
	 */
	public Jun2dDelaunayHalfEdge farestHalfEdgeFromVertex_(Jun2dDelaunayVertex aJun2dDelaunayVertex) {
		Jun2dDelaunayHalfEdge farest;
		double maxSquaredDistance;
		Jun2dDelaunayHalfEdge hEdge;

		farest = halfEdge;
		maxSquaredDistance = halfEdge.vertex().squaredDistance_(aJun2dDelaunayVertex);
		hEdge = halfEdge.next();

		do {
			double squaredDistance;

			squaredDistance = hEdge.vertex().squaredDistance_(aJun2dDelaunayVertex);

			if (maxSquaredDistance < squaredDistance) {
				maxSquaredDistance = squaredDistance;
				farest = hEdge;
			}

			hEdge = hEdge.next();
		} while (hEdge != halfEdge);

		return farest;
	}

	/**
	 * Answer the nearest half edge.
	 * 
	 * @param aJun2dDelaunayVertex jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @return jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayHalfEdge
	 * @category functions
	 */
	public Jun2dDelaunayHalfEdge nearestHalfEdgeFromVertex_(Jun2dDelaunayVertex aJun2dDelaunayVertex) {
		Jun2dDelaunayHalfEdge nearest;
		double minSquaredDistance;
		Jun2dDelaunayHalfEdge hEdge;

		nearest = halfEdge;
		minSquaredDistance = halfEdge.vertex().squaredDistance_(aJun2dDelaunayVertex);
		hEdge = halfEdge.next();

		do {
			double squaredDistance;

			squaredDistance = hEdge.vertex().squaredDistance_(aJun2dDelaunayVertex);

			if (squaredDistance < minSquaredDistance) {
				minSquaredDistance = squaredDistance;
				nearest = hEdge;
			}

			hEdge = hEdge.next();
		} while (hEdge != halfEdge);

		return nearest;
	}

	/**
	 * Enumerate the edges and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category enumerating
	 */
	public <R> R edgesDo_(final StBlockClosure1<R, Jun2dDelaunayEdge> aBlock) {
		return halfEdge.do_(new StBlockClosure1<R, Jun2dDelaunayHalfEdge>() {
			public R value(Jun2dDelaunayHalfEdge hEdge) {
				R returnObject = aBlock.value(hEdge.edge());
				if (returnObject != null) {
					return returnObject; // continuation exit
				}
				return null;
			}
		});
	}

	/**
	 * Enumerate the edges and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category enumerating
	 * @deprecated since Jun701, use Generics.
	 */
	public <R> R edgesDo_(final StBlockClosure<R, ?, ?, ?> aBlock) {
		return this.edgesDo_(new StBlockClosure1<R, Jun2dDelaunayEdge>() {
			public R value(Jun2dDelaunayEdge edge) {
				return aBlock.value_(edge);
			}
		});
	}

	/**
	 * Enumerate the half edges and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category enumerating
	 */
	public <R> R halfEdgesDo_(StBlockClosure1<R, Jun2dDelaunayHalfEdge> aBlock) {
		return halfEdge.do_(aBlock);
	}

	/**
	 * Enumerate the half edges and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category enumerating
	 * @deprecated since Jun701, use Generics.
	 */
	public <R> R halfEdgesDo_(final StBlockClosure<R, ?, ?, ?> aBlock) {
		return this.halfEdgesDo_(new StBlockClosure1<R, Jun2dDelaunayHalfEdge>() {
			public R value(Jun2dDelaunayHalfEdge edge) {
				return aBlock.value_(edge);
			}
		});
	}

	/**
	 * Enumerate the neighbors and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category enumerating
	 */
	public <R> R neighborsDo_(final StBlockClosure1<R, Jun2dDelaunayLoop> aBlock) {
		return halfEdge.do_(new StBlockClosure1<R, Jun2dDelaunayHalfEdge>() {
			public R value(Jun2dDelaunayHalfEdge hEdge) {
				Jun2dDelaunayLoop l = hEdge.pair().loop();
				if (l != null) {
					R returnObject = aBlock.value(l);
					if (returnObject != null) {
						return returnObject; // continuation exit
					}
				}

				return null;
			}
		});
	}

	/**
	 * Enumerate the neighbors and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category enumerating
	 * @deprecated since Jun701, use Generics.
	 */
	public <R> R neighborsDo_(final StBlockClosure<R, ?, ?, ?> aBlock) {
		return this.neighborsDo_(new StBlockClosure1<R, Jun2dDelaunayLoop>() {
			public R value(Jun2dDelaunayLoop loop) {
				return aBlock.value_(loop);
			}
		});
	}

	/**
	 * Enumerate the vertices and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category enumerating
	 */
	public <R> R verticesDo_(final StBlockClosure1<R, Jun2dDelaunayVertex> aBlock) {
		return halfEdge.do_(new StBlockClosure1<R, Jun2dDelaunayHalfEdge>() {
			public R value(Jun2dDelaunayHalfEdge hEdge) {
				R returnObject = aBlock.value(hEdge.vertex());
				if (returnObject != null) {
					return returnObject; // continuation exit
				}

				return null;
			}
		});
	}

	/**
	 * Enumerate the vertices and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category enumerating
	 * @deprecated since Jun701, use Generics.
	 */
	public <R> R verticesDo_(final StBlockClosure<R, ?, ?, ?> aBlock) {
		return this.verticesDo_(new StBlockClosure1<R, Jun2dDelaunayVertex>() {
			public R value(Jun2dDelaunayVertex vertex) {
				return aBlock.value_(vertex);
			}
		});
	}

	/**
	 * Answer true if the receiver contains the specified vertex, otherwise false.
	 * 
	 * @param aJun2dDelaunayVertex jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @return boolean
	 * @category testing
	 */
	public boolean contains_(Jun2dDelaunayVertex aJun2dDelaunayVertex) {
		return this.containsX_y_(aJun2dDelaunayVertex.x(), aJun2dDelaunayVertex.y());
	}

	/**
	 * Answer true if the receiver contains the specified point, otherwise false.
	 * 
	 * @param xNumber double
	 * @param yNumber double
	 * @return boolean
	 * @category testing
	 */
	public boolean containsX_y_(final double xNumber, final double yNumber) {
		if (this.area() > DECIMAL_12) {
			final StValueHolder<Double> minX = new StValueHolder<Double>(); // double
			final StValueHolder<Double> maxX = new StValueHolder<Double>(); // double
			final StValueHolder<Double> minY = new StValueHolder<Double>(); // double
			final StValueHolder<Double> maxY = new StValueHolder<Double>(); // double

			minX.value_(halfEdge.vertex().x());
			maxX.value_(halfEdge.vertex().x());
			minY.value_(halfEdge.vertex().y());
			maxY.value_(halfEdge.vertex().y());
			this.verticesDo_(new StBlockClosure1<Void, Jun2dDelaunayVertex>() {
				public Void value(Jun2dDelaunayVertex vertex) {
					double vx = vertex.x();
					double vy = vertex.y();

					if (vx < minX.value()) {
						minX.value_(vx);
					}
					if (maxX.value() < vx) {
						maxX.value_(vx);
					}
					if (vy < minY.value()) {
						minY.value_(vy);
					}
					if (maxY.value() < vy) {
						maxY.value_(vy);
					}

					return null;
				}
			});

			if (minX.value() - DECIMAL_12 <= xNumber && xNumber <= maxX.value() + DECIMAL_12 && minY.value() - DECIMAL_12 <= yNumber && yNumber <= maxY.value() + DECIMAL_12) {
				Boolean returnObject = this.halfEdgesDo_(new StBlockClosure1<Boolean, Jun2dDelaunayHalfEdge>() {
					public Boolean value(Jun2dDelaunayHalfEdge hEdge) {
						Jun2dDelaunayVertex v1 = hEdge.pair().vertex();
						Jun2dDelaunayVertex v2 = hEdge.vertex();

						if ((((v1.x() - xNumber) * (v2.y() - yNumber)) - ((v1.y() - yNumber) * (v2.x() - xNumber))) < 0) {
							return Boolean.FALSE;
						}

						return null;
					}
				});

				return (returnObject == Boolean.FALSE) ? false : true;
			}
		}

		return false;
	}

	/**
	 * Answer true if the receiver has a boundary vertex, otherwise false.
	 * 
	 * @return boolean
	 * @category testing
	 */
	public boolean hasBoundaryVertex() {
		Boolean returnObject = this.verticesDo_(new StBlockClosure1<Boolean, Jun2dDelaunayVertex>() {
			public Boolean value(Jun2dDelaunayVertex vertex) {
				if (vertex.isBoundaryVertex()) {
					return Boolean.TRUE;
				}

				return null;
			}
		});

		return (returnObject == Boolean.TRUE) ? true : false;
	}

	/**
	 * Answer true if the receiver intersects with the specified line segment, otherwise false.
	 * 
	 * @param aJun2dDelaunayVertex1 jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @param aJun2dDelaunayVertex2 jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @return boolean
	 * @category testing
	 */
	public boolean intersectsWithLineSegmentFrom_to_(final Jun2dDelaunayVertex aJun2dDelaunayVertex1, final Jun2dDelaunayVertex aJun2dDelaunayVertex2) {
		Boolean returnObject = halfEdge.do_(new StBlockClosure1<Boolean, Jun2dDelaunayHalfEdge>() {
			public Boolean value(Jun2dDelaunayHalfEdge hEdge) {
				if (hEdge.intersectsWithLineSegmentFrom_to_(aJun2dDelaunayVertex1, aJun2dDelaunayVertex2)) {
					return Boolean.TRUE;
				}

				return null;
			}
		});

		return (returnObject == Boolean.TRUE) ? true : false;
	}

	/**
	 * Answer true if the receiver is a triangle, otherwise false.
	 * 
	 * @return boolean
	 * @category testing
	 */
	public boolean isTriangle() {
		return (halfEdge.next().next().next() == halfEdge);
	}

	/**
	 * Display the receiver on the graphics.
	 * 
	 * @param aGraphics java.awt.Graphics
	 * @category displaying
	 */
	public void displayOn_(final Graphics aGraphics) {
		this.halfEdgesDo_(new StBlockClosure1<Void, Jun2dDelaunayHalfEdge>() {
			public Void value(Jun2dDelaunayHalfEdge hEdge) {
				hEdge.displayOn_(aGraphics);
				return null;
			}
		});
	}

	/**
	 * Print my string representation on aWriter.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException
	 * @see jp.co.sra.smalltalk.StObject#printOn_(java.io.Writer)
	 * @category printing
	 */
	public void printOn_(final Writer aWriter) throws IOException {
		aWriter.write("Loop(");
		this.verticesDo_(new StBlockClosure1<Void, Jun2dDelaunayVertex>() {
			public Void value(Jun2dDelaunayVertex vertex) {
				try {
					aWriter.write(Double.toString(vertex.x()));
					aWriter.write("@");
					aWriter.write(Double.toString(vertex.y()));
					aWriter.write(" ");
				} catch (IOException e) {
					//
				}
				return null;
			}
		});
		aWriter.write(")");
	}

	/**
	 * Search for a nearest half edge from the specified point.
	 * 
	 * @param xNumber double
	 * @param yNumber double
	 * @return jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayHalfEdge
	 * @category private
	 */
	protected Jun2dDelaunayHalfEdge nearestHalfEdgeFromX_y_(double xNumber, double yNumber) {
		Jun2dDelaunayHalfEdge nearest = halfEdge;
		double minSquaredDistance = halfEdge.edge().squaredDistanceFromX_y_(xNumber, yNumber);
		Jun2dDelaunayHalfEdge hEdge = halfEdge.next();

		do {
			double squaredDistance = hEdge.edge().squaredDistanceFromX_y_(xNumber, yNumber);
			if (squaredDistance < minSquaredDistance) {
				minSquaredDistance = squaredDistance;
				nearest = hEdge;
			}

			hEdge = hEdge.next();
		} while (hEdge != halfEdge);

		return nearest;
	}

}
