package jp.co.sra.jun.delaunay.twoD;

import java.io.IOException;
import java.io.Writer;

import jp.co.sra.smalltalk.StBlockClosure;
import jp.co.sra.smalltalk.StBlockClosure1;
import jp.co.sra.smalltalk.StValueHolder;

import jp.co.sra.jun.geometry.basic.Jun2dPoint;
import jp.co.sra.jun.geometry.basic.Jun3dPoint;
import jp.co.sra.jun.geometry.basic.JunPoint;

/**
 * Jun2dDelaunayVertex class
 * 
 *  @author    Ryouichi Matsuda
 *  @created   2002/01/14 (by Ryouichi Matsuda)
 *  @updated   2008/04/28 (by nisinaka)
 *  @version   758 (with StPL9.7) based on JunXXX for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2005 Information-technology Promotion Agency, Japan (IPA)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: Jun2dDelaunayVertex.java,v 9.10 2010/03/09 09:16:48 nisinaka Exp $
 */
public class Jun2dDelaunayVertex extends Jun2dDelaunayElement<Jun2dDelaunayVertex> {

	protected double x = Double.NaN;
	protected double y = Double.NaN;
	protected double z = Double.NaN;
	protected Jun2dDelaunayHalfEdge halfEdge;

	/**
	 * Create a new instance of Jun2dDelaunayVertex and initialize it.
	 *
	 * @category Instance creation
	 */
	public Jun2dDelaunayVertex() {

	}

	/**
	 * Create a new instance of <code>Jun2dDelaunayVertex</code> and initialize it.
	 * 
	 * @param xNumber double
	 * @param yNumber double
	 * @category Instance creation
	 */
	public Jun2dDelaunayVertex(double xNumber, double yNumber) {
		this.setX_(xNumber);
		this.setY_(yNumber);
	}

	/**
	 * Create a new instance of <code>Jun2dDelaunayVertex</code> and initialize it.
	 * 
	 * @param xNumber double
	 * @param yNumber double
	 * @param zNumber double
	 * @category Instance creation
	 */
	public Jun2dDelaunayVertex(double xNumber, double yNumber, double zNumber) {
		this.setX_(xNumber);
		this.setY_(yNumber);
		this.setZ_(zNumber);
	}

	/**
	 * Create a new instance of <code>Jun2dDelaunayVertex</code> and initialize it.
	 * 
	 * @param aPoint jp.co.sra.jun.geometry.basic.JunPoint
	 * @category Instance creation
	 */
	public Jun2dDelaunayVertex(JunPoint aPoint) {
		if (aPoint instanceof Jun2dPoint) {
			Jun2dPoint p = (Jun2dPoint) aPoint;
			this.setX_(p.x());
			this.setY_(p.y());
		}

		if (aPoint instanceof Jun3dPoint) {
			Jun3dPoint p = (Jun3dPoint) aPoint;
			this.setX_(p.x());
			this.setY_(p.y());
			this.setZ_(p.z());
		}
	}

	/**
	 * Answer my x coordinate.
	 * 
	 * @return double
	 * @category accessing
	 */
	public double x() {
		return x;
	}

	/**
	 * Answer my y coordinate.
	 * 
	 * @return double
	 * @category accessing
	 */
	public double y() {
		return y;
	}

	/**
	 * Answer my z coordinate.
	 * 
	 * @return double
	 * @category accessing
	 */
	public double z() {
		return z;
	}

	/**
	 * Answer my current half edge.
	 * 
	 * @return jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayHalfEdge
	 * @category accessing
	 */
	public Jun2dDelaunayHalfEdge halfEdge() {
		return halfEdge;
	}

	/**
	 * Set my new half edge.
	 * 
	 * @param aJun2dDelaunayHalfEdge jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayHalfEdge
	 * @category accessing
	 */
	public void halfEdge_(Jun2dDelaunayHalfEdge aJun2dDelaunayHalfEdge) {
		halfEdge = aJun2dDelaunayHalfEdge;
	}

	/**
	 * Answer a JunPoint representing the receiver.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.JunPoint
	 * @category accessing
	 */
	public JunPoint point() {
		if (Double.isNaN(z)) {
			return new Jun2dPoint(x, y);
		} else {
			return new Jun3dPoint(x, y, z);
		}
	}

	/**
	 * Convert the receiver as Jun2dPoint.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.Jun2dPoint
	 * @category converting
	 */
	public Jun2dPoint asJun2dPoint() {
		return new Jun2dPoint(x, y);
	}

	/**
	 * Convert the receiver as Jun3dPoint.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category converting
	 */
	public Jun3dPoint asJun3dPoint() {
		return new Jun3dPoint(x, y, (z == Double.NaN) ? 0.0d : z);
	}

	/**
	 * Enumerate the edges and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure1
	 * @return java.lang.Object
	 * @category enumerating
	 */
	public <R> R edgesDo_(final StBlockClosure1<R, Jun2dDelaunayEdge> aBlock) {
		return this.halfEdgesDo_(new StBlockClosure1<R, Jun2dDelaunayHalfEdge>() {
			public R value(Jun2dDelaunayHalfEdge hEdge) {
				R returnObject = aBlock.value(hEdge.edge);
				if (returnObject != null) {
					return returnObject; // continuation exit
				}

				return null;
			}
		});
	}

	/**
	 * Enumerate the edges and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category enumerating
	 * @deprecated since Jun701, use Generics.
	 */
	public <R> R edgesDo_(final StBlockClosure<R, ?, ?, ?> aBlock) {
		return this.edgesDo_(new StBlockClosure1<R, Jun2dDelaunayEdge>() {
			public R value(Jun2dDelaunayEdge edge) {
				return aBlock.value_(edge);
			}
		});
	}

	/**
	 * Enumerate the half edges and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure1
	 * @return java.lang.Object
	 * @category enumerating
	 */
	public <R> R halfEdgesDo_(StBlockClosure1<R, Jun2dDelaunayHalfEdge> aBlock) {
		Jun2dDelaunayHalfEdge hEdge = halfEdge;
		do {
			Jun2dDelaunayHalfEdge nextHalfEdge = hEdge.next().pair();
			R returnObject = aBlock.value(hEdge);
			if (returnObject != null) {
				return returnObject; // continuation exit
			}

			hEdge = nextHalfEdge;
		} while (hEdge != halfEdge);

		return null;
	}

	/**
	 * Enumerate the half edges and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category enumerating
	 * @deprecated since Jun701, use Generics.
	 */
	public <R> R halfEdgesDo_(final StBlockClosure<R, ?, ?, ?> aBlock) {
		return this.halfEdgesDo_(new StBlockClosure1<R, Jun2dDelaunayHalfEdge>() {
			public R value(Jun2dDelaunayHalfEdge hEdge) {
				return aBlock.value_(hEdge);
			}
		});
	}

	/**
	 * Enumerate the loops and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure1
	 * @return java.lang.Object
	 * @category enumerating
	 */
	public <R> R loopsDo_(final StBlockClosure1<R, Jun2dDelaunayLoop> aBlock) {
		return this.halfEdgesDo_(new StBlockClosure1<R, Jun2dDelaunayHalfEdge>() {
			public R value(Jun2dDelaunayHalfEdge hEdge) {
				Jun2dDelaunayLoop l = hEdge.loop();
				if (l != null) {
					R returnObject = aBlock.value(l);
					if (returnObject != null) {
						return returnObject; // continuation exit
					}
				}

				return null;
			}
		});
	}

	/**
	 * Enumerate the loops and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category enumerating
	 * @deprecated since Jun701, use Generics.
	 */
	public <R> R loopsDo_(final StBlockClosure<R, ?, ?, ?> aBlock) {
		return this.loopsDo_(new StBlockClosure1<R, Jun2dDelaunayLoop>() {
			public R value(Jun2dDelaunayLoop loop) {
				return aBlock.value_(loop);
			}
		});
	}

	/**
	 * Enumerate the neighbor vertexes and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure1
	 * @return java.lang.Object
	 * @category enumerating
	 */
	public <R> R neighborsDo_(final StBlockClosure1<R, Jun2dDelaunayVertex> aBlock) {
		return this.halfEdgesDo_(new StBlockClosure1<R, Jun2dDelaunayHalfEdge>() {
			public R value(Jun2dDelaunayHalfEdge hEdge) {
				R returnObject = aBlock.value(hEdge.pair().vertex());
				if (returnObject != null) {
					return returnObject; // continuation exit
				}

				return null;
			}
		});
	}

	/**
	 * Enumerate the neighbor vertexes and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category enumerating
	 * @deprecated since Jun701, use Generics.
	 */
	public <R> R neighborsDo_(final StBlockClosure<R, ?, ?, ?> aBlock) {
		return this.neighborsDo_(new StBlockClosure1<R, Jun2dDelaunayVertex>() {
			public R value(Jun2dDelaunayVertex vertex) {
				return aBlock.value_(vertex);
			}
		});
	}

	/**
	 * Calculate the angle with the specified vertexes.
	 * 
	 * @param aJun2dDelaunayVertex1 jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @param aJun2dDelaunayVertex2 jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @return double
	 * @category functions
	 */
	public double angleWith_with_(Jun2dDelaunayVertex aJun2dDelaunayVertex1, Jun2dDelaunayVertex aJun2dDelaunayVertex2) {
		return Math.acos(this.cosOfAngleWith_with_(aJun2dDelaunayVertex1, aJun2dDelaunayVertex2));
	}

	/**
	 * Calculate the area of the triangle made with the specified vertexes.
	 * 
	 * @param aJun2dDelaunayVertex1 jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @param aJun2dDelaunayVertex2 jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @return double
	 * @category functions
	 */
	public double areaWith_with_(Jun2dDelaunayVertex aJun2dDelaunayVertex1, Jun2dDelaunayVertex aJun2dDelaunayVertex2) {
		return ((aJun2dDelaunayVertex1.x() - x) * (aJun2dDelaunayVertex2.y() - y)) - ((aJun2dDelaunayVertex2.x() - x) * (aJun2dDelaunayVertex1.y() - y));
	}

	/**
	 * Calculate the cos of the angle with the specified vertexes.
	 * 
	 * @param aJun2dDelaunayVertex1 jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @param aJun2dDelaunayVertex2 jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @return double
	 * @category functions
	 */
	public double cosOfAngleWith_with_(Jun2dDelaunayVertex aJun2dDelaunayVertex1, Jun2dDelaunayVertex aJun2dDelaunayVertex2) {
		double d1;
		double d2;
		double dotProduct;

		d1 = this.squaredDistance_(aJun2dDelaunayVertex1);
		d2 = this.squaredDistance_(aJun2dDelaunayVertex2);
		dotProduct = ((aJun2dDelaunayVertex1.x() - x) * (aJun2dDelaunayVertex2.x() - x)) + ((aJun2dDelaunayVertex1.y() - y) * (aJun2dDelaunayVertex2.y() - y));

		return Math.min(Math.max(dotProduct / Math.sqrt(d1 * d2), -1.0d), 1.0d);
	}

	/**
	 * Calculate the distance from the specified vertex.
	 * 
	 * @param aJun2dDelaunayVertex jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @return double
	 * @category functions
	 */
	public double distance_(Jun2dDelaunayVertex aJun2dDelaunayVertex) {
		return Math.sqrt(this.squaredDistance_(aJun2dDelaunayVertex));
	}

	/**
	 * Calculate the dot product with the specified vertex.
	 * 
	 * @param aJun2dDelaunayVertex jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @return double
	 * @category functions
	 */
	public double dotProduct_(Jun2dDelaunayVertex aJun2dDelaunayVertex) {
		return (x * aJun2dDelaunayVertex.x()) + (y * aJun2dDelaunayVertex.y());
	}

	/**
	 * Find an edge to the specified vertex.
	 * 
	 * @param aJun2dDelaunayVertex jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @return jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayEdge
	 * @category functions
	 */
	public Jun2dDelaunayEdge edgeTo_(final Jun2dDelaunayVertex aJun2dDelaunayVertex) {
		return this.halfEdgesDo_(new StBlockClosure1<Jun2dDelaunayEdge, Jun2dDelaunayHalfEdge>() {
			public Jun2dDelaunayEdge value(Jun2dDelaunayHalfEdge hEdge) {
				if (aJun2dDelaunayVertex == hEdge.pair().vertex()) {
					return hEdge.edge; // continuation exit
				}

				return null;
			}
		});
	}

	/**
	 * Calculate the product with the specified vertex.
	 * 
	 * @param aJun2dDelaunayVertex jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @return double
	 * @category functions
	 */
	public double product_(Jun2dDelaunayVertex aJun2dDelaunayVertex) {
		return x * aJun2dDelaunayVertex.y() - y * aJun2dDelaunayVertex.x();
	}

	/**
	 * Calculate the squared distance from the specified vertex.
	 * 
	 * @param aJun2dDelaunayVertex jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @return double
	 * @category functions
	 */
	public double squaredDistance_(Jun2dDelaunayVertex aJun2dDelaunayVertex) {
		return this.squaredDistanceFromX_y_(aJun2dDelaunayVertex.x(), aJun2dDelaunayVertex.y());
	}

	/**
	 * Calculate the squared distance from the specified point.
	 * 
	 * @param xNumber double
	 * @param yNumber double
	 * @return double
	 * @category functions
	 */
	public double squaredDistanceFromX_y_(double xNumber, double yNumber) {
		return Math.pow(x - xNumber, 2) + Math.pow(y - yNumber, 2);
	}

	/**
	 * Print my string representation on aWriter.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws IOException DOCUMENT ME!
	 * @see jp.co.sra.smalltalk.StObject#printOn_(java.io.Writer)
	 * @category printing
	 */
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write("Vertex(");
		aWriter.write(Double.toString(x));
		aWriter.write("@");
		aWriter.write(Double.toString(y));
		aWriter.write(")");
	}

	/**
	 * Answer true if the receiver contains the specified vertex, otherwise false.
	 * 
	 * @param aJun2dDelaunayVertex jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @return boolean
	 * @category testing
	 */
	public boolean contains_(Jun2dDelaunayVertex aJun2dDelaunayVertex) {
		return (Math.pow(aJun2dDelaunayVertex.x() - x, 2) + Math.pow(aJun2dDelaunayVertex.y() - y, 2) < DECIMAL_12);
	}

	/**
	 * Answer true if the receiver contains the specified point, otherwise false.
	 * 
	 * @param xNumber double
	 * @param yNumber double
	 * @return boolean
	 * @category testing
	 */
	public boolean containsX_y_(double xNumber, double yNumber) {
		return (Math.pow(xNumber - x, 2) + Math.pow(yNumber - y, 2) < DECIMAL_12);
	}

	/**
	 * Answer true if the receiver is a boundary vertex, otherwise false.
	 * 
	 * @return boolean
	 * @category testing
	 */
	public boolean isBoundaryVertex() {
		return false;
	}

	/**
	 * Detect the half edge of the loop.
	 * 
	 * @param aJun2dDelaunayLoop jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayLoop
	 * @return jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayHalfEdge
	 * @category private
	 */
	protected Jun2dDelaunayHalfEdge detectHalfEdgeOfLoop_(final Jun2dDelaunayLoop aJun2dDelaunayLoop) {
		return this.halfEdgesDo_(new StBlockClosure1<Jun2dDelaunayHalfEdge, Jun2dDelaunayHalfEdge>() {
			public Jun2dDelaunayHalfEdge value(Jun2dDelaunayHalfEdge hEdge) {
				if (hEdge.loop() == aJun2dDelaunayLoop) {
					return hEdge; // continuation exit
				}

				return null;
			}
		});
	}

	/**
	 * Detect the half edge of the loop which contains the specified point.
	 * 
	 * @param xNumber double
	 * @param yNumber double
	 * @return jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayHalfEdge
	 * @category private
	 */
	protected Jun2dDelaunayHalfEdge detectHalfEdgeOfLoopContainsX_y_(final double xNumber, final double yNumber) {
		return this.halfEdgesDo_(new StBlockClosure1<Jun2dDelaunayHalfEdge, Jun2dDelaunayHalfEdge>() {
			public Jun2dDelaunayHalfEdge value(Jun2dDelaunayHalfEdge hEdge) {
				if (hEdge.loop().containsX_y_(xNumber, yNumber)) {
					return hEdge; // continuation exit
				}

				return null;
			}
		});
	}

	/**
	 * Find an half edge to the specified vertex.
	 * 
	 * @param aJun2dDelaunayVertex jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @return jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayHalfEdge
	 * @category private
	 */
	protected Jun2dDelaunayHalfEdge halfEdgeTo_(final Jun2dDelaunayVertex aJun2dDelaunayVertex) {
		return this.halfEdgesDo_(new StBlockClosure1<Jun2dDelaunayHalfEdge, Jun2dDelaunayHalfEdge>() {
			public Jun2dDelaunayHalfEdge value(Jun2dDelaunayHalfEdge hEdge) {
				if (aJun2dDelaunayVertex == hEdge.pair().vertex()) {
					return hEdge.pair(); // continuation exit
				}

				return null;
			}
		});
	}

	/**
	 * Find  an half edge toward the specified vertex.
	 * 
	 * @param aJun2dDelaunayVertex jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @return jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayHalfEdge
	 * @category private
	 */
	protected Jun2dDelaunayHalfEdge halfEdgeToward_(final Jun2dDelaunayVertex aJun2dDelaunayVertex) {
		final double scaleToTarget = this.distance_(aJun2dDelaunayVertex);
		final double vectorToTargetX = (aJun2dDelaunayVertex.x() - x) / scaleToTarget;
		final double vectorToTargetY = (aJun2dDelaunayVertex.y() - y) / scaleToTarget;

		return this.halfEdgesDo_(new StBlockClosure1<Jun2dDelaunayHalfEdge, Jun2dDelaunayHalfEdge>() {
			public Jun2dDelaunayHalfEdge value(Jun2dDelaunayHalfEdge hEdge) {
				Jun2dDelaunayVertex vertex;
				double scaleToNeighbor;
				double vectorToNeighborX;
				double vectorToNeighborY;

				vertex = hEdge.pair().vertex();
				scaleToNeighbor = distance_(vertex);
				vectorToNeighborX = (vertex.x() - x) / scaleToNeighbor;
				vectorToNeighborY = (vertex.y() - y) / scaleToNeighbor;

				if (hEdge.pair().vertex() == aJun2dDelaunayVertex) {
					return hEdge; // continuation exit
				}

				if (Math.abs((vectorToTargetX * vectorToNeighborY) - (vectorToNeighborX * vectorToTargetY)) < DECIMAL_12) {
					double t = (vectorToTargetX * vectorToNeighborX) + (vectorToTargetY * vectorToNeighborY);

					if ((-DECIMAL_12 <= t) && (t <= (1.0d + DECIMAL_12))) {
						return hEdge; // continuation exit
					}
				}

				return null;
			}
		});
	}

	/**
	 * Answer true if the receiver is on the specified line segment, otherwise false.
	 * 
	 * @param vertex1 jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @param vertex2 jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @return boolean
	 * @category private
	 */
	protected boolean isOnLineSegmentFrom_to_(Jun2dDelaunayVertex vertex1, Jun2dDelaunayVertex vertex2) {
		double yourScale;
		double yourVectorX;
		double yourVectorY;
		double myScale;
		double myVectorX;
		double myVectorY;

		if ((this == vertex1) || (this == vertex2)) {
			return true;
		}

		yourScale = vertex1.distance_(vertex2);
		yourVectorX = (vertex2.x() - vertex1.x()) / yourScale;
		yourVectorY = (vertex2.y() - vertex1.y()) / yourScale;
		myScale = vertex1.distance_(this);
		myVectorX = (this.x() - vertex1.x()) / myScale;
		myVectorY = (this.y() - vertex1.y()) / myScale;

		if (Math.abs((yourVectorX * myVectorY) - (myVectorX * yourVectorY)) < DECIMAL_12) {
			double t = (yourVectorX * myVectorX) + (yourVectorY * myVectorY);

			if ((0.0d <= t) && (t <= 1.0d)) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Find a nearest edge from the specified point.
	 * 
	 * @param xNumber double
	 * @param yNumber double
	 * @return jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayEdge
	 * @category private
	 */
	protected Jun2dDelaunayEdge nearestEdgeFromX_y_(double xNumber, double yNumber) {
		return (this.nearestHalfEdgeFromX_y_(xNumber, yNumber).edge());
	}

	/**
	 * Find a nearest half edge from the specified point.
	 * 
	 * @param xNumber double
	 * @param yNumber double
	 * @return jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayHalfEdge
	 * @category private
	 */
	protected Jun2dDelaunayHalfEdge nearestHalfEdgeFromX_y_(final double xNumber, final double yNumber) {
		final StValueHolder<Jun2dDelaunayHalfEdge> theNearest = new StValueHolder<Jun2dDelaunayHalfEdge>(); // Jun2dDelaunayHalfEdge
		final StValueHolder<Double> minSquaredDistance = new StValueHolder<Double>(); // double

		theNearest.value_(halfEdge);
		minSquaredDistance.value_(halfEdge.edge().squaredDistanceFromX_y_(xNumber, yNumber));
		this.halfEdgesDo_(new StBlockClosure1<Jun2dDelaunayHalfEdge, Jun2dDelaunayHalfEdge>() {
			public Jun2dDelaunayHalfEdge value(Jun2dDelaunayHalfEdge hEdge) {
				if (hEdge != halfEdge) {
					double squaredDistance = hEdge.edge().squaredDistanceFromX_y_(xNumber, yNumber);
					if (squaredDistance < minSquaredDistance.value()) {
						minSquaredDistance.value_(squaredDistance);
						theNearest.value_(hEdge);
					}
				}

				return null;
			}
		});

		return theNearest.value();
	}

	/**
	 * Find a nearest neighbor from the specified point.
	 * 
	 * @param xNumber double
	 * @param yNumber double
	 * @return jp.co.sra.jun.delaunay.twoD.Jun2dDelaunayVertex
	 * @category private
	 */
	protected Jun2dDelaunayVertex nearestNeighborFromX_y_(final double xNumber, final double yNumber) {
		final StValueHolder<Jun2dDelaunayVertex> theNearest = new StValueHolder<Jun2dDelaunayVertex>();
		final StValueHolder<Double> minSquaredDistance = new StValueHolder<Double>();

		theNearest.value_(this);
		minSquaredDistance.value_(Math.pow(x - xNumber, 2) + Math.pow(y - yNumber, 2));
		this.neighborsDo_(new StBlockClosure1<Void, Jun2dDelaunayVertex>() {
			public Void value(Jun2dDelaunayVertex vertex) {
				double squaredDistance = Math.pow(vertex.x() - xNumber, 2) + Math.pow(vertex.y() - yNumber, 2);
				if (squaredDistance < minSquaredDistance.value()) {
					minSquaredDistance.value_(squaredDistance);
					theNearest.value_(vertex);
				}

				return null;
			}
		});

		return theNearest.value();
	}

	/**
	 * Set my x coordinate.
	 * 
	 * @param xNumber double
	 * @category private
	 */
	protected void setX_(double xNumber) {
		x = xNumber;
	}

	/**
	 * Set my y coordinate.
	 * 
	 * @param yNumber double
	 * @category private
	 */
	protected void setY_(double yNumber) {
		y = yNumber;
	}

	/**
	 * Set my z coordinate.
	 * 
	 * @param zNumber double
	 * @category private
	 */
	protected void setZ_(double zNumber) {
		z = zNumber;
	}

}
