package jp.co.sra.jun.geometry.boundaries;

import java.awt.Rectangle;
import java.io.IOException;
import java.io.Writer;

import jp.co.sra.smalltalk.StObject;

import jp.co.sra.jun.geometry.basic.JunPoint;
import jp.co.sra.jun.goodies.lisp.JunLispCons;
import jp.co.sra.jun.goodies.lisp.JunLispList;

/**
 * JunBoundingBox class
 * 
 *  @author    nisinaka
 *  @created   1998/10/01 (by nisinaka)
 *  @updated   1999/08/05 (by nisinaka)
 *  @updated   2000/01/06 (by nisinaka)
 *  @updated   2004/10/18 (by Mitsuhiro Asada)
 *  @updated   2006/10/10 (by nisinaka)
 *  @updated   2009/03/28 (by Mitsuhiro Asada)
 *  @version   758 (with StPL9.7) based on Jun740 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2005 Information-technology Promotion Agency, Japan (IPA)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunBoundingBox.java,v 9.11 2010/03/09 09:16:50 nisinaka Exp $
 */
public abstract class JunBoundingBox extends JunBoundingObject {
	protected JunPoint _origin;
	protected JunPoint _corner;

	/**
	 * Create a new instance of <code>JunBoundingBox</code> and initialize it.
	 * 
	 * @category Instance creation
	 */
	public JunBoundingBox() {
		super();
	}

	/**
	 * Create a new instance of <code>JunBoundingBox</code> and initialize it.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category Instance creation
	 */
	public JunBoundingBox(JunLispList aList) {
		super(aList);
	}

	/**
	 * Initialize the receiver.
	 * 
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#initialize()
	 * @category initialize-release
	 */
	@Override
	protected void initialize() {
		super.initialize();
		_origin = null;
		_corner = null;
	}

	/**
	 * Answer the width of the receiver.
	 * 
	 * @return double
	 * @category accessing
	 */
	public double width() {
		return this._corner().x() - this._origin().x();
	}

	/**
	 * Answer the height of the receiver.
	 * 
	 * @return double
	 * @category accessing
	 */
	public double height() {
		return this._corner().y() - this._origin().y();
	}

	/**
	 * Answer the depth of the receiver.
	 * 
	 * @return double
	 * @category accessing
	 */
	public abstract double depth();

	/**
	 * Answer the max length.
	 * 
	 * @return double
	 * @category accessing
	 */
	public abstract double maxLength();

	/**
	 * Answer the min length.
	 * 
	 * @return double
	 * @category accessing
	 */
	public abstract double minLength();

	/**
	 * Answer my current origin point.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.JunPoint
	 * @category accessing
	 */
	protected JunPoint _origin() {
		return _origin;
	}

	/**
	 * Set my new origin point.
	 * 
	 * @param aPoint jp.co.sra.jun.geometry.basic.JunPoint
	 * @category accessing
	 */
	protected void _origin(JunPoint aPoint) {
		_origin = aPoint;
	}

	/**
	 * Answer my current corner point.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.JunPoint
	 * @category accessing
	 */
	protected JunPoint _corner() {
		return _corner;
	}

	/**
	 * Set my new corner point.
	 * 
	 * @param aPoint jp.co.sra.jun.geometry.basic.JunPoint
	 * @category accessing
	 */
	protected void _corner(JunPoint aPoint) {
		_corner = aPoint;
	}

	/**
	 * Answer true if the receiver is equal to the object while concerning an accuracy.
	 * 
	 * @param anObject java.lang.Object
	 * @return boolean
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#equal_(java.lang.Object)
	 * @category comparing
	 */
	@Override
	public boolean equal_(Object anObject) {
		if (anObject == null || this.getClass() != anObject.getClass()) {
			return false;
		}

		JunBoundingBox aBoundingBox = (JunBoundingBox) anObject;
		return this._origin().equal_(aBoundingBox._origin()) && this._corner().equal_(aBoundingBox._corner());
	}

	/**
	 * Answer true if the Object is equal to the receiver, otherwise false.
	 * 
	 * @param anObject java.lang.Object
	 * @return boolean
	 * @see jp.co.sra.jun.geometry.boundaries.JunBoundingBox#equals(java.lang.Object)
	 * @category comparing
	 */
	@Override
	public boolean equals(Object anObject) {
		if (anObject == null || this.getClass() != anObject.getClass()) {
			return false;
		}

		JunBoundingBox aBoundingBox = (JunBoundingBox) anObject;
		return this._origin().equals(aBoundingBox._origin()) && this._corner().equals(aBoundingBox._corner());
	}

	/**
	 * Compute a hash code for this object.
	 * 
	 * @return int
	 * @see java.lang.Object#hashCode()
	 * @category comparing
	 */
	@Override
	public int hashCode() {
		return this._origin().hashCode() | this._corner().hashCode();
	}

	/**
	 * Convert to a Rectangle.
	 * 
	 * @return java.awt.Rectangle
	 * @category converting
	 */
	public abstract Rectangle asRectangle();

	/**
	 * Convert to an array of Rectangle.
	 * 
	 * @return java.awt.Rectangle[]
	 * @category converting
	 */
	public abstract Rectangle[] asRectangles();

	/**
	 * Do an extra copy of the receiver.
	 * 
	 * @return jp.co.sra.smalltalk.StObject
	 * @see jp.co.sra.smalltalk.StObject#postCopy()
	 * @category copying
	 */
	@Override
	public StObject postCopy() {
		super.postCopy();
		_origin = (JunPoint) _origin.copy();
		_corner = (JunPoint) _corner.copy();
		return this;
	}

	/**
	 * Get the receiver's attributes from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#fromLispList(jp.co.sra.jun.goodies.lisp.JunLispList)
	 * @category lisp support
	 */
	@Override
	public void fromLispList(JunLispList aList) {
		super.fromLispList(aList);
		this.originFromLispList(aList);
		this.cornerFromLispList(aList);
	}

	/**
	 * Convert the receiver's attributes to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#toLispList()
	 * @category lisp support
	 */
	@Override
	public JunLispCons toLispList() {
		JunLispCons list = super.toLispList();
		list.add_(this.originToLispList());
		list.add_(this.cornerToLispList());
		return list;
	}

	/**
	 * Get the receiver's corner value from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected abstract void cornerFromLispList(JunLispList aList);

	/**
	 * Convert the receiver's corner value to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected JunLispCons cornerToLispList() {
		JunLispCons aList = this.lispCons();
		aList.head_($("corner"));
		aList.tail_(this._corner().toLispList());
		return aList;
	}

	/**
	 * Get the receiver's origin value from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected abstract void originFromLispList(JunLispList aList);

	/**
	 * Convert the receiver's origin value to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected JunLispCons originToLispList() {
		JunLispCons aList = this.lispCons();
		aList.head_($("origin"));
		aList.tail_(this._origin().toLispList());
		return aList;
	}

	/**
	 * Print my storable string representation on the writer.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException
	 * @see jp.co.sra.smalltalk.StObject#storeOn_(java.io.Writer)
	 * @category printing
	 */
	@Override
	public void storeOn_(Writer aWriter) throws IOException {
		aWriter.write('(');
		this._origin().printOn_(aWriter);
		aWriter.write(" box:");
		this._corner().printOn_(aWriter);
		aWriter.write(')');
	}

	/**
	 * Answer true if the receiver is a bounding box, otherwise false.
	 * 
	 * @return boolean
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#isBoundingBox()
	 * @category testing
	 */
	@Override
	public boolean isBoundingBox() {
		return true;
	}

}
