package jp.co.sra.jun.geometry.boxtree;

import java.awt.Color;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import jp.co.sra.smalltalk.SmalltalkException;
import jp.co.sra.smalltalk.StBlockClosure;
import jp.co.sra.smalltalk.StBlockClosure1;
import jp.co.sra.smalltalk.StBlockClosure2;
import jp.co.sra.smalltalk.StSymbol;

import jp.co.sra.jun.geometry.abstracts.JunGeometry;
import jp.co.sra.jun.geometry.basic.Jun3dPoint;
import jp.co.sra.jun.geometry.boundaries.Jun3dBoundingBox;
import jp.co.sra.jun.geometry.curves.Jun3dLine;
import jp.co.sra.jun.geometry.pluralities.Jun3dBoundingBoxes;
import jp.co.sra.jun.geometry.surfaces.Jun3dTriangle;
import jp.co.sra.jun.opengl.objects.JunOpenGL3dCompoundObject;
import jp.co.sra.jun.opengl.objects.JunOpenGL3dObject;
import jp.co.sra.jun.opengl.objects.JunOpenGL3dPolylineLoop;

/**
 * Jun3dBoundingBoxNode class
 * 
 *  @author    Mitsuhiro Asada
 *  @created   2007/05/28 (by m-asada)
 *  @updated   2008/04/15 (by nisinaka)
 *  @updated   2008/05/28 (by m-asada)
 *  @version   758 (with StPL9.7) based on Jun666 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: Jun3dBoundingBoxNode.java,v 9.13 2010/03/09 09:16:51 nisinaka Exp $
 */
public class Jun3dBoundingBoxNode extends JunBoundingBoxNode<Jun3dBoundingBoxNode> {
	/**
	 * Serial Version UID.
	 */
	private static final long serialVersionUID = 154605064818405668L;

	protected Jun3dBoundingBox boundingBox;
	protected StBlockClosure2<Boolean, JunGeometry, Jun3dBoundingBox> compareWithBlock;

	/**
	 * Create a new instance of <code>Jun3dBoundingBoxNode</code> and initialize it.
	 * 
	 * @category Instance creation
	 */
	protected Jun3dBoundingBoxNode() {
		super();
	}

	/**
	 * Create a new instance of <code>Jun3dBoundingBoxNode</code> and initialize it.
	 * 
	 * @param objectCollection java.util.Collection
	 * @param boundingBox jp.co.sra.jun.geometry.boundaries.Jun3dBoundingBox
	 * @category Instance creation
	 */
	public Jun3dBoundingBoxNode(Collection<JunGeometry> objectCollection, Jun3dBoundingBox boundingBox) {
		super();
		this.containedObjects_(objectCollection);
		this.boundingBox_(boundingBox);
	}

	/**
	 * Create a new instance of <code>Jun3dBoundingBoxNode</code> and initialize it.
	 * 
	 * @param segmentCollection jp.co.sra.jun.geometry.curves.Jun3dLine[]
	 * @category Instance creation
	 */
	public Jun3dBoundingBoxNode(Jun3dLine[] segmentCollection) {
		super();

		JunGeometry.HashEqualitySet<Jun3dLine> lineSet = new JunGeometry.HashEqualitySet<Jun3dLine>(segmentCollection.length);
		for (int i = 0; i < segmentCollection.length; i++) {
			lineSet.add(segmentCollection[i]);
		}
		Collection<JunGeometry> containedObjects = new ArrayList<JunGeometry>();
		containedObjects.addAll(lineSet);

		JunGeometry.HashEqualitySet<Jun3dPoint> pointSet = new JunGeometry.HashEqualitySet<Jun3dPoint>(segmentCollection.length + 1);
		for (int i = 0; i < segmentCollection.length; i++) {
			pointSet.add(segmentCollection[i].first());
			pointSet.add(segmentCollection[i].last());
		}
		Jun3dPoint[] pointCollection = pointSet.toArray(new Jun3dPoint[pointSet.size()]);
		Jun3dBoundingBox boundingBox = new Jun3dBoundingBox(pointCollection);

		this.containedObjects_(containedObjects);
		this.boundingBox_(boundingBox);
	}

	/**
	 * Create a new instance of <code>Jun3dBoundingBoxNode</code> and initialize it.
	 * 
	 * @param pointCollection jp.co.sra.jun.geometry.basic.Jun3dPoint[]
	 * @category Instance creation
	 */
	public Jun3dBoundingBoxNode(Jun3dPoint[] pointCollection) {
		super();

		JunGeometry.HashEqualitySet<Jun3dPoint> pointSet = new JunGeometry.HashEqualitySet<Jun3dPoint>(pointCollection.length);
		for (int i = 0; i < pointCollection.length; i++) {
			pointSet.add(pointCollection[i]);
		}
		Collection<JunGeometry> containedObjects = new ArrayList<JunGeometry>(pointSet);
		Jun3dBoundingBox boundingBox = new Jun3dBoundingBox(containedObjects.toArray(new Jun3dPoint[containedObjects.size()]));

		this.containedObjects_(containedObjects);
		this.boundingBox_(boundingBox);
	}

	/**
	 * Create a new instance of <code>Jun3dBoundingBoxNode</code> and initialize it.
	 * 
	 * @param patchCollection jp.co.sra.jun.geometry.surfaces.Jun3dTriangle[]
	 * @category Instance creation
	 */
	public Jun3dBoundingBoxNode(Jun3dTriangle[] patchCollection) {
		super();

		JunGeometry.HashEqualitySet<Jun3dTriangle> triangleSet = new JunGeometry.HashEqualitySet<Jun3dTriangle>(patchCollection.length);
		for (int i = 0; i < patchCollection.length; i++) {
			triangleSet.add(patchCollection[i]);
		}
		Collection<JunGeometry> containedObjects = new ArrayList<JunGeometry>(triangleSet);

		JunGeometry.HashEqualitySet<Jun3dPoint> pointSet = new JunGeometry.HashEqualitySet<Jun3dPoint>(patchCollection.length + 1);
		for (int i = 0; i < patchCollection.length; i++) {
			pointSet.add(patchCollection[i].p1());
			pointSet.add(patchCollection[i].p2());
			pointSet.add(patchCollection[i].p3());
		}
		Jun3dPoint[] pointCollection = pointSet.toArray(new Jun3dPoint[pointSet.size()]);
		Jun3dBoundingBox boundingBox = new Jun3dBoundingBox(pointCollection);

		this.containedObjects_(containedObjects);
		this.boundingBox_(boundingBox);
	}

	/**
	 * Initialize the receiver.
	 * 
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#initialize()
	 * @category initialize-release
	 */
	@Override
	protected void initialize() {
		super.initialize();

		boundingBox = null;
		compareWithBlock = new StBlockClosure2<Boolean, JunGeometry, Jun3dBoundingBox>() {
			public Boolean value(JunGeometry geometry, Jun3dBoundingBox boundingBox) {
				if (geometry.isPoint()) {
					return new Boolean(((Jun3dPoint) geometry).boundingBox().touches_(boundingBox));
				}
				if (geometry.isBoundingBox()) {
					return new Boolean(((Jun3dBoundingBox) geometry).boundingBox().touches_(boundingBox));
				}
				if (geometry.isLine()) {
					return new Boolean(((Jun3dLine) geometry).boundingBox().touches_(boundingBox));
				}
				if (geometry.isTriangle()) {
					return new Boolean(((Jun3dTriangle) geometry).boundingBox().touches_(boundingBox));
				}
				return null;
			}
		};
	}

	/**
	 * Answer the receiver's bounding box.
	 * 
	 * @return jp.co.sra.jun.geometry.boundaries.Jun3dBoundingBox
	 * @category accessing
	 */
	public Jun3dBoundingBox boundingBox() {
		if (boundingBox == null) {
			boundingBox = new Jun3dBoundingBox();
		}
		return boundingBox;
	}

	/**
	 * Set the receiver's bounding box.
	 * 
	 * @param aBoundingBox jp.co.sra.jun.geometry.boundaries.Jun3dBoundingBox
	 * @category accessing
	 */
	public void boundingBox_(Jun3dBoundingBox aBoundingBox) {
		boundingBox = aBoundingBox;
		this.flushSubdividedNodes();
	}

	/**
	 * Answer the receiver's compare with block.
	 * 
	 * @return jp.co.sra.smalltalk.StBlockClosure2
	 * @category accessing
	 */
	public StBlockClosure2<Boolean, JunGeometry, Jun3dBoundingBox> compareWithBlock() {
		return compareWithBlock;
	}

	/**
	 * Set the receiver's compare with block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @deprecated since Jun701, use Generics.
	 * @category accessing
	 */
	public void compareWithBlock_(final StBlockClosure<?, ?, ?, ?> aBlock) {
		this.compareWithBlock_(new StBlockClosure2<Boolean, JunGeometry, Jun3dBoundingBox>() {
			public Boolean value(JunGeometry aGeometry, Jun3dBoundingBox aBoundingBox) {
				return (Boolean) aBlock.value_value_(aGeometry, aBoundingBox);
			}
		});
	}

	/**
	 * Set the receiver's compare with block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure2
	 * @category accessing
	 */
	public void compareWithBlock_(StBlockClosure2<Boolean, JunGeometry, Jun3dBoundingBox> aBlock) {
		compareWithBlock = aBlock;
	}

	/**
	 * Answer the receiver's subdivided nodes.
	 * 
	 * @return jp.co.sra.jun.geometry.boxtree.JunBoundingBoxNode[]
	 * @see jp.co.sra.jun.geometry.boxtree.JunBoundingBoxNode#subdividedNodes()
	 * @category accessing
	 */
	@Override
	public Jun3dBoundingBoxNode[] subdividedNodes() {
		if (subdividedNodes == null) {
			this.subdivide();
		}
		return subdividedNodes.toArray(new Jun3dBoundingBoxNode[subdividedNodes.size()]);
	}

	/**
	 * Answer the subdivided nodes with the specified level value.
	 * 
	 * @param levelValue int
	 * @return jp.co.sra.jun.geometry.boxtree.JunBoundingBoxNode[]
	 * @see jp.co.sra.jun.geometry.boxtree.JunBoundingBoxNode#subdividedNodesLevel_(int)
	 * @category accessing
	 */
	@Override
	public Jun3dBoundingBoxNode[] subdividedNodesLevel_(int levelValue) {
		this.subdivideLevel_(levelValue);
		final Collection<Jun3dBoundingBoxNode> aCollection = new ArrayList<Jun3dBoundingBoxNode>();
		this.do_(new StBlockClosure1<Void, Jun3dBoundingBoxNode>() {
			public Void value(Jun3dBoundingBoxNode node) {
				if (node.boxSituation() == $("necessary")) {
					aCollection.add(node);
				}
				return null;
			}
		});
		return aCollection.toArray(new Jun3dBoundingBoxNode[aCollection.size()]);
	}

	/**
	 * Answer <code>true</code> if the receiver is equal to the object while concerning an accuracy.
	 * 
	 * @param anObject java.lang.Object
	 * @return boolean
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#equal_(java.lang.Object)
	 * @category comparing
	 */
	@Override
	public boolean equal_(Object anObject) {
		if (anObject == null || this.getClass() != anObject.getClass()) {
			return false;
		}
		return this.boundingBox().equal_(((Jun3dBoundingBoxNode) anObject).boundingBox());
	}

	/**
	 * Compute a hash code for this object.
	 * 
	 * @return int
	 * @see java.lang.Object#hashCode()
	 * @category comparing
	 */
	@Override
	public int hashCode() {
		return this.boundingBox().hashCode();
	}

	/**
	 * Convert to a <code>Jun3dBoundingBox</code>.
	 * 
	 * @return jp.co.sra.jun.geometry.boundaries.Jun3dBoundingBox
	 * @category converting
	 */
	public Jun3dBoundingBox asBoundingBox() {
		return this.boundingBox();
	}

	/**
	 * Convert to a <code>JunOpenGL3dObject</code>.
	 * 
	 * @return jp.co.sra.jun.opengl.objects.JunOpenGL3dObject
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#asJunOpenGL3dObject()
	 * @category converting
	 */
	@Override
	public JunOpenGL3dObject asJunOpenGL3dObject() {
		Color colorValue = Color.gray;
		if (boxSituation == $("through")) {
			colorValue = Color.cyan;
		}
		if (boxSituation == $("necessary")) {
			colorValue = Color.magenta;
		}
		if (boxSituation == $("unnecessary")) {
			colorValue = Color.yellow;
		}
		float alphaValue = 0.25f;
		JunOpenGL3dCompoundObject compoundObject = new JunOpenGL3dCompoundObject();
		compoundObject.add_(this.boundingBox().asJunOpenGL3dObjectColor_alpha_(colorValue, alphaValue));
		Jun3dPoint[][] points = this.boundingBox().pps();
		for (int i = 0; i < points.length; i++) {
			JunOpenGL3dPolylineLoop polylineLoop = new JunOpenGL3dPolylineLoop(points[i]);
			polylineLoop.paint_(colorValue);
			polylineLoop.lineWidth_(1);
			compoundObject.add_(polylineLoop);
		}
		return compoundObject;
	}

	/**
	 * Print my string representation on the writer.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException
	 * @see jp.co.sra.smalltalk.StObject#printOn_(java.io.Writer)
	 * @category printing
	 */
	@Override
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write("(");
		aWriter.write(this.shortPrintString());
		aWriter.write(Integer.toString(this.containedObjects().length));
		aWriter.write(", ");
		this.boundingBox().printOn_(aWriter);
		aWriter.write(")");
	}

	/**
	 * Answer the receiver's short print string.
	 * 
	 * @return java.lang.String
	 * @see jp.co.sra.jun.geometry.boxtree.JunBoundingBoxNode#shortPrintString()
	 * @category printing
	 */
	@Override
	public String shortPrintString() {
		return "3dbbn";
	}

	/**
	 * Answer the receiver's subdivide bounding box node with the specified level.
	 * 
	 * @param aNumber int
	 * @return jp.co.sra.jun.geometry.boundaries.JunBoundingBoxNode
	 * @throws jp.co.sra.smalltalk.SmalltalkException
	 * @see jp.co.sra.jun.geometry.boxtree.JunBoundingBoxNode#subdivideLevel_(int)
	 * @category subdividing
	 */
	@Override
	public Jun3dBoundingBoxNode subdivideLevel_(int aNumber) {
		if (this.levelNumber() == aNumber) {
			return this;
		}
		final Jun3dBoundingBoxNode self = this;
		if (subdividedNodes == null) {
			final List<Jun3dBoundingBoxNode> aCollection = new ArrayList<Jun3dBoundingBoxNode>(2);
			try {
				((Jun3dBoundingBoxes) this.boundingBox().perform_(this.messageToSubdivide().toString())).do_(new StBlockClosure1<Void, Jun3dBoundingBox>() {
					public Void value(Jun3dBoundingBox box) {
						JunGeometry[] containedObjects = self.containedObjects();
						Collection<JunGeometry> objectCollection = new ArrayList<JunGeometry>();
						for (int i = 0; i < containedObjects.length; i++) {
							if (Boolean.TRUE.equals(self.compare_with_(containedObjects[i], box))) {
								objectCollection.add(containedObjects[i]);
							}
						}
						Jun3dBoundingBoxNode subdividedNode = new Jun3dBoundingBoxNode();
						subdividedNode.containedObjects_(objectCollection);
						subdividedNode.boundingBox_(box);
						subdividedNode.vanishBoolean_(vanishBoolean);
						subdividedNode.messageToSubdivide_(messageToSubdivide);
						subdividedNode.compareWithBlock_(compareWithBlock);
						subdividedNode.containerNode_(self);
						if (objectCollection.isEmpty()) {
							subdividedNode.boxSituation_($("unnecessary"));
						} else {
							subdividedNode.boxSituation_($("necessary"));
						}
						aCollection.add(subdividedNode);
						return null;
					}
				});
			} catch (Exception e) {
				throw new SmalltalkException(e);
			}
			subdividedNodes = aCollection;
		}
		Jun3dBoundingBoxNode[] boxNodes = subdividedNodes.toArray(new Jun3dBoundingBoxNode[subdividedNodes.size()]);
		for (int i = 0; i < boxNodes.length; i++) {
			boxNodes[i] = boxNodes[i].subdivideLevel_(aNumber);
			subdividedNodes.set(i, boxNodes[i]);
		}
		Collection<StSymbol> boxSituations = new ArrayList<StSymbol>();
		for (int i = 0; i < boxNodes.length; i++) {
			boxSituations.add(boxNodes[i].boxSituation());
		}

		if (boxSituations.contains($("unnecessary")) || boxSituations.contains($("through"))) {
			this.boxSituation_($("through"));
		} else {
			boolean isAllNecessary = true;
			StSymbol[] symbols = boxSituations.toArray(new StSymbol[boxSituations.size()]);
			for (int i = 0; i < symbols.length; i++) {
				if (symbols[i] != $("necessary")) {
					isAllNecessary = false;
					break;
				}
			}
			if (isAllNecessary) {
				this.boxSituation_($("necessary"));
				if (this.vanishBoolean()) {
					this.flushSubdividedNodes();
				} else {
					for (int i = 0; i < boxNodes.length; i++) {
						boxNodes[i].boxSituation_($("unnecessary"));
					}
				}
			} else {
				this.boxSituation_($("through"));
			}
		}
		return this;
	}

	/**
	 * Answer true if the receiver is a 3d geometry element, otherwise false.
	 * 
	 * @return boolean
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#is3d()
	 * @category testing
	 */
	@Override
	public boolean is3d() {
		return true;
	}

	/**
	 * Compare the specified geometry object with bounding box.
	 * 
	 * @param object jp.co.sra.jun.geometry.abstracts.JunGeometry
	 * @param box jp.co.sra.jun.geometry.boundaries.Jun3dBoundingBox
	 * @return java.lang.Boolean
	 * @category private
	 */
	protected Boolean compare_with_(JunGeometry object, Jun3dBoundingBox box) {
		return compareWithBlock.value(object, box);
	}
}
