package jp.co.sra.jun.geometry.boxtree;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import jp.co.sra.smalltalk.StBlockClosure;
import jp.co.sra.smalltalk.StBlockClosure1;
import jp.co.sra.smalltalk.StBlockClosure2;
import jp.co.sra.smalltalk.StSymbol;

import jp.co.sra.jun.geometry.abstracts.JunGeometry;

/**
 * JunBoundingBoxNode class
 * 
 *  @author    Mitsuhiro Asada
 *  @created   2007/05/28 (by m-asada)
 *  @updated   2008/04/15 (by nisinaka)
 *  @updated   2008/05/28 (by m-asada)
 *  @version   758 (with StPL9.7) based on Jun666 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunBoundingBoxNode.java,v 9.12 2010/03/09 09:16:51 nisinaka Exp $
 */
public abstract class JunBoundingBoxNode<N extends JunBoundingBoxNode<N>> extends JunGeometry {

	protected List<N> subdividedNodes;
	protected N containerNode;
	protected Collection<JunGeometry> containedObjects;
	protected boolean vanishBoolean;
	protected StSymbol messageToSubdivide;
	protected StSymbol boxSituation;
	protected int levelNumber;

	/**
	 * Create a new instance of <code>JunBoundingBoxNode</code> and initialize it.
	 * 
	 * @category Instance creation
	 */
	public JunBoundingBoxNode() {
		super();
	}

	/**
	 * Initialize the receiver.
	 * 
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#initialize()
	 * @category initialize-release
	 */
	@Override
	protected void initialize() {
		super.initialize();

		subdividedNodes = null;
		containerNode = null;
		containedObjects = null;
		vanishBoolean = true;
		messageToSubdivide = $("subdivide2");
		boxSituation = $("necessary");
		levelNumber = Integer.MIN_VALUE;
		this.flushSubdividedNodes();
	}

	/**
	 * Answer the receiver's box situation.
	 * 
	 * @return jp.co.sra.smalltalk.StSymbol
	 * @category accessing
	 */
	public StSymbol boxSituation() {
		return boxSituation;
	}

	/**
	 * Set the receiver's box situation.
	 * 
	 * @param aSymbol jp.co.sra.smalltalk.StSymbol
	 * @category accessing
	 */
	public void boxSituation_(StSymbol aSymbol) {
		boxSituation = aSymbol;
	}

	/**
	 * Set the receiver's contained objects.
	 * 
	 * @param objectCollection java.util.Collection
	 * @category accessing
	 */
	public void containedObjects_(Collection<JunGeometry> objectCollection) {
		containedObjects = objectCollection;
		this.flushSubdividedNodes();
	}

	/**
	 * Answer the receiver's container node.
	 * 
	 * @return jp.co.sra.jun.geometry.boundaries.JunBoundingBoxNode
	 * @category accessing
	 */
	public N containerNode() {
		return containerNode;
	}

	/**
	 * Set the receiver's container node.
	 * 
	 * @param aBoundingBoxNode jp.co.sra.jun.geometry.boundaries.JunBoundingBoxNode
	 * @category accessing
	 */
	public void containerNode_(N aBoundingBoxNode) {
		containerNode = aBoundingBoxNode;
	}

	/**
	 * Answer the receiver's contained objects.
	 * 
	 * @return jp.co.sra.jun.geometry.abstracts.JunGeometry[]
	 * @category accessing
	 */
	public JunGeometry[] containedObjects() {
		if (containedObjects == null) {
			containedObjects = new ArrayList<JunGeometry>();
		}
		return containedObjects.toArray(new JunGeometry[containedObjects.size()]);
	}

	/**
	 * Answer the receiver's level number.
	 * 
	 * @return int
	 * @see jp.co.sra.jun.geometry.boxtree.JunBoundingBoxNode#levelNumber()
	 * @category accessing
	 */
	public int levelNumber() {
		if (levelNumber <= 0) {
			N boxNode = this.containerNode();
			int level = 0;
			while (boxNode != null) {
				boxNode = boxNode.containerNode();
				level = level + 1;
			}
			levelNumber = level;
		}
		return levelNumber;
	}

	/**
	 * Answer the receiver's messageToSubdivide.
	 * 
	 * @return jp.co.sra.smalltalk.StSymbol
	 * @category accessing
	 */
	public StSymbol messageToSubdivide() {
		return messageToSubdivide;
	}

	/**
	 * Set the receiver's message selector of subdivide.
	 * 
	 * @param messageSelector jp.co.sra.smalltalk.StSymbol
	 * @category accessing
	 */
	public void messageToSubdivide_(StSymbol messageSelector) {
		messageToSubdivide = messageSelector;
	}

	/**
	 * Answer the receiver's subdivided nodes.
	 * 
	 * @return jp.co.sra.jun.geometry.boxtree.JunBoundingBoxNode[]
	 * @category accessing
	 */
	public abstract N[] subdividedNodes();

	/**
	 * Answer the subdivided nodes with the specified level value.
	 * 
	 * @param levelValue int
	 * @return jp.co.sra.jun.geometry.boxtree.JunBoundingBoxNode[]
	 * @category accessing
	 */
	public abstract N[] subdividedNodesLevel_(int levelValue);

	/**
	 * Answer the receiver's vanish as <code>boolean</code>.
	 * 
	 * @return boolean
	 * @category accessing
	 */
	public boolean vanishBoolean() {
		return vanishBoolean;
	}

	/**
	 * Set the receiver's vanish as <code>boolean</code>.
	 * 
	 * @param aBoolean boolean
	 * @category accessing
	 */
	public void vanishBoolean_(boolean aBoolean) {
		vanishBoolean = aBoolean;
	}

	/**
	 * Answer <code>true</code> if the receiver is equal to the object.
	 * 
	 * @return boolean
	 * @param anObject java.lang.Object
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#equals(java.lang.Object)
	 * @category comparing
	 */
	@Override
	public boolean equals(Object anObject) {
		return this.equal_(anObject);
	}

	/**
	 * Enumerate each objects and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @deprecated since Jun701, use Generics.
	 * @category enumerating 
	 */
	public Object do_(final StBlockClosure<?, ?, ?, ?> aBlock) {
		return this.do_(new StBlockClosure2<Void, Jun2dBoundingBoxNode, Integer>() {
			public Void value(Jun2dBoundingBoxNode aNode, Integer aNumber) {
				switch (aBlock.numArgs()) {
					case 1:
						aBlock.valueWithArguments_(new Object[] { JunBoundingBoxNode.this });
						break;
					case 2:
						aBlock.valueWithArguments_(new Object[] { JunBoundingBoxNode.this, new Integer(JunBoundingBoxNode.this.levelNumber()) });
						break;
				}
				return null;
			}
		});
	}

	/**
	 * Enumerate each objects and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure1
	 * @return java.lang.Object
	 * @category enumerating 
	 */
	@SuppressWarnings("unchecked")
	public Object do_(StBlockClosure1<Void, N> aBlock) {
		aBlock.value((N) this);
		if (subdividedNodes != null && subdividedNodes.size() > 0) {
			N[] geometries = this.subdividedNodes();
			for (int i = 0; i < geometries.length; i++) {
				geometries[i].do_(aBlock);
			}
		}
		return null;
	}

	/**
	 * Enumerate each objects and evaluate the block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure2
	 * @return java.lang.Object
	 * @category enumerating 
	 */
	@SuppressWarnings("unchecked")
	public Object do_(StBlockClosure2<Void, N, Integer> aBlock) {
		aBlock.value((N) this, new Integer(this.levelNumber()));
		if (subdividedNodes != null && subdividedNodes.size() > 0) {
			N[] geometries = this.subdividedNodes();
			for (int i = 0; i < geometries.length; i++) {
				geometries[i].do_(aBlock);
			}
		}
		return null;
	}

	/**
	 * Flush the receiver's subdivided nodes.
	 * 
	 * @category flushing
	 */
	protected void flushSubdividedNodes() {
		subdividedNodes = null;
	}

	/**
	 * Answer the receiver's short print string.
	 * 
	 * @return java.lang.String
	 * @category printing
	 */
	public abstract String shortPrintString();

	/**
	 * Answer the receiver's subdivide bounding box node.
	 * 
	 * @return jp.co.sra.jun.geometry.boundaries.JunBoundingBoxNode
	 * @category subdividing
	 */
	public N subdivide() {
		return this.subdivideLevel_(1);
	}

	/**
	 * Answer the receiver's subdivide bounding box node with the specified level.
	 * 
	 * @param aNumber int
	 * @return jp.co.sra.jun.geometry.boundaries.JunBoundingBoxNode
	 * @throws jp.co.sra.smalltalk.SmalltalkException
	 * @category subdividing
	 */
	public abstract N subdivideLevel_(int aNumber);
}
