package jp.co.sra.jun.geometry.boxtree;

import jp.co.sra.jun.geometry.abstracts.JunGeometry;

/**
 * JunBoundingBoxTree class
 * 
 *  @author    Mitsuhiro Asada
 *  @created   2007/06/26 (by m-asada)
 *  @updated   2008/05/28 (by m-asada)
 *  @version   758 (with StPL9.7) based on Jun666 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunBoundingBoxTree.java,v 9.10 2010/03/09 09:16:51 nisinaka Exp $
 */
public abstract class JunBoundingBoxTree<N extends JunBoundingBoxNode<N>> extends JunGeometry {
	protected N rootNode;
	protected N[] subdividedNodes;
	protected int levelNumber;

	/**
	 * Create a new instance of <code>JunBoundingBoxTree</code> and initialize it.
	 * 
	 * @category Instance creation
	 */
	public JunBoundingBoxTree() {
		super();
	}

	/**
	 * Initialize the receiver.
	 * 
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#initialize()
	 * @category initialize-release
	 */
	@Override
	protected void initialize() {
		super.initialize();

		rootNode = null;
		subdividedNodes = null;
		levelNumber = Integer.MIN_VALUE;
	}

	/**
	 * Answer the receiver's contained points.
	 * 
	 * @return jp.co.sra.jun.geometry.abstracts.JunGeometry[]
	 * @category accessing
	 */
	public JunGeometry[] containedObjects() {
		return this.rootNode().containedObjects();
	}

	/**
	 * Answer the receiver's levelNumber.
	 * 
	 * @return int
	 * @category accessing
	 */
	public int levelNumber() {
		if (levelNumber < 0) {
			levelNumber = 0;
		}
		return levelNumber;
	}

	/**
	 * Answer the receiver's root node.
	 * 
	 * @return jp.co.sra.jun.geometry.boxtree.JunBoundingBoxNode
	 * @category accessing
	 */
	public N rootNode() {
		return rootNode;
	}

	/**
	 * Set the receiver's root node.
	 * 
	 * @param boundingNode jp.co.sra.jun.geometry.boxtree.JunBoundingBoxNode
	 * @category accessing
	 */
	public void rootNode_(N boundingNode) {
		rootNode = boundingNode;
	}

	/**
	 * Answer the receiver's subdivided nodes.
	 * 
	 * @return jp.co.sra.jun.geometry.boxtree.JunBoundingBoxNode[]
	 * @category accessing
	 */
	public N[] subdividedNodes() {
		if (subdividedNodes == null) {
			this.subdividedNodesLevel_(1);
		}
		return subdividedNodes;
	}

	/**
	 * Answer the subdivided nodes with the specified level value.
	 * 
	 * @param levelValue int
	 * @return jp.co.sra.jun.geometry.boxtree.JunBoundingBoxNode[]
	 * @category accessing
	 */
	public N[] subdividedNodesLevel_(int levelValue) {
		levelNumber = levelValue;
		subdividedNodes = this.rootNode().subdividedNodesLevel_(levelNumber);
		this.rootNode().flushSubdividedNodes();
		return subdividedNodes;
	}

	/**
	 * Answer <code>true</code> if the receiver is equal to the object.
	 * 
	 * @return boolean
	 * @param anObject java.lang.Object
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#equals(java.lang.Object)
	 * @category comparing
	 */
	@Override
	public boolean equals(Object anObject) {
		return this.equal_(anObject);
	}
}
