package jp.co.sra.jun.geometry.curves;

import java.awt.Point;
import java.io.IOException;
import java.io.Writer;

import jp.co.sra.smalltalk.SmalltalkException;
import jp.co.sra.smalltalk.StBlockClosure0;
import jp.co.sra.smalltalk.StBlockClosure1;

import jp.co.sra.jun.geometry.abstracts.JunCurve;
import jp.co.sra.jun.geometry.basic.JunAngle;
import jp.co.sra.jun.geometry.basic.JunPoint;
import jp.co.sra.jun.geometry.surfaces.JunPlane;
import jp.co.sra.jun.goodies.lisp.JunLispCons;
import jp.co.sra.jun.goodies.lisp.JunLispList;

/**
 * JunLine class
 * 
 *  @author    nisinaka
 *  @created   1998/10/09 (by nisinaka)
 *  @updated   2000/01/06 (by nisinaka)
 *  @updated   2004/10/20 (by Mitsuhiro Asada)
 *  @version   758 (with StPL9.7) based on Jun744 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2005 Information-technology Promotion Agency, Japan (IPA)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunLine.java,v 9.10 2010/03/09 09:16:51 nisinaka Exp $
 */
public abstract class JunLine extends JunCurve {
	/**
	 * Create a new instance of <code>JunLine</code> and initialize it.
	 * 
	 * @category Instance creation
	 */
	public JunLine() {
		super();
	}

	/**
	 * Create a new instance of <code>Jun2dLine</code> and initialize it.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category Instance creation
	 */
	public JunLine(JunLispList aList) {
		super(aList);
	}

	/**
	 * Answer my start point.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.JunPoint
	 * @category accessing
	 */
	protected abstract JunPoint _from();

	/**
	 * Answer my end point.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.JunPoint
	 * @category accessing
	 */
	protected abstract JunPoint _to();

	/**
	 * Initialize this line with two points.
	 * 
	 * @param fromPoint jp.co.sra.jun.geometry.basic.JunPoint
	 * @param toPoint jp.co.sra.jun.geometry.basic.JunPoint
	 * @category accessing
	 */
	public abstract void from_to_(JunPoint fromPoint, JunPoint toPoint);

	/**
	 * Answer the length of this curve.
	 * 
	 * @return double
	 * @category accessing
	 */
	public abstract double length();

	/**
	 * Convert the receiver as an arrays of <code>Point</code>.
	 * 
	 * @return java.awt.Point[][]
	 * @category converting
	 */
	public abstract Point[][] asPointArrays();

	/**
	 * Answer the angle with the specified line.
	 * 
	 * @param aLine jp.co.sra.jun.geometry.curves.JunLine
	 * @return jp.co.sra.jun.geometry.basic.JunAngle
	 * @throws java.lang.IllegalArgumentException
	 * @category functions
	 */
	public abstract JunAngle angleWithLine_(JunLine aLine);

	/**
	 * Convert the receiver's from value to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected JunLispCons fromToLispList() {
		JunLispCons aList = this.lispCons();
		aList.head_($("from"));
		aList.tail_(this._from().toLispList());
		return aList;
	}

	/**
	 * Get the receiver's from value from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.geometry.curves.JunLine#fromAndToFromLispList(jp.co.sra.jun.goodies.lisp.JunLispList)
	 * @category lisp support
	 */
	protected void fromAndToFromLispList(JunLispList aList) {
		JunLispList list = (JunLispList) aList.detect_ifNone_(new StBlockClosure1<Boolean, Object>() {
			public Boolean value(Object car) {
				return new Boolean(car instanceof JunLispCons && ((JunLispCons) car).head() == $("from"));
			}
		}, new StBlockClosure0<Object>());

		if (list == null) {
			return;
		}
		JunPoint fromPoint = (JunPoint) list.tail();

		list = (JunLispList) aList.detect_ifNone_(new StBlockClosure1<Boolean, Object>() {
			public Boolean value(Object car) {
				return new Boolean(car instanceof JunLispCons && ((JunLispCons) car).head() == $("to"));
			}
		}, new StBlockClosure0<Object>());

		if (list == null) {
			return;
		}
		JunPoint toPoint = (JunPoint) list.tail();

		this.from_to_(fromPoint, toPoint);
	}

	/**
	 * Get my attributes from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#fromLispList(jp.co.sra.jun.goodies.lisp.JunLispList)
	 * @category lisp support
	 */
	@Override
	public void fromLispList(JunLispList aList) {
		this.fromAndToFromLispList(aList);
	}

	/**
	 * Convert the receiver's attributes to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#toLispList()
	 * @category lisp support
	 */
	@Override
	public JunLispCons toLispList() {
		JunLispCons aList = super.toLispList();
		aList.add_(this.fromToLispList());
		aList.add_(this.toToLispList());
		return aList;
	}

	/**
	 * Convert the receiver's to value to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected JunLispCons toToLispList() {
		JunLispCons aList = this.lispCons();
		aList.head_($("to"));
		aList.tail_(this._to().toLispList());
		return aList;
	}

	/**
	 * Print my string representation on the writer.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException
	 * @see jp.co.sra.smalltalk.StObject#printOn_(java.io.Writer)
	 * @category printing
	 */
	@Override
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write('(');
		this._from().printOn_(aWriter);
		aWriter.write(" to: ");
		this._to().printOn_(aWriter);
		aWriter.write(')');
	}

	/**
	 * Print my storable string representation on the writer.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException
	 * @see jp.co.sra.smalltalk.StObject#storeOn_(java.io.Writer)
	 * @category printing
	 */
	@Override
	public void storeOn_(Writer aWriter) throws IOException {
		aWriter.write('(');
		aWriter.write(this._className().toString());
		aWriter.write(" from: ");
		this._from().printOn_(aWriter);
		aWriter.write(" to: ");
		this._to().printOn_(aWriter);
		aWriter.write(')');
	}

	/**
	 * Answer true if the receiver is a line geometry, otherwise false.
	 * 
	 * @return boolean
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#isLine()
	 * @category testing
	 */
	@Override
	public boolean isLine() {
		return true;
	}

	/**
	 * Answer the sign value of which side of the plane the receiver is located.
	 * 
	 * @param aPlane jp.co.sra.jun.geometry.surfaces.JunPlane
	 * @return int
	 * @throws jp.co.sra.smalltalk.SmalltalkException
	 * @category testing
	 */
	public int whichSideOf_(JunPlane aPlane) {
		throw SmalltalkException.ShouldNotImplement();
	}
}
