package jp.co.sra.jun.geometry.pluralities;

import java.util.Collection;

import jp.co.sra.jun.geometry.basic.Jun3dPoint;
import jp.co.sra.jun.geometry.boundaries.Jun3dBoundingBall;
import jp.co.sra.jun.geometry.boundaries.Jun3dBoundingBox;
import jp.co.sra.jun.geometry.surfaces.Jun3dTriangle;
import jp.co.sra.jun.geometry.transformations.Jun3dTransformation;
import jp.co.sra.jun.goodies.lisp.JunLispConverter;
import jp.co.sra.jun.goodies.lisp.JunLispList;

/**
 * Jun3dTriangles class
 * 
 *  @author    Mitsuhiro Asada
 *  @created   2007/06/05 (by m-asada)
 *  @updated   2008/04/16 (by nisinaka)
 *  @version   758 (with StPL9.7) based on Jun667 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: Jun3dTriangles.java,v 9.13 2010/03/09 09:16:54 nisinaka Exp $
 */
public class Jun3dTriangles extends JunTriangles<Jun3dTriangle> {

	static {
		JunLispConverter.Register(Jun3dTriangles.class);
	}

	/**
	 * Serial Version UID.
	 */
	private static final long serialVersionUID = 7498948414865798291L;

	/**
	 * Create a new instance of <code>Jun3dTriangles</code> and initialize it.
	 * 
	 * @category Instance creation
	 */
	public Jun3dTriangles() {
		super();
	}

	/**
	 * Create a new instance of <code>Jun3dTriangles</code> and initialize it.
	 * 
	 * @param triangleCollection jp.co.sra.jun.geometry.surfaces.Jun3dTriangle[]
	 * @category Instance creation
	 */
	public Jun3dTriangles(Jun3dTriangle[] triangleCollection) {
		super(triangleCollection);
	}

	/**
	 * Create a new instance of <code>Jun3dTriangles</code> and initialize it.
	 * 
	 * @param triangleCollection java.util.Collection
	 * @category Instance creation
	 */
	public Jun3dTriangles(Collection<Jun3dTriangle> triangleCollection) {
		super(triangleCollection);
	}

	/**
	 * Create a new instance of <code>Jun3dTriangles</code> and initialize it.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category Instance creation
	 */
	public Jun3dTriangles(JunLispList aList) {
		super(aList);
	}

	/**
	 * Answer the geometry objects.
	 * 
	 * @return jp.co.sra.jun.geometry.boundaries.Jun2dTriangle[]
	 * @see jp.co.sra.jun.geometry.pluralities.JunGeometries#geometries()
	 * @category accessing
	 */
	@Override
	public Jun3dTriangle[] geometries() {
		return this._geometries().toArray(new Jun3dTriangle[this._geometries().size()]);
	}

	/**
	 * Answer my current bounding ball.
	 * 
	 * @return jp.co.sra.jun.geometry.boundaries.Jun3dBoundingBall
	 * @category accessing
	 */
	public Jun3dBoundingBall boundingBall() {
		if (this.isEmpty()) {
			return new Jun3dBoundingBall((Jun3dPoint[]) null);
		}
		Jun3dBoundingBall boundingBall = null;
		Jun3dTriangle[] objects = this.geometries();
		for (int i = 0; i < objects.length; i++) {
			Jun3dTriangle each = objects[i];
			if (boundingBall == null) {
				boundingBall = each.boundingBall();
			} else {
				boundingBall = each.boundingBall().merge_(boundingBall);
			}
		}
		return boundingBall;
	}

	/**
	 * Answer my current bounding box.
	 * 
	 * @return jp.co.sra.jun.geometry.boundaries.Jun3dBoundingBox
	 * @category accessing
	 */
	public Jun3dBoundingBox boundingBox() {
		if (this.isEmpty()) {
			return new Jun3dBoundingBox();
		}
		Jun3dBoundingBox boundingBox = null;
		Jun3dTriangle[] objects = this.geometries();
		for (int i = 0; i < objects.length; i++) {
			Jun3dTriangle each = objects[i];
			if (boundingBox == null) {
				boundingBox = each.boundingBox();
			} else {
				boundingBox = each.boundingBox().merge_(boundingBox);
			}
		}
		return boundingBox;
	}

	/**
	 * Convert to a <code>Jun2dBoundingBall</code>.
	 * 
	 * @return jp.co.sra.jun.geometry.boundaries.Jun3dBoundingBall
	 * @category converting
	 */
	public Jun3dBoundingBall asBoundingBall() {
		return this.boundingBall();
	}

	/**
	 * Convert to a <code>Jun2dBoundingBox</code>.
	 * 
	 * @return jp.co.sra.jun.geometry.boundaries.Jun3dBoundingBox
	 * @category converting
	 */
	public Jun3dBoundingBox asBoundingBox() {
		return this.boundingBox();
	}

	/**
	 * Answer the receiver's subdivide triangles.
	 * 
	 * @return jp.co.sra.jun.geometry.pluralities.Jun3dTriangles
	 * @category subdividing
	 */
	public Jun3dTriangles subdivide() {
		Jun3dTriangles copy = new Jun3dTriangles();
		Jun3dTriangle[] triangles = this.triangles();
		for (int i = 0; i < triangles.length; i++) {
			copy.addAll_(triangles[i].subdivide().geometries());
		}
		return copy;
	}

	/**
	 * Answer the receiver's subdivide triangles with specified level.
	 * 
	 * @param anInteger int
	 * @return jp.co.sra.jun.geometry.pluralities.Jun3dTriangles
	 * @category subdividing
	 */
	public Jun3dTriangles subdivideLevel_(int anInteger) {
		Jun3dTriangles copy = new Jun3dTriangles();
		Jun3dTriangle[] triangles = this.triangles();
		for (int i = 0; i < triangles.length; i++) {
			copy.addAll_(triangles[i].subdivideLevel_(anInteger).triangles());
		}
		return copy;
	}

	/**
	 * Answer true if the receiver is a 3d geometry element.
	 * 
	 * @return boolean
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#is3d()
	 * @category testing
	 */
	public boolean is3d() {
		return true;
	}

	/**
	 * Apply a transformation 'aTransformation' to the receiver.
	 * 
	 * @return jp.co.sra.jun.geometry.pluralities.Jun3dTriangles
	 * @param aTransformation jp.co.sra.jun.geometry.transformations.Jun3dTransformation
	 * @category transforming
	 */
	public Jun3dTriangles transform_(Jun3dTransformation aTransformation) {
		Jun3dTriangles transformedCopy = new Jun3dTriangles();
		Jun3dTriangle[] geometries = this.geometries();
		for (int i = 0; i < geometries.length; i++) {
			this.add_(geometries[i].transform_(aTransformation));
		}
		return transformedCopy;
	}

}
