package jp.co.sra.jun.geometry.surfaces;

import java.io.IOException;
import java.io.Writer;

import jp.co.sra.smalltalk.StBlockClosure0;
import jp.co.sra.smalltalk.StBlockClosure1;

import jp.co.sra.jun.geometry.abstracts.JunSurface;
import jp.co.sra.jun.geometry.basic.JunPoint;
import jp.co.sra.jun.goodies.lisp.JunLispCons;
import jp.co.sra.jun.goodies.lisp.JunLispList;

/**
 * JunCircle class
 * 
 *  @author    Mitsuhiro Asada
 *  @created   2004/12/17 (by Mitsuhiro Asada)
 *  @updated   2009/03/18 (by Mitsuhiro Asada)
 *  @version   758 (with StPL9.7) based on Jun721 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2005 Information-technology Promotion Agency, Japan (IPA)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunCircle.java,v 9.9 2010/03/09 09:16:56 nisinaka Exp $
 */
public abstract class JunCircle extends JunSurface {

	protected double radius;

	/**
	 * Create a new instance of <code>JunCircle</code> and initialize it.
	 * 
	 * @category Instance creation
	 */
	public JunCircle() {
		super();
	}

	/**
	 * Create a new instance of <code>JunCircle</code> and initialize it.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category Instance creation
	 */
	public JunCircle(JunLispList aList) {
		super(aList);
	}

	/**
	 * Answer my center point.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.JunPoint
	 * @category accessing
	 */
	protected abstract JunPoint _center();

	/**
	 * Get the receiver's radius.
	 * 
	 * @return double
	 * @category accessing
	 */
	public double radius() {
		return radius;
	}

	/**
	 * Answer the this area size
	 * 
	 * @return double
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#area()
	 * @category accessing
	 */
	@Override
	public double area() {
		return Math.PI * Math.pow(this.radius(), 2);
	}

	/**
	 * Get the parameter r.
	 * 
	 * @return double
	 * @category accessing
	 */
	public double r() {
		return this.radius();
	}

	/**
	 * Answer true if the receiver is equal to the object while concerning an accuracy.
	 * 
	 * @param anObject java.lang.Object
	 * @return boolean
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#equal_(java.lang.Object)
	 * @category comparing
	 */
	@Override
	public boolean equal_(Object anObject) {
		if (this.getClass() != anObject.getClass()) {
			return false;
		}

		JunCircle aCircle = (JunCircle) anObject;
		return this.isEqualNumber_to_(this.radius(), aCircle.radius());
	}

	/**
	 * Answer true if the Object is equal to the receiver, otherwise false.
	 * 
	 * @param anObject java.lang.Object
	 * @return boolean
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#equals(java.lang.Object)
	 * @category comparing
	 */
	@Override
	public boolean equals(Object anObject) {
		if (this.getClass() != anObject.getClass()) {
			return false;
		}

		JunCircle aCircle = (JunCircle) anObject;
		return this.radius() == aCircle.radius();
	}

	/**
	 * Convert the receiver as an array of JunPlane.
	 * 
	 * @return jp.co.sra.jun.geometry.surfaces.JunPlane[]
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#asArrayOfPlanes()
	 * @category converting
	 */
	@Override
	public JunPlane[] asArrayOfPlanes() {
		return new JunPlane[] { this.asPlane() };
	}

	/**
	 * Convert to a JunPlane.
	 * 
	 * @return jp.co.sra.jun.geometry.surfaces.JunPlane
	 * @category converting
	 */
	public abstract JunPlane asPlane();

	/**
	 * Get the receiver's attributes from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#fromLispList(jp.co.sra.jun.goodies.lisp.JunLispList)
	 * @category lisp support
	 */
	@Override
	public void fromLispList(JunLispList aList) {
		this.radiusFromLispList(aList);
		this.centerFromLispList(aList);
	}

	/**
	 * Convert the receiver's attributes to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#toLispList()
	 * @category lisp support
	 */
	@Override
	public JunLispCons toLispList() {
		JunLispCons aList = super.toLispList();
		aList.add_(this.radiusToLispList());
		aList.add_(this.centerToLispList());
		return aList;
	}

	/**
	 * Get the receiver's center value from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected abstract void centerFromLispList(JunLispList aList);

	/**
	 * Convert the receiver's center value to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected JunLispCons centerToLispList() {
		JunLispCons aList = this.lispCons();
		aList.head_($("center"));
		aList.tail_(this._center().toLispList());
		return aList;
	}

	/**
	 * Get the receiver's radius value from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected void radiusFromLispList(JunLispList aList) {
		JunLispList list = (JunLispList) aList.detect_ifNone_(new StBlockClosure1<Boolean, Object>() {
			public Boolean value(Object car) {
				return new Boolean(car instanceof JunLispCons && ((JunLispCons) car).head() == $("radius"));
			}
		}, new StBlockClosure0<Object>());
		if (list == null) {
			return;
		}

		this.setRadius_(((Number) list.tail()).doubleValue());
	}

	/**
	 * Convert the receiver's radius value to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected JunLispCons radiusToLispList() {
		JunLispCons aList = this.lispCons();
		aList.head_($("radius"));
		aList.tail_(this.radius());
		return aList;
	}

	/**
	 * Print my string representation on the writer.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#printOn_(java.io.Writer)
	 * @category printing
	 */
	@Override
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write('(');
		this._center().printOn_(aWriter);
		aWriter.write(" radius: ");
		aWriter.write(String.valueOf(this.radius()));
		aWriter.write(')');
	}

	/**
	 * Print my storable string representation on the writer.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException
	 * @see jp.co.sra.smalltalk.StObject#storeOn_(java.io.Writer)
	 * @category printing
	 */
	@Override
	public void storeOn_(Writer aWriter) throws IOException {
		aWriter.write('(');
		aWriter.write(this._className().toString());
		aWriter.write(" center: ");
		this._center().printOn_(aWriter);
		aWriter.write(" radius: ");
		aWriter.write(String.valueOf(this.radius()));
		aWriter.write(')');
	}

	/**
	 * Answer true if the receiver is a circle, otherwise false.
	 * 
	 * @return boolean
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#isCircle()
	 * @category testing
	 */
	@Override
	public boolean isCircle() {
		return true;
	}

	/**
	 * Set the receiver's center point.
	 * 
	 * @param aPoint jp.co.sra.jun.geometry.basic.JunPoint
	 * @category private
	 */
	protected abstract void setCenter_(JunPoint aPoint);

	/**
	 * Set the receiver's radius.
	 * 
	 * @param aNumber double
	 * @category private
	 */
	protected void setR_(double aNumber) {
		this.setRadius_(aNumber);
	}

	/**
	 * Set the receiver's radius.
	 * 
	 * @param aNumber double
	 * @category private
	 */
	protected void setRadius_(double aNumber) {
		radius = aNumber;
	}
}
