package jp.co.sra.jun.geometry.surfaces;

import java.util.ArrayList;
import java.util.List;

import jp.co.sra.smalltalk.StBlockClosure0;
import jp.co.sra.smalltalk.StBlockClosure1;

import jp.co.sra.jun.geometry.abstracts.JunSurface;
import jp.co.sra.jun.geometry.basic.Jun3dPoint;
import jp.co.sra.jun.geometry.boundaries.Jun2dBoundingBox;
import jp.co.sra.jun.geometry.boundaries.Jun3dBoundingBox;
import jp.co.sra.jun.geometry.boundaries.JunBoundingBox;
import jp.co.sra.jun.geometry.curves.Jun3dLine;
import jp.co.sra.jun.goodies.lisp.JunLispCons;
import jp.co.sra.jun.goodies.lisp.JunLispConverter;
import jp.co.sra.jun.goodies.lisp.JunLispList;

/**
 * JunCompoundSurface class
 * 
 *  @author    Mitsuhiro Asada
 *  @created   2007/07/05 (by m-asada)
 *  @updated   2008/04/17 (by nisinaka)
 *  @updated   2009/03/19 (by Mitsuhiro Asada)
 *  @version   758 (with StPL9.7) based on Jun721 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunCompoundSurface.java,v 9.12 2010/03/09 09:16:56 nisinaka Exp $
 */
public class JunCompoundSurface extends JunSurface {

	static {
		JunLispConverter.Register(JunCompoundSurface.class);
	}

	/**
	 * Serial Version UID.
	 */
	private static final long serialVersionUID = 610532358558700436L;

	protected List<JunSurface> components;
	protected JunBoundingBox boundingBox;

	/**
	 * Create a new instance of <code>JunCompoundSurface</code> and initialize it.
	 * 
	 * @category Instance creation
	 */
	protected JunCompoundSurface() {
		super();
	}

	/**
	 * Create a new instance of <code>JunCompoundSurface</code> and initialize it.
	 * 
	 * @param anArrayOfJunSurface jp.co.sra.jun.geometry.abstracts.JunSurface[]
	 * @category Instance creation
	 */
	public JunCompoundSurface(JunSurface[] anArrayOfJunSurface) {
		super();
		this.setComponents_(anArrayOfJunSurface);
	}

	/**
	 * Create a new instance of <code>JunCompoundSurface</code> and initialize it.
	 * 
	 * @param anArrayOfJunSurface java.util.List
	 * @category Instance creation
	 */
	public JunCompoundSurface(List<JunSurface> anArrayOfJunSurface) {
		super();
		this.setComponents_(anArrayOfJunSurface);
	}

	/**
	 * Create a new instance of <code>JunCompoundSurface</code> and initialize it.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category Instance creation
	 */
	public JunCompoundSurface(JunLispList aList) {
		super(aList);
	}

	/**
	 * Answer the geometry objects as <code>List</code>.
	 * 
	 * @return java.util.List
	 * @category accessing
	 */
	protected List<JunSurface> _components() {
		if (components == null) {
			components = new ArrayList<JunSurface>();
		}
		return components;
	}

	/**
	 * Answer the receiver's components.
	 * 
	 * @return jp.co.sra.jun.geometry.abstracts.JunSurface[]
	 * @category accessing
	 */
	public JunSurface[] components() {
		return this._components().toArray(new JunSurface[this._components().size()]);
	}

	/**
	 * Answer the receiver's bounding box.
	 * 
	 * @return jp.co.sra.jun.geometry.boundaries.JunBoundingBox
	 * @category bounds accessing
	 */
	public JunBoundingBox boundingBox() {
		if (boundingBox == null) {
			boundingBox = this.preferredBoundingBox();
		}
		return boundingBox;
	}

	/**
	 * Answer true if the receiver is equal to the object while concerning an accuracy.
	 * 
	 * @param anObject java.lang.Object
	 * @return boolean
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#equal_(java.lang.Object)
	 * @category comparing
	 */
	@Override
	public boolean equal_(Object anObject) {
		if (anObject == null || this.getClass() != anObject.getClass()) {
			return false;
		}

		JunCompoundSurface another = (JunCompoundSurface) anObject;
		JunSurface[] myComponents = this.components();
		JunSurface[] anotherComponents = another.components();
		if (myComponents.length != anotherComponents.length) {
			return false;
		}

		for (int i = 0; i < myComponents.length; i++) {
			if (myComponents[i].equal_(anotherComponents[i]) == false) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Answer true if the Object is equal to the receiver, otherwise false.
	 * 
	 * @param anObject java.lang.Object
	 * @return boolean
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#equals(java.lang.Object)
	 * @category comparing
	 */
	@Override
	public boolean equals(Object anObject) {
		if (anObject == null || this.getClass() != anObject.getClass()) {
			return false;
		}

		JunCompoundSurface another = (JunCompoundSurface) anObject;
		JunSurface[] myComponents = this.components();
		JunSurface[] anotherComponents = another.components();
		if (myComponents.length != anotherComponents.length) {
			return false;
		}

		for (int i = 0; i < myComponents.length; i++) {
			if (myComponents[i].equals(anotherComponents[i]) == false) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Answer the distance from specified Point.
	 * 
	 * @param aPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @return double
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#distanceFromPoint_(jp.co.sra.jun.geometry.basic.Jun3dPoint)
	 * @category functions
	 */
	@Override
	public double distanceFromPoint_(Jun3dPoint aJun3dPoint) {
		return this.nearestPointFromPoint_(aJun3dPoint).distance_(aJun3dPoint);
	}

	/**
	 * Answer the intersecting point with the specified line.
	 * 
	 * @param aLine jp.co.sra.jun.geometry.curves.Jun3dLine
	 * @return jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category functions
	 */
	public Jun3dPoint intersectingPointWithLine_(Jun3dLine aJun3dLine) {
		JunSurface[] components = this.components();
		for (int i = 0; i < components.length; i++) {
			try {
				Jun3dPoint intersectingPoint = (Jun3dPoint) components[i].perform_with_("intersectingPointWithLine_", aJun3dLine);
				if (intersectingPoint != null) {
					return intersectingPoint;
				}
			} catch (Exception e) {
				System.out.println(e.getMessage());
				e.printStackTrace();
			}
		}
		return null;
	}

	/**
	 * Answer the intersecting point with the specified line segment.
	 * 
	 * @param aLine jp.co.sra.jun.geometry.curves.Jun3dLine
	 * @return jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category functions
	 */
	public Jun3dPoint intersectingPointWithLineSegment_(Jun3dLine aJun3dLine) {
		JunSurface[] components = this.components();
		for (int i = 0; i < components.length; i++) {
			try {
				Jun3dPoint intersectingPoint = (Jun3dPoint) components[i].perform_with_("intersectingPointWithLineSegment_", aJun3dLine);
				if (intersectingPoint != null) {
					return intersectingPoint;
				}
			} catch (Exception e) {
				System.out.println(e.getMessage());
				e.printStackTrace();
			}
		}
		return null;
	}

	/**
	 * Answer the distance from specified point.
	 * 
	 * @param aPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @return double
	 * @category functions
	 */
	public Jun3dPoint nearestPointFromPoint_(Jun3dPoint aJun3dPoint) {
		JunSurface[] components = this.components();
		Jun3dPoint thePoint = null;
		double minDistance;
		try {
			thePoint = (Jun3dPoint) components[0].perform_with_("nearestPointFromPoint_", aJun3dPoint);
			minDistance = aJun3dPoint.distance_(thePoint);
			for (int i = 1; i < components.length; i++) {
				Jun3dPoint point = (Jun3dPoint) components[i].perform_with_("nearestPointFromPoint_", aJun3dPoint);
				double distance = aJun3dPoint.distance_(point);
				if (minDistance > distance) {
					thePoint = point;
					minDistance = distance;
				}
			}
		} catch (Exception e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		}

		return thePoint;
	}

	/**
	 * Get the receiver's attributes from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#fromLispList(jp.co.sra.jun.goodies.lisp.JunLispList)
	 * @category lisp support
	 */
	@Override
	public void fromLispList(JunLispList aList) {
		this.componentsFromLispList(aList);
	}

	/**
	 * Convert the receiver's attributes to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#toLispList()
	 * @category lisp support
	 */
	@Override
	public JunLispCons toLispList() {
		JunLispCons aList = super.toLispList();
		aList.add_(this.componentsToLispList());
		return aList;
	}

	/**
	 * Get the receiver's components value from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected void componentsFromLispList(JunLispList aList) {
		JunLispList list = (JunLispList) aList.detect_ifNone_(new StBlockClosure1<Boolean, Object>() {
			public Boolean value(Object car) {
				return new Boolean(car instanceof JunLispCons && ((JunLispCons) car).head() == $("components"));
			}
		}, new StBlockClosure0<Object>());
		if (list == null) {
			return;
		}

		Object[] items = ((JunLispList) list.tail()).asArray();
		List<JunSurface> componentList = new ArrayList<JunSurface>(items.length);
		for (int i = 0; i < items.length; i++) {
			componentList.add(JunSurface.FromLispList((JunLispList) items[i]));
		}
		this.setComponents_(componentList);
	}

	/**
	 * Convert the receiver's components value to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected JunLispCons componentsToLispList() {
		JunLispCons list = this.lispCons();
		list.head_($("components"));
		JunSurface[] components = this.components();
		for (int i = 0; i < components.length; i++) {
			list.add_(components[i].toLispList());
		}
		return list;
	}

	/**
	 * Answer <code>true</code> if the receiver is a compound surface, otherwise <code>false</code>.
	 * 
	 * @return boolean
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#isCompoundSurface()
	 * @category testing
	 */
	@Override
	public boolean isCompoundSurface() {
		return true;
	}

	/**
	 * Answer the receiver's preferred bounding box.
	 * 
	 * @return jp.co.sra.jun.geometry.boundaries.JunBoundingBox
	 * @category private
	 */
	protected JunBoundingBox preferredBoundingBox() {
		JunSurface[] components = this.components();
		JunBoundingBox boundingBox = null;
		try {
			if (components[0].is2d()) {
				Jun2dBoundingBox bounds = ((Jun2dBoundingBox) components[0].perform_("boundingBox"));
				for (int i = 1; i < components.length; i++) {
					bounds.merge_(((Jun2dBoundingBox) components[i].perform_("boundingBox")));
				}
				boundingBox = bounds;
			} else {
				Jun3dBoundingBox bounds = ((Jun3dBoundingBox) components[0].perform_("boundingBox"));
				for (int i = 1; i < components.length; i++) {
					bounds.merge_(((Jun3dBoundingBox) components[i].perform_("boundingBox")));
				}
				boundingBox = bounds;
			}
		} catch (Exception e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		}

		return boundingBox;
	}

	/**
	 * Set the receiver's components.
	 * 
	 * @param anArrayOfJunSurface java.util.List
	 * @category private
	 */
	protected void setComponents_(List<JunSurface> anArrayOfJunSurface) {
		components = anArrayOfJunSurface;
		boundingBox = null;
	}

	/**
	 * Set the receiver's components.
	 * 
	 * @param anArrayOfJunSurface jp.co.sra.jun.geometry.abstracts.JunSurface[]
	 * @category private
	 */
	protected void setComponents_(JunSurface[] anArrayOfJunSurface) {
		List<JunSurface> aList = new ArrayList<JunSurface>();
		for (int i = 0; i < anArrayOfJunSurface.length; i++) {
			aList.add(anArrayOfJunSurface[i]);
		}
		this.setComponents_(aList);
	}

}
