package jp.co.sra.jun.geometry.surfaces;

import java.io.IOException;
import java.io.Writer;

import jp.co.sra.smalltalk.StBlockClosure0;
import jp.co.sra.smalltalk.StBlockClosure1;

import jp.co.sra.jun.geometry.abstracts.JunSurface;
import jp.co.sra.jun.geometry.basic.Jun3dPoint;
import jp.co.sra.jun.geometry.curves.Jun3dLine;
import jp.co.sra.jun.goodies.lisp.JunLispCons;
import jp.co.sra.jun.goodies.lisp.JunLispConverter;
import jp.co.sra.jun.goodies.lisp.JunLispList;

/**
 * JunSphereSurface class
 * 
 *  @author    nisinaka
 *  @created   1998/11/13 (by nisinaka)
 *  @updated   2006/10/10 (by nisinaka)
 *  @updated   2009/03/19 (by Mitsuhiro Asada)
 *  @version   758 (with StPL9.7) based on Jun721 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2005 Information-technology Promotion Agency, Japan (IPA)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunSphereSurface.java,v 9.12 2010/03/09 09:16:56 nisinaka Exp $
 */
public class JunSphereSurface extends JunSurface {

	static {
		JunLispConverter.Register(JunSphereSurface.class);
	}

	/**
	 * Serial Version UID.
	 */
	private static final long serialVersionUID = 6976802850751423411L;

	/** The center point. */
	protected Jun3dPoint center;

	/** The radius. */
	protected double radius;

	/**
	 * Create a new instance of JunSphereSurface and initialize it.
	 * 
	 * @param aPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @param aNumber double
	 * @return jp.co.sra.jun.geometry.surfaces.JunSphereSurface
	 * @category Instance creation
	 */
	public JunSphereSurface(Jun3dPoint aPoint, double aNumber) {
		super();
		this.center_(aPoint);
		this.radius_(aNumber);
	}

	/**
	 * Create a new instance of <code>JunSphereSurface</code> and initialize it.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category Instance creation
	 */
	public JunSphereSurface(JunLispList aList) {
		super(aList);
	}

	/**
	 * Create a new instance of JunSphereSurface and initialize it.
	 * 
	 * @param aPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @param aNumber double
	 * @return jp.co.sra.jun.geometry.surfaces.JunSphereSurface
	 * @deprecated since Jun482a, use the constructor.
	 * @category Instance creation
	 */
	public static JunSphereSurface Center_radius_(Jun3dPoint aPoint, double aNumber) {
		return new JunSphereSurface(aPoint, aNumber);
	}

	/**
	 * Initialize the receiver.
	 * 
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#initialize()
	 * @category initialize-release
	 */
	@Override
	protected void initialize() {
		super.initialize();
		center = null;
		radius = 0.0d;
	}

	/**
	 * Answer the center point of the receiver.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category accessing
	 */
	public Jun3dPoint center() {
		return center;
	}

	/**
	 * Set the center point of the receiver.
	 * 
	 * @param aPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category accessing
	 */
	public void center_(Jun3dPoint aPoint) {
		center = aPoint;
	}

	/**
	 * Answer the radius of the receiver.
	 * 
	 * @return double
	 * @category accessing
	 */
	public double radius() {
		return radius;
	}

	/**
	 * Set the radius of the receiver.
	 * 
	 * @param aNumber double
	 * @category accessing
	 */
	public void radius_(double aNumber) {
		radius = aNumber;
	}

	/**
	 * Answer true if the receiver is equal to the object while concerning an accuracy.
	 * 
	 * @param anObject java.lang.Object
	 * @return boolean
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#equal_(java.lang.Object)
	 * @category comparing
	 */
	@Override
	public boolean equal_(Object anObject) {
		if (anObject == null || this.getClass() != anObject.getClass()) {
			return false;
		}

		JunSphereSurface aSphereSurface = (JunSphereSurface) anObject;
		return this.isEqualNumber_to_(this.radius(), aSphereSurface.radius()) && this.center().equal_(aSphereSurface.center());
	}

	/**
	 * Answer true if the Object is equal to the receiver, otherwise false.
	 * 
	 * @param anObject java.lang.Object
	 * @return boolean
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#equals(java.lang.Object)
	 * @category comparing
	 */
	@Override
	public boolean equals(Object anObject) {
		if (anObject == null || this.getClass() != anObject.getClass()) {
			return false;
		}

		JunSphereSurface aSphereSurface = (JunSphereSurface) anObject;
		return this.radius() == aSphereSurface.radius() && this.center().equals(aSphereSurface.center());
	}

	/**
	 * Answer the cross points on the receiver with the specified Line.
	 * 
	 * @param aLine jp.co.sra.jun.geometry.curves.Jun3dLine
	 * @return jp.co.sra.jun.geometry.basic.Jun3dPoint[]
	 * @category functions
	 */
	public Jun3dPoint[] crossPointsWithLine_(Jun3dLine aLine) {
		Jun3dPoint nearestPointOnLine = aLine.nearestPointFromPoint_(center);
		double distance = center.distance_(nearestPointOnLine);
		if (distance > radius) {
			return new Jun3dPoint[0];
		}
		if (this.isEqualNumber_to_(distance, radius)) {
			Jun3dPoint[] points = { nearestPointOnLine };
			return points;
		}
		double shift = Math.sqrt((radius * radius) - (distance * distance));
		Jun3dPoint shiftVector = aLine.to().minus_(aLine.from()).unitVector().multipliedBy_(shift);
		Jun3dPoint[] points = { nearestPointOnLine.plus_(shiftVector), nearestPointOnLine.minus_(shiftVector) };
		return points;
	}

	/**
	 * Get the receiver's attributes from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#fromLispList(jp.co.sra.jun.goodies.lisp.JunLispList)
	 * @category lisp support
	 */
	@Override
	public void fromLispList(JunLispList aList) {
		this.centerFromLispList(aList);
		this.radiusFromLispList(aList);
	}

	/**
	 * Convert the receiver's attributes to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#toLispList()
	 * @category lisp support
	 */
	@Override
	public JunLispCons toLispList() {
		JunLispCons aList = super.toLispList();
		aList.add_(this.centerToLispList());
		aList.add_(this.radiusToLispList());
		return aList;
	}

	/**
	 * Get the receiver's center value from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected void centerFromLispList(JunLispList aList) {
		JunLispList list = (JunLispList) aList.detect_ifNone_(new StBlockClosure1<Boolean, Object>() {
			public Boolean value(Object car) {
				return new Boolean(car instanceof JunLispCons && ((JunLispCons) car).head() == $("center"));
			}
		}, new StBlockClosure0<Object>());
		if (list == null) {
			return;
		}

		this.center_(new Jun3dPoint((JunLispList) list.tail()));
	}

	/**
	 * Convert the receiver's center value to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected JunLispCons centerToLispList() {
		JunLispCons aList = this.lispCons();
		aList.head_($("center"));
		aList.tail_(this.center().toLispList());
		return aList;
	}

	/**
	 * Get the receiver's radius value from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected void radiusFromLispList(JunLispList aList) {
		JunLispList list = (JunLispList) aList.detect_ifNone_(new StBlockClosure1<Boolean, Object>() {
			public Boolean value(Object car) {
				return new Boolean(car instanceof JunLispCons && ((JunLispCons) car).head() == $("radius"));
			}
		}, new StBlockClosure0<Object>());
		if (list == null) {
			return;
		}

		this.radius_(((Number) list.tail()).doubleValue());
	}

	/**
	 * Convert the receiver's radius value to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected JunLispCons radiusToLispList() {
		JunLispCons aList = this.lispCons();
		aList.head_($("radius"));
		aList.tail_(this.radius());
		return aList;
	}

	/**
	 * Print my storable string representation on the writer.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException
	 * @see jp.co.sra.smalltalk.StObject#storeOn_(java.io.Writer)
	 * @category printing
	 */
	public void storeOn_(Writer aWriter) throws IOException {
		aWriter.write('(');
		aWriter.write(this._className().toString());
		aWriter.write(" center: ");
		this.center().storeOn_(aWriter);
		aWriter.write(" radius: ");
		aWriter.write(Double.toString(this.radius()));
		aWriter.write(')');
	}

}
