package jp.co.sra.jun.geometry.surfaces;

import java.io.IOException;
import java.io.Writer;
import java.util.Arrays;

import jp.co.sra.jun.geometry.abstracts.JunSurface;
import jp.co.sra.jun.geometry.basic.JunAngle;
import jp.co.sra.jun.geometry.basic.JunPoint;
import jp.co.sra.jun.geometry.curves.JunLine;
import jp.co.sra.jun.goodies.lisp.JunLispCons;
import jp.co.sra.jun.goodies.lisp.JunLispList;

/**
 * JunTriangle class
 * 
 *  @author    Mitsuhiro Asada
 *  @created   2004/12/14 (by Mitsuhiro Asada)
 *  @updated   2009/03/18 (by Mitsuhiro Asada)
 *  @version   758 (with StPL9.7) based on Jun721 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2005 Information-technology Promotion Agency, Japan (IPA)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunTriangle.java,v 9.9 2010/03/09 09:16:57 nisinaka Exp $
 */
public abstract class JunTriangle extends JunSurface {
	/**
	 * Create a new instance of <code>JunTriangle</code> and initialize it.
	 * 
	 * @category Instance creation
	 */
	public JunTriangle() {
		super();
	}

	/**
	 * Create a new instance of <code>JunTriangle</code> and initialize it.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category Instance creation
	 */
	public JunTriangle(JunLispList aList) {
		super(aList);
	}

	/**
	 * Answer my p1 point.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.JunPoint
	 * @category accessing
	 */
	protected abstract JunPoint _p1();

	/**
	 * Answer my p2 point.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.JunPoint
	 * @category accessing
	 */
	protected abstract JunPoint _p2();

	/**
	 * Answer my p3 point.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.JunPoint
	 * @category accessing
	 */
	protected abstract JunPoint _p3();

	/**
	 * Convert the receiver as an array of JunPlane.
	 * 
	 * @return jp.co.sra.jun.geometry.surfaces.JunPlane[]
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#asArrayOfPlanes()
	 * @category converting
	 */
	@Override
	public JunPlane[] asArrayOfPlanes() {
		return new JunPlane[] { this.asPlane() };
	}

	/**
	 * Convert to a JunPlane.
	 * 
	 * @return jp.co.sra.jun.geometry.surfaces.JunPlane
	 * @category converting
	 */
	public abstract JunPlane asPlane();

	/**
	 * Answer the angle with the specified line.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.JunAngle
	 * @param aLine jp.co.sra.jun.geometry.curves.JunLine
	 * @category functions
	 */
	public JunAngle angleWithLine_(JunLine aLine) {
		return this.asPlane().angleWithLine_(aLine);
	}

	/**
	 * Answer the angle with the specified plane.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.JunAngle
	 * @param aLine jp.co.sra.jun.geometry.surfaces.JunPlane
	 * @category functions
	 */
	public JunAngle angleWithLine_(JunPlane aPlane) {
		return this.asPlane().angleWithPlane_(aPlane);
	}

	/**
	 * Answer the angle with the specified triangle.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.JunAngle
	 * @param aLine jp.co.sra.jun.geometry.surfaces.JunTriangle
	 * @category functions
	 */
	public JunAngle angleWithTriangle_(JunTriangle aTriangle) {
		return this.asPlane().angleWithTriangle_(aTriangle);
	}

	/**
	 * Answer the receiver's area with sign.
	 * 
	 * @return double
	 * @category functions
	 */
	public abstract double areaWithSign();

	/**
	 * Get the receiver's attributes from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#fromLispList(jp.co.sra.jun.goodies.lisp.JunLispList)
	 * @category lisp support
	 */
	@Override
	public void fromLispList(JunLispList aList) {
		this.p1FromLispList(aList);
		this.p2FromLispList(aList);
		this.p3FromLispList(aList);
	}

	/**
	 * Convert the receiver's attributes to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#toLispList()
	 * @category lisp support
	 */
	@Override
	public JunLispCons toLispList() {
		JunLispCons aList = super.toLispList();
		aList.add_(this.p1ToLispList());
		aList.add_(this.p2ToLispList());
		aList.add_(this.p3ToLispList());
		return aList;
	}

	/**
	 * Get the receiver's p1 value from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected abstract void p1FromLispList(JunLispList aList);

	/**
	 * Convert the receiver's p1 value to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected JunLispCons p1ToLispList() {
		JunLispCons aList = this.lispCons();
		aList.head_($("p1"));
		aList.tail_(this._p1().toLispList());
		return aList;
	}

	/**
	 * Get the receiver's p2 value from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected abstract void p2FromLispList(JunLispList aList);

	/**
	 * Convert the receiver's p2 value to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected JunLispCons p2ToLispList() {
		JunLispCons aList = this.lispCons();
		aList.head_($("p2"));
		aList.tail_(this._p2().toLispList());
		return aList;
	}

	/**
	 * Get the receiver's p3 value from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected abstract void p3FromLispList(JunLispList aList);

	/**
	 * Convert the receiver's p3 value to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected JunLispCons p3ToLispList() {
		JunLispCons aList = this.lispCons();
		aList.head_($("p3"));
		aList.tail_(this._p3().toLispList());
		return aList;
	}

	/**
	 * Print my string representation on the writer.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException if failed.
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#printOn_(java.io.Writer)
	 * @category printing
	 */
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write('(');
		this._p1().printOn_(aWriter);
		aWriter.write(" triangle: ");
		this._p2().printOn_(aWriter);
		aWriter.write(" and: ");
		this._p3().printOn_(aWriter);
		aWriter.write(')');
	}

	/**
	 * Print my storable string representation on the writer.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException
	 * @see jp.co.sra.smalltalk.StObject#storeOn_(java.io.Writer)
	 * @category printing
	 */
	public void storeOn_(Writer aWriter) throws IOException {
		aWriter.write('(');
		aWriter.write(this._className().toString());
		aWriter.write(" on: ");
		this._p1().storeOn_(aWriter);
		aWriter.write(" on: ");
		this._p2().storeOn_(aWriter);
		aWriter.write(" on: ");
		this._p3().storeOn_(aWriter);
		aWriter.write(')');
	}

	/**
	 * Answer <code>true</code> if the receiver has an area, otherwise <code>false</code>.
	 * 
	 * @return boolean
	 * @category testing
	 */
	public boolean hasArea() {
		return (this._p1().equal_(this._p2()) || this._p2().equal_(this._p3()) || this._p3().equal_(this._p1())) == false;
	}

	/**
	 * Answer <code>true</code> if the receiver is a triangle, otherwise <code>false</code>.
	 * 
	 * @return boolean
	 * @see jp.co.sra.jun.geometry.abstracts.JunGeometry#isTriangle()
	 * @category testing
	 */
	public boolean isTriangle() {
		return true;
	}

	/**
	 * Answer the value which side of on a plane.
	 * 
	 * @param bisector jp.co.sra.jun.geometry.surfaces.JunPlane
	 * @return int
	 * @category testing
	 */
	public int whichSideOf_(JunPlane bisector) {
		JunPoint[] points = new JunPoint[] { this._p1(), this._p2(), this._p3() };
		int[] anArray = new int[points.length];
		for (int i = 0; i < points.length; i++) {
			anArray[i] = points[i].whichSideOf_(bisector);
		}
		boolean hasPlus = Arrays.binarySearch(anArray, 1) != -1;
		boolean hasMinus = Arrays.binarySearch(anArray, -1) != -1;
		if (hasPlus && hasMinus) {
			return 0;
		}
		boolean hasZero = Arrays.binarySearch(anArray, 0) != -1;
		if (hasZero) {
			return 0;
		}
		if (hasMinus && hasPlus == false) {
			return -1;
		}
		return 1;
	}

	/**
	 * Set the p1 point.
	 * 
	 * @param aPpoint jp.co.sra.jun.geometry.basic.JunPoint
	 * @category private
	 */
	protected abstract void setP1_(JunPoint aPpoint);

	/**
	 * Set the p2 point.
	 * 
	 * @param aPpoint jp.co.sra.jun.geometry.basic.JunPoint
	 * @category private
	 */
	protected abstract void setP2_(JunPoint aPpoint);

	/**
	 * Set the p3 point.
	 * 
	 * @param aPpoint jp.co.sra.jun.geometry.basic.JunPoint
	 * @category private
	 */
	protected abstract void setP3_(JunPoint aPpoint);
}
