package jp.co.sra.jun.geometry.transformations;

import java.io.IOException;
import java.io.Writer;

import jp.co.sra.smalltalk.StBlockClosure0;
import jp.co.sra.smalltalk.StBlockClosure1;

import jp.co.sra.jun.collections.sequences.JunDoubleMatrix;
import jp.co.sra.jun.geometry.abstracts.JunGeometry;
import jp.co.sra.jun.goodies.lisp.JunLispCons;
import jp.co.sra.jun.goodies.lisp.JunLispList;

/**
 * JunTransformation class
 * 
 *  @author    nisinaka
 *  @created   1998/09/28 (by nisinaka)
 *  @updated   2000/01/06 (by nisinaka)
 *  @updated   2009/03/24 (by m-asada)
 *  @version   758 (with StPL9.7) based on Jun721 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2005 Information-technology Promotion Agency, Japan (IPA)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunTransformation.java,v 9.9 2010/03/09 09:16:57 nisinaka Exp $
 */
public abstract class JunTransformation extends JunGeometry {
	/**
	 * Create a new instance of <code>JunTransformation</code> and initialize it.
	 * 
	 * @category Instance creation
	 */
	protected JunTransformation() {
		super();
	}

	/**
	 * Create a new instance of <code>JunTransformation</code> and initialize it.
	 * 
	 * @param anArray double[]
	 * @category Instance creation
	 */
	public JunTransformation(double[] anArray) {
		super();
		this.setFromArray_(anArray);
	}

	/**
	 * Create a new instance of <code>JunTransformation</code> and initialize it.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category Instance creation
	 */
	public JunTransformation(JunLispList aList) {
		super(aList);
	}

	/**
	 * Print my storable string representation on the writer.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException
	 * @see jp.co.sra.smalltalk.StObject#storeOn_(java.io.Writer)
	 * @category printing
	 */
	public void storeOn_(Writer aWriter) throws IOException {
		aWriter.write('(');
		aWriter.write(this._className().toString());
		aWriter.write(" fromArray: ");
		aWriter.write("#(");
		double[] array = this.asArray();
		if (array.length > 0) {
			aWriter.write(String.valueOf(array[0]));
		}
		for (int i = 1; i < array.length; i++) {
			aWriter.write(' ');
			aWriter.write(String.valueOf(array[i]));
		}
		aWriter.write(')');
		aWriter.write(')');
	}

	/**
	 * Convert this transformation as an array.
	 * 
	 * @return double[]
	 * @category converting
	 */
	public abstract double[] asArray();

	/**
	 * Convert this transformation as a matrix.
	 * 
	 * @return jp.co.sra.jun.collections.sequences.JunDoubleMatrix
	 * @category converting
	 */
	public abstract JunDoubleMatrix asMatrix();

	/**
	 * Answer the inverse of this transformation.
	 * 
	 * @return jp.co.sra.jun.geometry.transformations.JunTransformation
	 * @category transforming functions
	 */
	public abstract JunTransformation _inverse();
	
	/**
	 * Get the receiver's attributes from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#fromLispList(jp.co.sra.jun.goodies.lisp.JunLispList)
	 * @category lisp support
	 */
	@Override
	public void fromLispList(JunLispList aList) {
		super.fromLispList(aList);
		this.elementsFromLispList(aList);
	}

	/**
	 * Convert the receiver's attributes to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#toLispList()
	 * @category lisp support
	 */
	@Override
	public JunLispCons toLispList() {
		JunLispCons list = super.toLispList();
		list.add_(this.elementsToLispList());
		return list;
	}

	/**
	 * Get the receiver's elements from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected void elementsFromLispList(JunLispList aList) {
		JunLispList list = (JunLispList) aList.detect_ifNone_(new StBlockClosure1<Boolean, Object>() {
			public Boolean value(Object car) {
				return new Boolean(car instanceof JunLispCons && ((JunLispCons) car).head() == $("elements"));
			}
		}, new StBlockClosure0<Object>());
		if (list == null) {
			return;
		}

		Object[] values = ((JunLispList) list.tail()).asArray();
		double[] asArray = new double[values.length];
		for (int i = 0; i < values.length; i++) {
			asArray[i] = ((Number) values[i]).doubleValue();
		}
		this.setFromArray_(asArray);
	}

	/**
	 * Convert the receiver's elements to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected JunLispCons elementsToLispList() {
		JunLispCons list = this.lispCons();
		list.head_($("elements"));
		double[] asArray = this.asArray();
		Double[] values = new Double[asArray.length];
		for (int i = 0; i < values.length; i++) {
			values[i] = new Double(asArray[i]);
		}
		list.tail_(JunLispList.List_(values));
		return list;
	}
	
	/**
	 * Set the receiver's values.
	 * 
	 * @param anArray double[]
	 * @category private
	 */
	protected abstract void setFromArray_(double[] anArray);
}
