package jp.co.sra.jun.goodies.gauge;

import jp.co.sra.smalltalk.*;
import jp.co.sra.jun.system.framework.JunApplicationModel;
import jp.co.sra.jun.system.support.JunSystem;
import jp.co.sra.jun.goodies.track.JunSimpleSliderModel;

/**
 * JunSimpleGaugeModel class
 * 
 *  @author    Hoshi Takanori
 *  @created   2003/08/28 (by Hoshi Takanori)
 *  @updated   2008/05/11 (by Mitsuhiro Asada)
 *  @version   758 (with StPL9.7) based on Jun433 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2005 Information-technology Promotion Agency, Japan (IPA)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunSimpleGaugeModel.java,v 9.11 2010/03/09 09:17:10 nisinaka Exp $
 */
public class JunSimpleGaugeModel extends JunApplicationModel implements StBlockValued<Double> {

	protected static final double DefaultFromValue = 0;
	protected static final double DefaultToValue = 1;
	protected static final double DefaultRoundValue = 0.01;

	protected JunSimpleSliderModel simpleSlider;
	protected double fromValue = DefaultFromValue;
	protected double toValue = DefaultToValue;
	protected double roundValue = DefaultRoundValue;
	protected StBlockClosure1<String, Double> valueStringBlock;

	/**
	 * Create a new instance of <code>JunSimpleGaugeModel</code> and initialize it.
	 * 
	 * @param aNumber double
	 * @category Instance creation
	 */
	public JunSimpleGaugeModel(double aNumber) {
		super();
		this.value_(aNumber);
	}

	/**
	 * Create a new instance of <code>JunSimpleGaugeModel</code> and initialize it.
	 * 
	 * @param aNumber double
	 * @param from double
	 * @param to double
	 * @param round double
	 * @category Instance creation
	 */
	public JunSimpleGaugeModel(double aNumber, double from, double to, double round) {
		super();
		fromValue = from;
		toValue = to;
		roundValue = round;
		this.value_(aNumber);
	}

	/**
	 * Initialize the JunSimpleGaugeModel when created.
	 * 
	 * @see jp.co.sra.jun.system.framework.JunApplicationModel#initialize()
	 * @category initialize-release
	 */
	protected void initialize() {
		super.initialize();

		simpleSlider = null;
		fromValue = DefaultFromValue;
		toValue = DefaultToValue;
		roundValue = DefaultRoundValue;
		valueStringBlock = null;
	}

	/**
	 * Answer the simple slider.
	 *
	 * @return jp.co.sra.jun.goodies.track.JunSimpleSliderModel
	 * @category accessing
	 */
	public JunSimpleSliderModel simpleSlider() {
		if (simpleSlider == null) {
			final JunSimpleGaugeModel this_ = this;
			simpleSlider = new JunSimpleSliderModel();
			simpleSlider.compute_(new StBlockClosure1<Double, Double>() {
				public Double value(Double newValue) {
					this_.changed_($("value"));
					return null;
				}
			});
		}
		return simpleSlider;
	}

	/**
	 * Answer the from value.
	 *
	 * @return double
	 * @category accessing
	 */
	public double fromValue() {
		return fromValue;
	}

	/**
	 * Set the from value to aNumber.
	 *
	 * @param aNumber double
	 * @category accessing
	 */
	public void fromValue_(double aNumber) {
		if (fromValue != aNumber) {
			fromValue = aNumber;
			this.changed_($("value"));
		}
	}

	/**
	 * Answer the to value.
	 *
	 * @return double
	 * @category accessing
	 */
	public double toValue() {
		return toValue;
	}

	/**
	 * Set the to value to aNumber.
	 *
	 * @param aNumber double
	 * @category accessing
	 */
	public void toValue_(double aNumber) {
		if (toValue != aNumber) {
			toValue = aNumber;
			this.changed_($("value"));
		}
	}

	/**
	 * Answer the round value.
	 *
	 * @return double
	 * @category accessing
	 */
	public double roundValue() {
		return roundValue;
	}

	/**
	 * Set the round value to aNumber.
	 *
	 * @param aNumber double
	 * @category accessing
	 */
	public void roundValue_(double aNumber) {
		if (roundValue != aNumber) {
			roundValue = aNumber;
			this.changed_($("value"));
		}
	}

	/**
	 * Answer the step value.
	 *
	 * @return double
	 * @category accessing
	 */
	public double stepValue() {
		return (fromValue <= toValue) ? roundValue : -roundValue;
	}

	/**
	 * Answer the value as double.
	 *
	 * @return double
	 * @category accessing
	 */
	public double doubleValue() {
		double theValue = this.simpleSlider().doubleValue() * (toValue - fromValue);
		return Math.round(theValue / roundValue) * roundValue + fromValue;
	}

	/**
	 * Answer the value.
	 *
	 * @return java.lang.Double
	 * @see jp.co.sra.smalltalk.StValued#value()
	 * @category accessing
	 */
	public Double value() {
		return new Double(this.doubleValue());
	}

	/**
	 * Set the value to aNumber.
	 *
	 * @param aNumber double
	 * @category accessing
	 */
	public void value_(double aNumber) {
		double theValue = (aNumber - fromValue) / (toValue - fromValue);
		theValue = Math.max(0, Math.min(theValue, 1));
		if (theValue != this.simpleSlider().doubleValue()) {
			this.simpleSlider().value_(theValue);
		}
	}

	/**
	 * Answer the value string.
	 *
	 * @return java.lang.String
	 * @category accessing
	 */
	public String valueString() {
		return this.valueStringBlock().value(this.value());
	}

	/**
	 * Answer the value string block.
	 *
	 * @return jp.co.sra.smalltalk.StBlockClosure1
	 * @category accessing
	 */
	public StBlockClosure1<String, Double> valueStringBlock() {
		if (valueStringBlock == null) {
			valueStringBlock = this.defaultValueStringBlock();
		}
		return valueStringBlock;
	}

	/**
	 * Set the value string block to aBlock.
	 *
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @deprecated since Jun701, use Generics.
	 * @category accessing
	 */
	public void valueStringBlock_(final StBlockClosure<?, ?, ?, ?> aBlock) {
		this.valueStringBlock_(new StBlockClosure1<String, Double>() {
			public String value(Double aNumber) {
				return (String) aBlock.value_(aNumber);
			}
		});
	}

	/**
	 * Set the value string block to aBlock.
	 *
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure1
	 * @category accessing
	 */
	public void valueStringBlock_(StBlockClosure1<String, Double> aBlock) {
		valueStringBlock = aBlock;
	}

	/**
	 * Answer a StBlockValue that computes aBlock with the receiver's value as the argument.
	 *
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return jp.co.sra.smalltalk.StBlockValue
	 * @deprecated since Jun701, use Generics.
	 * @category constructing
	 */
	public StBlockValue<Object> compute_(final StBlockClosure<?, ?, ?, ?> aBlock) {
		return this.compute_(new StBlockClosure1<Object, Double>() {
			public Object value(Double aNumber) {
				return aBlock.value_(aNumber);
			}
		});
	}

	/**
	 * Answer a StBlockValue that computes aBlock with the receiver's value as the argument.
	 *
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure1
	 * @return jp.co.sra.smalltalk.StBlockValue
	 * @category constructing
	 */
	public <R> StBlockValue<R> compute_(StBlockClosure1<R, Double> aBlock) {
		return new StBlockValue<R>(aBlock, this);
	}

	/**
	 * Answer a default value string block.
	 *
	 * @return jp.co.sra.smalltalk.StBlockClosure1
	 * @category defaults
	 */
	protected StBlockClosure1<String, Double> defaultValueStringBlock() {
		return new StBlockClosure1<String, Double>() {
			public String value(Double newValue) {
				return formatValueString(newValue);
			}
		};
	}

	/**
	 * Answer a formatted value string.
	 *
	 * @param aNumber double
	 * @return java.lang.String
	 * @category private
	 */
	protected String formatValueString(double aNumber) {
		if (Math.floor(roundValue) == roundValue) {
			return Long.toString(Math.round(aNumber));
		} else {
			if (aNumber >= 0.0d) {
				return Double.toString(aNumber + roundValue / 10).substring(0, Double.toString(roundValue).length());
			} else {
				return Double.toString(aNumber - roundValue / 10).substring(0, Double.toString(roundValue).length() + 1);
			}
		}
	}

	/**
	 * Answer a default view.
	 * 
	 * @return jp.co.sra.smalltalk.StView
	 * @see jp.co.sra.smalltalk.StApplicationModel#defaultView()
	 * @category defaults
	 */
	public StView defaultView() {
		if (GetDefaultViewMode() == VIEW_AWT) {
			return new JunSimpleGaugeViewAwt(this);
		} else {
			return new JunSimpleGaugeViewSwing(this);
		}
	}

	/**
	 * Answer a window title.
	 * 
	 * @return java.lang.String
	 * @see jp.co.sra.smalltalk.StApplicationModel#windowTitle()
	 * @category interface opening
	 */
	protected String windowTitle() {
		return JunSystem.$String("Simple Gauge");
	}

}
