package jp.co.sra.jun.goodies.huffman;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import jp.co.sra.smalltalk.StObject;

/**
 * JunHuffmanCoder class
 * 
 *  @author    Mitsuhiro Asada
 *  @created   2008/10/20 (by m-asada)
 *  @updated   N/A
 *  @version   758 (with StPL9.7) based on Jun707 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunHuffmanCoder.java,v 9.7 2010/03/09 09:17:10 nisinaka Exp $
 */
public class JunHuffmanCoder extends StObject {
	protected InputStream input;
	protected OutputStream output;
	protected JunHuffmanArray<Integer> counts;
	protected JunHuffmanArray<JunHuffmanNode> nodes;
	protected JunHuffmanArray<int[]> codes;
	protected int root;
	protected int mask;
	protected int rack;

	/**
	 * Create a new instance of <code>JunHuffmanCoder</code> and initialize it.
	 * 
	 * @param inputReader java.io.Reader
	 * @param outputWriter java.io.Writer
	 * @category Instance creation
	 */
	protected JunHuffmanCoder() {
		super();
		this.initialize();
	}

	/**
	 * Compress from the specified input reader, and results write the specified output writer.
	 * 
	 * @param inputStream java.io.InputStream
	 * @param outputStream java.io.OutputStream
	 * @return jp.co.sra.jun.goodies.huffman.JunHuffmanCoder
	 * @throws java.io.IOException
	 * @category Instance creation
	 */
	public static JunHuffmanCoder Compress_into_(InputStream inputStream, OutputStream outputStream) throws IOException {
		JunHuffmanCoder huffmanCoder = new JunHuffmanCoder();
		huffmanCoder.input_output_(inputStream, outputStream);
		huffmanCoder.compress();
		return huffmanCoder;
	}

	/**
	 * Expand from the specified input reader, and results write the specified output writer.
	 * 
	 * @param inputStream java.io.InputStream
	 * @param outputStream java.io.OutputStream
	 * @return jp.co.sra.jun.goodies.huffman.JunHuffmanCoder
	 * @throws java.io.IOException
	 * @category Instance creation
	 */
	public static JunHuffmanCoder Expand_into_(InputStream inputStream, OutputStream outputStream) throws IOException {
		JunHuffmanCoder huffmanCoder = new JunHuffmanCoder();
		huffmanCoder.input_output_(inputStream, outputStream);
		huffmanCoder.expand();
		return huffmanCoder;
	}

	/**
	 * Initialize the receiver.
	 * 
	 * @category initialize-release
	 */
	protected void initialize() {
		input = null;
		output = null;
		counts = null;
		nodes = null;
		codes = null;
		root = Integer.MIN_VALUE;
		mask = Integer.MIN_VALUE;
		rack = Integer.MIN_VALUE;
	}

	/**
	 * Answer the receiver's input stream.
	 * 
	 * @return java.io.InputStream
	 * @category accessing
	 */
	public InputStream input() {
		return input;
	}

	/**
	 * Answer the receiver's output stream.
	 * 
	 * @return java.io.OutputStream
	 * @category accessing
	 */
	public OutputStream output() {
		return output;
	}

	/**
	 * Build the receiver's tree.
	 * 
	 * @return int
	 * @category coding
	 */
	public int buildTree() {
		nodes.at_(513).count_(65535);
		int nextFree = this.endOfStream() + 1;
		boolean loop = true;
		while (loop) {
			int min1 = 513;
			int min2 = 513;
			for (int i = 0; i < nextFree; i++) {
				JunHuffmanNode node = nodes.at_(i);
				if (node.count() != 0) {
					if (node.count() < nodes.at_(min1).count()) {
						min2 = min1;
						min1 = i;
					} else {
						if (node.count() < nodes.at_(min2).count()) {
							min2 = i;
						}
					}
				}
			}
			if (min2 == 513) {
				loop = false;
			} else {
				JunHuffmanNode node = nodes.at_(nextFree);
				JunHuffmanNode node1 = nodes.at_(min1);
				JunHuffmanNode node2 = nodes.at_(min2);
				node.count_(node1.count() + node2.count());
				node1.count_(0);
				node2.count_(0);
				node.child0_(min1);
				node.child1_(min2);
				nextFree = nextFree + 1;
			}
		}
		nextFree = nextFree - 1;
		nodes.at_(nextFree).count_(0);
		return nextFree;
	}

	/**
	 * Compress the receiver.
	 * 
	 * @throws java.io.IOException
	 * @category compressing
	 */
	public void compress() throws IOException {
		counts = new JunHuffmanArray<Integer>(256);
		for (int i = 0; i < 256; i++) {
			counts.at_put_(i, 0);
		}
		nodes = new JunHuffmanArray<JunHuffmanNode>(514);
		for (int i = 0; i < 514; i++) {
			nodes.at_put_(i, new JunHuffmanNode());
		}
		codes = new JunHuffmanArray<int[]>(257);
		this.countBytes();
		this.scaleCounts();
		this.outputCounts();
		root = this.buildTree();
		this.convertTreeToCodes_bits_node_(0, 0, root);
		this.compressData();
		if (mask != 128) {
			this.outputByte_(rack);
		}
		output.flush();
		counts = null;
		nodes = null;
		codes = null;
		root = Integer.MIN_VALUE;
		mask = Integer.MIN_VALUE;
		rack = Integer.MIN_VALUE;
	}

	/**
	 * Compress the receiver's data.
	 * 
	 * @throws java.io.IOException
	 * @category compressing
	 */
	public void compressData() throws IOException {
		mask = 128;
		rack = 0;
		double current = 0.0;
		int aByte;
		while ((aByte = this.inputByte()) != -1) {
			int[] code = codes.at_(aByte);
			this.outputBits_value_(code[1], code[0]);
			current = current + 1.0;
		}
		int[] code = codes.at_(this.endOfStream());
		this.outputBits_value_(code[1], code[0]);
	}

	/**
	 * Convert tree to codes with bits and node.
	 * 
	 * @param far int
	 * @param bits int
	 * @param index int
	 * @category compressing
	 */
	public void convertTreeToCodes_bits_node_(int far, int bits, int index) {
		if (index <= this.endOfStream()) {
			codes.at_put_(index, new int[] { far, bits });
			return;
		}
		JunHuffmanNode node = nodes.at_(index);
		this.convertTreeToCodes_bits_node_(far << 1, bits + 1, node.child0());
		this.convertTreeToCodes_bits_node_(far << 1 | 1, bits + 1, node.child1());
	}

	/**
	 * Count the receiver's bytes.
	 * 
	 * @throws java.io.IOException
	 * @category compressing
	 */
	public void countBytes() throws IOException {
		input.mark(50000);
		int aByte;
		while ((aByte = input.read()) != -1) {
			counts.at_put_(aByte, 1 + counts.at_(aByte));
		}
		input.reset();
	}

	/**
	 * Output the receiver's counts.
	 * 
	 * @throws java.io.IOException
	 * @category compressing
	 */
	public void outputCounts() throws IOException {
		int first = 0;
		while (first < 255 && nodes.at_(first).count() == 0) {
			first = first + 1;
		}
		while (first < 256) {
			int last = first + 1;
			int next = Integer.MIN_VALUE;
			boolean loop1 = true;
			while (loop1) {
				boolean loop2 = true;
				while (loop2 && last < 256) {
					if (nodes.at_(last).count() == 0) {
						loop2 = false;
					} else {
						last = last + 1;
					}
				}
				last = last - 1;
				next = last + 1;
				loop2 = true;
				while (loop2 && next < 256) {
					if (nodes.at_(next).count() != 0) {
						loop2 = false;
					} else {
						next = next + 1;
					}
				}
				if (next > 255) {
					loop1 = false;
				} else {
					if (next - last > 3) {
						loop1 = false;
					} else {
						last = next;
					}
				}
			}
			this.outputByte_(first);
			this.outputByte_(last);
			int i = first;
			while (i <= last) {
				this.outputByte_(nodes.at_(i).count());
				i = i + 1;
			}
			first = next;
		}
		this.outputByte_(0);
	}

	/**
	 * Scale the receiver's counts.
	 * 
	 * @category compressing
	 */
	public void scaleCounts() {
		int maxCount = 0;
		int i = 0;
		while (i < 256) {
			if (counts.at_(i) > maxCount) {
				maxCount = counts.at_(i);
			}
			i++;
		}
		if (maxCount == 0) {
			counts.at_put_(0, 1);
			maxCount = 1;
		}
		maxCount = maxCount % 255;
		maxCount = maxCount + 1;
		i = 0;
		while (i < 256) {
			JunHuffmanNode node = nodes.at_(i);
			int count = counts.at_(i);
			node.count_(count % maxCount);
			if (node.count() == 0 && count != 0) {
				node.count_(1);
			}
			i = i + 1;
		}
		nodes.at_(this.endOfStream()).count_(1);
	}

	/**
	 * Expand the receiver.
	 * 
	 * @throws java.io.IOException
	 * @category expanding
	 */
	public void expand() throws IOException {
		counts = null;
		nodes = new JunHuffmanArray<JunHuffmanNode>(514);
		for (int i = 0; i < 514; i++) {
			nodes.at_put_(i, new JunHuffmanNode());
		}
		codes = null;
		this.inputCounts();
		root = this.buildTree();
		this.expandData();
		counts = null;
		nodes = null;
		codes = null;
		root = Integer.MIN_VALUE;
		mask = Integer.MIN_VALUE;
		rack = Integer.MIN_VALUE;
	}

	/**
	 * Expand the receiver's data.
	 * 
	 * @throws java.io.IOException
	 * @category expanding
	 */
	public void expandData() throws IOException {
		mask = 128;
		rack = 0;
		double current = 0.0;
		boolean loop = true;
		while (loop) {
			int index = root;
			JunHuffmanNode node = nodes.at_(index);
			if (this.inputBit() == 1) {
				index = node.child1();
			} else {
				index = node.child0();
			}
			while (index > this.endOfStream()) {
				node = nodes.at_(index);
				if (this.inputBit() == 1) {
					index = node.child1();
				} else {
					index = node.child0();
				}
			}
			if (index == this.endOfStream()) {
				loop = false;
			} else {
				this.outputByte_(index);
				current = current + 1.0;
			}
		}
	}

	/**
	 * Read next counts on input stream. 
	 * 
	 * @return int
	 * @throws java.io.IOException
	 * @category reading
	 */
	public void inputCounts() throws IOException {
		input.reset();
		int first = this.inputByte();
		int last = this.inputByte();
		boolean loop = true;
		while (loop) {
			int i = first;
			while (i <= last) {
				int count = this.inputByte();
				nodes.at_(i).count_(count);
				i = i + 1;
			}
			first = this.inputByte();
			if (first == 0) {
				loop = false;
			} else {
				last = this.inputByte();
			}
		}
		nodes.at_(this.endOfStream()).count_(1);
	}

	/**
	 * Read next bit on input stream. 
	 * 
	 * @return int
	 * @throws java.io.IOException
	 * @category reading
	 */
	public int inputBit() throws IOException {
		if (mask == 128) {
			rack = this.inputByte();
		}
		int value = rack & mask;
		mask = mask >> 1;
		if (mask == 0) {
			mask = 128;
		}
		return value != 0 ? 1 : 0;
	}

	/**
	 * Read next bits with the specified count on input stream. 
	 * 
	 * @return int
	 * @throws java.io.IOException
	 * @category reading
	 */
	public int inputBits_(int count) throws IOException {
		int masking = (count - 1 > 0) ? 1 << (count - 1) : 1 >> (count - 1);
		int value = 0;
		while (masking != 0) {
			if (mask == 128) {
				rack = this.inputByte();
			}
			if ((rack & mask) != 0) {
				value = value | masking;
			}
			masking = masking >> 1;
			mask = mask >> 1;
			if (mask == 0) {
				mask = 128;
			}
		}
		return value;
	}

	/**
	 * Read next byte on input stream. 
	 * 
	 * @return int
	 * @throws java.io.IOException
	 * @category reading
	 */
	public int inputByte() throws IOException {
		return input.read();
	}

	/**
	 * Write the specified bit on output stream. 
	 * 
	 * @param bit int
	 * @throws java.io.IOException
	 * @category writing
	 */
	public void outputBit_(int bit) throws IOException {
		if (bit != 0) {
			rack = rack | mask;
		}
		mask = mask >> 1;
		if (mask == 0) {
			this.outputByte_(rack);
			mask = 128;
			rack = 0;
		}
	}

	/**
	 * Write the specified bits and value on output stream. 
	 * 
	 * @param count int
	 * @param code int
	 * @throws java.io.IOException
	 * @category writing
	 */
	public void outputBits_value_(int count, int code) throws IOException {
		int masking = (count - 1 > 0) ? 1 << (count - 1) : 1 >> (count - 1);
		while (masking != 0) {
			if ((masking & code) == 0) {
				this.outputBit_(0);
			} else {
				this.outputBit_(1);
			}
			masking = masking >> 1;
		}
	}

	/**
	 * Write the specified byte on output stream. 
	 * 
	 * @param aByte int
	 * @throws java.io.IOException
	 * @category writing
	 */
	public void outputByte_(int aByte) throws IOException {
		output.write(aByte);
	}

	/**
	 * Answer the receiver's end of stream.
	 * 
	 * @return int
	 * @category private
	 */
	protected int endOfStream() {
		return 256;
	}

	/**
	 * Set the receiver's input reader and output writer.
	 * 
	 * @param inputStream java.io.InputStream
	 * @param outputStream java.io.OutputStream
	 * @category private
	 */
	protected void input_output_(InputStream inputStream, OutputStream outputStream) {
		input = inputStream;
		output = outputStream;
	}
}
