package jp.co.sra.jun.goodies.morph;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.image.BufferedImage;
import java.lang.reflect.InvocationTargetException;

import jp.co.sra.smalltalk.StController;
import jp.co.sra.smalltalk.StRectangle;
import jp.co.sra.smalltalk.StSymbol;

import jp.co.sra.jun.geometry.curves.Jun2dLine;
import jp.co.sra.jun.system.framework.JunAbstractViewCanvas;

/**
 * JunMorphViewAwt class
 * 
 *  @author    Mitsuhiro Asada
 *  @created   2008/11/27 (by m-asada)
 *  @updated   N/A
 *  @version   758 (with StPL9.7) based on Jun714 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunMorphViewAwt.java,v 9.7 2010/03/09 09:17:17 nisinaka Exp $
 */
public class JunMorphViewAwt extends JunAbstractViewCanvas implements JunMorphView {
	protected StSymbol aspectMessage;
	protected BufferedImage aspectImage;
	protected BufferedImage cachedImage;

	/**
	 * Create a new instance of <code>JunMorphViewAwt</code> and initialize it.
	 * 
	 * @category Instance creation
	 */
	public JunMorphViewAwt() {
		super();
	}

	/**
	 * Create a new instance of <code>JunMorphViewAwt</code> and initialize it.
	 * 
	 * @param newModel jp.co.sra.jun.goodies.morph.JunMorphModel
	 * @category Instance creation
	 */
	public JunMorphViewAwt(JunMorphModel newModel) {
		super(newModel);
	}

	/**
	 * Initialize the receiver.
	 * 
	 * @see jp.co.sra.smalltalk.StViewJPanel#initialize()
	 * @category initialize-release
	 */
	@Override
	protected void initialize() {
		super.initialize();
		aspectMessage = null;
		aspectImage = null;
		cachedImage = null;
	}

	/**
	 * Build this component.
	 * 
	 * @see jp.co.sra.smalltalk.StViewJPanel#buildComponent()
	 * @category initialize-release
	 */
	public void buildComponent() {
		this.setBackground(Color.white);
	}

	/**
	 * Called by the garbage collector on an object when garbage collection determines
	 * that there are no more references to the object.
	 * 
	 * @see java.lang.Object#finalize()
	 * @throws java.lang.Throwable
	 * @category initialize-release
	 */
	@Override
	protected void finalize() throws Throwable {
		super.finalize();
		this.flushImage();
	}

	/**
	 * Answer the receiver's aspect image.
	 * 
	 * @return java.awt.image.BufferedImage
	 * @category accessing
	 */
	public BufferedImage aspectImage() {
		if (this.aspectMessage() == null) {
			this.flushImage();
		} else {
			String aMessage = this.aspectMessage().toString();
			if (this.model().respondsTo_(aMessage)) {
				BufferedImage anImage = null;
				try {
					anImage = (BufferedImage) this.model().perform_(aMessage);
				} catch (InvocationTargetException e) {
					e.printStackTrace();
				} catch (IllegalAccessException e) {
					e.printStackTrace();
				} catch (NoSuchMethodException e) {
					e.printStackTrace();
				}
				if (anImage == null) {
					this.flushImage();
				} else {
					if (anImage != aspectImage) {
						aspectImage = anImage;
						cachedImage = new BufferedImage(aspectImage.getWidth(), aspectImage.getHeight(), BufferedImage.TYPE_INT_ARGB);
						Graphics aGraphics = cachedImage.getGraphics();
						try {
							aGraphics.drawImage(aspectImage, 0, 0, this);
						} finally {
							if (aGraphics != null) {
								aGraphics.dispose();
							}
							aGraphics = null;
						}
					}
				}
			} else {
				this.flushImage();
			}
		}
		return cachedImage;
	}

	/**
	 * Answer the receiver's aspect lines.
	 * 
	 * @return jp.co.sra.jun.geometry.curves.Jun2dLine[]
	 * @see jp.co.sra.jun.goodies.morph.JunMorphView#aspectLines()
	 * @category accessing
	 */
	public Jun2dLine[] aspectLines() {
		if (this.isFirstImage()) {
			return this.getModel().fromLines();
		}
		if (this.isSecondImage()) {
			return this.getModel().toLines();
		}
		return null;
	}

	/**
	 * Answer the receiver's aspect message.
	 * 
	 * @return jp.co.sra.smalltalk.StSymbol
	 * @category accessing
	 */
	public StSymbol aspectMessage() {
		return aspectMessage;
	}

	/**
	 * Set the receiver's aspect message.
	 * 
	 * @param messageSelector jp.co.sra.smalltalk.StSymbol
	 * @category accessing
	 */
	public void aspectMessage_(StSymbol messageSelector) {
		aspectMessage = messageSelector;
	}

	/**
	 * Answer a default controller.
	 * 
	 * @return jp.co.sra.smalltalk.StController
	 * @see jp.co.sra.smalltalk.StViewJPanel#defaultController()
	 * @category controller accessing
	 */
	@Override
	protected StController defaultController() {
		return new JunMorphController();
	}

	/**
	 * Display the receiver on the graphics.
	 * 
	 * @param aGraphics java.awt.Graphics
	 * @see jp.co.sra.smalltalk.StDisplayable#displayOn_(java.awt.Graphics)
	 * @category displaying
	 */
	@Override
	public void displayOn_(Graphics aGraphics) {
		if (this.aspectImage() == null) {
			aGraphics.setColor(this.getBackground());
			aGraphics.fillRect(0, 0, this.getWidth(), this.getHeight());
		} else {
			BufferedImage anImage = this.aspectImage();
			aGraphics.drawImage(anImage, 0, 0, this);
			StRectangle[] areasOutside = new StRectangle(0, 0, this.getWidth(), this.getHeight()).areasOutside_(new StRectangle(0, 0, anImage.getWidth(), anImage.getHeight()));
			for (int i = 0; i < areasOutside.length; i++) {
				aGraphics.setColor(this.getBackground());
				aGraphics.fillRect(areasOutside[i].x(), areasOutside[i].y(), areasOutside[i].width(), areasOutside[i].height());
			}
		}

		Jun2dLine[] lines = this.aspectLines();
		if (lines == null) {
			return;
		}
		for (int i = 0; i < lines.length; i++) {
			Jun2dLine aLine = lines[i];
			this.displayOn_startPoint_endPoint_((Graphics2D) aGraphics, aLine.first().asPoint(), aLine.last().asPoint());
		}
	}

	/**
	 * Display the start point and end point on the graphics.
	 * 
	 * @param aGraphics java.awt.Graphics2D
	 * @param startPoint java.awt.Point
	 * @param endPoint java.awt.Point
	 * @see jp.co.sra.jun.goodies.morph.JunMorphView#displayOn_startPoint_endPoint_(java.awt.Graphics2D, java.awt.Point, java.awt.Point)
	 * @category displaying
	 */
	public void displayOn_startPoint_endPoint_(Graphics2D aGraphics, Point startPoint, Point endPoint) {
		if (startPoint.equals(endPoint) == false) {
			aGraphics.setStroke(new BasicStroke(1));
			aGraphics.setColor(Color.green);
			aGraphics.drawLine(startPoint.x, startPoint.y, endPoint.x, endPoint.y);
			aGraphics.setColor(Color.blue);
			Rectangle aBox = new Rectangle(endPoint, new Dimension(1, 1));
			aBox.grow(1, 1);
			aGraphics.fillRect(aBox.x, aBox.y, aBox.width, aBox.height);
		}
		Rectangle aBox = new Rectangle(startPoint, new Dimension(1, 1));
		aBox.grow(1, 1);
		aGraphics.setColor(Color.red);
		aGraphics.fillRect(aBox.x, aBox.y, aBox.width, aBox.height);
	}

	/**
	 * Flush the receiver's image.
	 * 
	 * @category flushing
	 */
	protected void flushImage() {
		aspectImage = null;
		cachedImage = null;
	}

	/**
	 * Answer the receiver's model as <code>JunMorphModel</code>.
	 * 
	 * @return jp.co.sra.jun.goodies.morph.JunMorphModel
	 * @see jp.co.sra.jun.goodies.morph.JunMorphView#getModel()
	 * @category model accessing
	 */
	public JunMorphModel getModel() {
		return (JunMorphModel) this.model();
	}

	/**
	 * Answer <code>true</code> if the receiver is first image, otherwise <code>false</code>.
	 * 
	 * @return boolean
	 * @category testing
	 */
	public boolean isFirstImage() {
		return this.aspectMessage() == $("firstImage");
	}

	/**
	 * Answer <code>true</code> if the receiver is second image, otherwise <code>false</code>.
	 * 
	 * @return boolean
	 * @category testing
	 */
	public boolean isSecondImage() {
		return this.aspectMessage() == $("secondImage");
	}
}
