package jp.co.sra.jun.goodies.print;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.print.PageFormat;
import java.util.Timer;
import java.util.TimerTask;

import javax.swing.BorderFactory;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.border.EtchedBorder;

import jp.co.sra.smalltalk.StComposedText;

/**
 * JunPageView class
 * 
 *  @author    nisinaka
 *  @created   2007/08/01 (by nisinaka)
 *  @updated   2008/08/14 (by nisinaka)
 *  @version   758 (with StPL9.7) based on JunXXX for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunPageView.java,v 9.7 2010/03/09 09:17:20 nisinaka Exp $
 */
public class JunPageView extends JDialog {

	protected JunPrintPreviewCanvas printPreviewCanvas;
	protected int pageIndex;
	protected JPanel canvas;

	private Timer _timer = null;
	private TimerTask _timerTask = null;
	private float _alphaOfPageID = Float.NaN;
	private Rectangle _boundsOfPageID = null;
	private StComposedText _textOfPageID = null;

	/**
	 * Create a new instance of JunPageView and initialize it.
	 *
	 * @param aPrintPreviewCanvas jp.co.sra.jun.goodies.print.JunPrintPreviewCanvas
	 * @category Instance creation
	 */
	public JunPageView(JunPrintPreviewCanvas aPrintPreviewCanvas) {
		super((Frame) null, null, true);

		printPreviewCanvas = aPrintPreviewCanvas;
		pageIndex = -1;

		this.setUndecorated(true);
		this.setResizable(false);
		this.getContentPane().add(this.canvas());
		this.pack();
	}

	/**
	 * Release the timer.
	 * 
	 * @category initialize-release
	 */
	protected void releaseTimer() {
		if (_timer != null) {
			_timer.cancel();
			_timer = null;
		}
	}

	/**
	 * Answer my current print preview canvas.
	 * 
	 * @return jp.co.sra.jun.goodies.print.JunPrintPreviewCanvas
	 * @category accessing
	 */
	protected JunPrintPreviewCanvas getPrintPreviewCanvas() {
		return printPreviewCanvas;
	}

	/**
	 * Answer my current page index.
	 * 
	 * @return int
	 * @category accessing
	 */
	protected int getPageIndex() {
		return pageIndex;
	}

	/**
	 * Set my new page index.
	 * 
	 * @param newPageIndex int
	 * @category accessing
	 */
	protected void setPageIndex(int newPageIndex) {
		pageIndex = newPageIndex;
		this.updateBounds();

		boolean beVisible = newPageIndex >= 0;
		if (beVisible) {
			try {
				this.startTimerTask();
				this.setVisible(true);
			} finally {
				this.stopTimerTask();
			}
		} else {
			this.setVisible(false);
		}
	}

	/**
	 * Answer my current timer.
	 * 
	 * @return java.util.Timer
	 * @category timer accessing
	 */
	protected Timer getTimer() {
		if (_timer == null) {
			_timer = new Timer();
		}
		return _timer;
	}

	/**
	 * Start the timer task.
	 * 
	 * @category timer accessing
	 */
	protected synchronized void startTimerTask() {
		if (_timerTask != null) {
			this.stopTimerTask();
		}

		_alphaOfPageID = 0.7f;
		_boundsOfPageID = null;
		_textOfPageID = null;

		_timerTask = new TimerTask() {
			public void run() {
				_alphaOfPageID -= 0.03f;

				if (_alphaOfPageID <= 0) {
					stopTimerTask();
					_alphaOfPageID = 0.0f;
				}

				canvas().repaint();
			}
		};
		this.getTimer().schedule(_timerTask, 500, 10);
	}

	/**
	 * Stop the timer task.
	 * 
	 * @category timer accessing
	 */
	protected synchronized void stopTimerTask() {
		if (_timerTask == null) {
			return;
		}

		_timerTask.cancel();
		_timerTask = null;
	}

	/**
	 * Answer my current canvas.
	 * 
	 * @return javax.swing.JPanel
	 * @category user interface
	 */
	protected JPanel canvas() {
		if (canvas == null) {
			canvas = new JPanel() {
				protected void paintComponent(Graphics aGraphics) {
					PageFormat pageFormat = getPrintPreviewCanvas().getPrintPreview().getPageFormat();
					int x = (int) Math.rint(pageFormat.getImageableX());
					int y = (int) Math.rint(pageFormat.getImageableY());
					int width = (int) Math.rint(pageFormat.getImageableWidth());
					int height = (int) Math.rint(pageFormat.getImageableHeight());
					int numberOfColumns = getPrintPreviewCanvas().getPrintPreview().getNumberOfColumns();
					int row = getPageIndex() / numberOfColumns;
					int column = getPageIndex() % numberOfColumns;

					if (_boundsOfPageID != null) {
						//aGraphics.clipRect(_boundsOfPageID.x, _boundsOfPageID.y, _boundsOfPageID.width, _boundsOfPageID.height);
					}
					if (_alphaOfPageID > 0) {
						Font font = StComposedText.DefaultFont().deriveFont(height / 2.0f);
						_textOfPageID = new StComposedText(JunPrintableObject.PageID(row, column), font);

						Point p = new Point(x + (width - _textOfPageID.width()) / 2, y + (height - _textOfPageID.height()) / 2);
						_boundsOfPageID = new Rectangle(p.x, p.y, _textOfPageID.width(), _textOfPageID.height());
					}

					super.paintComponent(aGraphics);

					Graphics2D g = null;
					try {
						g = (Graphics2D) aGraphics.create();

						g.clipRect(x, y, width, height);
						g.translate(x - width * column, y - height * row);
						double scale = getPrintPreviewCanvas().getPrintableObject().getScale();
						g.scale(scale, scale);

						getPrintPreviewCanvas().getPrintableObject().print(g);

					} finally {
						if (g != null) {
							g.dispose();
						}
					}

					if (_alphaOfPageID > 0) {
						g = (Graphics2D) aGraphics;
						g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, _alphaOfPageID));
						g.setColor(Color.black);
						_textOfPageID.displayOn_at_(g, _boundsOfPageID.getLocation());
					}
				}
			};
			canvas.setBackground(Color.white);
			canvas.setBorder(BorderFactory.createEtchedBorder(EtchedBorder.LOWERED));

			canvas.setFocusable(true);
			canvas.addKeyListener(new KeyAdapter() {
				public void keyReleased(KeyEvent e) {
					if (e.getKeyCode() == KeyEvent.VK_ALT) {
						getPrintPreviewCanvas().hidePageView();
					}
				}
			});
		}
		return canvas;
	}

	/**
	 * Update the bounds.
	 * 
	 * @category updating
	 */
	protected void updateBounds() {
		if (this.getPageIndex() < 0) {
			return;
		}

		PageFormat pageFormat = this.getPrintPreviewCanvas().getPrintPreview().getPageFormat();
		int width = (int) Math.rint(pageFormat.getWidth());
		int height = (int) Math.rint(pageFormat.getHeight());

		Rectangle pageBounds = this.getPrintPreviewCanvas().getPageBounds(this.getPageIndex());
		Point p = new Point(pageBounds.x + pageBounds.width / 2, pageBounds.y + pageBounds.height / 2);
		SwingUtilities.convertPointToScreen(p, this.getPrintPreviewCanvas());
		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		int x = Math.max(0, Math.min(p.x - width / 2, screenSize.width - width));
		int y = Math.max(0, Math.min(p.y - height / 2, screenSize.height - height));

		this.setBounds(x, y, width, height);
	}

}
