package jp.co.sra.jun.goodies.print;

import java.awt.Dimension;
import java.awt.print.PageFormat;
import java.awt.print.Paper;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;

import javax.print.PrintService;
import javax.print.attribute.AttributeSet;
import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.Size2DSyntax;
import javax.print.attribute.standard.JobName;
import javax.print.attribute.standard.MediaPrintableArea;
import javax.print.attribute.standard.MediaSize;
import javax.print.attribute.standard.MediaSizeName;
import javax.print.attribute.standard.OrientationRequested;
import javax.swing.JOptionPane;

import jp.co.sra.smalltalk.StView;

import jp.co.sra.jun.system.framework.JunApplicationModel;

/**
 * JunPrintPreview class
 * 
 *  @author    nisinaka
 *  @created   2007/07/18 (by nisinaka)
 *  @updated   2008/08/14 (by nisinaka)
 *  @version   758 (with StPL9.7) based on JunXXX for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunPrintPreview.java,v 9.9 2010/03/09 09:17:20 nisinaka Exp $
 */
public class JunPrintPreview extends JunApplicationModel {

	protected static PrintService CurrentPrintService = null;
	protected static PageFormat CurrentPageFormat = null;
	protected static int DefaultLargeNumberOfPages = 10;

	protected JunPrintableObject printableObject;
	protected PrinterJob printerJob;
	protected Dimension pageMatrix;
	protected boolean[] pageSelection;
	protected int largeNumberOfPages;

	/**
	 * Create a new instance of JunPrintPreview and initialize it.
	 *
	 * @param aPrintableObject jp.co.sra.jun.goodies.print.PrintableObject
	 * @category Instance creation
	 */
	public JunPrintPreview(JunPrintableObject aPrintableObject) {
		this.setPrintableObject(aPrintableObject);
	}

	/**
	 * Initialize the receiver.
	 * 
	 * @see jp.co.sra.smalltalk.StApplicationModel#initialize()
	 * @category initialize-release
	 */
	protected void initialize() {
		super.initialize();
		printableObject = null;
		printerJob = null;
		this.flushPageMatrix();
		pageSelection = null;
		largeNumberOfPages = DefaultLargeNumberOfPages;
	}

	/**
	 * Answer my current printable object.
	 * 
	 * @return jp.co.sra.jun.goodies.print.JunPrintableObject
	 * @category accessing
	 */
	public JunPrintableObject getPrintableObject() {
		return printableObject;
	}

	/**
	 * Set my new printable object.
	 * 
	 * @param aPrintableObject jp.co.sra.jun.goodies.print.JunPrintableObject
	 * @category accessing
	 */
	protected void setPrintableObject(JunPrintableObject aPrintableObject) {
		printableObject = aPrintableObject;
		this.flushPageMatrix();
	}

	/**
	 * Answer my current scale value.
	 * 
	 * @return double
	 * @category accessing
	 */
	protected double getScale() {
		return this.getPrintableObject().getScale();
	}

	/**
	 * Set my new scale value.
	 * 
	 * @param value double
	 * @category accessing
	 */
	protected void setScale(double value) {
		if (this.getPrintableObject().getScale() == value) {
			return;
		}

		this.getPrintableObject().setScale(value);
		this.flushPageMatrix();
		this.changed_($("scale"));
	}

	/**
	 * Answer true if the page ID should be printed, otherwise false.
	 * 
	 * @return boolean
	 * @category accessing
	 */
	protected boolean doPrintPageID() {
		return this.getPrintableObject().doPrintPageID();
	}

	/**
	 * Set whether the page ID should be printed or not.
	 * 
	 * @param doPrintPageID boolean
	 * @category accessing
	 */
	protected void doPrintPageID(boolean doPrintPageID) {
		if (this.doPrintPageID() == doPrintPageID) {
			return;
		}

		this.getPrintableObject().doPrintPageID(doPrintPageID);
		this.changed_($("printPageID"));
	}

	/**
	 * Answer true if the corner mark should be printed, otherwise false.
	 * 
	 * @return boolean
	 * @category accessing
	 */
	protected boolean doPrintCornerMarks() {
		return this.getPrintableObject().doPrintCornerMarks();
	}

	/**
	 * Set whether the corner marks should be printed or not.
	 * 
	 * @param doPrintCornerMarks boolean
	 * @category accessing
	 */
	protected void doPrintCornerMarks(boolean doPrintCornerMarks) {
		if (this.doPrintCornerMarks() == doPrintCornerMarks) {
			return;
		}

		this.getPrintableObject().doPrintCornerMarks(doPrintCornerMarks);
		this.changed_($("printCornerMarks"));
	}

	/**
	 * Answer my current printer job.
	 * 
	 * @return java.awt.print.PrinterJob
	 * @throws java.awt.print.PrinterException 
	 * @category accessing
	 */
	protected PrinterJob getPrinterJob() throws PrinterException {
		if (printerJob == null) {
			printerJob = PrinterJob.getPrinterJob();

			if (CurrentPrintService == null) {
				CurrentPrintService = printerJob.getPrintService();
			} else {
				printerJob.setPrintService(CurrentPrintService);
			}
		}
		return printerJob;
	}

	/**
	 * Answer my current print service.
	 * 
	 * @return javax.print.PrintService
	 * @throws java.awt.print.PrinterException 
	 * @category accessing
	 */
	protected PrintService getPrintService() throws PrinterException {
		return this.getPrinterJob().getPrintService();
	}

	/**
	 * Set my new print service.
	 * 
	 * @param aPrintService javax.print.PrintService
	 * @throws java.awt.print.PrinterException
	 * @category accessing
	 */
	protected void setPrintService(PrintService aPrintService) throws PrinterException {
		if (aPrintService == CurrentPrintService) {
			return;
		}

		this.getPrinterJob().setPrintService(aPrintService);
		CurrentPrintService = aPrintService;

		this.setPageFormat(this.getPrinterJob().validatePage(this.getPageFormat()));
	}

	/**
	 * Answer my current page format.
	 * 
	 * @return java.awt.print.PageFormat
	 * @category accessing
	 */
	protected PageFormat getPageFormat() {
		if (CurrentPageFormat == null) {
			CurrentPageFormat = this.defaultPageFormat();
			CurrentPageFormat.setOrientation(PageFormat.LANDSCAPE);
		}
		return CurrentPageFormat;
	}

	/**
	 * Set my new page format.
	 * 
	 * @param aPageFormat java.awt.print.PageFormat
	 * @throws java.awt.print.PrinterException
	 * @category accessing
	 */
	protected void setPageFormat(PageFormat aPageFormat) throws PrinterException {
		if (aPageFormat == CurrentPageFormat) {
			return;
		}

		CurrentPageFormat = aPageFormat;
		this.flushPageMatrix();
		this.changed_($("pageFormat"));
	}

	/**
	 * Answer my current media size name.
	 * 
	 * @return javax.print.attribute.standard.MediaSizeName
	 * @throws java.awt.print.PrinterException 
	 * @category accessing
	 */
	protected MediaSizeName getMediaSizeName() throws PrinterException {
		Paper paper = this.getPageFormat().getPaper();
		float w = (float) JunPrintableObject.POINTtoMM(paper.getWidth());
		float h = (float) JunPrintableObject.POINTtoMM(paper.getHeight());
		return MediaSize.findMedia(w, h, Size2DSyntax.MM);
	}

	/**
	 * Answer my current media size.
	 * 
	 * @return javax.print.attribute.standard.MediaSize
	 * @throws java.awt.print.PrinterException 
	 * @category accessing
	 */
	protected MediaSize getMediaSize() throws PrinterException {
		return MediaSize.getMediaSizeForName(this.getMediaSizeName());
	}

	/**
	 * Answer my current media printable area.
	 * 
	 * @return javax.print.attribute.standard.MediaPrintableArea
	 * @throws java.awt.print.PrinterException 
	 * @category accessing
	 */
	protected MediaPrintableArea getMediaPrintableArea() throws PrinterException {
		AttributeSet attributeSet = new HashPrintRequestAttributeSet();
		attributeSet.add(this.getMediaSizeName());
		MediaPrintableArea[] printableArea = (MediaPrintableArea[]) this.getPrinterJob().getPrintService().getSupportedAttributeValues(MediaPrintableArea.class, null, attributeSet);
		if (printableArea == null || printableArea.length == 0) {
			return null;
		}
		return printableArea[0];
	}

	/**
	 * Answer my current page matrix.
	 * 
	 * @return java.awt.Dimension
	 * @category accessing
	 */
	protected Dimension getPageMatrix() {
		if (pageMatrix == null) {
			pageMatrix = this.getPrintableObject().getPageMatrix(this.getPageFormat());
		}
		return pageMatrix;
	}

	/**
	 * Answer my current number of columns.
	 * 
	 * @return int
	 * @category accessing
	 */
	protected int getNumberOfColumns() {
		return this.getPageMatrix().width;
	}

	/**
	 * Answer my current number of rows.
	 * 
	 * @return int
	 * @category accessing
	 */
	protected int getNumberOfRows() {
		return this.getPageMatrix().height;
	}

	/**
	 * Answer my current number of pages.
	 * 
	 * @return int
	 * @category accessing
	 */
	protected int getNumberOfPages() {
		return this.getNumberOfColumns() * this.getNumberOfRows();
	}

	/**
	 * Answer the width of all pages.
	 * 
	 * @return double
	 * @category accessing
	 */
	protected double getWidthOfAllPages() {
		return this.getPageFormat().getWidth() * this.getNumberOfColumns();
	}

	/**
	 * Answer the height of all pages.
	 * 
	 * @return double
	 * @category accessing
	 */
	protected double getHeightOfAllPages() {
		return this.getPageFormat().getHeight() * this.getNumberOfRows();
	}

	/**
	 * Answer my current page selections.
	 * 
	 * @return boolean[]
	 * @category selection accessing
	 */
	protected boolean[] getPageSelections() {
		int numberOfPages = this.getNumberOfPages();
		if (pageSelection != null && pageSelection.length != numberOfPages) {
			pageSelection = null;
		}
		if (pageSelection == null) {
			pageSelection = new boolean[numberOfPages];
			for (int i = 0; i < numberOfPages; i++) {
				pageSelection[i] = true;
			}
		}
		return pageSelection;
	}

	/**
	 * Set whether the specified page is selected or not.
	 * 
	 * @param pageIndex int
	 * @param isSelected boolean
	 * @category selection accessing
	 */
	protected void setPageSelection(int pageIndex, boolean isSelected) {
		boolean wasSelected = this.isPageSelected(pageIndex);
		if (wasSelected == isSelected) {
			return;
		}

		this.getPageSelections()[pageIndex] = isSelected;
		this.changed_($("pageSelection"));
	}

	/**
	 * Clear my current page selections.
	 * 
	 * @category selection accessing
	 */
	protected void clearPageSelections() {
		boolean[] selections = this.getPageSelections();
		for (int i = 0; i < selections.length; i++) {
			selections[i] = true;
		}
		this.changed_($("pageSelection"));
	}

	/**
	 * Toggle the selection of the specified page.
	 * 
	 * @param pageIndex int
	 * @category selection accessing
	 */
	protected void togglePageSelection(int pageIndex) {
		boolean isSelected = this.isPageSelected(pageIndex);
		this.setPageSelection(pageIndex, !isSelected);
	}

	/**
	 * Answer the default page format of the current printer.
	 * 
	 * @return java.awt.print.PageFormat
	 * @category defaults
	 */
	protected PageFormat defaultPageFormat() {
		PageFormat pageFormat = null;
		try {
			pageFormat = this.getPrinterJob().defaultPage();
			Paper paper = pageFormat.getPaper();
			paper.setImageableArea(0, 0, paper.getWidth(), paper.getHeight());
			pageFormat.setPaper(paper);
			pageFormat = this.getPrinterJob().validatePage(pageFormat);
		} catch (PrinterException e) {
			e.printStackTrace();

			String message = e.getMessage();
			JOptionPane.showMessageDialog(null, message);
		}
		return pageFormat;
	}

	/**
	 * Flush the current page matrix.
	 * 
	 * @category flushing
	 */
	protected void flushPageMatrix() {
		pageMatrix = null;
	}

	/**
	 * Answer true if the specified page is currently selected, otherwise false.
	 * 
	 * @param pageIndex int
	 * @return boolean
	 * @category testing
	 */
	protected boolean isPageSelected(int pageIndex) {
		return this.getPageSelections()[pageIndex];
	}

	/**
	 * Answer true if the receiver has one or more selected pages, otherwise false.
	 * 
	 * @return boolean
	 * @category testing
	 */
	protected boolean hasPageSelection() {
		boolean[] selections = this.getPageSelections();
		for (int i = 0; i < selections.length; i++) {
			if (selections[i]) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Create a default view.
	 * 
	 * @see jp.co.sra.smalltalk.StApplicationModel#defaultView()
	 * @category interface opening
	 */
	public StView defaultView() {
		return new JunPrintPreviewViewSwing(this);
	}

	/**
	 * Answer a window title.
	 * 
	 * @return java.lang.String
	 * @see jp.co.sra.smalltalk.StApplicationModel#windowTitle()
	 * @category interface opening
	 */
	protected String windowTitle() {
		return $String("Print Preview");
	}

	/**
	 * Open the page dialog.
	 * 
	 * @throws java.awt.print.PrinterException 
	 * @category actions
	 */
	public void openPageDialog() throws PrinterException {
		PageFormat newPageFormat = null;
		if (false) {
			PageFormat pageFormat = this.getPageFormat();
			newPageFormat = this.getPrinterJob().pageDialog(pageFormat);
			if (newPageFormat == pageFormat) {
				return; // since it's canceled.
			}

			CurrentPrintService = this.getPrinterJob().getPrintService();
			// A workaround to update the instance variable of the printer job.
			this.getPrinterJob().setPrintService(CurrentPrintService);

		} else {
			PrintRequestAttributeSet attributeSet = this.createPrintRequestAttributeSet(this.getPageFormat());
			newPageFormat = this.getPrinterJob().pageDialog(attributeSet);
			if (newPageFormat == null) {
				return; // since it's canceled.
			}
		}

		this.setPageFormat(newPageFormat);
	}

	/**
	 * Scale the printable object to fit to one page.
	 * 
	 * @category actions
	 */
	public void fitToOnePage() {
		PageFormat pageFormat = this.getPageFormat();
		double scaleX = pageFormat.getImageableWidth() / this.getPrintableObject().getWidth();
		double scaleY = pageFormat.getImageableHeight() / this.getPrintableObject().getHeight();
		double scale = Math.floor(Math.min(scaleX, scaleY) * 100) / 100;
		this.setScale(scale);
	}

	/**
	 * Scale the printable object to fit to one row.
	 * 
	 * @category actions
	 */
	public void fitToOneRow() {
		this.fitToRows(1);
	}

	/**
	 * Scale the printable object to fit to the specified rows.
	 * 
	 * @param rows int
	 * @category actions
	 */
	public void fitToRows(int rows) {
		PageFormat pageFormat = this.getPageFormat();
		double totalPageHeight = pageFormat.getImageableHeight() * rows;
		double scale = totalPageHeight / this.getPrintableObject().getHeight();
		scale = Math.floor(scale * 100) / 100;
		this.setScale(scale);
	}

	/**
	 * Scale the printable object to fit to the columns.
	 * 
	 * @category actions
	 */
	public void fitToColumns() {
		this.fitToColumns(this.getNumberOfColumns());
	}

	/**
	 * Scale the printable object to fit to the specified columns.
	 * 
	 * @param columns int
	 * @category actions
	 */
	public void fitToColumns(int columns) {
		PageFormat pageFormat = this.getPageFormat();
		double totalPageWidth = pageFormat.getImageableWidth() * columns;
		double scale = totalPageWidth / this.getPrintableObject().getWidth();
		scale = Math.floor(scale * 100) / 100;
		this.setScale(scale);
	}

	/**
	 * Print the printable object in the current settings.
	 * 
	 * @throws java.awt.print.PrinterException 
	 * @category actions
	 */
	public void doPrint() throws PrinterException {
		// Setup the pages to print.
		int numberOfPages = 0;
		boolean[] selections = this.getPageSelections();
		for (int i = 0; i < selections.length; i++) {
			if (selections[i]) {
				numberOfPages++;
			}
		}
		if (numberOfPages == 0) {
			numberOfPages = this.getNumberOfPages();
			this.getPrintableObject().setPageNumbers(null);
		} else {
			int[] pageNumbers = new int[numberOfPages];
			int pageIndex = 0;
			for (int i = 0; i < selections.length; i++) {
				if (selections[i]) {
					pageNumbers[pageIndex] = i;
					pageIndex++;
				}
			}
			this.getPrintableObject().setPageNumbers(pageNumbers);
		}

		// Warn for the large number of the pages to print.
		if (numberOfPages >= largeNumberOfPages) {
			String message = $String("You are about to print <1p> pages.", null, String.valueOf(numberOfPages));
			message += "\n";
			message += $String("Are you sure you want to print all of them?");
			int option = JOptionPane.showConfirmDialog(null, message, $String("Print"), JOptionPane.YES_NO_OPTION);
			if (option == JOptionPane.NO_OPTION) {
				return;
			}
		}

		// OK, print.
		PrintRequestAttributeSet attributeSet = this.createPrintRequestAttributeSet(this.getPageFormat());
		attributeSet.add(new JobName(this.getPrintableObject().getJobName(), null));
		this.getPrinterJob().setPrintable(this.getPrintableObject(), this.getPageFormat());
		this.getPrinterJob().print(attributeSet);
	}

	/**
	 * Create a PrintRequestAttributeSet from a PageFormat.
	 * 
	 * @param aPageFormat java.awt.print.PageFormat
	 * @return javax.print.attribute.PrintRequestAttributeSet
	 * @throws java.awt.print.PrinterException
	 * @category private
	 */
	protected PrintRequestAttributeSet createPrintRequestAttributeSet(PageFormat aPageFormat) throws PrinterException {
		PrintRequestAttributeSet printRequestAttributeSet = new HashPrintRequestAttributeSet();

		printRequestAttributeSet.add(this.getMediaSizeName());

		switch (this.getPageFormat().getOrientation()) {
			case PageFormat.LANDSCAPE:
				printRequestAttributeSet.add(OrientationRequested.LANDSCAPE);
				break;
			case PageFormat.PORTRAIT:
				printRequestAttributeSet.add(OrientationRequested.PORTRAIT);
				break;
			case PageFormat.REVERSE_LANDSCAPE:
				printRequestAttributeSet.add(OrientationRequested.REVERSE_LANDSCAPE);
				break;
		}

		Paper paper = aPageFormat.getPaper();
		float x = (float) JunPrintableObject.POINTtoMM(paper.getImageableX());
		float y = (float) JunPrintableObject.POINTtoMM(paper.getImageableY());
		float w = (float) JunPrintableObject.POINTtoMM(paper.getImageableWidth());
		float h = (float) JunPrintableObject.POINTtoMM(paper.getImageableHeight());
		printRequestAttributeSet.add(new MediaPrintableArea(x, y, w, h, MediaPrintableArea.MM));

		return printRequestAttributeSet;
	}

}
