package jp.co.sra.jun.goodies.print;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.event.ActionEvent;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.print.PageFormat;

import javax.swing.AbstractAction;
import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.JComponent;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;

import jp.co.sra.smalltalk.DependentEvent;
import jp.co.sra.smalltalk.StImage;
import jp.co.sra.smalltalk.StSymbol;
import jp.co.sra.smalltalk.SystemInterface;

import jp.co.sra.jun.system.framework.JunAbstractViewJPanel;

/**
 * JunPrintPreviewCanvas class
 * 
 *  @author    nisinaka
 *  @created   2007/07/19 (by nisinaka)
 *  @updated   2008/08/14 (by nisinaka)
 *  @version   758 (with StPL9.7) based on JunXXX for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunPrintPreviewCanvas.java,v 9.7 2010/03/09 09:17:20 nisinaka Exp $
 */
public class JunPrintPreviewCanvas extends JunAbstractViewJPanel {

	protected double scale;
	protected StImage printableImage;
	protected Rectangle[] pageBounds;
	protected Rectangle selectionBounds;
	protected Shape clippingBounds;

	private JunPageView _pageView;

	/**
	 * Create a new instance of JunPrintPreviewCanvas and initialize it.
	 * 
	 * @param jp.co.sra.jun.goodies.print.PrintPreview
	 * @category Instance creation
	 */
	public JunPrintPreviewCanvas(JunPrintPreview aPrintPreview) {
		super(aPrintPreview);
	}

	/**
	 * Initialize the receiver.
	 * 
	 * @see jp.co.sra.smalltalk.StViewJPanel#initialize()
	 * @category initialize-release
	 */
	protected void initialize() {
		super.initialize();
		scale = Double.NaN;
		printableImage = null;
		pageBounds = null;
		selectionBounds = null;
		_pageView = null;
		this.setBackground(Color.white);

		InputMap inputMap = this.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW);
		ActionMap actionMap = this.getActionMap();
		inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_ALT, InputEvent.ALT_MASK, false), "altPressed");
		actionMap.put("altPressed", new AbstractAction("altPressed") {
			public void actionPerformed(ActionEvent e) {
				Point p = SystemInterface._MousePoint();
				SwingUtilities.convertPointFromScreen(p, JunPrintPreviewCanvas.this);
				JunPrintPreviewCanvas.this.showPageView(p);
			}
		});
	}

	/**
	 * Notifies this component that it no longer has a parent component.
	 * 
	 * @see javax.swing.JComponent#removeNotify()
	 * @category initialize-release
	 */
	public void removeNotify() {
		super.removeNotify();

		if (_pageView != null) {
			_pageView.releaseTimer();
			_pageView.dispose();
			_pageView = null;
		}
	}

	/**
	 * Answer my model as JunPrintPreview.
	 * 
	 * @return jp.co.sra.jun.goodies.print.JunPrintPreview
	 * @see jp.co.sra.jun.goodies.print.JunPrintPreviewCanvas#getPrintPreview()
	 * @category model accessing
	 */
	public JunPrintPreview getPrintPreview() {
		return (JunPrintPreview) this.model();
	}

	/**
	 * Answer the printable object.
	 * 
	 * @return jp.co.sra.jun.goodies.print.JunPrintableObject
	 * @category model accessing
	 */
	public JunPrintableObject getPrintableObject() {
		return this.getPrintPreview().getPrintableObject();
	}

	/**
	 * Answer my current selection bounds.
	 * 
	 * @return java.awt.Rectangle
	 * @see jp.co.sra.jun.goodies.print.JunPrintPreviewCanvas#getSelectionBounds()
	 * @category accessing
	 */
	public Rectangle getSelectionBounds() {
		return selectionBounds;
	}

	/**
	 * Set my current selection bounds.
	 * 
	 * @param aRectangle java.awt.Rectangle
	 * @see jp.co.sra.jun.goodies.print.JunPrintPreviewCanvas#setSelectionBounds(java.awt.Rectangle)
	 * @category accessing
	 */
	public void setSelectionBounds(Rectangle aRectangle) {
		selectionBounds = aRectangle;

		if (selectionBounds != null) {
			boolean[] selections = this.getPrintPreview().getPageSelections();
			for (int pageIndex = 0; pageIndex < selections.length; pageIndex++) {
				Rectangle bounds = this.getPageBounds(pageIndex);
				if (selectionBounds.width > 0 && selectionBounds.height > 0) {
					selections[pageIndex] = bounds.intersects(selectionBounds);
				} else {
					selections[pageIndex] = bounds.intersectsLine(selectionBounds.x, selectionBounds.y, selectionBounds.x + selectionBounds.width, selectionBounds.y + selectionBounds.height);
				}
			}
		}

		this.repaint();
	}

	/**
	 * Answer the index of the page at the specified point.
	 * If the point is not on a page, return -1.
	 * 
	 * @param aPoint java.awt.Point
	 * @return int
	 * @see jp.co.sra.jun.goodies.print.JunPrintPreviewCanvas#getPageIndexAt(java.awt.Point)
	 * @category accessing
	 */
	public int getPageIndexAt(Point aPoint) {
		int numberOfPages = this.getPrintPreview().getNumberOfPages();
		for (int pageIndex = 0; pageIndex < numberOfPages; pageIndex++) {
			if (this.getPageBounds(pageIndex).contains(aPoint)) {
				return pageIndex;
			}
		}
		return -1;
	}

	/**
	 * Answer my current scale.
	 * 
	 * @return double
	 * @category accessing
	 */
	protected double getScale() {
		if (Double.isNaN(scale)) {
			Dimension size = this.getSize();
			double totalWidth = this.getPrintPreview().getWidthOfAllPages();
			double totalHeight = this.getPrintPreview().getHeightOfAllPages();
			scale = Math.min(size.width / totalWidth, size.height / totalHeight);
		}
		return scale;
	}

	/**
	 * Answer my current printable image.
	 * 
	 * @return jp.co.sra.smalltalk.StImage
	 * @category accessing
	 */
	protected StImage getPrintableImage() {
		if (printableImage == null) {
			int width = (int) Math.rint(this.getPrintableObject().getScaledWidth() * this.getScale());
			int height = (int) Math.rint(this.getPrintableObject().getScaledHeight() * this.getScale());
			StImage image = new StImage(width, height);
			Graphics2D g = null;
			try {
				g = (Graphics2D) image.image().getGraphics();
				g.scale(this.getScale(), this.getScale());
				g.scale(this.getPrintableObject().getScale(), this.getPrintableObject().getScale());
				this.getPrintableObject().print(g);
			} finally {
				if (g != null) {
					g.dispose();
				}
			}

			printableImage = image;
		}
		return printableImage;
	}

	/**
	 * Answer the bounds of the specified page.
	 * 
	 * @param row int
	 * @param column int
	 * @return java.awt.Rectangle
	 * @category accessing
	 */
	protected Rectangle getPageBounds(int pageIndex) {
		return this.getPageBounds()[pageIndex];
	}

	/**
	 * Answer my current page bounds.
	 * 
	 * @return java.awt.Rectangle[]
	 * @category accessing
	 */
	protected Rectangle[] getPageBounds() {
		if (pageBounds == null) {
			double totalPageWidth = this.getPrintPreview().getWidthOfAllPages();
			double totalPageHeight = this.getPrintPreview().getHeightOfAllPages();
			double scale = this.getScale();
			Dimension size = this.getSize();
			int maxX = Math.min((int) Math.rint(totalPageWidth * scale), size.width - 1);
			int maxY = Math.min((int) Math.rint(totalPageHeight * scale), size.height - 1);

			int numberOfRows = this.getPrintPreview().getNumberOfRows();
			int numberOfColumns = this.getPrintPreview().getNumberOfColumns();
			Rectangle[] bounds = new Rectangle[numberOfRows * numberOfColumns];

			for (int row = 0; row < numberOfRows; row++) {
				int y1 = (int) Math.rint((double) maxY * row / numberOfRows);
				int y2 = (int) Math.rint((double) maxY * (row + 1) / numberOfRows);

				for (int column = 0; column < numberOfColumns; column++) {
					int x1 = (int) Math.rint((double) maxX * column / numberOfColumns);
					int x2 = (int) Math.rint((double) maxX * (column + 1) / numberOfColumns);

					int pageIndex = column + row * numberOfColumns;
					bounds[pageIndex] = new Rectangle(x1, y1, x2 - x1, y2 - y1);
				}
			}

			pageBounds = bounds;
		}
		return pageBounds;
	}

	/**
	 * Set my bounds.
	 * 
	 * @param x int
	 * @param y int
	 * @param width int
	 * @param height int
	 * @see java.awt.Component#setBounds(int, int, int, int)
	 * @category bounds accessing
	 */
	public void setBounds(int x, int y, int width, int height) {
		this.flushBounds();
		super.setBounds(x, y, width, height);
	}

	/**
	 * Flush the bounds.
	 * 
	 * @category flushing
	 */
	protected void flushBounds() {
		scale = Double.NaN;
		printableImage = null;
		pageBounds = null;
	}

	/**
	 * Display the receiver on the graphics.
	 * 
	 * @param aGraphics java.awt.Graphics
	 * @see jp.co.sra.smalltalk.StViewJPanel#displayOn_(java.awt.Graphics)
	 * @category displaying
	 */
	public void displayOn_(Graphics aGraphics) {
		Graphics2D g = (Graphics2D) aGraphics;

		this.displayPagesOn_(g);
		this.displaySelectionBoundsOn_(g);
	}

	/**
	 * Display the pages on the graphics.
	 * 
	 * @param aGraphics java.awt.Graphics2D
	 * @category displaying
	 */
	protected void displayPagesOn_(Graphics2D aGraphics) {
		int numberOfColumns = this.getPrintPreview().getNumberOfColumns();
		int numberOfRows = this.getPrintPreview().getNumberOfRows();
		for (int row = 0; row < numberOfRows; row++) {
			for (int column = 0; column < numberOfColumns; column++) {
				this.displayPageOn_(aGraphics, row, column);
			}
		}
	}

	/**
	 * Display the specified page on the graphics.
	 * 
	 * @param aGraphics java.awt.Graphics2D
	 * @param row int
	 * @param column int
	 * @category displaying
	 */
	protected void displayPageOn_(Graphics2D aGraphics, int row, int column) {
		int pageIndex = column + row * this.getPrintPreview().getNumberOfColumns();
		Rectangle pageBounds = this.getPageBounds(pageIndex);

		PageFormat pageFormat = this.getPrintPreview().getPageFormat();
		double scale = this.getScale();
		int x = pageBounds.x + (int) Math.rint(pageFormat.getImageableX() * scale);
		int y = pageBounds.y + (int) Math.rint(pageFormat.getImageableY() * scale);
		int width = (int) Math.rint(pageFormat.getImageableWidth() * scale);
		int height = (int) Math.rint(pageFormat.getImageableHeight() * scale);

		aGraphics.setClip(x, y, width, height);

		int dx = width * column;
		int dy = height * row;
		aGraphics.translate(-dx, -dy);
		this.getPrintableImage().displayOn_at_(aGraphics, new Point(x, y));
		aGraphics.translate(dx, dy);

		aGraphics.setClip(pageBounds.x, pageBounds.y, pageBounds.width + 1, pageBounds.height + 1);

		if (this.getPrintPreview().hasPageSelection() && this.getPrintPreview().isPageSelected(pageIndex) == false) {
			aGraphics.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.5f));
			aGraphics.setColor(Color.gray);
			aGraphics.fillRect(pageBounds.x, pageBounds.y, pageBounds.width, pageBounds.height);
			aGraphics.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER));
		}

		aGraphics.drawRect(pageBounds.x, pageBounds.y, pageBounds.width, pageBounds.height);
	}

	/**
	 * Display the selection bounds on the graphics.
	 * 
	 * @param aGraphics java.awt.Graphics2D
	 * @category displaying
	 */
	protected void displaySelectionBoundsOn_(Graphics2D aGraphics) {
		if (selectionBounds == null) {
			return;
		}

		aGraphics.setClip(0, 0, this.getWidth(), this.getHeight());
		aGraphics.setColor(Color.red);
		aGraphics.drawRect(selectionBounds.x, selectionBounds.y, selectionBounds.width, selectionBounds.height);
	}

	/**
	 * Update for the change notice.
	 * 
	 * @param e jp.co.sra.smalltalk.DependentEvent
	 * @see jp.co.sra.smalltalk.StViewJPanel#update_(jp.co.sra.smalltalk.DependentEvent)
	 * @category updating
	 */
	public void update_(DependentEvent e) {
		StSymbol aspect = e.getAspect();
		if (aspect == $("pageFormat") || aspect == $("scale")) {
			this.flushBounds();
		}

		super.update_(e);
	}

	/**
	 * Show the page view at the specified point.
	 * 
	 * @param aPoint java.awt.Point
	 * @category page view accessing
	 */
	public void showPageView(Point aPoint) {
		int pageIndex = this.getPageIndexAt(aPoint);
		this.getPageView().setPageIndex(pageIndex);
	}

	/**
	 * Hide the page view.
	 * 
	 * @category page view accessing
	 */
	public void hidePageView() {
		this.getPageView().setPageIndex(-1);
	}

	/**
	 * Answer my current page view.
	 * 
	 * @return jp.co.sra.jun.goodies.print.JunPageView
	 * @category page view accessing
	 */
	protected JunPageView getPageView() {
		if (_pageView == null) {
			_pageView = new JunPageView(this);
		}
		return _pageView;
	}

}
