package jp.co.sra.jun.goodies.print;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.geom.AffineTransform;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;
import java.awt.print.PageFormat;
import java.awt.print.Paper;
import java.awt.print.Printable;
import java.awt.print.PrinterException;

import jp.co.sra.smalltalk.StComposedText;

import jp.co.sra.jun.system.support.JunSystem;

/**
 * JunPrintableObject class
 * 
 *  @author    nisinaka
 *  @created   2007/07/18 (by nisinaka)
 *  @updated   2008/08/14 (by nisinaka)
 *  @version   758 (with StPL9.7) based on JunXXX for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunPrintableObject.java,v 9.7 2010/03/09 09:17:21 nisinaka Exp $
 */
public abstract class JunPrintableObject implements Printable {

	protected double scale = 1.0;
	protected int[] pageNumbers = null;
	protected boolean doPrintPageID = true;
	protected boolean doPrintCornerMarks = true;

	/**
	 * Answer my job name.
	 * 
	 * @return java.lang.String
	 * @category accessing
	 */
	protected String getJobName() {
		return JunSystem.DefaultBaseName();
	}

	/**
	 * Answer my width.
	 * 
	 * @return double
	 * @category accessing
	 */
	protected abstract double getWidth();

	/**
	 * Answer my height.
	 * 
	 * @return double
	 * @category accessing
	 */
	protected abstract double getHeight();

	/**
	 * Answer my scale.
	 * 
	 * @return double
	 * @category accessing
	 */
	protected double getScale() {
		return scale;
	}

	/**
	 * Set my scale.
	 * 
	 * @param value double
	 * @category accessing
	 */
	protected void setScale(double value) {
		scale = value;
	}

	/**
	 * Answer my current scaled width.
	 * 
	 * @return double
	 * @category accessing
	 */
	protected double getScaledWidth() {
		return this.getWidth() * this.getScale();
	}

	/**
	 * Answer my current scaled height.
	 * 
	 * @return double
	 * @category accessing
	 */
	protected double getScaledHeight() {
		return this.getHeight() * this.getScale();
	}

	/**
	 * Answer my current page numbers.
	 * 
	 * @return int[]
	 * @category accessing
	 */
	protected int[] getPageNumbers() {
		return pageNumbers;
	}

	/**
	 * Set my new page numbers.
	 * 
	 * @param newPageNumbers int[]
	 * @category accessing
	 */
	protected void setPageNumbers(int[] newPageNumbers) {
		pageNumbers = newPageNumbers;
	}

	/**
	 * Answer true if the page ID should be printed, otherwise false.
	 * 
	 * @return boolean
	 * @category accessing
	 */
	protected boolean doPrintPageID() {
		return doPrintPageID;
	}

	/**
	 * oSet whether the page ID should be printed or not.
	 * 
	 * @param newDoPrintPageID boolean
	 * @category accessing
	 */
	protected void doPrintPageID(boolean newDoPrintPageID) {
		doPrintPageID = newDoPrintPageID;
	}

	/**
	 * Answer true if the corner marks should be printed, otherwise false.
	 * 
	 * @return boolean
	 * @category accessing
	 */
	protected boolean doPrintCornerMarks() {
		return doPrintCornerMarks;
	}

	/**
	 * Set whether the corner marks should be printed or not.
	 * 
	 * @param newDoPrintCornerMarks boolean
	 * @category accessing
	 */
	protected void doPrintCornerMarks(boolean newDoPrintCornerMarks) {
		doPrintCornerMarks = newDoPrintCornerMarks;
	}

	/**
	 * Answer the matrix of the pages.
	 * 
	 * @param aPageFormat java.awt.print.PageFormat
	 * @return java.awt.Dimension
	 * @category accessing
	 */
	protected Dimension getPageMatrix(PageFormat aPageFormat) {
		double pageWidth = aPageFormat.getImageableWidth();
		double pageHeight = aPageFormat.getImageableHeight();
		double scaledWidth = this.getWidth() * this.getScale();
		double scaledHeight = this.getHeight() * this.getScale();
		int numberOfColumns = (int) Math.ceil(scaledWidth / pageWidth);
		int numberOfRows = (int) Math.ceil(scaledHeight / pageHeight);
		return new Dimension(numberOfColumns, numberOfRows);
	}

	/**
	 * Prints the page at the specified index into the specified {@link Graphics} context in the specified format.
	 * 
	 * @param aGraphics java.awt.Graphics
	 * @param aPageFormat java.awt.print.PageFormat
	 * @param pageIndex int
	 * @return int
	 * @see java.awt.print.Printable#print(java.awt.Graphics, java.awt.print.PageFormat, int)
	 * @category printing
	 */
	public int print(Graphics aGraphics, PageFormat aPageFormat, int pageIndex) throws PrinterException {
		Dimension matrix = this.getPageMatrix(aPageFormat);
		int numberOfTotalPages = matrix.width * matrix.height;

		int[] pageNumbers = this.getPageNumbers();
		if (pageNumbers != null) {
			numberOfTotalPages = Math.min(numberOfTotalPages, pageNumbers.length);
		}

		if (pageIndex >= numberOfTotalPages) {
			return Printable.NO_SUCH_PAGE;
		}

		if (pageNumbers != null) {
			pageIndex = pageNumbers[pageIndex];
		}

		int row = pageIndex / matrix.width;
		int column = pageIndex % matrix.width;

		Graphics2D g = (Graphics2D) aGraphics;
		g.setClip(new Rectangle2D.Double(aPageFormat.getImageableX(), aPageFormat.getImageableY(), aPageFormat.getImageableWidth(), aPageFormat.getImageableHeight()));

		AffineTransform transform = g.getTransform();
		try {
			double dx = aPageFormat.getImageableX() - aPageFormat.getImageableWidth() * column;
			double dy = aPageFormat.getImageableY() - aPageFormat.getImageableHeight() * row;

			g.translate(dx, dy);
			g.scale(this.getScale(), this.getScale());

			this.print(g);

		} finally {
			g.setTransform(transform);
		}

		// Print the Page ID
		if (this.doPrintPageID()) {
			int x = (int) Math.rint(aPageFormat.getImageableX() + aPageFormat.getImageableWidth());
			int y = (int) Math.rint(aPageFormat.getImageableY() + aPageFormat.getImageableHeight());
			StComposedText pageID = new StComposedText(PageID(row, column));
			pageID.displayOn_at_(aGraphics, new Point(x - pageID.width(), y - pageID.height()));
		}

		// Print the corner marks
		if (this.doPrintCornerMarks()) {
			double delta = MMtoPOINT(10);
			double originX = aPageFormat.getImageableX();
			double originY = aPageFormat.getImageableY();
			double cornerX = aPageFormat.getImageableX() + aPageFormat.getImageableWidth();
			double cornerY = aPageFormat.getImageableY() + aPageFormat.getImageableHeight();

			g.draw(new Line2D.Double(originX, originY + delta, originX, originY));
			g.draw(new Line2D.Double(originX, originY, originX + delta, originY));

			g.draw(new Line2D.Double(cornerX - delta, originY, cornerX, originY));
			g.draw(new Line2D.Double(cornerX, originY, cornerX, originY + delta));

			g.draw(new Line2D.Double(cornerX, cornerY - delta, cornerX, cornerY));
			g.draw(new Line2D.Double(cornerX, cornerY, cornerX - delta, cornerY));

			g.draw(new Line2D.Double(originX + delta, cornerY, originX, cornerY));
			g.draw(new Line2D.Double(originX, cornerY, originX, cornerY - delta));
		}

		return Printable.PAGE_EXISTS;
	}

	/**
	 * Print the receiver on the graphics.
	 * 
	 * @param aGraphics java.awt.Graphics
	 * @category printing
	 */
	protected abstract void print(Graphics aGraphics);

	////////////////////////////////////////////////////////////////////////////////

	private static double DPI = 72;
	private static double MMPI = 25.4;

	/**
	 * Convert the unit of the value from point to inch.
	 * 
	 * @param value double
	 * @return double
	 * @category Utilities
	 */
	public static double POINTtoINCH(double value) {
		return value / DPI;
	}

	/**
	 * Convert the unit of the value from inch to point.
	 * 
	 * @param value double
	 * @return double
	 * @category Utilities
	 */
	public static double INCHtoPOINT(double value) {
		return value * DPI;
	}

	/**
	 * Convert the unit of the value from point to mm.
	 * 
	 * @param value double
	 * @return double
	 * @category Utilities
	 */
	public static double POINTtoMM(double value) {
		return POINTtoINCH(value) * MMPI;
	}

	/**
	 * Convert the unit of the value from mm to point.
	 * 
	 * @param value double
	 * @return double
	 * @category Utilities
	 */
	public static double MMtoPOINT(double value) {
		return INCHtoPOINT(value) / MMPI;
	}

	/**
	 * Answer the ID string of the page.
	 * 
	 * @param row int
	 * @param column int
	 * @return java.lang.String
	 * @category Utilities
	 */
	public static String PageID(int row, int column) {
		StringBuffer buf = new StringBuffer();

		int c = column;
		do {
			buf.append((char) ('A' + (c % 26)));
			c = c / 26 - 1;
		} while (c >= 0);

		buf = buf.reverse();

		buf.append('-');
		buf.append(row + 1);

		return buf.toString();
	}

	/**
	 * Print the information about the paper.
	 * 
	 * @param aPaper java.awt.print.Paper
	 * @category Utilities
	 */
	public static void _Print(Paper aPaper) {
		System.out.print("Paper (");
		System.out.print("width=" + POINTtoMM(aPaper.getWidth()));
		System.out.print(",");
		System.out.print("height=" + POINTtoMM(aPaper.getHeight()));
		System.out.print(" [");
		System.out.print("x=" + POINTtoMM(aPaper.getImageableX()));
		System.out.print(",");
		System.out.print("y=" + POINTtoMM(aPaper.getImageableY()));
		System.out.print(",");
		System.out.print("w=" + POINTtoMM(aPaper.getImageableWidth()));
		System.out.print(",");
		System.out.print("h=" + POINTtoMM(aPaper.getImageableHeight()));
		System.out.print(",");
		System.out.print("R=" + POINTtoMM(aPaper.getWidth() - aPaper.getImageableX() - aPaper.getImageableWidth()));
		System.out.print(",");
		System.out.print("B=" + POINTtoMM(aPaper.getHeight() - aPaper.getImageableY() - aPaper.getImageableHeight()));
		System.out.println("])");
	}

}
