package jp.co.sra.jun.goodies.tables;

import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.TreeSet;
import java.util.Map.Entry;

import jp.co.sra.smalltalk.StBlockClosure;
import jp.co.sra.smalltalk.StBlockClosure3;
import jp.co.sra.smalltalk.StSymbol;

import jp.co.sra.jun.goodies.lisp.JunLispCons;
import jp.co.sra.jun.goodies.lisp.JunLispList;
import jp.co.sra.jun.system.framework.JunAbstractObject;
import jp.co.sra.jun.system.generics.JunPairObject;
import jp.co.sra.jun.system.generics.JunTrioObject;

/**
 * JunAdjacencyMatrix class
 * 
 *  @author    nisinaka
 *  @created   2006/04/06 (by nisinaka)
 *  @updated   2008/05/21 (by nisinaka)
 *  @version   758 (with StPL9.7) based on Jun697 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunAdjacencyMatrix.java,v 9.16 2010/03/09 09:17:24 nisinaka Exp $
 */
public class JunAdjacencyMatrix<N, A> extends JunAbstractObject {

	protected boolean isUndirectedGraph;
	protected JunAdjacencyTable<N, A> adjacencyTable;
	protected ArrayList<N> cachedNodes;
	protected ArrayList<Arc> cachedArcs;
	protected ArrayList<N> cachedRoots;
	protected HashMap<N, ArrayList<Arc>> cachedNodeToArcs;

	/**
	 * Create a new instance of <code>JunAdjacencyMatrix</code> and initialize it.
	 * 
	 * @category Instance creation
	 */
	public JunAdjacencyMatrix() {
		super();
	}

	/**
	 * Create a new instance of <code>JunAdjacencyMatrix</code> and initialize it.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category Instance creation
	 */
	public JunAdjacencyMatrix(JunLispList aList) {
		this();
		this.fromLispList(aList);
	}

	/**
	 * Initialize the receiver.
	 * 
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#initialize()
	 * @category initialize-release
	 */
	protected void initialize() {
		super.initialize();

		isUndirectedGraph = false;
		adjacencyTable = new JunAdjacencyTable<N, A>();
		this.flushCachedAll();
	}

	/**
	 * Set the receiver to be a directed graph.
	 * 
	 * @category accessing
	 */
	public void beDirectedGraph() {
		if (this.isDirectedGraph()) {
			return;
		}

		isUndirectedGraph = false;
	}

	/**
	 * Set the receiver to be an undirected graph.
	 * 
	 * @category accessing
	 */
	public void beUndirectedGraph() {
		if (this.isUndirectedGraph()) {
			return;
		}

		isUndirectedGraph = true;

		ArrayList<Arc> arcs = this._arcs();
		for (int i = 0; i < arcs.size(); i++) {
			Arc arc = arcs.get(i);
			JunAttributeTable<StSymbol, A> first2last = arc.attribute;
			JunAttributeTable<StSymbol, A> last2first = adjacencyTable.at_with_(arc.last, arc.first);
			JunAttributeTable<StSymbol, A> attributes = first2last;
			if (first2last != null && last2first != null) {
				attributes = first2last.merge_(last2first);
			}
			adjacencyTable.at_with_put_(arc.first, arc.last, attributes);
			adjacencyTable.at_with_put_(arc.last, arc.first, attributes);
		}
		this.flushCachedAll();
	}

	/**
	 * Answer the number of nodes.
	 * 
	 * @return int
	 * @category accessing
	 */
	public int numberOfNodes() {
		return this._nodes().size();
	}

	/**
	 * Answer the number of arcs.
	 * 
	 * @return int
	 * @category accessing
	 */
	public int numberOfArcs() {
		int numberOfArcs = this._arcs().size();
		if (this.isUndirectedGraph()) {
			numberOfArcs /= 2;
		}
		return numberOfArcs;
	}

	/**
	 * Answer the number of roots.
	 * 
	 * @return int
	 * @category accessing
	 */
	public int numberOfRoots() {
		return this._roots().size();
	}

	/**
	 * Answer the minimum number of arcs.
	 * 
	 * @return int
	 * @category accessing
	 */
	public int minimumNumberOfArcs() {
		return 0;
	}

	/**
	 * Answer the maximum number of arcs.
	 * 
	 * @return int
	 * @category accessing
	 */
	public int maximumNumberOfArcs() {
		int maximumNumberOfArcs = this.size() * (this.size() - 1);
		if (this.isUndirectedGraph()) {
			maximumNumberOfArcs /= 2;
		}
		return maximumNumberOfArcs;
	}

	/**
	 * Answer my size.
	 * 
	 * @return int
	 * @category accessing
	 */
	public int size() {
		return adjacencyTable.size();
	}

	/**
	 * Answer my current nodes.
	 * 
	 * @return java.lang.Object[]
	 * @category accessing nodes
	 */
	public Object[] nodes() {
		return this._nodes().toArray();
	}

	/**
	 * Answer my current nodes as an array list.
	 * 
	 * @return java.util.ArrayList
	 * @category accessing nodes
	 */
	public ArrayList<N> _nodes() {
		if (cachedNodes == null) {
			TreeSet<N> keys = new TreeSet<N>(new Comparator<N>() {
				@SuppressWarnings("unchecked")
				public int compare(N o1, N o2) {
					if (o1 instanceof Comparable) {
						return ((Comparable) o1).compareTo(o2);
					} else if (o2 instanceof Comparable) {
						return ((Comparable) o2).compareTo(o1) * -1;
					} else {
						return o1.toString().compareTo(o2.toString());
					}
				}
			});
			keys.addAll(adjacencyTable.keys());
			cachedNodes = new ArrayList<N>(keys);
		}
		return cachedNodes;
	}

	/**
	 * Answer the nodes in breadth first order.
	 * 
	 * @return java.lang.Object[]
	 * @category accessing nodes
	 */
	public Object[] nodesBreadthFirst() {
		return this.nodesBreadthFirst_(null);
	}

	/**
	 * Answer the nodes in breadth first order.
	 * 
	 * @return java.util.ArrayList
	 * @category accessing nodes
	 */
	public ArrayList<N> _nodesBreadthFirst() {
		return this._nodesBreadthFirst_(null);
	}

	/**
	 * Answer the nodes in breadth first order.
	 * 
	 * @param startNode java.lang.Object
	 * @return java.lang.Object[]
	 * @category accessing nodes
	 */
	public Object[] nodesBreadthFirst_(N startNode) {
		return this._nodesBreadthFirst_(startNode).toArray();
	}

	/**
	 * Answer the nodes in breadth first order as an array list.
	 * 
	 * @param startNode java.lang.Object
	 * @return java.util.ArrayList
	 * @category accessing nodes
	 */
	public ArrayList<N> _nodesBreadthFirst_(N startNode) {
		final ArrayList<N> aList = new ArrayList<N>(this.numberOfNodes());
		this.nodesBreadthFirstDo_startNode_(new StBlockClosure3<Void, N, Integer, Integer>() {
			public Void value(N node, Integer indent, Integer sequence) {
				aList.add(node);
				return null;
			}
		}, startNode);
		return aList;
	}

	/**
	 * Answer the nodes in depth first order.
	 * 
	 * @return java.lang.Object[]
	 * @category accessing nodes
	 */
	public Object[] nodesDepthFirst() {
		return this.nodesDepthFirst_(null);
	}

	/**
	 * Answer the nodes in depth first order as an array list.
	 * 
	 * @return java.util.ArrayList
	 * @category accessing nodes
	 */
	public ArrayList<N> _nodesDepthFirst() {
		return this._nodesDepthFirst_(null);
	}

	/**
	 * Answer the nodes in depth first order.
	 * 
	 * @param startNode java.lang.Object
	 * @return java.lang.ArrayList
	 * @category accessing nodes
	 */
	public Object[] nodesDepthFirst_(N startNode) {
		return this._nodesDepthFirst_(startNode).toArray();
	}

	/**
	 * Answer the nodes in depth first order as an array list.
	 * 
	 * @param startNode java.lang.Object
	 * @return java.util.ArrayList
	 * @category accessing nodes
	 */
	public ArrayList<N> _nodesDepthFirst_(N startNode) {
		final ArrayList<N> aList = new ArrayList<N>(this.numberOfNodes());
		this.nodesDepthFirstDo_startNode_(new StBlockClosure3<Void, N, Integer, Integer>() {
			public Void value(N node, Integer indent, Integer sequence) {
				aList.add(node);
				return null;
			}
		}, startNode);
		return aList;
	}

	/**
	 * Answer the nodes in partial order.
	 * 
	 * @return java.lang.Object[]
	 * @category accessing nodes
	 */
	public Object[] nodesPartialOrder() {
		return this.nodesPartialOrder_(null);
	}

	/**
	 * Answer the nodes in partial order as an array list.
	 * 
	 * @return java.util.ArrayList
	 * @category accessing nodes
	 */
	public ArrayList<N> _nodesPartialOrder() {
		return this._nodesPartialOrder_(null);
	}

	/**
	 * Answer the nodes in partial order.
	 * 
	 * @param startNode java.lang.Object
	 * @return java.lang.Object[]
	 * @category accessing nodes
	 */
	public Object[] nodesPartialOrder_(N startNode) {
		return this._nodesPartialOrder_(startNode).toArray();
	}

	/**
	 * Answer the nodes in partial order as an array list.
	 * 
	 * @param startNode java.lang.Object
	 * @return java.util.ArrayList
	 * @category accessing nodes
	 */
	public ArrayList<N> _nodesPartialOrder_(N startNode) {
		final ArrayList<N> aList = new ArrayList<N>(this.numberOfNodes());
		this.nodesPartialOrderDo_startNode_(new StBlockClosure3<Void, N, Integer, Integer>() {
			public Void value(N node, Integer indent, Integer sequence) {
				aList.add(node);
				return null;
			}
		}, startNode);
		return aList;
	}

	/**
	 * Answer the nodes in topological sort.
	 * 
	 * @return java.lang.Object[]
	 * @category accessing nodes
	 */
	public Object[] nodesTopologicalSort() {
		return this.nodesPartialOrder();
	}

	/**
	 * Answer the nodes in topological sort as na array list.
	 * 
	 * @return java.util.ArrayList
	 * @category accessing nodes
	 */
	public ArrayList<N> _nodesTopologicalSort() {
		return this._nodesPartialOrder();
	}

	/**
	 * Answer the nodes in topological sort.
	 * 
	 * @param startNode java.lang.Object
	 * @return java.lang.Object[]
	 * @category accessing nodes
	 */
	public Object[] nodesTopologicalSort_(N startNode) {
		return this.nodesPartialOrder_(startNode);
	}

	/**
	 * Answer the nodes in topological sort as an array list.
	 * 
	 * @param startNode java.lang.Object
	 * @return java.util.ArrayList
	 * @category accessing nodes
	 */
	public ArrayList<N> _nodesTopologicalSort_(N startNode) {
		return this._nodesPartialOrder_(startNode);
	}

	/**
	 * Answer my current roots.
	 * 
	 * @return java.lang.Object[]
	 * @category accessing nodes
	 */
	public Object[] roots() {
		return this._roots().toArray();
	}

	/**
	 * Answer my current roots as an array list.
	 * 
	 * @return java.util.ArrayList
	 * @category accessing nodes
	 */
	public ArrayList<N> _roots() {
		if (cachedRoots == null) {
			if (this.isEmpty()) {
				cachedRoots = new ArrayList<N>();
			} else {
				HashSet<N> aSet = new HashSet<N>();
				ArrayList<Arc> arcs = this._arcs();
				for (int i = 0; i < arcs.size(); i++) {
					aSet.add(arcs.get(i).last);
				}

				ArrayList<N> aList = new ArrayList<N>();
				ArrayList<N> nodes = this._nodes();
				for (int i = 0; i < nodes.size(); i++) {
					N node = nodes.get(i);
					if (aSet.contains(node) == false) {
						aList.add(node);
					}
				}

				cachedRoots = aList;
			}
		}
		return cachedRoots;
	}

	/**
	 * Answer the shortest paths of all nodes.
	 * 
	 * @return java.util.Map
	 * @category accessing nodes
	 */
	public Map<N, Map<N, Integer>> shortestPaths() {
		HashMap<N, Map<N, Integer>> aMap = new HashMap<N, Map<N, Integer>>(this.numberOfNodes());
		ArrayList<N> nodes = this._nodes();
		for (int i = 0; i < nodes.size(); i++) {
			N node = nodes.get(i);
			aMap.put(node, this.shortestPathFrom_(node));
		}
		return aMap;
	}

	/**
	 * Answer the shortest path from the specified node.
	 * 
	 * @param startNode java.lang.Object
	 * @return java.util.Map
	 * @category  accessing nodes
	 */
	@SuppressWarnings("unchecked")
	public Map<N, Integer> shortestPathFrom_(N startNode) {
		return (Map<N, Integer>) this.visitDijkstra_(startNode).at_($("distanceTable"));
	}

	/**
	 * Answer the reachable nodes from the specified node.
	 * 
	 * @param startNode java.lang.Object
	 * @return java.lang.Object[]
	 * @category accessing nodes
	 */
	public Object[] reachableNodesFrom_(N startNode) {
		return this._reachableNodesFrom_(startNode).toArray();
	}

	/**
	 * Answer the reachable nodes from the specified node as an array list.
	 * 
	 * @param startNode java.lang.Object
	 * @return java.util.ArrayList
	 * @category accessing nodes
	 */
	public ArrayList<N> _reachableNodesFrom_(final N startNode) {
		final ArrayList<N> aList = new ArrayList<N>(this.numberOfNodes());
		this.visitDepthFirst_nodeDo_arcDo_(startNode, new StBlockClosure3<Void, N, Integer, Integer>() {
			public Void value(N node, Integer indent, Integer sequence) {
				if (node.equals(startNode) == false) {
					aList.add(node);
				}
				return null;
			}
		}, null);
		return aList;
	}

	/**
	 * Answer my current arcs.
	 * 
	 * @return java.lang.Object[]
	 * @deprecated since Jun701, use _arcs()
	 * @category accessing arcs
	 */
	public Object[] arcs() {
		ArrayList<Arc> arcs = this._arcs();
		Object[] objects = new Object[arcs.size()];
		for (int i = 0; i < objects.length; i++) {
			objects[i] = arcs.get(i).toArray();
		}
		return objects;
	}

	/**
	 * Answer my current arcs.
	 * 
	 * @return java.util.ArrayList
	 * @category accessing arcs
	 */
	public ArrayList<Arc> _arcs() {
		if (cachedArcs == null) {
			TreeSet<Arc> arcs = new TreeSet<Arc>(new Comparator<Arc>() {
				@SuppressWarnings("unchecked")
				public int compare(Arc arc1, Arc arc2) {
					N first1 = arc1.first;
					N first2 = arc2.first;
					N last1 = arc1.last;
					N last2 = arc2.last;

					N key1 = first1;
					N key2 = first2;
					if (first1.equals(first2)) {
						key1 = last1;
						key2 = last2;
					}

					if (key1 instanceof Comparable) {
						return ((Comparable) key1).compareTo(key2);
					} else if (key2 instanceof Comparable) {
						return ((Comparable) key2).compareTo(key1) * -1;
					} else {
						return key1.toString().compareTo(key2.toString());
					}
				}
			});

			Iterator<Map.Entry<N, JunReferenceTable<N, JunAttributeTable<StSymbol, A>>>> i = adjacencyTable.entries().iterator();
			while (i.hasNext()) {
				Map.Entry<N, JunReferenceTable<N, JunAttributeTable<StSymbol, A>>> firstEntry = i.next();
				N first = firstEntry.getKey();

				Iterator<Map.Entry<N, JunAttributeTable<StSymbol, A>>> j = firstEntry.getValue().entries().iterator();
				while (j.hasNext()) {
					Map.Entry<N, JunAttributeTable<StSymbol, A>> lastEntry = j.next();
					N last = lastEntry.getKey();

					JunAttributeTable<StSymbol, A> attributes = adjacencyTable.at_with_(first, last);
					if (attributes != null) {
						arcs.add(new Arc(first, attributes, last));
					}
				}
			}

			cachedArcs = new ArrayList<Arc>(arcs);
		}
		return cachedArcs;
	}

	/**
	 * Answer the arcs of the specified node.
	 * 
	 * @param aNode java.lang.Object
	 * @return java.lang.Object[]
	 * @deprecated since Jun701, use _arcsOf_(N)
	 * @category accessing arcs
	 */
	public Object[] arcsOf_(N aNode) {
		ArrayList<Arc> arcs = this._arcsOf_(aNode);
		Object[] objects = new Object[arcs.size()];
		for (int i = 0; i < objects.length; i++) {
			Arc arc = arcs.get(i);
			objects[i] = new Object[] { arc.first, arc.attribute, arc.last };
		}
		return objects;
	}

	/**
	 * Answer the arcs of the specified node.
	 * 
	 * @param aNode java.lang.Object
	 * @return java.util.ArrayList
	 * @category accessing arcs
	 */
	public ArrayList<Arc> _arcsOf_(N aNode) {
		if (cachedNodeToArcs == null) {
			cachedNodeToArcs = new HashMap<N, ArrayList<Arc>>();
		}
		ArrayList<Arc> arcs = cachedNodeToArcs.get(aNode);
		if (arcs == null) {
			TreeSet<Arc> aTreeSet = new TreeSet<Arc>(new Comparator<Arc>() {
				@SuppressWarnings("unchecked")
				public int compare(Arc a1, Arc a2) {
					N last1 = a1.last;
					N last2 = a2.last;

					if (last1 instanceof Comparable) {
						return ((Comparable) last1).compareTo(last2);
					} else if (last2 instanceof Comparable) {
						return ((Comparable) last2).compareTo(last2) * -1;
					} else {
						return last1.toString().compareTo(last2.toString());
					}
				}
			});
			aTreeSet.addAll(this._downArcsOfNode_(aNode));
			arcs = new ArrayList<Arc>(aTreeSet);
			cachedNodeToArcs.put(aNode, arcs);
		}
		return arcs;
	}

	/**
	 * Answer the down arcs of the node.
	 * 
	 * @param aNode java.lang.Object
	 * @return java.lang.Object[]
	 * @deprecated since Jun701, use _downArcsOf_(N)
	 * @category accessing arcs
	 */
	public Object[] downArcsOfNode_(N aNode) {
		ArrayList<Arc> arcs = this._downArcsOfNode_(aNode);
		Object[] objects = new Object[arcs.size()];
		for (int i = 0; i < objects.length; i++) {
			Arc arc = arcs.get(i);
			objects[i] = new Object[] { arc.first, arc.attribute, arc.last };
		}
		return objects;
	}

	/**
	 * Answer the down arcs of the node.
	 * 
	 * @param aNode java.lang.Object
	 * @return java.util.ArrayList
	 * @category accessing arcs
	 */
	public ArrayList<Arc> _downArcsOfNode_(N aNode) {
		ArrayList<Arc> aList = new ArrayList<Arc>();
		ArrayList<Arc> arcs = this._arcs();
		for (int i = 0; i < arcs.size(); i++) {
			Arc arc = arcs.get(i);
			if (arc.first.equals(aNode)) {
				aList.add(arc);
			}
		}
		return aList;
	}

	/**
	 * Answer the up arcs of the node.
	 * 
	 * @param aNode java.lang.Object
	 * @return java.lang.Object[]
	 * @deprecated since Jun701, use _upArcsOf_(N)
	 * @category accessing arcs
	 */
	public Object[] upArcsOfNode_(N aNode) {
		ArrayList<Arc> arcs = this._upArcsOfNode_(aNode);
		Object[] objects = new Object[arcs.size()];
		for (int i = 0; i < objects.length; i++) {
			Arc arc = arcs.get(i);
			objects[i] = new Object[] { arc.first, arc.attribute, arc.last };
		}
		return objects;
	}

	/**
	 * Answer the up arcs of the node.
	 * 
	 * @param aNode java.lang.Object
	 * @return java.util.ArrayList
	 * @category accessing arcs
	 */
	public ArrayList<Arc> _upArcsOfNode_(N aNode) {
		ArrayList<Arc> aList = new ArrayList<Arc>();
		ArrayList<Arc> arcs = this._arcs();
		for (int i = 0; i < arcs.size(); i++) {
			Arc arc = arcs.get(i);
			if (arc.last.equals(aNode)) {
				aList.add(arc);
			}
		}
		return aList;
	}

	/**
	 * Answer the attributes of the arc between the nodes.
	 * 
	 * @param first java.lang.Object
	 * @param last java.lang.Object
	 * @return jp.co.sra.jun.goodies.tables.JunAttributeTable
	 * @category accessing arcs
	 */
	public JunAttributeTable<StSymbol, A> between_and_(N first, N last) {
		return adjacencyTable.at_with_(first, last);
	}

	/**
	 * Add the node.
	 * 
	 * @param node java.lang.Object
	 * @return jp.co.sra.jun.goodies.tables.JunReferenceTable
	 * @category adding
	 */
	public JunReferenceTable<N, JunAttributeTable<StSymbol, A>> add_(N node) {
		JunReferenceTable<N, JunAttributeTable<StSymbol, A>> table = adjacencyTable.add_(node);
		if (table != null) {
			this.flushCachedNodes();
		}
		return table;
	}

	/**
	 * Remove the node.
	 * 
	 * @param node java.lang.Object
	 * @return jp.co.sra.jun.goodies.tables.JunReferenceTable
	 * @category removing
	 */
	public Object remove_(N node) {
		Object result = adjacencyTable.remove_(node);
		if (result != null) {
			this.flushCachedAll();
			result = node;
		}
		return result;
	}

	/**
	 * Connect the first object with the last object.
	 * 
	 * @param first java.lang.Object
	 * @param last java.lang.Object
	 * @return java.lang.Boolean
	 * @category connecting 
	 */
	public Boolean connect_with_(N first, N last) {
		Boolean triplet = adjacencyTable.connect_with_undirected_(first, last, this.isUndirectedGraph());
		if (triplet != null && triplet.booleanValue()) {
			this.flushCachedAll();
		}
		return triplet;
	}

	/**
	 * Connect the first object with the last object, and specifies its attribute.
	 * 
	 * @param first java.lang.Object
	 * @param last java.lang.Object
	 * @param attribute java.util.Map.Entry
	 * @return java.lang.Boolean
	 * @category connecting
	 */
	public Boolean connect_with_attribute_(N first, N last, Map.Entry<StSymbol, A> attribute) {
		return this.connect_with_key_value_(first, last, attribute.getKey(), attribute.getValue());
	}

	/**
	 * Connect the first object with the last object, and specifies its attribute.
	 * 
	 * @param first java.lang.Object
	 * @param last java.lang.Object
	 * @param key java.lang.Object
	 * @param value java.lang.Object
	 * @return java.lang.Boolean
	 * @category connecting
	 */
	public Boolean connect_with_key_value_(N first, N last, StSymbol key, A value) {
		Boolean triplet = this.connect_with_(first, last);
		if (triplet != null) {
			JunAttributeTable<StSymbol, A> attributes = adjacencyTable.at_with_(first, last);
			attributes.at_put_(key, value);
			if (this.isUndirectedGraph()) {
				attributes = adjacencyTable.at_with_(last, first);
				attributes.at_put_(key, value);
			}
		}
		return triplet;
	}

	/**
	 * Connect the first object with the last object, and specifies its cost.
	 * 
	 * @param first java.lang.Object
	 * @param last java.lang.Object
	 * @param value int
	 * @return java.lang.Boolean
	 * @category connecting
	 */
	public Boolean connect_with_cost_(N first, N last, A value) {
		return this.connect_with_key_value_(first, last, $("cost"), value);
	}

	/**
	 * Disconnect the first object with the last object.
	 * 
	 * @param first java.lang.Object
	 * @param last java.lang.Object
	 * @return java.lang.Boolean
	 * @category connecting 
	 */
	public Boolean disconnect_with_(N first, N last) {
		Boolean triplet = adjacencyTable.disconnect_with_undirected_(first, last, this.isUndirectedGraph());
		if (triplet != null && triplet.booleanValue()) {
			this.flushCachedAll();
		}
		return triplet;
	}

	/**
	 * Visit the nodes in the breadth first order.
	 * 
	 * @param aNode java.lang.Object
	 * @param nodeBlock jp.co.sra.smalltalk.StBlockClosure
	 * @param arcBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return jp.co.sra.jun.goodies.tables.JunAttributeTable
	 * @category visiting
	 */
	@SuppressWarnings("unchecked")
	public JunAttributeTable<StSymbol, Object> visitBreadthFirst_nodeDo_arcDo_(N aNode, StBlockClosure<Void, N, Integer, Integer> nodeBlock, StBlockClosure<Void, Arc, Integer, Integer> arcBlock) {
		JunAttributeTable<StSymbol, Object> visitingAttributes = new JunAttributeTable<StSymbol, Object>();
		visitingAttributes.at_put_($("nodeSequenceNumber"), new Integer(0));
		visitingAttributes.at_put_($("arcSequenceNumber"), new Integer(0));
		visitingAttributes.at_put_($("nodesBreadthFirst"), new ArrayList<JunTrioObject<N, Integer, Integer>>(this.numberOfNodes()));
		visitingAttributes.at_put_($("arcsBreadthFirst"), new ArrayList<JunTrioObject<Arc, Integer, Integer>>(this.numberOfArcs()));
		visitingAttributes.at_put_($("closedPath"), new ArrayList<N>());
		HashMap<N, StSymbol> visitingTable = new HashMap<N, StSymbol>();
		ArrayList<N> nodes = this._nodes();
		for (int i = 0; i < nodes.size(); i++) {
			visitingTable.put(nodes.get(i), $("unvisited"));
		}
		visitingAttributes.at_put_($("visitingTable"), visitingTable);

		LinkedList<JunPairObject<N, Integer>> graphQueue = new LinkedList<JunPairObject<N, Integer>>();

		if (aNode == null) {
			ArrayList<N> roots = this._roots();
			for (int i = 0; i < roots.size(); i++) {
				N fromNode = roots.get(i);
				if (((HashMap<N, StSymbol>) visitingAttributes.at_($("visitingTable"))).get(fromNode) == $("unvisited")) {
					this.visitBreadthFirst_fromNode_graphQueue_nodeDo_arcDo_(visitingAttributes, fromNode, graphQueue, nodeBlock, arcBlock);
				}
			}
			for (int i = 0; i < nodes.size(); i++) {
				N fromNode = nodes.get(i);
				if (((HashMap<N, StSymbol>) visitingAttributes.at_($("visitingTable"))).get(fromNode) == $("unvisited")) {
					this.visitBreadthFirst_fromNode_graphQueue_nodeDo_arcDo_(visitingAttributes, fromNode, graphQueue, nodeBlock, arcBlock);
				}
			}
		} else {
			if (((HashMap<N, StSymbol>) visitingAttributes.at_($("visitingTable"))).get(aNode) == $("unvisited")) {
				this.visitBreadthFirst_fromNode_graphQueue_nodeDo_arcDo_(visitingAttributes, aNode, graphQueue, nodeBlock, arcBlock);
			}
		}

		return visitingAttributes;
	}

	/**
	 * Visit the nodes in the breadth first order.
	 * 
	 * @param visitingAttributes jp.co.sra.jun.goodies.tables.JunAttributeTable
	 * @param fromNode java.lang.Object
	 * @param graphQueue java.util.LinkedList
	 * @param nodeBlock jp.co.sra.smalltalk.StBlockClosure
	 * @param arcBlock jp.co.sra.smalltalk.StBlockClosure
	 * @category visiting
	 */
	@SuppressWarnings("unchecked")
	protected void visitBreadthFirst_fromNode_graphQueue_nodeDo_arcDo_(JunAttributeTable<StSymbol, Object> visitingAttributes, N fromNode, LinkedList<JunPairObject<N, Integer>> graphQueue, StBlockClosure<Void, N, Integer, Integer> nodeBlock,
			StBlockClosure<Void, Arc, Integer, Integer> arcBlock) {
		ArrayList<JunTrioObject<N, Integer, Integer>> nodesBreadthFirst = (ArrayList<JunTrioObject<N, Integer, Integer>>) visitingAttributes.at_($("nodesBreadthFirst"));
		ArrayList<JunTrioObject<Arc, Integer, Integer>> arcsBreadthFirst = (ArrayList<JunTrioObject<Arc, Integer, Integer>>) visitingAttributes.at_($("arcsBreadthFirst"));
		ArrayList<N> closedPath = (ArrayList<N>) visitingAttributes.at_($("closedPath"));
		HashMap<N, StSymbol> visitingTable = (HashMap<N, StSymbol>) visitingAttributes.at_($("visitingTable"));
		visitingTable.put(fromNode, $("entered"));
		graphQueue.add(new JunPairObject<N, Integer>(fromNode, 0));
		while (graphQueue.isEmpty() == false) {
			JunPairObject<N, Integer> aPair = graphQueue.removeFirst();
			N theNode = aPair.first();
			int indentLevel = aPair.second();
			visitingAttributes.at_put_($("nodeSequenceNumber"), new Integer(((Number) visitingAttributes.at_($("nodeSequenceNumber"))).intValue() + 1));
			JunTrioObject<N, Integer, Integer> nodeArgs = new JunTrioObject<N, Integer, Integer>(theNode, indentLevel, (Integer) visitingAttributes.at_($("nodeSequenceNumber")));
			nodesBreadthFirst.add(nodeArgs);
			if (nodeBlock != null) {
				nodeBlock.value(nodeArgs.first(), nodeArgs.second(), nodeArgs.third());
			}
			visitingTable.put(theNode, $("visited"));
			ArrayList<Arc> arcs = this._arcsOf_(theNode);
			for (int i = 0; i < arcs.size(); i++) {
				Arc arc = arcs.get(i);
				N toNode = arc.last;
				visitingAttributes.at_put_($("arcSequenceNumber"), new Integer(((Number) visitingAttributes.at_($("arcSequenceNumber"))).intValue() + 1));
				JunTrioObject<Arc, Integer, Integer> arcArgs = new JunTrioObject<Arc, Integer, Integer>(arc, indentLevel, (Integer) visitingAttributes.at_($("arcSequenceNumber")));
				arcsBreadthFirst.add(arcArgs);
				if (arcBlock != null) {
					arcBlock.value(arcArgs.first(), arcArgs.second(), arcArgs.third());
				}
				if (visitingTable.get(toNode) == $("unvisited")) {
					graphQueue.add(new JunPairObject<N, Integer>(toNode, new Integer(indentLevel + 1)));
					visitingTable.put(toNode, $("entered"));
				} else {
					closedPath.add(toNode);
				}
			}
		}
	}

	/**
	 * Visit the nodes in the depth first order.
	 * 
	 * @param aNode java.lang.Object
	 * @param nodeBlock jp.co.sra.smalltalk.StBlockClosure
	 * @param arcBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return jp.co.sra.jun.goodies.tables.JunAttributeTable
	 * @category visiting
	 */
	@SuppressWarnings("unchecked")
	public JunAttributeTable<StSymbol, Object> visitDepthFirst_nodeDo_arcDo_(N aNode, StBlockClosure3<Void, N, Integer, Integer> nodeBlock, StBlockClosure3<Void, Arc, Integer, Integer> arcBlock) {
		JunAttributeTable<StSymbol, Object> visitingAttributes = new JunAttributeTable<StSymbol, Object>();
		visitingAttributes.at_put_($("nodeSequenceNumber"), new Integer(0));
		visitingAttributes.at_put_($("arcSequenceNumber"), new Integer(0));
		visitingAttributes.at_put_($("nodesDepthFirst"), new ArrayList<Object[]>(this.numberOfNodes()));
		visitingAttributes.at_put_($("arcsDepthFirst"), new ArrayList<Object[]>(this.numberOfArcs()));
		visitingAttributes.at_put_($("topologicalSort"), new ArrayList<N>(this.numberOfNodes()));
		visitingAttributes.at_put_($("closedPath"), new ArrayList<N>());
		HashMap<N, StSymbol> visitingTable = new HashMap<N, StSymbol>();
		ArrayList<N> nodes = this._nodes();
		for (int i = 0; i < nodes.size(); i++) {
			visitingTable.put(nodes.get(i), $("unvisited"));
		}
		visitingAttributes.at_put_($("visitingTable"), visitingTable);

		if (aNode == null) {
			ArrayList<N> roots = this._roots();
			for (int i = 0; i < roots.size(); i++) {
				N fromNode = roots.get(i);
				if (((HashMap<N, StSymbol>) visitingAttributes.at_($("visitingTable"))).get(fromNode) == $("unvisited")) {
					this.visitDepthFirst_fromNode_indentLevel_nodeDo_arcDo_(visitingAttributes, fromNode, 0, nodeBlock, arcBlock);
				}
			}
			for (int i = 0; i < nodes.size(); i++) {
				N fromNode = nodes.get(i);
				if (((HashMap<N, StSymbol>) visitingAttributes.at_($("visitingTable"))).get(fromNode) == $("unvisited")) {
					this.visitDepthFirst_fromNode_indentLevel_nodeDo_arcDo_(visitingAttributes, fromNode, 0, nodeBlock, arcBlock);
				}
			}
		} else {
			if (((HashMap<N, StSymbol>) visitingAttributes.at_($("visitingTable"))).get(aNode) == $("unvisited")) {
				this.visitDepthFirst_fromNode_indentLevel_nodeDo_arcDo_(visitingAttributes, aNode, 0, nodeBlock, arcBlock);
			}
		}

		return visitingAttributes;
	}

	/**
	 * Visit the nodes in the depth first order.
	 * 
	 * @param visitingAttributes jp.co.sra.jun.goodies.tables.JunAttributeTable
	 * @param fromNode java.lang.Object
	 * @param indentLevel int
	 * @param nodeBlock jp.co.sra.smalltalk.StBlockClosure
	 * @param arcBlock jp.co.sra.smalltalk.StBlockClosure
	 * @category visiting
	 */
	@SuppressWarnings("unchecked")
	protected void visitDepthFirst_fromNode_indentLevel_nodeDo_arcDo_(JunAttributeTable<StSymbol, Object> visitingAttributes, N fromNode, int indentLevel, StBlockClosure<Void, N, Integer, Integer> nodeBlock,
			StBlockClosure<Void, Arc, Integer, Integer> arcBlock) {
		ArrayList<JunTrioObject<N, Integer, Integer>> nodesDepthFirst = (ArrayList<JunTrioObject<N, Integer, Integer>>) visitingAttributes.at_($("nodesDepthFirst"));
		ArrayList<JunTrioObject<Arc, Integer, Integer>> arcsDepthFirst = (ArrayList<JunTrioObject<Arc, Integer, Integer>>) visitingAttributes.at_($("arcsDepthFirst"));
		ArrayList<N> topologicalSort = (ArrayList<N>) visitingAttributes.at_($("topologicalSort"));
		ArrayList<N> closedPath = (ArrayList<N>) visitingAttributes.at_($("closedPath"));
		HashMap<N, StSymbol> visitingTable = (HashMap<N, StSymbol>) visitingAttributes.at_($("visitingTable"));
		visitingAttributes.at_put_($("nodeSequenceNumber"), new Integer(((Number) visitingAttributes.at_($("nodeSequenceNumber"))).intValue() + 1));
		JunTrioObject<N, Integer, Integer> nodeArgs = new JunTrioObject<N, Integer, Integer>(fromNode, indentLevel, (Integer) visitingAttributes.at_($("nodeSequenceNumber")));
		nodesDepthFirst.add(nodeArgs);
		if (nodeBlock != null) {
			nodeBlock.value(nodeArgs.first(), nodeArgs.second(), nodeArgs.third());
		}
		visitingTable.put(fromNode, $("visited"));
		ArrayList<Arc> arcs = this._arcsOf_(fromNode);
		for (int i = 0; i < arcs.size(); i++) {
			Arc arc = arcs.get(i);
			N toNode = arc.last;
			visitingAttributes.at_put_($("arcSequenceNumber"), new Integer(((Number) visitingAttributes.at_($("arcSequenceNumber"))).intValue() + 1));
			JunTrioObject<Arc, Integer, Integer> arcArgs = new JunTrioObject<Arc, Integer, Integer>(arc, indentLevel, (Integer) visitingAttributes.at_($("arcSequenceNumber")));
			arcsDepthFirst.add(arcArgs);
			if (arcBlock != null) {
				arcBlock.value(arcArgs.first(), arcArgs.second(), arcArgs.third());
			}
			if (visitingTable.get(toNode) == $("unvisited")) {
				this.visitDepthFirst_fromNode_indentLevel_nodeDo_arcDo_(visitingAttributes, toNode, indentLevel + 1, nodeBlock, arcBlock);
			} else {
				closedPath.add(toNode);
			}
		}
		topologicalSort.add(fromNode);
	}

	/**
	 * Visit the nodes in Dijkstra's algorithm.
	 * 
	 * @param aNode java.lang.Object
	 * @return jp.co.sra.jun.goodies.tables.JunAttributeTable
	 * @category visiting
	 */
	public JunAttributeTable<StSymbol, Object> visitDijkstra_(N aNode) {
		return this.visitDijkstra_attributeSymbol_(aNode, $("cost"));
	}

	/**
	 * Visit the nodes in Dijkstra's algorithm.
	 * 
	 * @param aNode java.lang.Object
	 * @param aSymbol jp.co.sra.smalltalk.StSymbol
	 * @return jp.co.sra.jun.goodies.tables.JunAttributeTable
	 * @category visiting
	 */
	public JunAttributeTable<StSymbol, Object> visitDijkstra_attributeSymbol_(N aNode, StSymbol aSymbol) {
		HashMap<N, Integer> distanceTable = new HashMap<N, Integer>();
		ArrayList<N> nodes = this._nodes();
		for (int i = 0; i < nodes.size(); i++) {
			distanceTable.put(nodes.get(i), null);
		}

		JunAttributeTable<StSymbol, Object> visitingAttributes = new JunAttributeTable<StSymbol, Object>();
		visitingAttributes.at_put_($("distanceTable"), distanceTable);
		visitingAttributes.at_put_($("visitedNodes"), new HashSet<N>());
		visitingAttributes.at_put_($("unvisitedNodes"), new HashSet<N>());

		if (aNode != null) {
			this.visitDijkstra_fromNode_attributeSymbol_(visitingAttributes, aNode, aSymbol);
		}

		return visitingAttributes;
	}

	/**
	 * Visit the nodes in Dijkstra's algorithm.
	 * 
	 * @param visitingAttributes jp.co.sra.jun.goodies.tables.JunAttributeTable
	 * @param fromNode java.lang.Object
	 * @param aSymbol jp.co.sra.smalltalk.StSymbol
	 * @category visiting
	 */
	@SuppressWarnings("unchecked")
	protected void visitDijkstra_fromNode_attributeSymbol_(JunAttributeTable<StSymbol, Object> visitingAttributes, N fromNode, StSymbol aSymbol) {
		Collection<N> visitedNodes = (Collection<N>) visitingAttributes.at_($("visitedNodes"));
		Collection<N> unvisitedNodes = (Collection<N>) visitingAttributes.at_($("unvisitedNodes"));
		Map<N, Integer> distanceTable = (Map<N, Integer>) visitingAttributes.at_($("distanceTable"));
		distanceTable.put(fromNode, new Integer(0));
		visitedNodes.add(fromNode);
		ArrayList<N> reachableNodes = this._reachableNodesFrom_(fromNode);
		for (int i = 0; i < reachableNodes.size(); i++) {
			distanceTable.put(reachableNodes.get(i), null);
			unvisitedNodes.add(reachableNodes.get(i));
		}
		ArrayList<Arc> arcs = this._arcsOf_(fromNode);
		for (int i = 0; i < arcs.size(); i++) {
			Arc arc = arcs.get(i);
			N toNode = arc.last;
			JunAttributeTable<StSymbol, A> attributes = arc.attribute;
			int length = 1;
			if (attributes.at_(aSymbol) != null) {
				length = ((Number) attributes.at_(aSymbol)).intValue();
				length = Math.max(length, 0);
			}
			distanceTable.put(toNode, new Integer(length));
		}
		while (unvisitedNodes.isEmpty() == false) {
			N aNode = null;
			Number minDistance = null;
			Iterator<N> iterator = unvisitedNodes.iterator();
			while (iterator.hasNext()) {
				N node = iterator.next();
				Integer theDistance = distanceTable.get(node);
				if (theDistance != null) {
					if (minDistance == null) {
						minDistance = theDistance;
						aNode = node;
					} else {
						if (theDistance.intValue() < minDistance.intValue()) {
							minDistance = theDistance;
							aNode = node;
						}
					}
				}
			}
			if (aNode == null) {
				aNode = unvisitedNodes.iterator().next();
			}
			unvisitedNodes.remove(aNode);
			visitedNodes.add(aNode);

			arcs = this._arcsOf_(aNode);
			for (int i = 0; i < arcs.size(); i++) {
				Arc arc = arcs.get(i);
				N toNode = arc.last;
				if (unvisitedNodes.contains(toNode)) {
					JunAttributeTable<StSymbol, A> attributes = arc.attribute;
					int length = 1;
					if (attributes.at_(aSymbol) != null) {
						length = ((Number) attributes.at_(aSymbol)).intValue();
						length = Math.max(length, 0);
					}
					Integer toDistance = distanceTable.get(toNode);
					if (toDistance == null) {
						Number aDistance = distanceTable.get(aNode);
						if (aDistance == null) {
							//
						} else {
							length += aDistance.intValue();
						}
					} else {
						Integer aDistance = distanceTable.get(aNode);
						if (aDistance == null) {
							length = Math.min(toDistance, length);
						} else {
							length = Math.min(toDistance, length + aDistance);
						}
					}
					distanceTable.put(toNode, length);
				}
			}
		}
	}

	/**
	 * Enumerate nodes in breadth first order and evaluate the block with the node.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @category enumerating nodes
	 */
	public void nodesBreadthFirstDo_(StBlockClosure3<Void, N, Integer, Integer> aBlock) {
		this.nodesBreadthFirstDo_startNode_(aBlock, null);
	}

	/**
	 * Enumerate nodes in breadth first order and evaluate the block with the node.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @category enumerating nodes
	 * @deprecated since Jun701, use Generics.
	 */
	public void nodesBreadthFirstDo_(final StBlockClosure<?, ?, ?, ?> aBlock) {
		this.nodesBreadthFirstDo_(new StBlockClosure3<Void, N, Integer, Integer>() {
			public Void value(N node, Integer indent, Integer sequence) {
				aBlock.value_value_value_(node, indent, sequence);
				return null;
			}
		});
	}

	/**
	 * Enumerate nodes in breadth first order and evaluate the block with the node.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @param startNode java.lang.Object
	 * @category enumerating nodes
	 */
	@SuppressWarnings("unchecked")
	public void nodesBreadthFirstDo_startNode_(StBlockClosure3<Void, N, Integer, Integer> aBlock, N startNode) {
		JunAttributeTable<StSymbol, Object> attributes = this.visitBreadthFirst_nodeDo_arcDo_(startNode, null, null);
		ArrayList<JunTrioObject<N, Integer, Integer>> aList = (ArrayList<JunTrioObject<N, Integer, Integer>>) attributes.at_($("nodesBreadthFirst"));
		for (int i = 0; i < aList.size(); i++) {
			JunTrioObject<N, Integer, Integer> nodeArgs = aList.get(i);
			aBlock.value(nodeArgs.first(), nodeArgs.second(), nodeArgs.third());
		}
	}

	/**
	 * Enumerate nodes in depth first order and evaluate the block with the node.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @category enumerating nodes
	 */
	public void nodesDepthFirstDo_(StBlockClosure3<Void, N, Integer, Integer> aBlock) {
		this.nodesDepthFirstDo_startNode_(aBlock, null);
	}

	/**
	 * Enumerate nodes in depth first order and evaluate the block with the node.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @category enumerating nodes
	 * @deprecated since Jun701, use Generics.
	 */
	public void nodesDepthFirstDo_(final StBlockClosure<?, ?, ?, ?> aBlock) {
		this.nodesDepthFirstDo_(new StBlockClosure3<Void, N, Integer, Integer>() {
			public Void value(N node, Integer indent, Integer sequence) {
				aBlock.value_value_value_(node, indent, sequence);
				return null;
			}
		});
	}

	/**
	 * Enumerate nodes in depth first order and evaluate the block with the node.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @param startNode java.lang.Object
	 * @category enumerating nodes
	 */
	@SuppressWarnings("unchecked")
	public void nodesDepthFirstDo_startNode_(StBlockClosure3<Void, N, Integer, Integer> aBlock, N startNode) {
		JunAttributeTable<StSymbol, Object> attributes = this.visitDepthFirst_nodeDo_arcDo_(startNode, null, null);
		ArrayList<JunTrioObject<N, Integer, Integer>> aList = (ArrayList<JunTrioObject<N, Integer, Integer>>) attributes.at_($("nodesDepthFirst"));
		for (int i = 0; i < aList.size(); i++) {
			JunTrioObject<N, Integer, Integer> nodeArgs = aList.get(i);
			aBlock.value(nodeArgs.first(), nodeArgs.second(), nodeArgs.third());
		}
	}

	/**
	 * Enumerate nodes in partial order and evaluate the block with the node.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @category enumerating nodes
	 */
	public void nodesPartialOrderDo_(StBlockClosure3<Void, N, Integer, Integer> aBlock) {
		this.nodesPartialOrderDo_startNode_(aBlock, null);
	}

	/**
	 * Enumerate nodes in partial order and evaluate the block with the node.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @category enumerating nodes
	 * @deprecated since Jun701, use Generics.
	 */
	public void nodesPartialOrderDo_(final StBlockClosure<?, ?, ?, ?> aBlock) {
		this.nodesPartialOrderDo_(new StBlockClosure3<Void, N, Integer, Integer>() {
			public Void value(N node, Integer indent, Integer sequence) {
				aBlock.value_value_value_(node, indent, sequence);
				return null;
			}
		});
	}

	/**
	 * Enumerate nodes in partial order and evaluate the block with the node.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @param startNode java.lang.Object
	 * @category enumerating nodes
	 */
	@SuppressWarnings("unchecked")
	public void nodesPartialOrderDo_startNode_(StBlockClosure3<Void, N, Integer, Integer> aBlock, N startNode) {
		final HashMap<N, Integer[]> visitingDictionary = new HashMap<N, Integer[]>(this.size());
		JunAttributeTable<StSymbol, Object> visitingAttributes = this.visitDepthFirst_nodeDo_arcDo_(startNode, new StBlockClosure3<Void, N, Integer, Integer>() {
			public Void value(N node, Integer indent, Integer sequence) {
				visitingDictionary.put(node, new Integer[] { indent, sequence });
				return null;
			}
		}, null);

		ArrayList<N> aList = (ArrayList<N>) visitingAttributes.at_($("topologicalSort"));
		for (int i = 0; i < aList.size(); i++) {
			N node = aList.get(i);
			Integer[] anArray = visitingDictionary.get(node);
			Integer indent = anArray[0];
			Integer sequence = anArray[1];
			aBlock.value(node, indent, sequence);
		}
	}

	/**
	 * Enumerate nodes in topological sort order and evaluate the block with the node.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @category enumerating nodes
	 */
	public void nodesTopologicalSortDo_(StBlockClosure3<Void, N, Integer, Integer> aBlock) {
		this.nodesPartialOrderDo_(aBlock);
	}

	/**
	 * Enumerate nodes in topological sort order and evaluate the block with the node.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @category enumerating nodes
	 * @deprecated since Jun701, use Generics.
	 */
	public void nodesTopologicalSortDo_(final StBlockClosure<?, ?, ?, ?> aBlock) {
		this.nodesTopologicalSortDo_(new StBlockClosure3<Void, N, Integer, Integer>() {
			public Void value(N node, Integer indent, Integer sequence) {
				aBlock.value_value_value_(node, indent, sequence);
				return null;
			}
		});
	}

	/**
	 * Enumerate nodes in topological sort order and evaluate the block with the node.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @param startNode java.lang.Object
	 * @category enumerating nodes
	 */
	public void nodesTopologicalSortDo_startNode_(StBlockClosure3<Void, N, Integer, Integer> aBlock, N startNode) {
		this.nodesPartialOrderDo_startNode_(aBlock, startNode);
	}

	/**
	 * Enumerate arcs in the breadth first order and evaluate the block with the arc.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @category enumerating arcs
	 */
	public void arcsBreadthFirstDo_(StBlockClosure3<Void, Arc, Integer, Integer> aBlock) {
		this.arcsBreadthFirstDo_startNode_(aBlock, null);
	}

	/**
	 * Enumerate arcs in the breadth first order and evaluate the block with the arc.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @category enumerating arcs
	 * @deprecated since Jun701, use Generics.
	 */
	public void arcsBreadthFirstDo_(final StBlockClosure<?, ?, ?, ?> aBlock) {
		this.arcsBreadthFirstDo_(new StBlockClosure3<Void, Arc, Integer, Integer>() {
			public Void value(Arc arc, Integer indent, Integer sequence) {
				aBlock.value_value_value_(arc, indent, sequence);
				return null;
			}
		});
	}

	/**
	 * Enumerate arcs in the breadth first order and evaluate the block with the arc.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @param startNode java.lang.Object
	 * @category enumerating arcs
	 */
	@SuppressWarnings("unchecked")
	public void arcsBreadthFirstDo_startNode_(StBlockClosure3<Void, Arc, Integer, Integer> aBlock, N startNode) {
		JunAttributeTable<StSymbol, Object> attributes = this.visitBreadthFirst_nodeDo_arcDo_(startNode, null, null);
		ArrayList<JunTrioObject<Arc, Integer, Integer>> aList = (ArrayList<JunTrioObject<Arc, Integer, Integer>>) attributes.at_($("arcsBreadthFirst"));
		for (int i = 0; i < aList.size(); i++) {
			JunTrioObject<Arc, Integer, Integer> arcArgs = aList.get(i);
			aBlock.value(arcArgs.first(), arcArgs.second(), arcArgs.third());
		}
	}

	/**
	 * Enumerate arcs in the depth first order and evaluate the block with the arc.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @category enumerating arcs
	 */
	public void arcsDepthFirstDo_(StBlockClosure3<Void, Arc, Integer, Integer> aBlock) {
		this.arcsDepthFirstDo_startNode_(aBlock, null);
	}

	/**
	 * Enumerate arcs in the depth first order and evaluate the block with the arc.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @category enumerating arcs
	 * @deprecated since Jun701, use Generics.
	 */
	public void arcsDepthFirstDo_(final StBlockClosure<?, ?, ?, ?> aBlock) {
		this.arcsDepthFirstDo_(new StBlockClosure3<Void, Arc, Integer, Integer>() {
			public Void value(Arc arc, Integer indent, Integer sequence) {
				aBlock.value_value_value_(arc, indent, sequence);
				return null;
			}
		});
	}

	/**
	 * Enumerate arcs in the depth first order and evaluate the block with the arc.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @param startNode java.lang.Object
	 * @category enumerating arcs
	 */
	@SuppressWarnings("unchecked")
	public void arcsDepthFirstDo_startNode_(StBlockClosure3<Void, Arc, Integer, Integer> aBlock, N startNode) {
		JunAttributeTable<StSymbol, Object> attributes = this.visitDepthFirst_nodeDo_arcDo_(startNode, null, null);
		ArrayList<JunTrioObject<Arc, Integer, Integer>> aList = (ArrayList<JunTrioObject<Arc, Integer, Integer>>) attributes.at_($("arcsDepthFirst"));
		for (int i = 0; i < aList.size(); i++) {
			JunTrioObject<Arc, Integer, Integer> arcArgs = aList.get(i);
			aBlock.value(arcArgs.first(), arcArgs.second(), arcArgs.third());
		}
	}

	/**
	 * Flush the cached information.
	 * 
	 * @category flushing
	 */
	protected void flushCachedAll() {
		this.flushCachedNodes();
		this.flushCachedArcs();
	}

	/**
	 * Flush the cached nodes.
	 * 
	 * @category flushing
	 */
	protected void flushCachedNodes() {
		cachedNodes = null;
		cachedRoots = null;
	}

	/**
	 * Flush the cached arcs.
	 * 
	 * @category flushing
	 */
	protected void flushCachedArcs() {
		cachedArcs = null;
		cachedRoots = null;
		cachedNodeToArcs = null;
	}

	/**
	 * Answer true if the receiver is set to be a directed graph, otherwise false.
	 * 
	 * @return boolean
	 * @category testing
	 */
	public boolean isDirectedGraph() {
		return !this.isUndirectedGraph();
	}

	/**
	 * Answer true if the receiver is set to be an undirected graph, otherwise false.
	 * 
	 * @return boolean
	 * @category testing
	 */
	public boolean isUndirectedGraph() {
		return isUndirectedGraph;
	}

	/**
	 * Answer true if the receiver has no entities, otherwise false.
	 * 
	 * @return boolean
	 * @category testing
	 */
	public boolean isEmpty() {
		return adjacencyTable.isEmpty();
	}

	/**
	 * Answer true if the receiver is isomorphic to the specified matrix, otherwise false.
	 * 
	 * @param anotherMatrix jp.co.sra.jun.goodies.tables.JunAdjacencyMatrix
	 * @return boolean
	 * @category testing
	 */
	public boolean isIsomorphicTo_(JunAdjacencyMatrix<? extends Object, ? extends Object> aMatrix) {
		if (aMatrix == null) {
			return false;
		}
		if (this.isEmpty() && aMatrix.isEmpty()) {
			return true;
		}
		if (this.numberOfNodes() != aMatrix.numberOfNodes()) {
			return false;
		}
		if (this.numberOfArcs() != aMatrix.numberOfArcs()) {
			return false;
		}
		if (this.numberOfRoots() != aMatrix.numberOfRoots()) {
			return false;
		}
		if (this.numberOfArcs() == 0 && aMatrix.numberOfArcs() == 0) {
			return true;
		}

		JunAdjacencyMatrix<Object, Object> anotherMatrix = new JunAdjacencyMatrix<Object, Object>(aMatrix.toLispList());

		ArrayList<N> nodes = this._nodes();
		for (int i = 0; i < nodes.size(); i++) {
			N node = nodes.get(i);
			ArrayList<?> anotherNodes = anotherMatrix._nodes();
			for (int j = 0; j < anotherNodes.size(); j++) {
				Object anotherNode = anotherNodes.get(j);
				if (this._downArcsOfNode_(node).size() == anotherMatrix._downArcsOfNode_(anotherNode).size() && this._upArcsOfNode_(node).size() == anotherMatrix._upArcsOfNode_(anotherNode).size()) {
					if (this.privateIsomorphicTo_with_and_(anotherMatrix, node, anotherNode)) {
						return true;
					}
				}
			}
		}

		return false;
	}

	/**
	 * Update the unique numbers.
	 * 
	 * @param uniqueNumberTable java.util.Map
	 * @category updating
	 */
	public void _updateUniqueNumbers(Map<N, N> uniqueNumberTable) {
		ArrayList<Entry<N, JunReferenceTable<N, JunAttributeTable<StSymbol, A>>>> adjacencies = adjacencyTable.adjacencies();
		for (int i = 0; i < adjacencies.size(); i++) {
			N oldUniqueNumber = adjacencies.get(i).getKey();
			JunReferenceTable<N, JunAttributeTable<StSymbol, A>> referenceTable = adjacencies.get(i).getValue();
			N newUniqueNumber = uniqueNumberTable.get(oldUniqueNumber);
			adjacencyTable.removeKey_(oldUniqueNumber);
			adjacencyTable.at_put_(newUniqueNumber, referenceTable);

			ArrayList<Entry<N, JunAttributeTable<StSymbol, A>>> references = referenceTable.references();
			for (int j = 0; j < references.size(); j++) {
				oldUniqueNumber = references.get(j).getKey();
				JunAttributeTable<StSymbol, A> attributeTable = references.get(j).getValue();
				newUniqueNumber = uniqueNumberTable.get(oldUniqueNumber);
				referenceTable.removeKey_(oldUniqueNumber);
				referenceTable.at_put_(newUniqueNumber, attributeTable);
			}
		}
		this.flushCachedAll();
	}

	/**
	 * Print my string representation on the writer.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException if failed.
	 * @see jp.co.sra.smalltalk.StObject#printOn_(java.io.Writer)
	 * @category printing
	 */
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write('(');
		aWriter.write(this._className().toString());

		aWriter.write(" (nodes");
		ArrayList<N> nodes = this._nodes();
		int size = nodes.size();
		if (size <= this.printMax()) {
			for (int i = 0; i < size; i++) {
				this.printNodeOn_(nodes.get(i), aWriter);
			}
		} else {
			for (int i = 0; i < this.printMax() / 2; i++) {
				this.printNodeOn_(nodes.get(i), aWriter);
			}
			aWriter.write(" ...");
			for (int i = size - this.printMax() / 2; i < size; i++) {
				this.printNodeOn_(nodes.get(i), aWriter);
			}
		}
		aWriter.write(')');

		aWriter.write(" (arcs");
		ArrayList<Arc> arcs = this._arcs();
		size = arcs.size();
		if (size <= this.printMax()) {
			for (int i = 0; i < size; i++) {
				this.printArcOn_(arcs.get(i), aWriter);
			}
		} else {
			for (int i = 0; i < this.printMax() / 2; i++) {
				this.printArcOn_(arcs.get(i), aWriter);
			}
			aWriter.write(" ...");
			for (int i = size - this.printMax() / 2; i < size; i++) {
				this.printArcOn_(arcs.get(i), aWriter);
			}
		}
		aWriter.write(')');

		aWriter.write(')');
	}

	/**
	 * Print the string representation of the node on the writer.
	 * 
	 * @param node java.lang.Object
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException
	 * @category printing
	 */
	protected void printNodeOn_(N node, Writer aWriter) throws IOException {
		aWriter.write(' ');
		aWriter.write(node.toString());
	}

	/**
	 * Print the string representation of the arc on the writer.
	 * 
	 * @param arc jp.co.sra.jun.goodies.tables.JunAdjacencyMatrix.Arc
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException
	 * @category printing
	 */
	protected void printArcOn_(Arc arc, Writer aWriter) throws IOException {
		aWriter.write(" (");
		aWriter.write(arc.first.toString());
		aWriter.write("=>");
		aWriter.write(arc.last.toString());
		aWriter.write(')');
	}

	/**
	 * Answer the maximum number for printing elements.
	 * 
	 * @return int
	 * @see jp.co.sra.jun.goodies.tables.JunAttributeTable#printMax()
	 * @category printing
	 */
	protected int printMax() {
		return 100;
	}

	/**
	 * Answer the StSymbol which represents the kind of the receiver.
	 * 
	 * @return jp.co.sra.smalltalk.StSymbol
	 * @category lisp support
	 */
	public StSymbol kindName() {
		return this._className();
	}

	/**
	 * Get my attributes from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	public void fromLispList(JunLispList aList) {
		this.undirectedGraphFromLispList(aList);
		this.adjacencyTableFromLispList(aList);
	}

	/**
	 * Convert my attributes to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispCons
	 * @category lisp support
	 */
	public JunLispCons toLispList() {
		JunLispCons aList = new JunLispCons(this.kindName());
		aList.add_(this.undirectedGraphToLispList());
		aList.add_(this.adjacencyTableToLispList());
		return aList;
	}

	/**
	 * Get my undirectedGraph from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected void undirectedGraphFromLispList(JunLispList aList) {
		JunLispList undirectedGraphList = aList._findSublistWithHead($("undirectedGraph"));
		if (undirectedGraphList == null) {
			return;
		}

		Boolean aBoolean = (Boolean) undirectedGraphList.tail();
		if (aBoolean.booleanValue()) {
			this.beUndirectedGraph();
		} else {
			this.beDirectedGraph();
		}
	}

	/**
	 * Convert my undirectedGraph to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispCons
	 * @category lisp support
	 */
	protected JunLispCons undirectedGraphToLispList() {
		return new JunLispCons($("undirectedGraph"), Boolean.valueOf(isUndirectedGraph));
	}

	/**
	 * Get my adjacency table from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected void adjacencyTableFromLispList(JunLispList aList) {
		JunLispList adjacencyTableList = aList._findSublistWithHead($("adjacencyTable"));
		if (adjacencyTableList == null) {
			return;
		}

		adjacencyTable = new JunAdjacencyTable<N, A>((JunLispList) adjacencyTableList.tail());
	}

	/**
	 * Convert my adjacency table to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispCons
	 * @category lisp support
	 */
	protected JunLispCons adjacencyTableToLispList() {
		return new JunLispCons($("adjacencyTable"), adjacencyTable.toLispList());
	}

	/**
	 * Private method to check the two matrixes are isomorphic.
	 * 
	 * @param anotherMatrix jp.co.sra.jun.goodies.tables.JunAdjacencyMatrix
	 * @param aNode java.lang.Object
	 * @param anotherNode java.lang.Object
	 * @return boolean
	 * @category private
	 */
	@SuppressWarnings("unchecked")
	protected boolean privateIsomorphicTo_with_and_(JunAdjacencyMatrix<Object, Object> anotherMatrix, Object aNode, Object anotherNode) {
		JunAdjacencyMatrix<Object, Object> thisMatrix = new JunAdjacencyMatrix<Object, Object>(this.toLispList());

		final Collection<Integer> thisIndentCollection = new ArrayList<Integer>(this.numberOfNodes());
		JunAttributeTable<StSymbol, Object> aTable = thisMatrix.visitDepthFirst_nodeDo_arcDo_(aNode, new StBlockClosure3<Void, Object, Integer, Integer>() {
			public Void value(Object node, Integer indent, Integer sequence) {
				thisIndentCollection.add(indent);
				return null;
			}
		}, null);
		final Collection<Integer> anotherIndentCollection = new ArrayList<Integer>(anotherMatrix.numberOfNodes());
		JunAttributeTable<StSymbol, Object> anotherTable = anotherMatrix.visitDepthFirst_nodeDo_arcDo_(anotherNode, new StBlockClosure3<Void, Object, Integer, Integer>() {
			public Void value(Object node, Integer indent, Integer sequence) {
				anotherIndentCollection.add(indent);
				return null;
			}
		}, null);
		if (thisIndentCollection.equals(anotherIndentCollection) == false) {
			return false;
		}

		Collection<Map.Entry<Object, StSymbol>> thisEntryCollection = new ArrayList<Map.Entry<Object, StSymbol>>(this.numberOfNodes());
		Iterator<Map.Entry<Object, StSymbol>> iterator = ((Map<Object, StSymbol>) aTable.at_($("visitingTable"))).entrySet().iterator();
		while (iterator.hasNext()) {
			Map.Entry<Object, StSymbol> entry = iterator.next();
			if (entry.getValue() == $("unvisited")) {
				thisEntryCollection.add(entry);
			}
		}
		Collection<Map.Entry<Object, StSymbol>> anotherEntryCollection = new ArrayList<Map.Entry<Object, StSymbol>>(this.numberOfNodes());
		iterator = ((Map<Object, StSymbol>) anotherTable.at_($("visitingTable"))).entrySet().iterator();
		while (iterator.hasNext()) {
			Map.Entry<Object, StSymbol> entry = iterator.next();
			if (entry.getValue() == $("unvisited")) {
				anotherEntryCollection.add(entry);
			}
		}
		if (thisEntryCollection.size() != anotherEntryCollection.size()) {
			return false;
		}

		thisEntryCollection.clear();
		iterator = ((Map<Object, StSymbol>) aTable.at_($("visitingTable"))).entrySet().iterator();
		while (iterator.hasNext()) {
			Map.Entry<Object, StSymbol> entry = iterator.next();
			if (entry.getValue() == $("visited")) {
				thisEntryCollection.add(entry);
			}
		}
		anotherEntryCollection.clear();
		iterator = ((Map<Object, StSymbol>) anotherTable.at_($("visitingTable"))).entrySet().iterator();
		while (iterator.hasNext()) {
			Map.Entry<Object, StSymbol> entry = iterator.next();
			if (entry.getValue() == $("visited")) {
				anotherEntryCollection.add(entry);
			}
		}

		JunAdjacencyMatrix<Object, Object> aCopiedMatrix = (JunAdjacencyMatrix<Object, Object>) this.copy();
		JunAdjacencyMatrix<Object, Object> anotherCopiedMatrix = (JunAdjacencyMatrix<Object, Object>) anotherMatrix.copy();
		iterator = thisEntryCollection.iterator();
		while (iterator.hasNext()) {
			Map.Entry<Object, StSymbol> entry = iterator.next();
			aCopiedMatrix.remove_(entry.getKey());
		}
		iterator = anotherEntryCollection.iterator();
		while (iterator.hasNext()) {
			Map.Entry<Object, StSymbol> entry = iterator.next();
			anotherCopiedMatrix.remove_(entry.getKey());
		}

		return aCopiedMatrix.isIsomorphicTo_(anotherCopiedMatrix);
	}

	////////////////////////////////////////////////////////////////////////////////

	public class Arc {
		protected N first;
		protected N last;
		protected JunAttributeTable<StSymbol, A> attribute;

		Arc(N first, JunAttributeTable<StSymbol, A> attribute, N last) {
			this.first = first;
			this.last = last;
			this.attribute = attribute;
		}

		public N first() {
			return first;
		}

		public N last() {
			return last;
		}

		public JunAttributeTable<StSymbol, A> attribute() {
			return attribute;
		}

		public Object[] toArray() {
			return new Object[] { first, attribute, last };
		}
	}

}
