package jp.co.sra.jun.goodies.tables;

import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import jp.co.sra.smalltalk.StBlockClosure;
import jp.co.sra.smalltalk.StBlockClosure0;
import jp.co.sra.smalltalk.StBlockClosure1;
import jp.co.sra.smalltalk.StObject;
import jp.co.sra.smalltalk.StSymbol;

import jp.co.sra.jun.goodies.lisp.JunLispCons;
import jp.co.sra.jun.goodies.lisp.JunLispConverter;
import jp.co.sra.jun.goodies.lisp.JunLispList;
import jp.co.sra.jun.goodies.lisp.JunLispParser;
import jp.co.sra.jun.system.framework.JunAbstractObject;

/**
 * JunAttributeTable class
 * 
 *  @author    nisinaka
 *  @created   2006/04/05 (by nisinaka)
 *  @updated   2008/05/21 (by nisinaka)
 *  @version   758 (with StPL9.7) based on Jun721 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunAttributeTable.java,v 9.15 2010/03/09 09:17:24 nisinaka Exp $
 */
public class JunAttributeTable<K, V> extends JunAbstractObject {

	protected HashMap<K, V> attributeTable;

	/**
	 * Create a new instance of JunAttributeTable from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @return jp.co.sra.jun.goodies.tables.JunAttributeTable
	 * @category Lisp support
	 */
	public static <K, V> JunAttributeTable<K, V> FromLispList_(JunLispList aList) {
		final JunAttributeTable<K, V> targetAttributes = new JunAttributeTable<K, V>();

		JunLispCons pair = aList._findSublistWithHead($("attributeTable"));
		if (pair != null) {
			((JunLispCons) pair.tail()).do_(new StBlockClosure1<Void, Object>() {
				@SuppressWarnings("unchecked")
				public Void value(Object anObject) {
					if (anObject instanceof JunLispCons) {
						JunLispCons cons = (JunLispCons) anObject;
						Object object = null;
						if (cons.tail() instanceof JunLispList) {
							object = JunLispConverter.Convert((JunLispList) cons.tail());
						}
						if (object == null) {
							object = cons.tail();
						}

						targetAttributes.at_put_((K) cons.head(), (V) object);
					}
					return null;
				}
			});
		}

		return targetAttributes;
	}

	/**
	 * Create a new instance of JunAttributeTable from the map.
	 * 
	 * @param aMap java.util.Map<K, V>
	 * @return jp.co.sra.jun.goodies.tables.JunAttributeTable
	 * @category Instance creation
	 */
	public static <K, V> JunAttributeTable<K, V> FromMap_(Map<K, V> aMap) {
		return new JunAttributeTable<K, V>(aMap);
	}

	/**
	 * Create a new instance of JunAttributeTable from the string.
	 * 
	 * @param aString java.lang.String
	 * @return jp.co.sra.jun.goodies.tables.JunAttributeTable
	 * @category Instance creation
	 */
	public static <K, V> JunAttributeTable<K, V> FromString_(String aString) {
		JunLispList aList = (JunLispList) JunLispParser.Parse_(aString);
		return JunAttributeTable.<K, V> FromLispList_(aList);
	}

	/**
	 * Create a new instance of JunAttributeTable and initialize it.
	 *
	 * @category Instance creation
	 */
	public JunAttributeTable() {
		this(0);
	}

	/**
	 * Create a new instance of JunAttributeTable and initialize it.
	 *
	 * @param size int
	 * @category Instance creation
	 */
	public JunAttributeTable(int size) {
		super();
		this.initialize(size);
	}

	/**
	 * Create a new instance of JunAttributeTable and initialize it.
	 *
	 * @param aMap java.util.Map
	 * @category Instance creation
	 */
	public JunAttributeTable(Map<K, V> aMap) {
		this(aMap.size());

		Iterator<Map.Entry<K, V>> i = aMap.entrySet().iterator();
		while (i.hasNext()) {
			Map.Entry<K, V> e = i.next();
			this.at_put_(e.getKey(), e.getValue());
		}
	}

	/**
	 * Create a new instance of JunAttributeTable and initialize it.
	 *
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category Instance creation
	 */
	public JunAttributeTable(JunLispList aList) {
		this();
		this.fromLispList(aList);
	}

	/**
	 * Initialize the receiver.
	 * 
	 * @param size int
	 * @category initialize-release
	 */
	protected void initialize(int size) {
		attributeTable = new HashMap<K, V>(Math.max(size, 0));
	}

	/**
	 * Answer the current attribute table.
	 * 
	 * @return java.util.Map
	 * @category accessing
	 */
	public Map<K, V> attributeTable() {
		return attributeTable;
	}

	/**
	 * Answer the size of the receiver.
	 * 
	 * @return int
	 * @category accessing
	 */
	public int size() {
		return attributeTable.size();
	}

	/**
	 * Answer the keys of the attribute table.
	 * 
	 * @return java.util.Set
	 * @category accessing
	 */
	public Set<K> keys() {
		return attributeTable.keySet();
	}

	/**
	 * Answer the entries of the attribute table.
	 * This method corresponds to 'associations' in Smalltalk version.
	 * 
	 * @return java.util.Set
	 * @category accessing
	 */
	public Set<Map.Entry<K, V>> entries() {
		return attributeTable.entrySet();
	}

	/**
	 * Answer my current bindings.
	 * 
	 * @return java.util.ArrayList
	 * @category accessing
	 */
	public ArrayList<Map.Entry<K, V>> bindings() {
		TreeSet<Map.Entry<K, V>> bindings = new TreeSet<Map.Entry<K, V>>(new Comparator<Map.Entry<K, V>>() {
			@SuppressWarnings("unchecked")
			public int compare(Map.Entry<K, V> o1, Map.Entry<K, V> o2) {
				K key1 = o1.getKey();
				K key2 = o2.getKey();
				if (key1 instanceof Comparable) {
					return ((Comparable<K>) key1).compareTo(key2);
				} else if (key2 instanceof Comparable) {
					return ((Comparable<K>) key2).compareTo(key1) * -1;
				} else {
					return key1.toString().compareTo(key2.toString());
				}
			}
		});
		bindings.addAll(this.entries());
		return new ArrayList<Map.Entry<K, V>>(bindings);
	}

	/**
	 * Answer the current value for the key.
	 * 
	 * @param key java.lang.Object
	 * @return java.lang.Object
	 * @category accessing
	 */
	public V at_(K key) {
		return attributeTable.get(key);
	}

	/**
	 * Answer the value at key.  If key is not found, answer the result of
	 * evaluation valueBlock.
	 * 
	 * @param key java.lang.Object
	 * @param valueBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category accessing
	 */
	public V at_ifAbsent_(K key, StBlockClosure0<V> valueBlock) {
		V obj = attributeTable.get(key);
		if (obj == null) {
			return valueBlock.value();
		} else {
			return obj;
		}
	}

	/**
	 * Answer the value at key.  If key is not found, answer the result of
	 * evaluation valueBlock.
	 * 
	 * @param key java.lang.Object
	 * @param valueBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category accessing
	 * @deprecated since Jun701, use Generics.
	 */
	public V at_ifAbsent_(K key, final StBlockClosure<?, ?, ?, ?> valueBlock) {
		return this.at_ifAbsent_(key, new StBlockClosure0<V>() {
			@SuppressWarnings("unchecked")
			public V value() {
				return (V) valueBlock.value();
			}
		});
	}

	/**
	 * Answer the value at key.  If key is not found, create a new entry
	 * for key, set its value to be the result of evaluating valueBlock,
	 * and answer the value.
	 * 
	 * @param key java.lang.Object
	 * @param valueBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category accessing
	 */
	public V at_ifAbsentPut_(final K key, final StBlockClosure0<V> valueBlock) {
		return this.at_ifAbsent_(key, new StBlockClosure0<V>() {
			public V value() {
				V value = valueBlock.value();
				JunAttributeTable.this.at_put_(key, value);
				return value;
			}
		});
	}

	/**
	 * Answer the value at key.  If key is not found, create a new entry
	 * for key, set its value to be the result of evaluating valueBlock,
	 * and answer the value.
	 * 
	 * @param key java.lang.Object
	 * @param valueBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.lang.Object
	 * @category accessing
	 * @deprecated since Jun701, use Generics.
	 */
	public V at_ifAbsentPut_(final K key, final StBlockClosure<?, ?, ?, ?> valueBlock) {
		return this.at_ifAbsentPut_(key, new StBlockClosure0<V>() {
			@SuppressWarnings("unchecked")
			public V value() {
				return (V) valueBlock.value();
			}
		});
	}

	/**
	 * Set the new value for the key.
	 * 
	 * @param key java.lang.Object
	 * @param value java.lang.Object
	 * @category accessing
	 */
	public void at_put_(K key, V value) {
		attributeTable.put(key, value);
	}

	/**
	 * Merge with the specified attributes.
	 * 
	 * @param anotherAttributes jp.co.sra.jun.goodies.tables.JunAttributeTable
	 * @return jp.co.sra.jun.goodies.tables.JunAttributeTable
	 * @category accessing
	 */
	@SuppressWarnings("unchecked")
	public JunAttributeTable<K, V> merge_(JunAttributeTable<K, V> anotherAttributes) {
		JunAttributeTable<K, V> mergedAttributes = (JunAttributeTable<K, V>) this.copy();

		Iterator<Map.Entry<K, V>> i = anotherAttributes.entries().iterator();
		while (i.hasNext()) {
			Map.Entry<K, V> entry = i.next();
			if (mergedAttributes.includesKey_(entry.getKey()) == false) {
				mergedAttributes.at_put_(entry.getKey(), entry.getValue());
			}
		}

		return mergedAttributes;
	}

	/**
	 * Add the entry.
	 * 
	 * @param anEntry java.util.Map.Entry
	 * @category adding
	 */
	public void add_(Map.Entry<K, V> anEntry) {
		this.at_put_(anEntry.getKey(), anEntry.getValue());
	}

	/**
	 * Remove the entry with the specified key.
	 * 
	 * @param key java.lang.Object
	 * @return java.lang.Object
	 * @category removing 
	 */
	public V removeKey_(K key) {
		return attributeTable.remove(key);
	}

	/**
	 * Remove all entries.
	 * 
	 * @category removing
	 */
	public void removeAll() {
		attributeTable.clear();
	}

	/**
	 * Convert to a map.
	 * 
	 * @return java.util.Map
	 * @category converting
	 */
	public Map<K, V> asDictionary() {
		return this.toMap();
	}

	/**
	 * Convert the receiver to the string.
	 * 
	 * @return java.lang.String
	 * @category converting
	 */
	public String asString() {
		return this.saveString();
	}

	/**
	 * Convert to a map.
	 * 
	 * @return java.util.Map
	 * @category converting
	 */
	public Map<K, V> toMap() {
		return new HashMap<K, V>(this.attributeTable());
	}

	/**
	 * Answer true if the receiver has no entities, otherwise false.
	 * 
	 * @return boolean
	 * @category testing
	 */
	public boolean isEmpty() {
		return this.size() == 0;
	}

	/**
	 * Answer true if the receiver has an entry with the specified key, otherwise false.
	 * 
	 * @param key java.lang.Object
	 * @return boolean
	 * @category testing
	 */
	public boolean includesKey_(K key) {
		return attributeTable.containsKey(key);
	}

	/**
	 * Answer a copy of receiver.
	 * 
	 * @return jp.co.sra.smalltalk.StObject
	 * @see jp.co.sra.smalltalk.StObject#copy()
	 * @category copying
	 */
	@SuppressWarnings("unchecked")
	public StObject copy() {
		try {
			JunAttributeTable<K, V> anAttributeTable = this.getClass().newInstance();
			anAttributeTable.fromLispList(this.toLispList());
			return anAttributeTable.postCopy();
		} catch (Exception e) {
			return null;
		}
	}

	/**
	 * Print my string representation on the writer.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException if failed.
	 * @see jp.co.sra.smalltalk.StObject#printOn_(java.io.Writer)
	 * @category printing
	 */
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write('(');
		aWriter.write(this._className().toString());
		if (this.isEmpty() == false) {
			aWriter.write(" (bindings");

			ArrayList<Map.Entry<K, V>> bindings = this.bindings();
			int size = bindings.size();
			if (size <= this.printMax()) {
				for (int i = 0; i < size; i++) {
					this.printEntryOn_(bindings.get(i), aWriter);
				}
			} else {
				for (int i = 0; i < this.printMax() / 2; i++) {
					this.printEntryOn_(bindings.get(i), aWriter);
				}
				aWriter.write(" ...");
				for (int i = size - this.printMax() / 2; i < size; i++) {
					this.printEntryOn_(bindings.get(i), aWriter);
				}
			}

			aWriter.write(')');
		}
		aWriter.write(')');
	}

	/**
	 * Print the entry on the writer.
	 * 
	 * @param anEntry java.util.Map.Entry
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException
	 * @category printing
	 */
	protected void printEntryOn_(Map.Entry<K, V> anEntry, Writer aWriter) throws IOException {
		aWriter.write(" (");
		aWriter.write(anEntry.getKey().toString());
		aWriter.write("->");
		aWriter.write(anEntry.getValue().toString());
		aWriter.write(')');
	}

	/**
	 * Answer the maximum number for printing elements.
	 * 
	 * @return int
	 * @category printing
	 */
	protected int printMax() {
		return 30;
	}

	/**
	 * Answer the save string.
	 * 
	 * @return java.lang.String
	 * @category printing
	 */
	protected String saveString() {
		JunLispCons list = JunLispCons.Cell();
		list.head_(this._className());
		list.add_(this.attributeTableToLispList());
		return list.saveString();
	}

	/**
	 * Print my storable string representation on aWriter.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException
	 * @see jp.co.sra.smalltalk.StObject#storeOn_(java.io.Writer)
	 * @category printing
	 */
	public void storeOn_(Writer aWriter) throws IOException {
		aWriter.write("((");
		aWriter.write(this._className().toString());
		aWriter.write(" fromString: '");
		aWriter.write(this.saveString());
		aWriter.write("') yourself)");
	}

	/**
	 * Answer the StSymbol which represents the kind of the receiver.
	 * 
	 * @return jp.co.sra.smalltalk.StSymbol
	 * @category lisp support
	 */
	public StSymbol kindName() {
		return this._className();
	}

	/**
	 * Get my attributes from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	public void fromLispList(JunLispList aList) {
		this.sizeFromLispList(aList);
		this.attributeTableFromLispList(aList);
	}

	/**
	 * Convert my attributes to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispCons
	 * @category lisp support
	 */
	public JunLispCons toLispList() {
		JunLispCons aList = new JunLispCons(this.kindName());
		aList.add_(this.sizeToLispList());
		aList.add_(this.attributeTableToLispList());
		return aList;
	}

	/**
	 * Get my size from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected void sizeFromLispList(JunLispList aList) {
		JunLispList sizeList = aList._findSublistWithHead($("size"));
		if (sizeList == null) {
			return;
		}

		this.initialize(((Number) sizeList.tail()).intValue());
	}

	/**
	 * Convert my size to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispCons
	 * @category lisp support
	 */
	protected JunLispCons sizeToLispList() {
		return new JunLispCons($("size"), new Integer(this.size()));
	}

	/**
	 * Get my attribute table from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	@SuppressWarnings("unchecked")
	protected void attributeTableFromLispList(JunLispList aList) {
		JunLispList attributeTableList = aList._findSublistWithHead($("attributeTable"));
		if (attributeTableList == null) {
			return;
		}

		Object[] entities = ((JunLispList) attributeTableList.tail()).asArray();
		for (int i = 0; i < entities.length; i++) {
			JunLispCons cons = (JunLispCons) entities[i];
			V value = (V) cons.tail();
			if (value instanceof JunLispCons) {
				// TODO: respondsTo: #fromLispList:
			}
			this.at_put_((K) cons.head(), value);
		}
	}

	/**
	 * Convert my attribute table to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispCons
	 * @category lisp support
	 */
	protected JunLispCons attributeTableToLispList() {
		JunLispCons attributeTableList = new JunLispCons($("attributeTable"));

		ArrayList<Map.Entry<K, V>> bindings = this.bindings();
		for (int i = 0; i < bindings.size(); i++) {
			Object value = bindings.get(i).getValue();
			if (value instanceof StObject && ((StObject) value).respondsTo_("toLispList")) {
				try {
					value = ((StObject) value).perform_("toLispList");
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			attributeTableList.add_(new JunLispCons(bindings.get(i).getKey(), value));
		}

		return attributeTableList;
	}

}
