package jp.co.sra.jun.goodies.utilities;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.IndexColorModel;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;

import jp.co.sra.smalltalk.StBlockClosure;
import jp.co.sra.smalltalk.StBlockClosure1;
import jp.co.sra.smalltalk.StDisplayable;
import jp.co.sra.smalltalk.StImage;
import jp.co.sra.smalltalk.StObject;

import jp.co.sra.jun.geometry.basic.Jun2dPoint;
import jp.co.sra.jun.geometry.curves.Jun2dLine;
import jp.co.sra.jun.goodies.cursors.JunCursors;
import jp.co.sra.jun.goodies.files.JunFileModel;
import jp.co.sra.jun.goodies.image.framework.JunImageDisplayModel;
import jp.co.sra.jun.goodies.image.streams.JunBmpImageStream;
import jp.co.sra.jun.goodies.image.streams.JunImageStream;
import jp.co.sra.jun.goodies.movie.framework.JunMovieHandle;
import jp.co.sra.jun.goodies.multimedia.JunMultiMediaImageModel;
import jp.co.sra.jun.graphics.navigator.JunFileRequesterDialog;
import jp.co.sra.jun.system.framework.JunAbstractObject;
import jp.co.sra.jun.system.support.JunSystem;

/**
 * JunImageUtility class
 * 
 *  @author    nisinaka
 *  @created   2005/10/26 (by nisinaka)
 *  @updated   2006/04/07 (by m-asada)
 *  @updated   2008/05/08 (by m-asada)
 *  @updated   2008/11/04 (by m-asada)
 *  @version   758 (with StPL9.7) based on Jun746 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2005 Information-technology Promotion Agency, Japan (IPA)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunImageUtility.java,v 9.14 2010/03/09 09:17:27 nisinaka Exp $
 */
public class JunImageUtility extends JunAbstractObject {

	/**
	 * Answer the blend color with the specified two colors.
	 * 
	 * @param firstColor java.awt.Color
	 * @param secondColor java.awt.Color
	 * @return java.awt.Color
	 * @category Blending
	 */
	public static Color Blend_and_(Color firstColor, Color secondColor) {
		return Blend_and_weight_and_(firstColor, secondColor, 0.5, 0.5);
	}

	/**
	 * Answer the blend color with the specified two colors and weight value.
	 * 
	 * @param firstColor java.awt.Color
	 * @param secondColor java.awt.Color
	 * @param firstWeight double
	 * @return java.awt.Color
	 * @category Blending
	 */
	public static Color Blend_and_weight_(Color firstColor, Color secondColor, double firstWeight) {
		double normalizedWeight = Math.max(0.0, Math.min(firstWeight, 1.0));
		return Blend_and_weight_and_(firstColor, secondColor, firstWeight, normalizedWeight);
	}

	/**
	 * Answer the blend color with the specified two colors and weight values.
	 * 
	 * @param firstColor java.awt.Color
	 * @param secondColor java.awt.Color
	 * @param firstWeight double
	 * @param secondWeight double
	 * @return java.awt.Color
	 * @category Blending
	 */
	public static Color Blend_and_weight_and_(Color firstColor, Color secondColor, double firstWeight, double secondWeight) {
		double totalWeight = firstWeight + secondWeight;
		double redValue = (firstColor.getRed() * firstWeight + secondColor.getRed() * secondWeight) / totalWeight;
		double greenValue = (firstColor.getGreen() * firstWeight + secondColor.getGreen() * secondWeight) / totalWeight;
		double blueValue = (firstColor.getBlue() * firstWeight + secondColor.getBlue() * secondWeight) / totalWeight;
		return new Color((int) Math.round(redValue), (int) Math.round(greenValue), (int) Math.round(blueValue));
	}

	/**
	 * Answer the blend image with the specified two image and weight value.
	 * 
	 * @param firstImage java.awt.image.BufferedImage
	 * @param secondImage java.awt.image.BufferedImage
	 * @param firstWeight double
	 * @return java.awt.image.BufferedImage
	 * @category Blending
	 */
	public static BufferedImage BlendImage_andImage_weight_(BufferedImage firstImage, BufferedImage secondImage, double firstWeight) {
		double normalizedWeight = Math.max(0.0, Math.min(firstWeight, 1.0));
		return BlendImage_andImage_weight_and_(firstImage, secondImage, firstWeight, normalizedWeight);
	}

	/**
	 * Answer the blend image with the specified two image and weight value.
	 * 
	 * @param firstImage java.awt.image.BufferedImage
	 * @param secondImage java.awt.image.BufferedImage
	 * @param firstWeight double
	 * @param secondWeight double
	 * @return java.awt.image.BufferedImage
	 * @category Blending
	 */
	public static BufferedImage BlendImage_andImage_weight_and_(BufferedImage firstImage, BufferedImage secondImage, double firstWeight, double secondWeight) {
		ColorModel firstPalette = firstImage.getColorModel();
		ColorModel secondPalette = secondImage.getColorModel();
		ColorModel blendPalette = firstPalette;
		BufferedImage blendImage = new BufferedImage(blendPalette, blendPalette.createCompatibleWritableRaster(firstImage.getWidth(), firstImage.getHeight()), false, null);
		int w = secondImage.getWidth() - 1;
		int h = secondImage.getHeight() - 1;
		for (int y = 0; y < firstImage.getHeight(); y++) {
			for (int x = 0; x < firstImage.getWidth(); x++) {
				Color firstColor = new Color(firstPalette.getRed(firstImage.getRGB(x, y)), firstPalette.getGreen(firstImage.getRGB(x, y)), firstPalette.getBlue(firstImage.getRGB(x, y)));
				Color colorValue;
				if ((0 <= x && x <= w) && (0 <= y && y <= h)) {
					Color secondColor = new Color(secondPalette.getRed(secondImage.getRGB(x, y)), firstPalette.getGreen(secondImage.getRGB(x, y)), firstPalette.getBlue(secondImage.getRGB(x, y)));
					colorValue = Blend_and_weight_and_(firstColor, secondColor, firstWeight, secondWeight);
				} else {
					colorValue = firstColor;
				}
				int indexValue;
				if (blendPalette instanceof IndexColorModel) {
					indexValue = StImage._IndexOfNearestColor(colorValue, (IndexColorModel) blendPalette);
				} else {
					indexValue = colorValue.getRGB();
				}
				blendImage.setRGB(x, y, indexValue);
			}
		}
		return blendImage;
	}

	/**
	 * Create an image from display.
	 * 
	 * @return java.awt.image.BufferedImage
	 * @category Creating
	 */
	public static BufferedImage FromDisplay() {
		return FromDisplay_(new Rectangle(new Point(0, 0), Toolkit.getDefaultToolkit().getScreenSize()));
	}

	/**
	 * Create an image from display with the specified bounds.
	 * 
	 * @param aRectangle java.awt.Rectangle
	 * @return java.awt.image.BufferedImage
	 * @category Creating
	 */
	public static BufferedImage FromDisplay_(Rectangle aRectangle) {
		return StImage._OfArea(aRectangle).image();
	}

	/**
	 * Create an image from the file.
	 * 
	 * @return java.awt.image.BufferedImage
	 * @category Creating
	 */
	public static BufferedImage FromFile() {
		JunFileModel.FileType[] fileTypes = new JunFileModel.FileType[] {
				JunFileModel.FileType.All($String("All files")),
				new JunFileModel.FileType($String("Gazou file"), JunSystem.DefaultImageExtensionPatterns()),
				new JunFileModel.FileType($String("<1p> files", null, "GIF"), new String[] { "*.gif", "*.GIF", "*.giff", "*.GIFF" }),
				new JunFileModel.FileType($String("<1p> files", null, "JPEG"), new String[] { "*.jpg", "*.JPG", "*.jpeg", "*.JPEG" }),
				new JunFileModel.FileType($String("<1p> files", null, "PNG"), new String[] { "*.png", "*.PNG" }),
				new JunFileModel.FileType($String("<1p> files", null, "BMP"), new String[] { "*.bmp", "*.BMP" }) };
		File aFile = JunFileRequesterDialog.RequestFile($String("Select a gazou file."), fileTypes, fileTypes[1]);
		if (aFile == null) {
			return null;
		}
		return JunImageUtility.FromFile_(aFile);
	}

	/**
	 * Create an image from the specified file.
	 * 
	 * @param aFile java.io.File
	 * @return java.awt.image.BufferedImage
	 * @category Creating
	 */
	public static BufferedImage FromFile_(File aFile) {
		StImage anImage = null;
		Class<? extends JunImageStream> aClass = JunImageStream.ImageStreamClassForFile_(aFile);
		if (aClass == null) {
			anImage = JunMovieHandle.Filename_(aFile).asImage();
		} else {
			JunImageStream imageStream = null;
			try {
				imageStream = (JunImageStream) StObject._PerformWith(aClass, "On_", new FileInputStream(aFile));
				anImage = imageStream.nextImage();
			} catch (NoSuchMethodException e) {
				System.out.println(e.getMessage());
				e.printStackTrace();
			} catch (IOException e) {
				System.out.println(e.getMessage());
				e.printStackTrace();
			} finally {
				if (imageStream != null) {
					try {
						imageStream.close();
					} catch (IOException e) {
						System.out.println(e.getMessage());
						e.printStackTrace();
					} finally {
						imageStream = null;
					}
				}
			}
		}
		return anImage.image();
	}

	/**
	 * Create an image from the specified filename.
	 * 
	 * @param aFilename java.lang.String
	 * @return java.awt.image.BufferedImage
	 * @category Creating
	 */
	public static BufferedImage FromFilename_(String aFilename) {
		return FromFile_(new File(aFilename));
	}

	/**
	 * Create an image from user.
	 * 
	 * @return java.awt.image.BufferedImage
	 * @category Creating
	 */
	public static BufferedImage FromUser() {
		return StImage._FromUser().image();
	}

	/**
	 * Create an image with the specified extent.
	 * 
	 * @param imageExtent java.awt.Dimension
	 * @return java.awt.image.BufferedImage
	 * @category Creating
	 */
	public static BufferedImage ImageExtent_(Dimension imageExtent) {
		return ImageExtent_displayBlock_(imageExtent, new StBlockClosure1<Void, Graphics>() {
			public Void value(Graphics graphicsContext) {
				return null;
			}
		});
	}

	/**
	 * Create an image with the specified extent.
	 * Apply the display block to the image.
	 * 
	 * @param imageExtent java.awt.Dimension
	 * @param displayBlock jp.co.sra.smalltalk.StBlockClosure
	 * @return java.awt.image.BufferedImage
	 * @deprecated since Jun701, use Generics.
	 * @category Creating
	 */
	public static BufferedImage ImageExtent_displayBlock_(Dimension imageExtent, final StBlockClosure<?, ?, ?, ?> displayBlock) {
		return ImageExtent_displayBlock_(imageExtent, new StBlockClosure1<Void, Graphics>() {
			public Void value(Graphics aGraphics) {
				displayBlock.value_(aGraphics);
				return null;
			}
		});
	}

	/**
	 * Create an image with the specified extent.
	 * Apply the display block to the image.
	 * 
	 * @param imageExtent java.awt.Dimension
	 * @param displayBlock jp.co.sra.smalltalk.StBlockClosure1
	 * @return java.awt.image.BufferedImage
	 * @category Creating
	 */
	public static BufferedImage ImageExtent_displayBlock_(Dimension imageExtent, StBlockClosure1<?, Graphics> displayBlock) {
		BufferedImage anImage = null;

		Graphics aGraphics = null;
		try {
			anImage = new BufferedImage(imageExtent.width, imageExtent.height, BufferedImage.TYPE_INT_ARGB);
			aGraphics = anImage.getGraphics();
			aGraphics.setColor(Color.white);
			aGraphics.fillRect(0, 0, imageExtent.width, imageExtent.height);
			aGraphics.setColor(Color.black);
			if (displayBlock != null) {
				displayBlock.value(aGraphics);
			}
		} finally {
			if (aGraphics != null) {
				aGraphics.dispose();
			}
		}

		return anImage;
	}

	/**
	 * Answer the default image quality value.
	 * 
	 * @return float
	 * @category Defaults
	 */
	public static float DefaultImageQuality() {
		return 0.95f;
	}

	/**
	 * Answer the morphing image from the specified images and lines.
	 * 
	 * @param firstImage java.awt.image.BufferedImage
	 * @param secondImage java.awt.image.BufferedImage
	 * @param fromLine jp.co.sra.jun.geometry.curves.Jun2dLine
	 * @param toLine jp.co.sra.jun.geometry.curves.Jun2dLine
	 * @return java.awt.image.BufferedImage
	 * @category Morphing
	 */
	public static BufferedImage MorphImage_andImage_fromLine_toLine_(BufferedImage firstImage, BufferedImage secondImage, Jun2dLine fromLine, Jun2dLine toLine) {
		return MorphImage_andImage_fromLines_toLines_(firstImage, secondImage, new Jun2dLine[] { fromLine }, new Jun2dLine[] { toLine });
	}

	/**
	 * Answer the morphing image from the specified images and lines.
	 * 
	 * @param firstImage java.awt.image.BufferedImage
	 * @param secondImage java.awt.image.BufferedImage
	 * @param fromLine jp.co.sra.jun.geometry.curves.Jun2dLine
	 * @param toLine jp.co.sra.jun.geometry.curves.Jun2dLine
	 * @param firstWeight double
	 * @param secondWeight double
	 * @return java.awt.image.BufferedImage
	 * @category Morphing
	 */
	public static BufferedImage MorphImage_andImage_fromLines_toLines_weight_and_(BufferedImage firstImage, BufferedImage secondImage, Jun2dLine fromLine, Jun2dLine toLine, double firstWeight, double secondWeight) {
		return MorphImage_andImage_fromLines_toLines_weight_and_(firstImage, secondImage, new Jun2dLine[] { fromLine }, new Jun2dLine[] { toLine }, firstWeight, secondWeight);
	}

	/**
	 * Answer the morphing image from the specified images and lines.
	 * 
	 * @param firstImage java.awt.image.BufferedImage
	 * @param secondImage java.awt.image.BufferedImage
	 * @param fromLines jp.co.sra.jun.geometry.curves.Jun2dLine[]
	 * @param toLines jp.co.sra.jun.geometry.curves.Jun2dLine[]
	 * @return java.awt.image.BufferedImage
	 * @category Morphing
	 */
	public static BufferedImage MorphImage_andImage_fromLines_toLines_(BufferedImage firstImage, BufferedImage secondImage, Jun2dLine[] fromLines, Jun2dLine[] toLines) {
		return MorphImage_andImage_fromLines_toLines_weight_and_(firstImage, secondImage, fromLines, toLines, 0.5, 0.5);
	}

	/**
	 * Answer the morphing image from the specified images and lines.
	 * 
	 * @param firstImage java.awt.image.BufferedImage
	 * @param secondImage java.awt.image.BufferedImage
	 * @param fromLines jp.co.sra.jun.geometry.curves.Jun2dLine[]
	 * @param toLines jp.co.sra.jun.geometry.curves.Jun2dLine[]
	 * @param firstWeight double
	 * @param secondWeight double
	 * @return java.awt.image.BufferedImage
	 * @category Morphing
	 */
	public static BufferedImage MorphImage_andImage_fromLines_toLines_weight_and_(BufferedImage firstImage, BufferedImage secondImage, Jun2dLine[] fromLines, Jun2dLine[] toLines, double firstWeight, double secondWeight) {
		double totalWeight = firstWeight + secondWeight;
		double secondPerTotal = secondWeight / totalWeight;
		double firstPerTotal = firstWeight / totalWeight;
		ArrayList<Jun2dLine> tweenLines = new ArrayList<Jun2dLine>(toLines.length);
		for (int i = 0; i < fromLines.length; i++) {
			Jun2dLine fromLine = fromLines[i];
			Jun2dLine toLine = toLines[i];
			tweenLines.add(fromLine.first().to_(toLine.first()).atT_(secondPerTotal).to_(fromLine.last().to_(toLine.last()).atT_(secondPerTotal)));
		}
		BufferedImage warpedFirstImage = WarpImage_fromLines_toLines_(firstImage, fromLines, tweenLines.toArray(new Jun2dLine[tweenLines.size()]));
		tweenLines = new ArrayList<Jun2dLine>(fromLines.length);
		for (int i = 0; i < toLines.length; i++) {
			Jun2dLine toLine = toLines[i];
			Jun2dLine fromLine = fromLines[i];
			tweenLines.add(toLine.first().to_(fromLine.first()).atT_(firstPerTotal).to_(toLine.last().to_(fromLine.last()).atT_(firstPerTotal)));
		}
		BufferedImage warpedSecondImage = WarpImage_fromLines_toLines_(secondImage, toLines, tweenLines.toArray(new Jun2dLine[tweenLines.size()]));
		return BlendImage_andImage_weight_and_(warpedFirstImage, warpedSecondImage, firstWeight, secondWeight);
	}

	/**
	 * Request image file, and answer it.
	 * 
	 * @return jp.co.sra.smalltalk.StImage
	 * @category Reading
	 */
	public static StImage RequestImage() {
		File aFile = RequestImageFile();
		if (aFile == null) {
			return null;
		}
		JunMultiMediaImageModel imageModel = new JunMultiMediaImageModel(aFile);
		if (imageModel == null) {
			return null;
		}
		StImage anImage = imageModel.visual().asImage();
		return anImage;
	}

	/**
	 * Request image file, and answer it.
	 * 
	 * @return java.io.File
	 * @category Reading
	 */
	public static File RequestImageFile() {
		JunFileModel.FileType[] fileTypes = new JunFileModel.FileType[] {
				new JunFileModel.FileType(JunSystem.$String("<1p> files", null, JunSystem.$String("Image")), JunSystem.DefaultImageExtensionPatterns()),
				new JunFileModel.FileType(JunSystem.$String("<1p> files", null, JunSystem.$String("All")), new String[] { "*" }) };
		String message = JunSystem.$String("Select a <1p> file.", null, JunSystem.$String("Image"));
		File file = JunFileRequesterDialog.Request(message, fileTypes, fileTypes[0]);
		if (file == null) {
			return null;
		}
		return file;
	}

	/**
	 * Open the new application window with the specified image.
	 * 
	 * @param anImage java.awt.Image
	 * @return jp.co.sra.jun.goodies.image.framework.JunImageDisplayModel
	 * @category Showing
	 */
	public static JunImageDisplayModel Show_(Image anImage) {
		return JunImageDisplayModel.Show_(anImage);
	}

	/**
	 * Open the new application window with the specified image.
	 * 
	 * @param anImage jp.co.sra.smalltalk.StDisplayable
	 * @return jp.co.sra.jun.goodies.image.framework.JunImageDisplayModel
	 * @category Showing
	 */
	public static JunImageDisplayModel Show_(StDisplayable anImage) {
		return JunImageDisplayModel.Show_(anImage);
	}

	/**
	 * Open the new application window with the specified image at the specified point.
	 * 
	 * @param anImage java.awt.Image
	 * @param aPoint java.awt.Point
	 * @return jp.co.sra.jun.goodies.image.framework.JunImageDisplayModel
	 * @category Showing
	 */
	public static JunImageDisplayModel Show_at_(Image anImage, Point aPoint) {
		return JunImageDisplayModel.Show_at_(anImage, aPoint);
	}

	/**
	 * Open the new application window with the specified image at the specified point.
	 * 
	 * @param anImage jp.co.sra.smalltalk.StDisplayable
	 * @param aPoint java.awt.Point
	 * @return jp.co.sra.jun.goodies.image.framework.JunImageDisplayModel
	 * @category Showing
	 */
	public static JunImageDisplayModel Show_at_(StDisplayable anImage, Point aPoint) {
		return JunImageDisplayModel.Show_at_(anImage, aPoint);
	}

	/**
	 * Open the new application window with the specified image in the specified area.
	 * 
	 * @param anImage java.awt.Image
	 * @param aRectangle java.awt.Rectangle
	 * @return jp.co.sra.jun.goodies.image.framework.JunImageDisplayModel
	 * @category Showing
	 */
	public static JunImageDisplayModel Show_in_(Image anImage, Rectangle aRectangle) {
		return JunImageDisplayModel.Show_in_(anImage, aRectangle);
	}

	/**
	 * Open the new application window with the specified image in the specified area.
	 * 
	 * @param anImage jp.co.sra.smalltalk.StDisplayable
	 * @param aRectangle java.awt.Rectangle
	 * @return jp.co.sra.jun.goodies.image.framework.JunImageDisplayModel
	 * @category Showing
	 */
	public static JunImageDisplayModel Show_in_(StDisplayable anImage, Rectangle aRectangle) {
		return JunImageDisplayModel.Show_in_(anImage, aRectangle);
	}

	/**
	 * Answer the warp image with the specified lines.
	 * 
	 * @param sourceImage java.awt.image.BufferedImage
	 * @param destinationImage java.awt.image.BufferedImage
	 * @param sourceLine jp.co.sra.jun.geometry.curves.Jun2dLine
	 * @param destinationLine jp.co.sra.jun.geometry.curves.Jun2dLine
	 * @return java.awt.image.BufferedImage
	 * @category Warping
	 */
	public static BufferedImage WarpImage_fromLine_toLine_(BufferedImage sourceImage, Jun2dLine sourceLine, Jun2dLine destinationLine) {
		return WarpImage_fromLines_toLines_(sourceImage, new Jun2dLine[] { sourceLine }, new Jun2dLine[] { destinationLine });
	}

	/**
	 * Answer the warp image with the specified lines.
	 * 
	 * @param sourceImage java.awt.image.BufferedImage
	 * @param destinationImage java.awt.image.BufferedImage
	 * @param sourceLines jp.co.sra.jun.geometry.curves.Jun2dLine[]
	 * @param destinationLines jp.co.sra.jun.geometry.curves.Jun2dLine[]
	 * @return java.awt.image.BufferedImage
	 * @category Warping
	 */
	public static BufferedImage WarpImage_fromLines_toLines_(BufferedImage sourceImage, Jun2dLine[] sourceLines, Jun2dLine[] destinationLines) {
		BufferedImage destinationImage = new BufferedImage(sourceImage.getWidth(), sourceImage.getHeight(), sourceImage.getType());
		destinationImage.setData(sourceImage.getData());
		return WarpImage_toImage_fromLines_toLines_(sourceImage, destinationImage, sourceLines, destinationLines);
	}

	/**
	 * Answer the warp image from the destination image with lines.
	 * 
	 * @param sourceImage java.awt.image.BufferedImage
	 * @param destinationImage java.awt.image.BufferedImage
	 * @param sourceLine jp.co.sra.jun.geometry.curves.Jun2dLine
	 * @param destinationLine jp.co.sra.jun.geometry.curves.Jun2dLine
	 * @return java.awt.image.BufferedImage
	 * @category Warping
	 */
	public static BufferedImage WarpImage_toImage_fromLine_toLine_(BufferedImage sourceImage, BufferedImage destinationImage, Jun2dLine sourceLine, Jun2dLine destinationLine) {
		return WarpImage_toImage_fromLines_toLines_(sourceImage, destinationImage, new Jun2dLine[] { sourceLine }, new Jun2dLine[] { destinationLine });
	}

	/**
	 * Answer the warp image from the destination image with lines.
	 * 
	 * @param sourceImage java.awt.image.BufferedImage
	 * @param destinationImage java.awt.image.BufferedImage
	 * @param sourceLines jp.co.sra.jun.geometry.curves.Jun2dLine[]
	 * @param destinationLines jp.co.sra.jun.geometry.curves.Jun2dLine[]
	 * @return java.awt.image.BufferedImage
	 * @category Warping
	 */
	public static BufferedImage WarpImage_toImage_fromLines_toLines_(BufferedImage sourceImage, BufferedImage destinationImage, Jun2dLine[] sourceLines, Jun2dLine[] destinationLines) {
		int w = sourceImage.getWidth() - 1;
		int h = sourceImage.getHeight() - 1;
		// ColorModel s = sourceImage.getColorModel();
		ColorModel d = destinationImage.getColorModel();
		int dw = destinationImage.getWidth();
		int dh = destinationImage.getHeight();
		for (int y = 0; y < dh; y++) {
			for (int x = 0; x < dw; x++) {
				Jun2dPoint aPoint = WarpPoint_fromLines_toLines_(new Jun2dPoint(x, y), destinationLines, sourceLines);
				int xValue = (int) aPoint.rounded().x();
				int yValue = (int) aPoint.rounded().y();
				if ((0 <= xValue && xValue <= w) && (0 <= yValue && yValue <= h)) {
					Color colorValue = new Color(sourceImage.getRGB(xValue, yValue));
					int indexValue;
					if (d instanceof IndexColorModel) {
						indexValue = StImage._IndexOfNearestColor(colorValue, (IndexColorModel) d);
					} else {
						indexValue = colorValue.getRGB();
					}
					destinationImage.setRGB(x, y, indexValue);
				}
			}
		}
		return destinationImage;
	}

	/**
	 * Answer the warp point with the specified lines.
	 * 
	 * @param sourcePoint jp.co.sra.jun.geometry.basic
	 * @param sourceLines jp.co.sra.jun.geometry.curves.Jun2dLine[]
	 * @param destinationLines jp.co.sra.jun.geometry.curves.Jun2dLine[]
	 * @return jp.co.sra.jun.geometry.basic
	 * @category Warping
	 */
	public static Jun2dPoint WarpPoint_fromLines_toLines_(Jun2dPoint sourcePoint, Jun2dLine[] sourceLines, Jun2dLine[] destinationLines) {
		return WarpPoint_fromLines_toLines_powerValue_(sourcePoint, sourceLines, destinationLines, 0);
	}

	/**
	 * Answer the warp point with the specified lines.
	 * 
	 * @param sourcePoint jp.co.sra.jun.geometry.basic
	 * @param sourceLines jp.co.sra.jun.geometry.curves.Jun2dLine[]
	 * @param destinationLines jp.co.sra.jun.geometry.curves.Jun2dLine[]
	 * @param powerFactor double
	 * @return jp.co.sra.jun.geometry.basic
	 * @category Warping
	 */
	public static Jun2dPoint WarpPoint_fromLines_toLines_powerValue_(Jun2dPoint sourcePoint, Jun2dLine[] sourceLines, Jun2dLine[] destinationLines, double powerFactor) {
		double accuracyValue = 0.000001;
		double powerValue = Math.max(0.0, Math.min(powerFactor, 2.0));
		ArrayList<Jun2dPoint> pointCollection = new ArrayList<Jun2dPoint>(sourceLines.length);
		ArrayList<Double> weightCollection = new ArrayList<Double>(sourceLines.length);
		for (int i = 0; i < sourceLines.length; i++) {
			Jun2dLine sourceLine = sourceLines[i];
			Jun2dLine destinationLine = destinationLines[i];
			Jun2dPoint pointA = sourceLine.first();
			Jun2dPoint pointB = sourceLine.last();
			if (pointA.equal_(pointB)) {
				pointB = pointA.plus_(new Jun2dPoint(0.0, 1.0));
			}
			double lengthAB = pointB.distance_(pointA);
			Jun2dPoint vectorAB = pointB.minus_(pointA);
			Jun2dPoint vectorAP = sourcePoint.minus_(pointA);
			Jun2dLine bisectorAB = pointB.bisector_(pointA);
			Jun2dPoint unitVectorAC = bisectorAB.last().minus_(bisectorAB.first());
			Jun2dPoint pointC = pointA.to_(pointA.plus_(unitVectorAC)).atT_(lengthAB);
			Jun2dPoint vectorAC = pointC.minus_(pointA);
			double squaredLengthAB = lengthAB * lengthAB;
			double valueU = vectorAP.dotProduct_(vectorAC) / squaredLengthAB;
			double valueV = vectorAP.dotProduct_(vectorAB) / squaredLengthAB;
			double valueL = Math.pow(lengthAB, powerValue);
			pointA = destinationLine.first();
			pointB = destinationLine.last();
			if (pointA.equal_(pointB)) {
				pointB = pointA.plus_(new Jun2dPoint(0.0d, 1.0d));
			}
			lengthAB = pointB.distance_(pointA);
			bisectorAB = pointB.bisector_(pointA);
			unitVectorAC = bisectorAB.last().minus_(bisectorAB.first());
			Jun2dPoint pointO = pointA.to_(pointB).atT_(valueV);
			Jun2dPoint pointP = pointO.to_(pointO.plus_(unitVectorAC)).atT_(valueU * lengthAB);
			pointCollection.add(pointP);
			if (Math.abs(valueU) < accuracyValue) {
				weightCollection.add(valueL);
			} else {
				weightCollection.add(valueL / (valueU * valueU));
			}
		}
		Jun2dPoint sigmaPoint = new Jun2dPoint(0.0, 0.0);
		double sigmaWeight = 0.0d;
		for (int i = 0; i < pointCollection.size(); i++) {
			sigmaPoint = sigmaPoint.plus_((pointCollection.get(i).minus_(sourcePoint).multipliedBy_(weightCollection.get(i))));
			sigmaWeight = sigmaWeight + weightCollection.get(i);
		}
		Jun2dPoint warpedPoint;
		if (Math.abs(sigmaWeight) < accuracyValue) {
			warpedPoint = sourcePoint;
		} else {
			warpedPoint = sourcePoint.plus_(sigmaPoint.dividedBy_(sigmaWeight));
		}
		return warpedPoint;
	}

	/**
	 * Request new image file, and answer it.
	 * 
	 * @return java.io.File
	 * @category Writing
	 */
	public static File RequestNewImageFile() {
		String[] values = new String[] { "*.jpg", "*.jpeg", "*.bmp", "*.png", "*.JPG", "*.JPEG", "*.BMP", "*.PNG" };
		JunFileModel.FileType[] fileTypes = new JunFileModel.FileType[] { new JunFileModel.FileType(JunSystem.$String("<1p> files", null, JunSystem.$String("Image")), values) };
		String message = JunSystem.$String("Input a <1p> file.", null, JunSystem.$String("Image"));
		File file = JunFileRequesterDialog.RequestNewFile(message, new File(JunSystem.DefaultBaseName() + ".jpg"), fileTypes, fileTypes[0]);
		if (file == null) {
			return null;
		}
		return file;
	}

	/**
	 * Request new image file, then write the specified image to a file.
	 * 
	 * @param anImage java.awt.Image
	 * @throws java.io.IOException
	 * @category Writing
	 */
	public static File RequestNewImageFileThenWriteImage_(Image anImage) throws IOException {
		return RequestNewImageFileThenWriteImage_(new StImage(anImage));
	}

	/**
	 * Request new image file, then write the specified image to a file.
	 * 
	 * @param anImage jp.co.sra.smalltalk.StImage
	 * @throws java.io.IOException
	 * @category Writing
	 */
	public static File RequestNewImageFileThenWriteImage_(StImage anImage) throws IOException {
		File aFile = RequestNewImageFile();
		if (aFile == null) {
			return null;
		}
		return WriteImage_to_(anImage, aFile);
	}

	/**
	 * Write the specified image to a file.
	 * 
	 * @param anImage java.awt.Image
	 * @param aFile java.io.File
	 * @return java.io.File
	 * @throws java.io.IOException
	 * @category Writing
	 */
	public static File WriteImage_to_(Image anImage, File aFile) throws IOException {
		return WriteImage_to_(new StImage(anImage), aFile);
	}

	/**
	 * Write the specified image to a file.
	 * 
	 * @param anImage jp.co.sra.smalltalk.StImage
	 * @param aFile java.io.File
	 * @return java.io.File
	 * @throws java.io.IOException
	 * @category Writing
	 */
	public static File WriteImage_to_(StImage anImage, File aFile) throws IOException {
		File theFile = aFile;
		JunCursors cursor = new JunCursors(JunCursors.WriteCursor());
		JunImageStream aStream = null;
		try {
			cursor._show();
			Class<? extends JunImageStream> aClass = JunImageStream.ImageStreamClassForFile_(aFile);
			if (aClass == null) {
				aClass = JunBmpImageStream.class;
				theFile = new File(aFile.getAbsoluteFile() + ".bmp");
			}

			aStream = (JunImageStream) StObject._PerformWith(aClass, "On_", new FileOutputStream(theFile));
			if (aStream.respondsTo_("quality_")) {
				aStream.perform_with_("quality_", new Float(DefaultImageQuality()));
			}
			aStream.nextPutImage_(anImage);
		} catch (NoSuchMethodException e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
		} catch (InvocationTargetException e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
		} finally {
			if (aStream != null) {
				aStream.flush();
				aStream.close();
			}
			cursor._restore();
		}
		return theFile;
	}
}
