package jp.co.sra.jun.obj;

import java.awt.Image;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.imageio.ImageIO;

import jp.co.sra.smalltalk.SmalltalkException;
import jp.co.sra.smalltalk.StReadStream;
import jp.co.sra.smalltalk.StSymbol;

import jp.co.sra.jun.geometry.basic.Jun2dPoint;
import jp.co.sra.jun.geometry.basic.Jun3dPoint;
import jp.co.sra.jun.opengl.objects.JunOpenGL3dCompoundObject;
import jp.co.sra.jun.opengl.objects.JunOpenGL3dObject;
import jp.co.sra.jun.opengl.objects.JunOpenGL3dPolygon;
import jp.co.sra.jun.opengl.texture.JunOpenGLTexture;

/**
 * JunOBJParser class
 * 
 *  @author    Albert Kerssies
 *  @created   2009/03/25 (by nisinaka)
 *  @updated   N/A
 *  @version   758 (with StPL9.7) based on Jun732 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunOBJParser.java,v 9.7 2010/03/09 09:17:32 nisinaka Exp $
 */
public class JunOBJParser extends JunAbstractOBJParser {

	protected Map<String, JunOBJMaterial> materials;
	protected List<Jun3dPoint> positions;
	protected List<Jun3dPoint> normalCoordinates;
	protected List<Jun2dPoint> textureCoordinates;
	protected Map<String, Image> images;
	protected List<JunOBJGroup> groups;
	protected JunOBJMesh currentMesh;

	/**
	 * Parse the object and generate.
	 * 
	 * @param anObject java.lang.Object
	 * @return java.lang.Object
	 * @category 
	 */
	public static JunOpenGL3dObject Parse_(Object anObject) {
		JunOBJParser parser = new JunOBJParser();
		return parser.parse_(MakeStream_(anObject));
	}

	/**
	 * Initialize the receiver.
	 * 
	 * @see jp.co.sra.jun.obj.JunAbstractOBJParser#initialize()
	 * @category initialize-release
	 */
	@Override
	protected void initialize() {
		super.initialize();

		materials = new HashMap<String, JunOBJMaterial>();
		positions = new ArrayList<Jun3dPoint>();
		normalCoordinates = new ArrayList<Jun3dPoint>();
		textureCoordinates = new ArrayList<Jun2dPoint>();
		images = new HashMap<String, Image>();
		groups = new ArrayList<JunOBJGroup>();
	}

	/**
	 * Initialize the keyword table.
	 * 
	 * @see jp.co.sra.jun.obj.JunAbstractOBJParser#initKeyTable()
	 * @category initialize-release
	 */
	@Override
	protected void initKeyTable() {
		super.initKeyTable();

		keyTable.put("v", "parseVertex");
		keyTable.put("vt", "parseTextureCoordinate");
		keyTable.put("vn", "parseNormal");
		keyTable.put("mtllib", "parseMaterialFile");
		keyTable.put("usemtl", "parseUseMaterial");
		keyTable.put("f", "parseFace");
		keyTable.put("g", "parseGroup");
	}

	/**
	 * Set my new material.
	 * 
	 * @param aList java.util.List<JunOBJMaterial>
	 * @category accessing
	 */
	public void materials_(List<JunOBJMaterial> aList) {
		for (JunOBJMaterial each : aList) {
			materials.put(each.name(), each);
		}
	}

	/**
	 * Answer the position at the specified index.
	 * 
	 * @param index int
	 * @return jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category accessing
	 */
	public Jun3dPoint positionAt_(int index) {
		if (index <= 0 || positions.size() < index) {
			return null;
		}

		return positions.get(index - 1);
	}

	/**
	 * Answer the position at the specified index.
	 * 
	 * @param index int
	 * @return jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category accessing
	 */
	public Jun3dPoint normalCoordinateAt_(int index) {
		if (index <= 0 || normalCoordinates.size() < index) {
			return null;
		}

		return normalCoordinates.get(index - 1);
	}

	/**
	 * Answer the texture coordinate at the specified index.
	 * 
	 * @param index int
	 * @return jp.co.sra.jun.geometry.basic.Jun2dPoint
	 * @category accessing
	 */
	public Jun2dPoint textureCoordinateAt_(int index) {
		if (index <= 0 || textureCoordinates.size() < index) {
			return null;
		}

		return textureCoordinates.get(index - 1);
	}

	/**
	 * Create a default material.
	 * 
	 * @return jp.co.sra.jun.obj.JunOBJMaterial
	 * @category accessing
	 */
	protected JunOBJMaterial defaultMaterial() {
		return new JunOBJMaterial("defalut");
	}

	/**
	 * Parse the stream and generate an object.
	 * 
	 * @param sourceStream jp.co.sra.smalltalk.StReadStream
	 * @return jp.co.sra.jun.opengl.objects.JunOpenGL3dObject
	 * @see jp.co.sra.jun.obj.JunAbstractOBJParser#parse_(jp.co.sra.smalltalk.StReadStream)
	 * @category parsing
	 */
	@Override
	protected JunOpenGL3dObject parse_(StReadStream sourceStream) {
		if (sourceStream.atEnd()) {
			return new JunOpenGL3dCompoundObject();
		}

		super.parse_(sourceStream);
		return this.generatedJunOpenGL3dObject();
	}

	/**
	 * Parse the face.
	 * 
	 * 		f int ..., or f int/int ..., or f int/int/int.
	 * 
	 * int/int/int corresponds to the indices of a vertex position, the optional texture coordinate of the vertex and its optional normal
	 * 
	 * @category parsing
	 */
	public void parseFace() {
		JunOBJIndexedFace face = new JunOBJIndexedFace();
		StSymbol prevType = null;
		JunOBJIndexedVertex vertex = null;

		this.nextToken();
		while (tokenType != $("eolOrf")) {
			if (tokenType == $("number")) {
				if (prevType != $("slash")) {
					vertex = new JunOBJIndexedVertex();
					face.addIndexedVertex_(vertex);
				}
				vertex.addIndex_(((Number) token).intValue());
			} else if (tokenType != $("slash")) {
				this.failed_("Not a valid face token");
				return;
			}

			prevType = tokenType;
			this.nextToken();
		}

		if (face.size() == 0) {
			this.failed_("No data for a face found");
			return;
		}
		if (face.hasEqualSizedVertices() == false) {
			this.failed_("Face with different kind of vertices");
			return;
		}

		currentMesh.addIndexedFace_(face);
	}

	/**
	 * Parse the group.
	 * 
	 * 		g string
	 * 
	 * Group name.
	 * 
	 * @category parsing
	 */
	public void parseGroup() {
		groups.add(new JunOBJGroup(this.nextName()));
	}

	/**
	 * Parse the material file.
	 * 
	 * 		mtllib string.mtl
	 * 
	 * Material file string.mtl
	 * 
	 * @category parsing
	 */
	public void parseMaterialFile() {
		this.nextToken();
		if (tokenType == $("number")) {
			this.unNextToken();
			this.nextName();
		}

		String name = (String) token;
		List<JunOBJMaterial> materialColl = JunMTLParser.Parse_(new File(this.directory(), name));
		if (materialColl != null) {
			this.materials_(materialColl);
		}
	}

	/**
	 * Parse the normal.
	 * 
	 * 		vn float float float
	 * 
	 * Vertex normal x y z
	 * 
	 * @category parsing
	 */
	public void parseNormal() {
		Number[] numbers = this.nextNumbers_(3);
		normalCoordinates.add(new Jun3dPoint(numbers[0], numbers[1], numbers[2]));
	}

	/**
	 * Parse the texture coordinate.
	 * 
	 * 		vt float float float
	 * 
	 * Texture cooridnate tu tv
	 * 
	 * @category parsing
	 */
	public void parseTextureCoordinate() {
		Number[] numbers = this.nextNumbers_(2);
		textureCoordinates.add(new Jun2dPoint(numbers[0], numbers[1]));
	}

	/**
	 * Parse the use material.
	 * 
	 * 		usemtl string
	 * 
	 * Selection of material with name string.
	 * 
	 * @category parsing
	 */
	public void parseUseMaterial() {
		String materialName = (String) this.nextToken();
		JunOBJMaterial material = materials.get(materialName);
		if (material == null) {
			material = this.defaultMaterial();
		}

		currentMesh = new JunOBJMesh(material);
		this.addMesh_(currentMesh);
	}

	/**
	 * Parse the vertex.
	 * 
	 * 		v float float float
	 * 
	 * Vertex position x y z
	 * 
	 * @category parsing
	 */
	public void parseVertex() {
		Number[] numbers = this.nextNumbers_(3);
		positions.add(new Jun3dPoint(numbers[0], numbers[1], numbers[2]));
	}

	/**
	 * Generate a JunOpenGL3dObject from the current information.
	 * 
	 * @return jp.co.sra.jun.obj.JunOBJParser.generatedJunOpenGL3dObject()
	 * @category generating
	 */
	protected JunOpenGL3dObject generatedJunOpenGL3dObject() {
		if (groups.size() == 0) {
			return new JunOpenGL3dCompoundObject();
		} else if (groups.size() == 1) {
			return groups.get(0).asJunOpenGL3dObjectUsing_(this);
		}

		List<JunOpenGL3dObject> components = new ArrayList<JunOpenGL3dObject>();
		for (JunOBJGroup g : groups) {
			components.add(g.asJunOpenGL3dObjectUsing_(this));
		}
		return new JunOpenGL3dCompoundObject(components);
	}

	/**
	 * Generate a JunOpenGL3dObject from the specified face.
	 * 
	 * @param aJunOBJIndexedFace jp.co.sra.jun.obj.JunOBJIndexedFace
	 * @return jp.co.sra.jun.opengl.objects.JunOpenGL3dObject
	 * @category generating
	 */
	protected JunOpenGL3dObject junOpenGL3dObjectFromFace_(JunOBJIndexedFace aJunOBJIndexedFace) {
		Integer[] indices = aJunOBJIndexedFace.positionIndices();
		Jun3dPoint[] vertexes = new Jun3dPoint[indices.length];
		for (int i = 0; i < indices.length; i++) {
			vertexes[i] = this.positionAt_(indices[i]);
			if (vertexes[i] == null) {
				this.generatingError_("Wrong position index");
			}
		}

		JunOpenGL3dPolygon polygon = new JunOpenGL3dPolygon(vertexes);

		if (aJunOBJIndexedFace.hasNormals()) {
			indices = aJunOBJIndexedFace.normalIndices();
			Jun3dPoint[] normalVectors = new Jun3dPoint[indices.length];
			for (int i = 0; i < indices.length; i++) {
				normalVectors[i] = this.normalCoordinateAt_(indices[i]);
				if (normalVectors[i] == null) {
					this.generatingError_("Wrong normal index");
				}
			}

			polygon.normalVectors_(normalVectors);
		}

		return polygon;
	}

	/**
	 * Generate a JunOpenGL3dObject from the specified group.
	 * 
	 * @param aJunOBJGroup jp.co.sra.jun.obj.JunOBJGroup
	 * @return jp.co.sra.jun.opengl.objects.JunOpenGL3dObject
	 * @category generating
	 */
	protected JunOpenGL3dObject junOpenGL3dObjectFromGroup_(JunOBJGroup aJunOBJGroup) {
		JunOpenGL3dCompoundObject openGL3dObject = new JunOpenGL3dCompoundObject();
		openGL3dObject.name_(aJunOBJGroup.name());

		ArrayList<JunOpenGL3dObject> components = new ArrayList<JunOpenGL3dObject>();
		for (JunOBJMesh m : aJunOBJGroup.meshes()) {
			components.add(m.asJunOpenGL3dObjectUsing_(this));
		}
		openGL3dObject.components_(components);

		return openGL3dObject;
	}

	/**
	 * Generate a JunOpenGL3dObject from the specified mesh.
	 * 
	 * @param aJunOBJMesh jp.co.sra.jun.obj.JunOBJMesh
	 * @return jp.co.sra.jun.opengl.objects.JunOpenGL3dObject
	 * @category generating
	 */
	protected JunOpenGL3dObject junOpenGL3dObjectFromMesh_(JunOBJMesh aJunOBJMesh) {
		JunOpenGL3dCompoundObject openGL3dObject = new JunOpenGL3dCompoundObject();

		ArrayList<JunOpenGL3dObject> components = new ArrayList<JunOpenGL3dObject>();
		for (JunOBJIndexedFace f : aJunOBJMesh.faces()) {
			components.add(f.asJunOpenGL3dObjectUsing_(this));
		}
		openGL3dObject.components_(components);
		openGL3dObject.paint_(aJunOBJMesh.material().colorValue());
		openGL3dObject.texture_(this.junTextureFrom_(aJunOBJMesh));

		return openGL3dObject;
	}

	/**
	 * Generate a JunOpenGLTexture from the specified mesh.
	 * 
	 * @param aJunOBJMesh jp.co.sra.jun.obj.JunOBJMesh
	 * @return jp.co.sra.jun.opengl.texture.JunOpenGLTexture
	 * @category generating
	 */
	protected JunOpenGLTexture junTextureFrom_(JunOBJMesh aJunOBJMesh) {
		String textureName = aJunOBJMesh.material().textureImageName();
		if (textureName == null) {
			return null;
		}

		Image im = this.getImage_(textureName);
		if (im == null) {
			return null;
		}

		JunOpenGLTexture texture = new JunOpenGLTexture(im);
		Integer[] textureIndices = aJunOBJMesh.textureIndices();
		ArrayList<Jun2dPoint> texCoordinates = new ArrayList<Jun2dPoint>();
		for (int index : textureIndices) {
			if (index >= 0) {
				texCoordinates.add(this.textureCoordinateAt_(index));
			}
		}
		texture.coordinates_(texCoordinates.toArray(new Jun2dPoint[texCoordinates.size()]));

		return texture;
	}

	/**
	 * Generate an exception.
	 * 
	 * @param aString java.lang.String
	 * @throws jp.co.sra.smalltalk.SmalltalkException
	 * @category error handling
	 */
	protected void generatingError_(String aString) throws SmalltalkException {
		throw SmalltalkException.Error(aString + " found while generating 3d object.");
	}

	/**
	 * Add the specified mesh to my latest group.
	 * 
	 * @param aJunOBJMesh jp.co.sra.jun.obj.JunOBJMesh
	 * @category private
	 */
	protected void addMesh_(JunOBJMesh aJunOBJMesh) {
		groups.get(groups.size() - 1).addMesh(aJunOBJMesh);
	}

	/**
	 * Answer the image which corresponds to the string.
	 * 
	 * @param aString java.lang.String
	 * @return java.awt.Image
	 * @category private
	 */
	protected Image getImage_(String aString) {
		Image im = images.get(aString);
		if (im != null) {
			return im;
		}

		try {
			// TODO: should the directory be same as of the source obj file.
			im = ImageIO.read(new File(aString));
		} catch (IOException e) {
			e.printStackTrace();
		}
		if (im != null) {
			images.put(aString, im);
		}
		return im;
	}

}
