package jp.co.sra.jun.obj;

import java.io.File;

import jp.co.sra.smalltalk.SmalltalkException;
import jp.co.sra.smalltalk.StBlockClosure1;
import jp.co.sra.smalltalk.StNumber;
import jp.co.sra.smalltalk.StObject;
import jp.co.sra.smalltalk.StReadStream;
import jp.co.sra.smalltalk.StSymbol;

/**
 * JunOBJScanner class
 * 
 *  @author    Albert Kerssies
 *  @created   2009/03/25 (by nisinaka)
 *  @updated   N/A
 *  @version   758 (with StPL9.7) based on Jun732 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunOBJScanner.java,v 9.7 2010/03/09 09:17:32 nisinaka Exp $
 */
public class JunOBJScanner extends StObject {

	protected static final char CR = '\r';
	protected static final char LF = '\n';

	protected StReadStream source;
	protected int mark;
	protected Object token;
	protected StSymbol tokenType;
	protected JunOBJScannerTable typeTable;
	protected StBlockClosure1<Void, String> failBlock;
	protected char endChar;
	protected char separator;

	/**
	 * Create a new instance of JunOBJScanner and initialize it.
	 *
	 * @category Instance creation
	 */
	public JunOBJScanner() {
		this.initialize();
	}

	/**
	 * Initialize the receiver.
	 * 
	 * @category initialize-release
	 */
	protected void initialize() {
		typeTable = JunOBJScannerTable.ScannerTable();
		this.initFailBlock();
		endChar = '!';
	}

	/**
	 * Initialize the fail block.
	 * 
	 * @category initialize-release
	 */
	protected void initFailBlock() {
		failBlock = new StBlockClosure1<Void, String>() {
			public Void value(String errorMessage) {
				String label = errorMessage + " near " + token;

				StringBuilder string = new StringBuilder();
				while (!source.atEnd()) {
					string.append(source.next());
				}

				String message = string.toString();
				if (message.length() == 0) {
					message = "--> end of file";
				} else {
					message = "--> " + message.substring(0, 29) + "...";
				}

				throw SmalltalkException.Error(label + "\n" + message);
			}
		};
	}

	/**
	 * Initialize with the input stream.
	 * 
	 * @param inputStream jp.co.sra.smalltalk.StReadStream
	 * @category initialize-release
	 */
	protected void on_(StReadStream inputStream) {
		source = inputStream;
		mark = source.position();
		this.findSeparator();
	}

	/**
	 * Answer my file.
	 * 
	 * @return java.io.File
	 * @category accessing
	 */
	public File file() {
		File aFile = source.file();
		if (aFile == null) {
			aFile = new File(".");
		}
		return aFile;
	}

	/**
	 * Answer my directory.
	 * 
	 * @return java.io.File
	 * @category accessing
	 */
	public File directory() {
		return this.file().getParentFile();
	}

	/**
	 * Scan the next token.
	 * 
	 * @return java.lang.Object
	 * @category scanning
	 */
	protected Object nextToken() {
		this.basicNextToken();

		String tokenTypeString = tokenType.toString();
		if (tokenTypeString.charAt(0) == 'x') {
			this.multiChar_(tokenTypeString);
		} else {
			this.singleChar_(tokenTypeString);
		}

		return token;
	}

	/**
	 * Scan the next name.
	 * 
	 * @return java.lang.String
	 * @category scanning
	 */
	protected String nextName() {
		this.basicNextToken();
		this.xName();
		return (String) token;
	}

	/**
	 * Scan the next token type.
	 * 
	 * @category scanning
	 */
	protected void basicNextToken() {
		mark = source.position();
		char c = this.peekChar();
		tokenType = typeTable.at_(c);
		while (tokenType == $("xSpace")) {
			this.nextChar();
			c = this.peekChar();
			tokenType = typeTable.at_(c);
		}
	}

	/**
	 * Get a next token with multiple characters.
	 * 
	 * @param type java.lang.String
	 * @throws jp.co.sra.smalltalk.SmalltalkException 
	 * @category scanning
	 */
	protected void multiChar_(String type) {
		try {
			this.perform_(type);
		} catch (Exception e) {
			throw new SmalltalkException(e);
		}
	}

	/**
	 * Get a next token which is just a single character.
	 * 
	 * @param type java.lang.Object
	 * @category scanning
	 */
	protected void singleChar_(Object type) {
		if (type != $("eolOrf")) {
			this.nextChar();
		}
		token = type;
	}

	/**
	 * Skip the rest of the line.
	 * 
	 * @category scanning
	 */
	protected void skipLine() {
		source.skipThrough_(separator);
	}

	/**
	 * Push back the token to the stream.
	 * 
	 * @category scanning
	 */
	protected void unNextToken() {
		source.position_(mark);
	}

	/**
	 * Read a name from the stream.
	 * 
	 * @param aStream jp.co.sra.jun.obj.JunOBJScanner.nameFrom_(StReadStream)
	 * @return java.lang.String
	 * @category reading
	 */
	protected String nameFrom_(StReadStream aStream) {
		StringBuilder buffer = new StringBuilder();
		char c = this.peekChar();
		tokenType = typeTable.at_(c);
		while (tokenType != $("eolOrf")) {
			buffer.append(this.nextChar());
			c = this.peekChar();
			tokenType = typeTable.at_(c);
		}

		return buffer.toString();
	}

	/**
	 * Read a number from the stream.
	 * 
	 * @param aStream jp.co.sra.smalltalk.StReadStream
	 * @return java.lang.Number
	 * @category reading
	 */
	protected Number numberFrom_(StReadStream aStream) {
		return StNumber.ReadFrom_(aStream);
	}

	/**
	 * Read a string from the stream.
	 * 
	 * @param aStream jp.co.sra.smalltalk.StReadStream
	 * @return java.lang.String
	 * @category reading 
	 */
	protected String stringFrom_(StReadStream aStream) {
		StringBuilder buffer = new StringBuilder();
		char c = this.peekChar();
		tokenType = typeTable.at_(c);
		while (tokenType != $("xSpace") && tokenType != $("eolOrf")) {
			buffer.append(this.nextChar());
			c = this.peekChar();
			tokenType = typeTable.at_(c);
		}

		return buffer.toString();
	}

	/**
	 * Handle the error with the fail block.
	 * 
	 * @param aString java.lang.String
	 * @category error handling
	 */
	protected void failed_(String aString) {
		failBlock.value(aString);
	}

	/**
	 * Determine the separator in the source stream.
	 * 
	 * @category private
	 */
	protected void findSeparator() {
		separator = 0;
		while (separator == 0 && source.atEnd() == false) {
			char c = source.next();
			if (c == LF) {
				separator = LF;
			} else if (c == CR) {
				c = source.next();
				if (c == LF) {
					separator = LF;
				} else {
					separator = CR;
				}
			}
		}

		source.position_(0);
	}

	/**
	 * Get the next character from the source.
	 * 
	 * @return char
	 * @category private
	 */
	protected char nextChar() {
		char c = source.next();
		if (c == 0) {
			return endChar;
		}
		return c;
	}

	/**
	 * Get the next numbers.
	 * 
	 * @param size int
	 * @return java.lang.Number[]
	 * @category private
	 */
	protected Number[] nextNumbers_(int size) {
		Number[] array = new Number[size];
		for (int i = 0; i < size; i++) {
			this.nextToken();
			if (tokenType != $("number")) {
				this.failed_("Not a number");
			}
			array[i] = (Number) token;
		}
		return array;
	}

	/**
	 * Peek the next character from the source.
	 * 
	 * @return char
	 * @category private
	 */
	protected char peekChar() {
		char c = source.peek();
		if (c == 0) {
			return endChar;
		}
		return c;
	}

	/**
	 * Push back a character to the source stream.
	 * 
	 * @category private
	 */
	protected void unNextChar() {
		source.skip_(-1);
	}

	/**
	 * Get a binary token.
	 * 
	 * @category x
	 */
	public void xBinary() {
		this.failed_("Syntax error " + source.peek());
	}

	/**
	 * Get a comment token.
	 * 
	 * @category x
	 */
	public void xComment() {
		this.skipLine();
		this.nextToken();
	}

	/**
	 * Get a digit token.
	 * 
	 * @category x
	 */
	public void xDigit() {
		tokenType = $("number");
		token = this.numberFrom_(source);
	}

	/**
	 * Get a name token.
	 * 
	 * @category x
	 */
	public void xName() {
		tokenType = $("string");
		token = this.nameFrom_(source);
	}

	/**
	 * Get a sign token.
	 * 
	 * @category x
	 */
	public void xSign() {
		char sign = this.nextChar();
		char c = this.peekChar();
		if (Character.isDigit(c)) {
			tokenType = $("number");
			Number number = this.numberFrom_(source);
			if (sign == '-') {
				// I'm so sad to do this. :-(
				if (number instanceof Integer) {
					number = new Integer(-number.intValue());
				} else if (number instanceof Float) {
					number = new Float(-number.floatValue());
				} else if (number instanceof Double) {
					number = new Double(-number.doubleValue());
				} else if (number instanceof Long) {
					number = new Long(-number.longValue());
				}
			}
			token = number;
		} else {
			this.unNextChar();
			tokenType = $("string");
			token = this.stringFrom_(source);
		}
	}

	/**
	 * Get a string token.
	 * 
	 * @category x
	 */
	public void xString() {
		tokenType = $("string");
		token = this.stringFrom_(source);
	}

}
