package jp.co.sra.jun.obj;

import java.awt.Color;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeSet;

import jp.co.sra.smalltalk.StBlockClosure1;
import jp.co.sra.smalltalk.StImage;
import jp.co.sra.smalltalk.StObject;

import jp.co.sra.jun.geometry.basic.Jun2dPoint;
import jp.co.sra.jun.geometry.basic.Jun3dPoint;
import jp.co.sra.jun.goodies.image.streams.JunBmpImageStream;
import jp.co.sra.jun.goodies.image.streams.JunImageStream;
import jp.co.sra.jun.opengl.objects.JunOpenGL3dObject;
import jp.co.sra.jun.opengl.objects.JunOpenGL3dVertexesObject;
import jp.co.sra.jun.opengl.texture.JunOpenGLTexture;
import jp.co.sra.jun.system.support.JunSystem;

/**
 * JunOBJWriter class
 * 
 *  @author    Albert Kerssies
 *  @created   2009/08/18 (by nisinaka)
 *  @updated   N/A
 *  @version   758 (with StPL9.7) based on Jun732 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunOBJWriter.java,v 9.3 2010/03/09 09:17:32 nisinaka Exp $
 */
public class JunOBJWriter extends StObject {

	protected File modelFile;
	protected JunOpenGL3dObject modelObject;
	protected String modelName;
	protected JunOBJFaceGroup currentGroup;
	protected JunOBJMaterial currentMaterial;
	protected List<JunOBJFaceGroup> faceGroups;
	protected Map<Jun3dPoint, Integer> vertices;
	protected Map<Jun2dPoint, Integer> uvCoordinates;
	protected Map<Jun3dPoint, Integer> normalCoordinates;
	protected Map<JunMaterialKey, JunOBJMaterial> materials;
	protected Map<StImage, String> textures;
	protected StBlockClosure1<Jun2dPoint, Jun3dPoint> textureCoordinateBlock;
	protected int textureCoordinateIndex;

	/**
	 * Create a new instance of JunOBJWriter and initialize it.
	 *
	 * @param aFile java.io.File
	 * @param anObject jp.co.sra.jun.opengl.objects.JunOpenGL3dObject
	 * @category Instance creation
	 */
	public JunOBJWriter(File aFile, JunOpenGL3dObject anObject) {
		this.initialize();
		modelFile = aFile;
		modelObject = anObject;
	}

	/**
	 * Initialize the receiver.
	 * 
	 * @category initialize-release
	 */
	protected void initialize() {
		modelFile = null;
		modelObject = null;
		modelName = null;
		currentGroup = null;
		currentMaterial = null;
		faceGroups = new ArrayList<JunOBJFaceGroup>();
		vertices = new HashMap<Jun3dPoint, Integer>();
		uvCoordinates = new HashMap<Jun2dPoint, Integer>();
		normalCoordinates = new HashMap<Jun3dPoint, Integer>();
		materials = new HashMap<JunMaterialKey, JunOBJMaterial>();
		textures = new HashMap<StImage, String>();
		this.setTextureCoordinateBlock();
	}

	/**
	 * Answer my model name.
	 * 
	 * @return java.lang.String
	 * @category accessing
	 */
	public String modelName() {
		if (modelName == null) {
			String aString = modelFile.getName();

			// "return string without the .obj extension and replace spaces with an underscore"
			if (aString.endsWith(".obj")) {
				aString = aString.substring(0, aString.length() - 4);
			}
			aString = aString.replace(' ', '_');

			modelName = aString;
		}
		return modelName;
	}

	/**
	 * Answer my current texture coordinate index.
	 * 
	 * @return int
	 * @category accessing
	 */
	public int textureCoordinateIndex() {
		return textureCoordinateIndex;
	}

	/**
	 * Set my new texture coordinate index.
	 * 
	 * @param index int
	 * @category accessing
	 */
	public void textureCoordinateIndex_(int index) {
		textureCoordinateIndex = index;
	}

	/**
	 * Answer my material filename as a string.
	 * 
	 * @return java.lang.String
	 * @category accessing
	 */
	protected String materialFilenameAsString() {
		return this.modelName() + ".mtl";
	}

	/**
	 * Answer my current directory.
	 * 
	 * @return java.io.File
	 * @category accessing
	 */
	protected File directory() {
		return modelFile.getParentFile();
	}

	/**
	 * Answer my current texture.
	 * 
	 * @return jp.co.sra.jun.opengl.texture.JunOpenGLTexture
	 * @category accessing
	 */
	protected JunOpenGLTexture texture() {
		return currentMaterial.texture();
	}

	/**
	 * Answer the index of the specified vertex.
	 * 
	 * @param aPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @return int
	 * @category accessing
	 */
	protected int vertexIndexFor_(Jun3dPoint aPoint) {
		if (vertices.containsKey(aPoint) == false) {
			vertices.put(aPoint, vertices.size() + 1);
		}
		return vertices.get(aPoint);
	}

	/**
	 * Answer the index of the specified UV coordinate.
	 * 
	 * @param aPoint jp.co.sra.jun.geometry.basic.Jun2dPoint
	 * @return int
	 * @category accessing
	 */
	protected int uvIndexFor_(Jun2dPoint aPoint) {
		if (aPoint == null) {
			return -1;
		}

		if (uvCoordinates.containsKey(aPoint) == false) {
			uvCoordinates.put(aPoint, uvCoordinates.size() + 1);
		}
		return uvCoordinates.get(aPoint);
	}

	/**
	 * Answer the index of the specified normal vector.
	 * 
	 * @param aPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @return int
	 * @category accessing
	 */
	protected int normalIndexFor_(Jun3dPoint aPoint) {
		if (aPoint == null) {
			return -1;
		}

		if (normalCoordinates.containsKey(aPoint) == false) {
			normalCoordinates.put(aPoint, normalCoordinates.size() + 1);
		}
		return normalCoordinates.get(aPoint);
	}

	/**
	 * Do the saving.
	 * 
	 * @exception java.io.IOException
	 * @category actions
	 */
	public void save() throws IOException {
		String name = this.materialFilenameAsString();

		BufferedWriter aWriter = new BufferedWriter(new FileWriter(modelFile));
		try {
			this.writeModelHeaderOn_(aWriter);

			aWriter.write("mtllib ");
			aWriter.write(name);
			aWriter.newLine();
			aWriter.newLine();

			// ensure adding a group
			if (modelObject.hasName() == false) {
				this.addGroupNamed_(this.modelName());
			}
			modelObject.toOBJUsing_(this);

			// vertexes etc. are written once
			this.writeCoordinatesOn_(aWriter);
			this.writeGroupsOn_(aWriter);

			if (materials.isEmpty()) {
				return;
			}
			File mtlFile = new File(this.directory(), name);
			BufferedWriter mtlWriter = new BufferedWriter(new FileWriter(mtlFile));
			try {
				this.writeMaterialsOn_(mtlWriter);
			} finally {
				mtlWriter.flush();
				mtlWriter.close();
			}
			this.writeTexturesOn_(aWriter);

		} finally {
			if (aWriter != null) {
				aWriter.flush();
				aWriter.close();
			}
		}
	}

	/**
	 * Set my current material.
	 * 
	 * @param aMaterial jp.co.sra.jun.obj.JunOBJMaterial
	 * @category actions
	 */
	public void setMaterial_(JunOBJMaterial aMaterial) {
		currentMaterial = aMaterial;
	}

	/**
	 * Derive the material for the specified JunOpenGL3dObject.
	 * 
	 * @param anOpenGL3dObject jp.co.sra.jun.opengl.objects.JunOpenGL3dObject
	 * @return jp.co.sra.jun.obj.JunOBJMaterial
	 * @category actions
	 */
	public JunOBJMaterial deriveMaterialFor_(JunOpenGL3dObject anOpenGL3dObject) {
		JunMaterialKey materialKey = this.materialKeyFor_(anOpenGL3dObject);
		JunOBJMaterial material = materials.get(materialKey);
		if (material == null) {
			material = this.newMaterialFor_withTextureImageName_(anOpenGL3dObject, materialKey.textureImageName());
			materials.put(materialKey, material);
		}

		currentMaterial = material;

		// "Be aware, equal materialKey's can differ in textureObjects (especially in their coordinates), so be sure to set the texture in the currentMaterial as well."
		currentMaterial.texture_(anOpenGL3dObject.texture());

		return material;
	}

	/**
	 * Derive the colors based material for the specified JunOpenGL3dVertexesObject.
	 * 
	 * @param anOpenGL3dVertexesObject jp.co.sra.jun.opengl.objects.JunOpenGL3dVertexesObject
	 * @return jp.co.sra.jun.obj.JunOBJMaterial
	 * @category actions
	 */
	public JunOBJMaterial deriveColorsBasedMaterialFor_(JunOpenGL3dVertexesObject anOpenGL3dVertexesObject) {
		// "anOpenGL3dVertexesObject has colors but no paint, alpha or texture to derive a material from. This method averages the colors to approximate an overall paint"

		JunMaterialKey materialKey = this.colorsBasedMaterialKeyFor_(anOpenGL3dVertexesObject);
		JunOBJMaterial material = materials.get(materialKey);
		if (material == null) {
			material = this.newMaterialForPaint_(materialKey.colorValue());
			materials.put(materialKey, material);
		}

		currentMaterial = material;
		return material;
	}

	/**
	 * Add a named group.
	 * 
	 * @param aString java.lang.String
	 * @category adding
	 */
	public void addGroupNamed_(String aString) {
		currentGroup = new JunOBJFaceGroup(aString);
		faceGroups.add(currentGroup);
	}

	/**
	 * Add a polygon with its normal vectors.
	 * 
	 * @param vertexes jp.co.sra.jun.geometry.basic.Jun3dPoint[]
	 * @param normalVectors jp.co.sra.jun.geometry.basic.Jun3dPoint[]
	 * @category adding
	 */
	public void addPolygon_normalVectors_(Jun3dPoint[] vertexes, Jun3dPoint[] normalVectors) {
		StBlockClosure1<Jun2dPoint, Jun3dPoint> texBlock = null;
		JunOpenGLTexture texture = this.texture();
		if (texture != null && texture.hasCoordinates()) {
			texBlock = textureCoordinateBlock;
		}

		int size = vertexes.length;
		ArrayList<int[]> faceElements = new ArrayList<int[]>(size);
		for (int i = 0; i < size; i++) {
			Jun3dPoint a3dPoint = vertexes[i];
			int[] faceElement = new int[3];
			faceElement[0] = this.vertexIndexFor_(a3dPoint);
			Jun2dPoint a2dPoint = (texBlock != null) ? texBlock.value(a3dPoint) : null;
			faceElement[1] = this.uvIndexFor_(a2dPoint);
			Jun3dPoint normalVector = (normalVectors != null && normalVectors.length > i) ? normalVectors[i] : null;
			faceElement[2] = this.normalIndexFor_(normalVector);

			faceElements.add(faceElement);
		}

		this.addFaceElements_(faceElements);
	}

	/**
	 * Add the specified face elements.
	 * 
	 * @param faceElements java.util.ArrayList<int[]>
	 * @category adding
	 */
	protected void addFaceElements_(ArrayList<int[]> faceElements) {
		currentGroup.add_for_(faceElements, currentMaterial.name());
	}

	/**
	 * Answer true if the specified material is equal to the current material, otherwise false.
	 * 
	 * @param aMaterial jp.co.sra.jun.obj.JunOBJMaterial
	 * @return boolean
	 * @category testing
	 */
	public boolean currentMaterialEquals_(JunOBJMaterial aMaterial) {
		return currentMaterial.materialKey().equals(aMaterial.materialKey());
	}

	/**
	 * Write the model header on the specified writer.
	 * 
	 * @param aWriter java.io.BufferedWriter
	 * @throws java.io.IOException 
	 * @category writing
	 */
	protected void writeModelHeaderOn_(BufferedWriter aWriter) throws IOException {
		aWriter.write("# Alias OBJ Model File");
		aWriter.newLine();
		aWriter.write("# Exported from " + JunSystem.System() + " version " + JunSystem.Version());
		aWriter.newLine();
		aWriter.newLine();
	}

	/**
	 * Write the coordinates on the specified writer.
	 * 
	 * @param aWriter java.io.BufferedWriter
	 * @throws java.io.IOException
	 * @category writing
	 */
	protected void writeCoordinatesOn_(BufferedWriter aWriter) throws IOException {
		TreeSet<Map.Entry<Jun3dPoint, Integer>> vTreeSet = new TreeSet<Map.Entry<Jun3dPoint, Integer>>(new Comparator<Map.Entry<Jun3dPoint, Integer>>() {
			public int compare(Map.Entry<Jun3dPoint, Integer> o1, Map.Entry<Jun3dPoint, Integer> o2) {
				return o1.getValue().compareTo(o2.getValue());
			}
		});
		vTreeSet.addAll(vertices.entrySet());
		for (Map.Entry<Jun3dPoint, Integer> e : vTreeSet) {
			this.writeVertexOn_(aWriter, e.getKey());
		}
		aWriter.newLine();

		TreeSet<Map.Entry<Jun2dPoint, Integer>> vtTreeSet = new TreeSet<Map.Entry<Jun2dPoint, Integer>>(new Comparator<Map.Entry<Jun2dPoint, Integer>>() {
			public int compare(Map.Entry<Jun2dPoint, Integer> o1, Map.Entry<Jun2dPoint, Integer> o2) {
				return o1.getValue().compareTo(o2.getValue());
			}
		});
		vtTreeSet.addAll(uvCoordinates.entrySet());
		for (Map.Entry<Jun2dPoint, Integer> e : vtTreeSet) {
			this.writeUVCoordinateOn_(aWriter, e.getKey());
		}
		aWriter.newLine();

		TreeSet<Map.Entry<Jun3dPoint, Integer>> vnTreeSet = new TreeSet<Map.Entry<Jun3dPoint, Integer>>(new Comparator<Map.Entry<Jun3dPoint, Integer>>() {
			public int compare(Map.Entry<Jun3dPoint, Integer> o1, Map.Entry<Jun3dPoint, Integer> o2) {
				return o1.getValue().compareTo(o2.getValue());
			}
		});
		vnTreeSet.addAll(normalCoordinates.entrySet());
		for (Map.Entry<Jun3dPoint, Integer> e : vnTreeSet) {
			this.writeNormalVectorOn_(aWriter, e.getKey());
		}
		aWriter.newLine();
	}

	/**
	 * Write the vertex on the specified writer.
	 * 
	 * @param aWriter java.io.BufferedWriter
	 * @param aPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @throws java.io.IOException
	 * @category writing
	 */
	protected void writeVertexOn_(BufferedWriter aWriter, Jun3dPoint aPoint) throws IOException {
		aWriter.write("v ");
		aWriter.write(String.valueOf((float) aPoint.x()));
		aWriter.write(' ');
		aWriter.write(String.valueOf((float) aPoint.y()));
		aWriter.write(' ');
		aWriter.write(String.valueOf((float) aPoint.z()));
		aWriter.newLine();
	}

	/**
	 * Write the UV coordinate on the specified writer.
	 * 
	 * @param aWriter java.io.BufferedWriter
	 * @param aPoint jp.co.sra.jun.geometry.basic.Jun2dPoint
	 * @throws java.io.IOException
	 * @category writing
	 */
	protected void writeUVCoordinateOn_(BufferedWriter aWriter, Jun2dPoint aPoint) throws IOException {
		aWriter.write("vt ");
		aWriter.write(String.valueOf((float) aPoint.x()));
		aWriter.write(' ');
		aWriter.write(String.valueOf((float) aPoint.y()));
		aWriter.newLine();
	}

	/**
	 * Write the normal vector on the specified writer.
	 * 
	 * @param aWriter java.io.BufferedWriter
	 * @param aPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @throws java.io.IOException
	 * @category writing
	 */
	protected void writeNormalVectorOn_(BufferedWriter aWriter, Jun3dPoint aPoint) throws IOException {
		aWriter.write("vn ");
		aWriter.write(String.valueOf((float) aPoint.x()));
		aWriter.write(' ');
		aWriter.write(String.valueOf((float) aPoint.y()));
		aWriter.write(' ');
		aWriter.write(String.valueOf((float) aPoint.z()));
		aWriter.newLine();
	}

	/**
	 * Write the groups on the specified writer.
	 * 
	 * @param aWriter java.io.BufferedWriter
	 * @throws java.io.IOException
	 * @category writing
	 */
	protected void writeGroupsOn_(BufferedWriter aWriter) throws IOException {
		for (JunOBJFaceGroup group : faceGroups) {
			aWriter.write("g ");
			aWriter.write(group.name());
			aWriter.newLine();

			for (Map.Entry<String, List<List<int[]>>> e : group.faces().entrySet()) {
				aWriter.write("usemtl ");
				aWriter.write(e.getKey());
				aWriter.newLine();

				for (List<int[]> elements : e.getValue()) {
					this.writeFaceElementsOn_(aWriter, elements);
				}
				aWriter.newLine();
			}
		}
	}

	/**
	 * Write the face elements on the specified writer.
	 * 
	 * @param aWriter java.io.BufferedWriter
	 * @param elements java.util.List<int[]>
	 * @throws java.io.IOException
	 * @category writing
	 */
	protected void writeFaceElementsOn_(BufferedWriter aWriter, List<int[]> elements) throws IOException {
		// "aCollection is a Collection of faceElements; a faceElement is represented by an Array with 3 elements with the first element representing a vertex, the second a uvCoordinate and the third element a normalVector"

		// "There can be less normalVectors than vertices, in that case the last of the normalVectors applies to the remaining faces"

		// "f int/int/int, or f int//int, ..."
		aWriter.write('f');

		int uvCoordinate = -1;
		int normalVector = -1;
		for (int[] array : elements) {
			aWriter.write(' ');

			aWriter.write(String.valueOf(array[0]));
			aWriter.write('/');

			uvCoordinate = array[1];
			if (uvCoordinate > 0) {
				aWriter.write(String.valueOf(uvCoordinate));
			}
			aWriter.write('/');

			if (array[2] > 0) {
				normalVector = array[2];
			}
			if (normalVector > 0) {
				aWriter.write(String.valueOf(normalVector));
			}
		}

		aWriter.newLine();
	}

	/**
	 * Write the materials on the specified writer.
	 * 
	 * @param aWriter java.io.BufferedWriter
	 * @throws java.io.IOException
	 * @category writing
	 */
	protected void writeMaterialsOn_(BufferedWriter aWriter) throws IOException {
		this.writeMaterialHeaderOn_(aWriter);

		for (JunOBJMaterial material : materials.values()) {
			material.writeOn_(aWriter);
		}
	}

	/**
	 * Write the material header on the specified writer.
	 * 
	 * @param aWriter java.io.BufferedWriter
	 * @throws java.io.IOException
	 * @category writing
	 */
	protected void writeMaterialHeaderOn_(BufferedWriter aWriter) throws IOException {
		aWriter.write("# Alias OBJ Material File");
		aWriter.newLine();
		aWriter.write("# Exported from " + JunSystem.System() + " version " + JunSystem.Version());
		aWriter.newLine();
		aWriter.newLine();
	}

	/**
	 * Write the texture on the specified writer.
	 * 
	 * @param aWriter java.io.BufferedWriter
	 * @throws java.io.IOException
	 * @category writing
	 */
	protected void writeTexturesOn_(BufferedWriter aWriter) throws IOException {
		for (Map.Entry<StImage, String> e : textures.entrySet()) {
			this.writeTexture_withName_(e.getKey(), e.getValue());
		}
	}

	/**
	 * Write the texture image with the name.
	 * 
	 * @param textureImage jp.co.sra.smalltalk.StImage
	 * @param name java.lang.String
	 * @throws IOException 
	 * @category writing
	 */
	protected void writeTexture_withName_(StImage textureImage, String name) throws IOException {
		File aFile = new File(this.directory(), name + ".bmp");

		JunImageStream aStream = null;
		try {
			aStream = JunBmpImageStream.On_(new FileOutputStream(aFile));
			aStream.nextPutImage_(textureImage);
		} finally {
			if (aStream != null) {
				aStream.flush();
				aStream.close();
			}
		}
	}

	/**
	 * Set the texture coordinate block.
	 * 
	 * @category private
	 */
	private void setTextureCoordinateBlock() {
		textureCoordinateBlock = new StBlockClosure1<Jun2dPoint, Jun3dPoint>() {
			public Jun2dPoint value(Jun3dPoint aPoint) {
				JunOpenGLTexture texture = JunOBJWriter.this.texture();
				Jun2dPoint[] points = texture.coordinates();
				if (textureCoordinateIndex() > points.length - 1) {
					textureCoordinateIndex_(0);
				}
				Jun2dPoint point = points[textureCoordinateIndex()];
				textureCoordinateIndex_(textureCoordinateIndex() + 1);
				return point;
			}
		};
	}

	/**
	 * Create a texture image name for the specified JunOpenGL3dObject.
	 * 
	 * @param anOpenGL3dObject jp.co.sra.jun.opengl.objects.JunOpenGL3dObject
	 * @return java.lang.String
	 * @category private
	 */
	private String textureImageNameFor_(JunOpenGL3dObject anOpenGL3dObject) {
		if (anOpenGL3dObject == null || anOpenGL3dObject.texture() == null) {
			return "";
		}

		if (textures.containsKey(anOpenGL3dObject.texture().image()) == false) {
			textures.put(anOpenGL3dObject.texture().image(), this.modelName() + "_Tex_" + String.valueOf(textures.size() + 1));
		}
		return textures.get(anOpenGL3dObject.texture().image());
	}

	/**
	 * Create a material key for the specified JunOpenGL3dObject.
	 * 
	 * @param anOpenGL3dObject jp.co.sra.jun.opengl.objects.JunOpenGL3dObject
	 * @return jp.co.sra.jun.obj.JunMaterialKey
	 * @category private material
	 */
	private JunMaterialKey materialKeyFor_(JunOpenGL3dObject anOpenGL3dObject) {
		return new JunMaterialKey(anOpenGL3dObject.paint(), anOpenGL3dObject.alpha(), this.textureImageNameFor_(anOpenGL3dObject));
	}

	/**
	 * Create a colors based material key for the specified JunOpenGL3dVertexesObject.
	 * 
	 * @param anOpenGL3dVertexesObject jp.co.sra.jun.opengl.objects.JunOpenGL3dVertexesObject
	 * @return jp.co.sra.jun.obj.JunMaterialKey
	 * @category private material
	 */
	private JunMaterialKey colorsBasedMaterialKeyFor_(JunOpenGL3dVertexesObject anOpenGL3dVertexesObject) {
		int red = 0;
		int green = 0;
		int blue = 0;

		Color[] colors = anOpenGL3dVertexesObject.colors();
		for (Color c : colors) {
			red += c.getRed();
			green += c.getGreen();
			blue += c.getBlue();
		}

		int n = colors.length;
		Color aColor = new Color(red / n, green / n, blue / n);
		return new JunMaterialKey(aColor);
	}

	/**
	 * Create a new material for the specified JunOpenGL3dObject with the specified texture image name.
	 * 
	 * @param anOpenGL3dObject jp.co.sra.jun.opengl.objects.JunOpenGL3dObject
	 * @param textureImageName java.lang.String
	 * @return jp.co.sra.jun.obj.JunOBJMaterial
	 * @category private material
	 */
	private JunOBJMaterial newMaterialFor_withTextureImageName_(JunOpenGL3dObject anOpenGL3dObject, String textureImageName) {
		return this.newMaterialForPaint_transparency_textureImageName(anOpenGL3dObject.paint(), anOpenGL3dObject.alpha(), textureImageName);
	}

	/**
	 * Create a new material with the specified color value.
	 * 
	 * @param colorValue java.lang.Color
	 * @return jp.co.sra.jun.obj.JunOBJMaterial
	 * @category private material
	 */
	private JunOBJMaterial newMaterialForPaint_(Color colorValue) {
		return this.newMaterialForPaint_transparency_textureImageName(colorValue, 1.0f, "");
	}

	/**
	 * Create a new material with the specified attributes.
	 * 
	 * @param colorValue java.lang.Color
	 * @param transparency float
	 * @param textureImageName java.lang.String
	 * @return jp.co.sra.jun.obj.JunOBJMaterial
	 * @category private material
	 */
	private JunOBJMaterial newMaterialForPaint_transparency_textureImageName(Color colorValue, float transparency, String textureImageName) {
		JunOBJMaterial material = new JunOBJMaterial();
		material.colorValue_(colorValue);
		material.transparency_(transparency);
		material.textureImageName_(textureImageName);
		material.name_(this.modelName() + "_Mtl_" + String.valueOf(materials.size() + 1));
		return material;
	}

}
