package jp.co.sra.jun.opengl.catalog;

import java.io.File;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Map;

import jp.co.sra.smalltalk.StBlockClosure1;
import jp.co.sra.smalltalk.StReadStream;
import jp.co.sra.smalltalk.StSymbol;

import jp.co.sra.jun.geometry.basic.Jun2dPoint;
import jp.co.sra.jun.goodies.lisp.JunLispCons;
import jp.co.sra.jun.goodies.lisp.JunLispList;
import jp.co.sra.jun.goodies.lisp.JunLispParser;
import jp.co.sra.jun.goodies.tables.JunAttributeTable;
import jp.co.sra.jun.opengl.display.JunOpenGLDisplayLight;
import jp.co.sra.jun.opengl.display.JunOpenGLDisplayModel;
import jp.co.sra.jun.opengl.objects.JunOpenGL3dObject;
import jp.co.sra.jun.system.framework.JunAbstractObject;

/**
 * JunOpenGLCatalogElement class
 * 
 *  @author    nisinaka
 *  @created   2009/06/15 (by nisinaka)
 *  @updated   N/A
 *  @version   758 (with StPL9.7) based on Jun771 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunOpenGLCatalogElement.java,v 9.3 2010/03/09 09:17:33 nisinaka Exp $
 */
public class JunOpenGLCatalogElement extends JunAbstractObject {

	protected JunOpenGL3dObject displayObject;
	protected JunAttributeTable<StSymbol, Object> preferenceTable;

	/**
	 * Create a new instance of JunOpenGLCatalogElement and initialize it.
	 *
	 * @param anObject jp.co.sra.jun.opengl.objects.JunOpenGL3dObject
	 * @category Instance creation
	 */
	public JunOpenGLCatalogElement(JunOpenGL3dObject anObject) {
		this.displayObject_(anObject);
	}

	/**
	 * Create a new instance of JunOpenGLCatalogElement and initialize it with the specified lisp list.
	 *
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category Instance creation
	 */
	public JunOpenGLCatalogElement(JunLispList aList) {
		super(aList);
	}

	/**
	 * Create a new instance of JunOpenGLCatalogElement and initialize it.
	 *
	 * @param aFile java.io.File
	 * @throws java.io.IOException
	 * @category Instance creation
	 */
	public static JunOpenGLCatalogElement FromFile(File aFile) throws IOException {
		JunOpenGL3dObject displayObject = null;

		StReadStream readStream = null;
		try {
			readStream = new StReadStream(aFile);
			displayObject = JunOpenGL3dObject.LoadFrom_(readStream);

		} finally {
			if (readStream != null) {
				readStream.close();
			}
		}

		return new JunOpenGLCatalogElement(displayObject);
	}

	/**
	 * Initialize the receiver.
	 * 
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#initialize()
	 * @category initialize-release
	 */
	@Override
	protected void initialize() {
		super.initialize();
		displayObject = null;
		preferenceTable = null;
	}

	/**
	 * Answer my current display object.
	 * 
	 * @return jp.co.sra.jun.opengl.objects.JunOpenGL3dObject
	 * @category accessing
	 */
	public JunOpenGL3dObject displayObject() {
		return displayObject;
	}

	/**
	 * Set my new display object.
	 * 
	 * @param anObject jp.co.sra.jun.opengl.objects.JunOpenGL3dObject
	 * @category accessing
	 */
	public void displayObject_(JunOpenGL3dObject anObject) {
		displayObject = anObject;
	}

	/**
	 * Answer the name of my current display object.
	 * 
	 * @return java.lang.String
	 * @category accessing
	 */
	public String displayObjectName() {
		if (this.hasDisplayObject() == false) {
			return "";
		}
		return this.displayObject().name();
	}

	/**
	 * Set the name of my current display object.
	 * 
	 * @param aString java.lang.String
	 * @category accessing
	 */
	public void displayObjectName_(String aString) {
		if (this.hasDisplayObject()) {
			this.displayObject().name_(aString);
		}
	}

	/**
	 * Answer my current preference table.
	 * 
	 * @return jp.co.sra.jun.goodies.tables.JunAttributeTable<StSymbol, Object>
	 * @category preferences
	 */
	public JunAttributeTable<StSymbol, Object> preferenceTable() {
		if (preferenceTable == null) {
			preferenceTable = new JunAttributeTable<StSymbol, Object>();
		}
		return preferenceTable;
	}

	/**
	 * Set my new preference table.
	 * 
	 * @param anAttributeTable jp.co.sra.jun.goodies.tables.JunAttributeTable<StSymbol, Object?>
	 * @category preferences
	 */
	public void preferenceTable_(JunAttributeTable<StSymbol, Object> anAttributeTable) {
		preferenceTable = anAttributeTable;
	}

	/**
	 * Answer my current projection table.
	 * 
	 * @return java.util.Map<StSymbol, Object>
	 * @category preferences
	 */
	public Map<StSymbol, Object> projectionTable() {
		String aString = (String) this.preferenceTable().at_($("projectionTable"));
		if (aString == null) {
			return null;
		}

		JunLispCons aList = (JunLispCons) JunLispParser.Parse_(aString);
		if (aList.head() != $("projectionTable")) {
			return null;
		}

		return JunAttributeTable.<StSymbol, Object> FromLispList_((JunLispList) aList.tail()).toMap();
	}

	/**
	 * Set my new projection table.
	 * 
	 * @param aMap java.util.Map<StSymbol, Object>
	 * @category preferences
	 */
	public void projectionTable_(Map<StSymbol, Object> aMap) {
		JunLispCons list = this.lispCons();
		list.head_($("projectionTable"));
		list.tail_(JunAttributeTable.FromMap_(aMap).toLispList());

		this.preferenceTable().at_put_($("projectionTable"), "\n" + list.saveString());
	}

	/**
	 * Answer my current display lights.
	 * 
	 * @return jp.co.sra.jun.opengl.display.JunOpenGLDisplayLight[]
	 * @category preferences
	 */
	public JunOpenGLDisplayLight[] displayLights() {
		String aString = (String) this.preferenceTable().at_($("displayLights"));
		if (aString == null) {
			return null;
		}

		JunLispCons aList = (JunLispCons) JunLispParser.Parse_(aString);
		if (aList.head() != $("displayLights")) {
			return null;
		}

		Object[] lists = ((JunLispList) aList.tail()).asArray();
		JunOpenGLDisplayLight[] lights = new JunOpenGLDisplayLight[lists.length];
		for (int i = 0; i < lists.length; i++) {
			lights[i] = new JunOpenGLDisplayLight((JunLispList) lists[i]);
		}
		return lights;
	}

	/**
	 * Set my new display lights.
	 * 
	 * @param displayLights jp.co.sra.jun.opengl.display.JunOpenGLDisplayLight[]
	 * @category preferences
	 */
	public void displayLights_(JunOpenGLDisplayLight[] displayLights) {
		JunLispCons list = this.lispCons();
		list.head_($("displayLights"));

		ArrayList<JunLispList> lists = new ArrayList<JunLispList>();
		for (JunOpenGLDisplayLight light : displayLights) {
			final ArrayList<Object> collection = new ArrayList<Object>();
			light.toLispList().do_(new StBlockClosure1<Void, Object>() {
				public Void value(Object anObject) {
					if (anObject instanceof JunLispList == false || ((JunLispList) anObject).head() != $("lightName")) {
						collection.add(anObject);
					}
					return null;
				}
			});
			lists.add(JunLispCons.List_(collection));
		}

		list.tail_(JunLispCons.List_(lists));

		this.preferenceTable().at_put_($("displayLights"), "\n" + list.saveString());
	}

	/**
	 * Answer my current movement vector.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.Jun2dPoint
	 * @category preferences
	 */
	public Jun2dPoint movementVector() {
		String aString = (String) this.preferenceTable().at_($("movementVector"));
		if (aString == null) {
			return null;
		}

		JunLispCons aList = (JunLispCons) JunLispParser.Parse_(aString);
		if (aList.head() != $("movementVector")) {
			return null;
		}

		return new Jun2dPoint((JunLispList) aList.tail());
	}

	/**
	 * Set my new movement vector.
	 * 
	 * @param aPoint jp.co.sra.jun.geometry.basic.Jun2dPoint
	 * @category preferences
	 */
	public void movementVector_(Jun2dPoint aPoint) {
		if (aPoint == null) {
			this.preferenceTable().removeKey_($("movementVector"));
			return;
		}

		JunLispCons list = this.lispCons();
		list.head_($("movementVector"));
		list.tail_(aPoint.toLispList());

		this.preferenceTable().at_put_($("movementVector"), "\n" + list.saveString());
	}

	/**
	 * Apply the preferences from the specified display model.
	 * 
	 * @param aDisplayModel jp.co.sra.jun.opengl.display.JunOpenGLDisplayModel
	 * @category preferences
	 */
	public void applyPreferecesFrom_(JunOpenGLDisplayModel aDisplayModel) {
		if (aDisplayModel == null || aDisplayModel.displayObject() == null) {
			return;
		}

		this.projectionTable_(aDisplayModel.projectionTable());
		this.displayLights_(aDisplayModel.displayLights());
		this.movementVector_(aDisplayModel.movementVector());
	}

	/**
	 * Apply the preferences to the specified display model.
	 * 
	 * @param aDisplayModel jp.co.sra.jun.opengl.display.JunOpenGLDisplayModel
	 * @category preferences
	 */
	public void applyPreferencesTo_(JunOpenGLDisplayModel aDisplayModel) {
		aDisplayModel.displayObject_(this.displayObject());

		Map<StSymbol, Object> projectionTable = this.projectionTable();
		if (projectionTable == null) {
			aDisplayModel.resetView();
		} else {
			aDisplayModel.setProjectionTable_(projectionTable);
		}

		JunOpenGLDisplayLight[] displayLights = this.displayLights();
		if (displayLights == null) {
			aDisplayModel.displayLights_(new JunOpenGLDisplayModel().displayLights());
		} else {
			aDisplayModel.displayLights_(displayLights);
		}

		Jun2dPoint movementVector = this.movementVector();
		if (movementVector == null) {
			aDisplayModel.flushMovementVector();
			aDisplayModel.movementVector_(null);
		} else {
			aDisplayModel.flushMovementVector();
			aDisplayModel.movementVector_(movementVector);
		}

		aDisplayModel.updateMenuIndication();
	}

	/**
	 * Answer true if the receiver has a display object, otherwise false.
	 * 
	 * @return boolean
	 * @category testing
	 */
	public boolean hasDisplayObject() {
		return displayObject != null;
	}

	/**
	 * Print my string representation on the writer.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws java.io.IOException
	 * @see jp.co.sra.smalltalk.StObject#printOn_(java.io.Writer)
	 * @category printing
	 */
	@Override
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write("CatalogElement(");
		aWriter.write(this.displayObjectName());
		aWriter.write(")");
	}

	/**
	 * Get the receiver's attributes from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#fromLispList(jp.co.sra.jun.goodies.lisp.JunLispList)
	 * @category lisp support
	 */
	@Override
	public void fromLispList(JunLispList aList) {
		super.fromLispList(aList);

		this.displayObjectFromLispList(aList);
		this.preferenceTableFromLispList(aList);
	}

	/**
	 * Convert the receiver's attributes to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#toLispList()
	 * @category lisp support
	 */
	@Override
	public JunLispCons toLispList() {
		JunLispCons list = super.toLispList();

		if (this.hasDisplayObject()) {
			list.add_(this.displayObjectToLispList());
		}
		if (this.preferenceTable().isEmpty() == false) {
			list.add_(this.preferenceTableToLispList());
		}

		return list;
	}

	/**
	 * Get the receiver's display object from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected void displayObjectFromLispList(JunLispList aList) {
		JunLispList displayObjectList = aList._findSublistWithHead($("displayObject"));
		if (displayObjectList == null) {
			return;
		}

		this.displayObject_(JunOpenGL3dObject.ObjectFromLispList_((JunLispCons) displayObjectList.tail()));
	}

	/**
	 * Convert the receiver's display object to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispCons
	 * @category lisp support
	 */
	protected JunLispCons displayObjectToLispList() {
		JunLispCons list = this.lispCons();
		list.head_($("displayObject"));
		list.tail_(this.displayObject().toLispList());
		return list;
	}

	/**
	 * Get the receiver's preference table from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected void preferenceTableFromLispList(JunLispList aList) {
		JunLispList preferenceTableList = aList._findSublistWithHead($("preferenceTable"));
		if (preferenceTableList == null) {
			return;
		}

		this.preferenceTable_(JunAttributeTable.<StSymbol, Object> FromLispList_((JunLispList) preferenceTableList.tail()));
	}

	/**
	 * Convert the receiver's preference table to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispCons
	 * @category lisp support
	 */
	protected JunLispCons preferenceTableToLispList() {
		JunLispCons list = this.lispCons();
		list.head_($("preferenceTable"));
		list.tail_(this.preferenceTable().toLispList());
		return list;
	}

}
