package jp.co.sra.jun.system.dialog;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.Proxy;
import java.net.URL;
import java.net.URLConnection;

import jp.co.sra.smalltalk.StSimpleDialog;
import jp.co.sra.smalltalk.StView;

import jp.co.sra.jun.system.framework.JunDialog;

/**
 * JunInternetConnectionDialog class
 * 
 *  @author    nisinaka
 *  @created   2008/08/15 (by nisinaka)
 *  @updated   N/A
 *  @version   758 (with StPL9.7) based on Jun704 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunInternetConnectionDialog.java,v 9.7 2010/03/09 09:17:47 nisinaka Exp $
 */
public class JunInternetConnectionDialog extends StSimpleDialog {

	protected static boolean DefaultUseHttpProxy = false;
	protected static String DefaultProxyHost = "proxy.example.com";
	protected static int DefaultPortNumber = 80;

	protected Proxy proxy;
	protected boolean useHttpProxy;
	protected String proxyHost;
	protected int portNumber;

	/**
	 * Open a dialog for requesting a proxy information.
	 * 
	 * @return java.net.Proxy
	 * @category Utilities
	 */
	public static Proxy Request() {
		JunInternetConnectionDialog aDialog = new JunInternetConnectionDialog();
		aDialog.openDialog();
		return aDialog.accept().value() ? aDialog.getProxy() : null;
	}

	/**
	 * Initialize the receiver.
	 * 
	 * @see jp.co.sra.smalltalk.StSimpleDialog#initialize()
	 * @category initialize-release
	 */
	@Override
	protected void initialize() {
		super.initialize();

		this.flushProxy();
		useHttpProxy = DefaultUseHttpProxy;
		proxyHost = DefaultProxyHost;
		portNumber = DefaultPortNumber;
	}

	/**
	 * Answer my current proxy.
	 * 
	 * @return java.net.Proxy
	 * @category accessing
	 */
	public Proxy getProxy() {
		if (proxy == null) {
			if (this.useHttpProxy()) {
				proxy = new Proxy(Proxy.Type.HTTP, new InetSocketAddress(this.proxyHost(), this.portNumber()));
			} else {
				proxy = Proxy.NO_PROXY;
			}

			DefaultUseHttpProxy = this.useHttpProxy();
			DefaultProxyHost = this.proxyHost();
			DefaultPortNumber = this.portNumber();
		}
		return proxy;
	}

	/**
	 * Answer my current proxy host.
	 * 
	 * @return java.lang.String
	 * @category accessing
	 */
	protected String proxyHost() {
		if (proxyHost == null) {
			proxyHost = DefaultProxyHost;
		}
		return proxyHost;
	}

	/**
	 * Set my new proxy host.
	 * 
	 * @param aString java.lang.String
	 * @category accessing
	 */
	protected void proxyHost_(String aString) {
		if (this.proxyHost().equals(aString)) {
			return;
		}

		proxyHost = aString;

		this.flushProxy();
		this.changed_($("proxyHost"));
	}

	/**
	 * Answer my current port number.
	 * 
	 * @return int
	 * @category accessing
	 */
	protected int portNumber() {
		if (portNumber <= 0) {
			portNumber = DefaultPortNumber;
		}
		return portNumber;
	}

	/**
	 * Set my new port number.
	 * 
	 * @param aNumber int
	 * @category accessing
	 */
	protected void portNumber_(int aNumber) {
		if (this.portNumber() == aNumber) {
			return;
		}

		portNumber = aNumber;

		this.flushProxy();
		this.changed_($("portNumber"));
	}

	/**
	 * Answer true if the current specification is to use HTTP proxy, otherwise false.
	 * 
	 * @return boolean
	 * @category accessing
	 */
	protected boolean useHttpProxy() {
		return useHttpProxy;
	}

	/**
	 * Set whether the current specification is to use HTTP proxy or not.
	 * 
	 * @param aBoolean boolean
	 * @category accessing
	 */
	protected void useHttpProxy_(boolean aBoolean) {
		if (useHttpProxy == aBoolean) {
			return;
		}

		useHttpProxy = aBoolean;

		this.flushProxy();
		this.changed_($("useHttpProxy"));
	}

	/**
	 * Flush the current proxy.
	 * 
	 * @category flushing
	 */
	protected void flushProxy() {
		proxy = null;
	}

	/**
	 * Answer true if the action for the accept button can be activated.
	 * 
	 * @category actions
	 */
	protected boolean acceptButtonPressed() {
		try {
			URL aURL = new URL("http://www.microsoft.com");
			URLConnection aURLConnection = aURL.openConnection(this.getProxy());
			aURLConnection.connect();
			return true;

		} catch (IOException e) {
			JunDialog.Warn_(this.errorMessage());
		}

		return false;
	}

	/**
	 * Answer true if the action for the cancel button can be activated.
	 * 
	 * @category actions
	 */
	protected boolean cancelButtonPressed() {
		this.flushProxy();
		return true;
	}

	/**
	 * Action for the direct connection button.
	 * 
	 * @category actions
	 */
	protected void directConnectionButtonPressed() {
		this.useHttpProxy_(false);
	}

	/**
	 * Action for the use HTTP proxy button.
	 * 
	 * @category actions
	 */
	protected void useHttpProxyButtonPressed() {
		this.useHttpProxy_(true);
	}

	/**
	 * Answer a default view.
	 * 
	 * @return jp.co.sra.smalltalk.StView
	 * @see jp.co.sra.smalltalk.StSimpleDialog#defaultView()
	 * @category interface opening
	 */
	@Override
	public StView defaultView() {
		if (GetDefaultViewMode() == VIEW_AWT) {
			return new JunInternetConnectionDialogViewAwt(this);
		} else {
			return new JunInternetConnectionDialogViewSwing(this);
		}
	}

	/**
	 * Answer a window title.
	 * 
	 * @return java.lang.String
	 * @see jp.co.sra.smalltalk.StSimpleDialog#windowTitle()
	 * @category interface opening
	 */
	@Override
	protected String windowTitle() {
		return $String("Select your internet connection.");
	}

	/**
	 * Answer the error message.
	 * 
	 * @return java.lang.String
	 * @category private
	 */
	protected String errorMessage() {
		String message = $String("Can't establish internet connection.");
		message += "\n";
		message += $String("Set your internet connection correctly.");
		return message;
	}

}
