package jp.co.sra.jun.system.framework;

import java.awt.Color;
import java.awt.Frame;
import java.awt.TextArea;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import jp.co.sra.smalltalk.StBlockClosure1;
import jp.co.sra.smalltalk.StObject;
import jp.co.sra.smalltalk.StSymbol;

import jp.co.sra.jun.goodies.lisp.JunLispCons;
import jp.co.sra.jun.goodies.lisp.JunLispList;
import jp.co.sra.jun.goodies.lisp.JunLispNil;
import jp.co.sra.jun.goodies.tables.JunAttributeTable;
import jp.co.sra.jun.system.support.JunSystem;

/**
 * JunAbstractObject class
 * 
 *  @author    Mitsuhiro Asada
 *  @created   2008/01/30 (by m-asada)
 *  @updated   N/A
 *  @version   758 (with StPL9.7) based on Jun719 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunAbstractObject.java,v 9.10 2010/03/09 09:17:48 nisinaka Exp $
 */
public class JunAbstractObject extends StObject {

	/**
	 * Show a window which displays the specified Lisp List.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category Lisp support
	 */
	public static JunLispList ShowLispList_(JunLispList aList) {
		TextArea textArea = new TextArea(5, 40);
		textArea.setEditable(false);
		textArea.append(aList.saveString());

		Frame aFrame = new Frame();
		aFrame.add("Center", textArea);
		aFrame.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				e.getWindow().dispose();
			}
		});
		aFrame.setTitle($String("Lisp List (S expression)"));
		aFrame.pack();
		aFrame.setSize(400, 300);
		aFrame.setVisible(true);

		return aList;
	}

	/**
	 * Create a JunAttributeTable from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @return jp.co.sra.jun.goodies.tables.JunAttributeTable<Object, Object>
	 * @category Lisp support
	 */
	public static JunAttributeTable<Object, Object> TableFromLispList_(JunLispList aList) {
		final JunAttributeTable<Object, Object> aTable = new JunAttributeTable<Object, Object>();
		aList.do_(new StBlockClosure1<Void, Object>() {
			public Void value(Object anObject) {
				if (anObject instanceof JunLispCons) {
					JunLispCons pair = (JunLispCons) anObject;
					aTable.at_put_(pair.head(), pair.tail());
				}
				return null;
			}
		});
		return aTable;
	}

	/**
	 * Create a Color from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category Lisp support
	 */
	public static Color ColorValueFromLispList_(JunLispList aList) {
		assert aList.head() == $("ColorValue");
		JunAttributeTable<Object, Object> aTable = TableFromLispList_((JunLispList) aList.tail());
		float r = ((Number) aTable.at_($("red"))).floatValue();
		float g = ((Number) aTable.at_($("green"))).floatValue();
		float b = ((Number) aTable.at_($("blue"))).floatValue();
		return new Color(r, g, b);
	}

	/**
	 * Convert the color to a lisp list.
	 * 
	 * @param aColor java.awt.Color
	 * @return jp.co.sra.jun.goodies.lisp.JunLispCons
	 * @category Lisp support
	 */
	public static JunLispCons ColorValueToLispList_(Color aColor) {
		JunLispCons list = JunLispCons.Cell();
		list.head_($("ColorValue"));
		float[] rgb = aColor.getRGBColorComponents(null);
		list.add_(new JunLispCons($("red"), rgb[0]));
		list.add_(new JunLispCons($("green"), rgb[1]));
		list.add_(new JunLispCons($("blue"), rgb[2]));
		return list;
	}

	/**
	 * Create a new instance of <code>JunAbstractObject</code> and initialize it.
	 * 
	 * @category Instance creation
	 */
	public JunAbstractObject() {
		super();
		this.initialize();
	}

	/**
	 * Create a new instance of <code>JunAbstractObject</code> and initialize it.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category Instance creation
	 */
	public JunAbstractObject(JunLispList aList) {
		this();
		this.fromLispList(aList);
	}

	/**
	 * Initialize the receiver when created.
	 * 
	 * @category initialize-release
	 */
	protected void initialize() {
		// Nothing to do by default.
	}

	/**
	 * Answer the string of default base name.
	 *
	 * @return java.lang.String
	 * @category defaults
	 */
	public String defaultBaseName() {
		return JunSystem.DefaultBaseName();
	}

	/**
	 * Answer the kind name of the element.
	 * 
	 * @return jp.co.sra.smalltalk.StSymbol
	 * @category lisp support
	 */
	protected StSymbol kindName() {
		return this._className();
	}

	/**
	 * Answer a new lisp cons cell.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispCons
	 * @category lisp support
	 */
	protected JunLispCons lispCons() {
		return JunLispCons.Cell();
	}

	/**
	 * Create an instance of JunLispNil.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispNil
	 * @category lisp support
	 */
	protected JunLispNil lispNil() {
		return JunLispList.NullList();
	}

	/**
	 * Get the receiver's attributes from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	public void fromLispList(JunLispList aList) {
		// Nothing to do by default.
	}

	/**
	 * Convert the receiver's attributes to a lisp list.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	public JunLispCons toLispList() {
		JunLispCons list = this.lispCons();
		list.head_(this.kindName());
		return list;
	}

	/**
	 * Create a JunAttributeTable from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @return jp.co.sra.jun.goodies.tables.JunAttributeTable<Object, Object>
	 * @category lisp support
	 */
	protected JunAttributeTable<Object, Object> tableFromLispList_(JunLispList aList) {
		return TableFromLispList_(aList);
	}

	/**
	 * Create a Color from the lisp list.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	protected Color colorValueFromLispList(JunLispList aList) {
		return ColorValueFromLispList_(aList);
	}

	/**
	 * Convert the color to a lisp list.
	 * 
	 * @param aColor java.awt.Color
	 * @return jp.co.sra.jun.goodies.lisp.JunLispCons
	 * @category lisp support
	 */
	protected JunLispCons colorValueToLispList(Color aColor) {
		return ColorValueToLispList_(aColor);
	}

	/**
	 * Show a window which displays the Lisp List of the receiver.
	 * 
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	public JunLispList showLispList() {
		return this.showLispList_(this.toLispList());
	}

	/**
	 * Show a window which displays the specified Lisp List.
	 * 
	 * @param aList jp.co.sra.jun.goodies.lisp.JunLispList
	 * @return jp.co.sra.jun.goodies.lisp.JunLispList
	 * @category lisp support
	 */
	public JunLispList showLispList_(JunLispList aList) {
		return ShowLispList_(aList);
	}

}
