package jp.co.sra.jun.threeDImage.basic;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;

import jp.co.sra.jun.geometry.basic.Jun3dPoint;
import jp.co.sra.jun.system.framework.JunAbstractObject;
import jp.co.sra.jun.threeDImage.colors.JunColorValue;
import jp.co.sra.jun.threeDImage.support.Jun3dGraphicsContext;

/**
 * Jun3dImage class
 * 
 *  @author    kondo
 *  @created   1999/09/06 (by kondo)
 *  @updated   N/A
 *  @version   758 (with StPL9.7) based on Jun697 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2005 Information-technology Promotion Agency, Japan (IPA)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: Jun3dImage.java,v 9.10 2010/03/09 09:17:52 nisinaka Exp $
 */
public class Jun3dImage extends JunAbstractObject {

	private Color background;
	private Color[][] colors;
	private Jun3dPoint extent;
	private int yStep;

	/**
	 * Create a new instance of <code>Jun3dImage</code> and initialize it.
	 * 
	 * @param aJun3dPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @return jp.co.sra.jun.threeDImage.basic.Jun3dImage
	 * @category Instance creation
	 * @deprecated since Jun701, using constructor
	 */
	public static Jun3dImage Extent_(Jun3dPoint aJun3dPoint) {
		return new Jun3dImage(aJun3dPoint);
	}

	/**
	 * Create a new instance of <code>Jun3dImage</code> and initialize it.
	 * 
	 * @param aJun3dPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @return jp.co.sra.jun.threeDImage.basic.Jun3dImage
	 * @category Instance creation
	 */
	public Jun3dImage(Jun3dPoint aJun3dPoint) {
		super();
		this.extent_(aJun3dPoint);
	}

	/**
	 * Please refer to the corresponding method in Smalltalk.
	 * 
	 * @param anInteger1 int
	 * @param anInteger2 int
	 * @param anInteger3 int
	 * @param aJunColorValue java.awt.Color
	 * @category private
	 */
	public void atX_y_z_put_(int anInteger1, int anInteger2, int anInteger3, Color aJunColorValue) {
		this.colors[this.indexAtX_y_(anInteger1, anInteger2)][anInteger3 + 1] = aJunColorValue;
	}

	/**
	 * Answer the receiver's background color.
	 * 
	 * @return java.awt.Color
	 * @category accessing
	 */
	public Color background() {
		return this.background;
	}

	/**
	 * Set the receiver's background color.
	 * 
	 * @param aColor java.awt.Color
	 * @category accessing
	 */
	public void background_(Color aColor) {
		background = aColor;
	}

	/**
	 * Answer the receiver's depth.
	 * 
	 * @return int
	 * @category accessing
	 */
	public int depth() {
		return (int) this.extent().z();
	}

	/**
	 * Answer the receiver's extent.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category accessing
	 */
	public Jun3dPoint extent() {
		return extent;
	}

	/**
	 * Please refer to the corresponding method in Smalltalk.
	 * 
	 * @return jp.co.sra.jun.threeDImage.support.Jun3dGraphicsContext
	 * @category accessing
	 */
	public Jun3dGraphicsContext graphicsContext() {
		return new Jun3dGraphicsContext(this);
	}

	/**
	 * Answer the receiver's height.
	 * 
	 * @return int
	 * @category accessing
	 */
	public int height() {
		return (int) this.extent().y();
	}

	/**
	 * Answer the receiver's width.
	 * 
	 * @return int
	 * @category accessing
	 */
	public int width() {
		return (int) this.extent().x();
	}

	/**
	 * Please refer to the corresponding method in Smalltalk.
	 * 
	 * @param anInteger1 int
	 * @param anInteger2 int
	 * @param anInteger3 int
	 * @return java.awt.Color
	 * @category private
	 */
	private Color atX_y_z_(int anInteger1, int anInteger2, int anInteger3) {
		Color color = this.colors[this.indexAtX_y_(anInteger1, anInteger2)][anInteger3 + 1];

		if (color == null) {
			return this.background();
		} else {
			return color;
		}
	}

	/**
	 * Please refer to the corresponding method in Smalltalk.
	 * 
	 * @param anArrayOfColor java.awt.Color[]
	 * @return java.awt.Color
	 * @category private
	 */
	private Color composeColorFromColors_(Color[] anArrayOfColor) {
		Color color = JunColorValue.Transparent();
		for (int index = 0; index < anArrayOfColor.length; index++) {
			Color c = anArrayOfColor[index];
			Color underColor;
			if (c == null) {
				underColor = JunColorValue.Transparent();
			} else {
				underColor = c;
			}
			color = JunColorValue.BlendOver_(color, underColor);
			if (JunColorValue.IsSolid_(color)) {
				return color;
			}
		}

		return JunColorValue.BlendOver_(color, this.background());
	}

	/**
	 * Answer the receiver's default background color.
	 * 
	 * @return java.awt.Color
	 * @category defaults
	 */
	public Color defaultBackground() {
		return Color.white;
	}

	/**
	 * Display the receiver on the graphics at the specified point.
	 * 
	 * @param aGraphics java.awt.Graphics
	 * @param aPoint java.awt.Point
	 * @see jp.co.sra.smalltalk.StDisplayable#displayOn_at_(java.awt.Graphics, java.awt.Point)
	 * @category displaying
	 */
	public void displayOn_at_(Graphics aGraphics, Point aPoint) {
		for (int x = 0; x < this.width(); x++) {
			for (int y = 0; y < this.height(); y++) {
				int index = this.indexAtX_y_(x, y);
				Color color = this.composeColorFromColors_(colors[index]);
				aGraphics.setColor(color);
				aGraphics.fillRect(aPoint.x + x, aPoint.y + y, 1, 1);
			}
		}
	}

	/**
	 * Set the receiver's extent.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category private
	 */
	private void extent_(Jun3dPoint aJun3dPoint) {
		extent = aJun3dPoint;
		yStep = (int) aJun3dPoint.x();

		int size = (int) (aJun3dPoint.x() * aJun3dPoint.y());
		int zSize = (int) aJun3dPoint.z();
		colors = new Color[size][zSize];
	}

	/**
	 * Please refer to the corresponding method in Smalltalk.
	 * 
	 * @param anInteger1 int
	 * @param anInteger2 int
	 * @return int
	 * @category private
	 */
	private int indexAtX_y_(int anInteger1, int anInteger2) {
		return anInteger1 + 1 + (anInteger2 * this.yStep);
	}

	/**
	 * Initialize the receiver when created.
	 * 
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#initialize()
	 * @category initialize-release
	 */
	protected void initialize() {
		super.initialize();
		this.background = this.defaultBackground();
	}

	/**
	 * Please refer to the corresponding method in Smalltalk.
	 * 
	 * @param aJun3dPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @return java.awt.Color
	 * @category bits access
	 */
	protected Color valueAtPoint_(Jun3dPoint aJun3dPoint) {
		return this.atX_y_z_((int) aJun3dPoint.x(), (int) aJun3dPoint.y(), (int) aJun3dPoint.z());
	}

	/**
	 * Please refer to the corresponding method in Smalltalk.
	 * 
	 * @param aJun3dPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @param aJunColorValue java.awt.Color
	 * @category bits access
	 */
	protected void valueAtPoint_put_(Jun3dPoint aJun3dPoint, Color aJunColorValue) {
		this.atX_y_z_put_((int) aJun3dPoint.x(), (int) aJun3dPoint.y(), (int) aJun3dPoint.z(), aJunColorValue);
	}

}
