package jp.co.sra.jun.threeDImage.support;

import java.awt.Color;

import jp.co.sra.smalltalk.StBlockClosure1;
import jp.co.sra.smalltalk.StBlockClosure2;

import jp.co.sra.jun.geometry.abstracts.JunGeometry;
import jp.co.sra.jun.geometry.basic.Jun3dPoint;
import jp.co.sra.jun.geometry.boundaries.Jun3dBoundingBox;
import jp.co.sra.jun.solid.abstracts.JunSolidModel;
import jp.co.sra.jun.system.framework.JunAbstractObject;
import jp.co.sra.jun.threeDImage.basic.Jun3dImage;

/**
 * Jun3dGraphicsContext class
 * 
 *  @author    kondo
 *  @created   1999/09/06 (by kondo)
 *  @updated   2008/05/27 (by Mitsuhiro Asada)
 *  @version   758 (with StPL9.7) based on Jun697 for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2005 Information-technology Promotion Agency, Japan (IPA)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: Jun3dGraphicsContext.java,v 9.10 2010/03/09 09:17:52 nisinaka Exp $
 */
public class Jun3dGraphicsContext extends JunAbstractObject {

	public static final double ACCURACY = JunGeometry.ACCURACY;

	private Jun3dImage medium;
	private Color paint;
	private Jun3dBoundingBox clippingBounds;
	private Jun3dPoint offset;
	private Jun3dPoint scale;

	/**
	 * Create a new instance of <code>Jun3dGraphicsContext</code> and initialize it.
	 * 
	 * @param aJun3dImage jp.co.sra.jun.threeDImage.basic.Jun3dImage
	 * @return jp.co.sra.jun.threeDImage.support.Jun3dGraphicsContext
	 * @deprecated since Jun701, using constructor
	 * @category Instance creation
	 */
	public static Jun3dGraphicsContext On_(Jun3dImage aJun3dImage) {
		return new Jun3dGraphicsContext(aJun3dImage);
	}

	/**
	 * Create a new instance of <code>Jun3dGraphicsContext</code> and initialize it.
	 * 
	 * @param aJun3dImage jp.co.sra.jun.threeDImage.basic.Jun3dImage
	 * @category Instance creation
	 */
	public Jun3dGraphicsContext(Jun3dImage aJun3dImage) {
		super();
		this.setMedium_(aJun3dImage);
	}

	/**
	 * Answer the accuracy.
	 * 
	 * @return double
	 * @category private
	 */
	public double accuracy() {
		return ACCURACY;
	}

	/**
	 * Answer the receiver's clipping bounds.
	 * 
	 * @return jp.co.sra.jun.geometry.boundaries.Jun3dBoundingBox
	 * @category accessing
	 */
	public Jun3dBoundingBox clippingBounds() {
		return clippingBounds;
	}

	/**
	 * Set the receiver's clipping bounds.
	 * 
	 * @param aJun3dBoundingBox jp.co.sra.jun.geometry.boundaries.Jun3dBoundingBox
	 * @category accessing
	 */
	public void clippingBounds_(Jun3dBoundingBox aJun3dBoundingBox) {
		clippingBounds = aJun3dBoundingBox;
	}

	/**
	 * Answer the default clipping bounds.
	 * 
	 * @return jp.co.sra.jun.geometry.boundaries.Jun3dBoundingBox
	 * @category defaults
	 */
	public Jun3dBoundingBox defaultClippingBounds() {
		return Jun3dBoundingBox.Origin_extent_(Jun3dPoint.Zero(), this.medium().extent());
	}

	/**
	 * Answer the default offset point.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category defaults
	 */
	public Jun3dPoint defaultOffset() {
		return Jun3dPoint.Zero();
	}

	/**
	 * Answer the default paint color.
	 * 
	 * @return java.awt.Color
	 * @category defaults
	 */
	public Color defaultPaint() {
		return this.medium().background();
	}

	/**
	 * Answer the default scale size.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category defaults
	 */
	public Jun3dPoint defaultScale() {
		return Jun3dPoint.Unity();
	}

	/**
	 * Please refer to the corresponding method in Smalltalk.
	 * 
	 * @param aJunSolid jp.co.sra.jun.solid.abstracts.JunSolidModel
	 * @param aJun3dPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category displaying
	 */
	public void displaySolid_at_(JunSolidModel aJunSolid, Jun3dPoint aJun3dPoint) {
		final Jun3dGraphicsContext this_ = this;
		final JunSolidModel aJunSolid_ = aJunSolid;
		final Color paint_ = this.paint();
		final double xOffset = this.offset.x() + aJun3dPoint.x();
		final double yOffset = this.offset.y() + aJun3dPoint.y();
		final double zOffset = this.offset.z() + aJun3dPoint.z();
		final double xScale = this.scale().x();
		final double yScale = this.scale().y();
		final double zScale = this.scale().z();
		final Jun3dBoundingBox bounds = this.clippingBounds().intersect_(Jun3dBoundingBox.Origin_extent_(Jun3dPoint.Zero(), this.medium().extent()));

		final StBlockClosure1<Void, StBlockClosure2<Void, Integer, Double>> xIterationBlock;
		if (Math.abs(xScale) > this.accuracy()) {
			xIterationBlock = new StBlockClosure1<Void, StBlockClosure2<Void, Integer, Double>>() {
				public Void value(StBlockClosure2<Void, Integer, Double> block) {
					for (int mediumX = (int) Math.round(bounds.origin().x()); mediumX < Math.round(bounds.corner().x()) - 1; mediumX++) {
						block.value(new Integer(mediumX), new Double((mediumX - xOffset) / xScale));
					}
					return null;
				}
			};
		} else {
			xIterationBlock = new StBlockClosure1<Void, StBlockClosure2<Void, Integer, Double>>() {
				public Void value(StBlockClosure2<Void, Integer, Double> block) {
					block.value((int) Math.round(bounds.center().x()), new Double(bounds.center().x() - xOffset));
					return null;
				}
			};
		}

		final StBlockClosure1<Void, StBlockClosure2<Void, Integer, Double>> yIterationBlock;
		if (Math.abs(yScale) > this.accuracy()) {
			yIterationBlock = new StBlockClosure1<Void, StBlockClosure2<Void, Integer, Double>>() {
				public Void value(StBlockClosure2<Void, Integer, Double> block) {
					for (int mediumY = (int) Math.round(bounds.origin().y()); mediumY < Math.round(bounds.corner().y()) - 1; mediumY++) {
						block.value(new Integer(mediumY), new Double((mediumY - yOffset) / yScale));
					}
					return null;
				}
			};
		} else {
			yIterationBlock = new StBlockClosure1<Void, StBlockClosure2<Void, Integer, Double>>() {
				public Void value(StBlockClosure2<Void, Integer, Double> block) {
					block.value((int) Math.round(bounds.center().y()), new Double(bounds.center().y() - yOffset));
					return null;
				}
			};
		}

		StBlockClosure1<Void, StBlockClosure2<Void, Integer, Double>> zIterationBlock;
		if (Math.abs(zScale) > this.accuracy()) {
			zIterationBlock = new StBlockClosure1<Void, StBlockClosure2<Void, Integer, Double>>() {
				public Void value(StBlockClosure2<Void, Integer, Double> block) {
					for (int mediumZ = (int) Math.round(bounds.origin().z()); mediumZ < Math.round(bounds.corner().z()) - 1; mediumZ++) {
						block.value(new Integer(mediumZ), new Double((mediumZ - zOffset) / zScale));
					}
					return null;
				}
			};
		} else {
			zIterationBlock = new StBlockClosure1<Void, StBlockClosure2<Void, Integer, Double>>() {
				public Void value(StBlockClosure2<Void, Integer, Double> block) {
					block.value((int) Math.round(bounds.center().z()), new Double(bounds.center().z() - zOffset));
					return null;
				}
			};

		}

		zIterationBlock.value(new StBlockClosure2<Void, Integer, Double>() {
			public Void value(final Integer mediumZ, final Double solidZ) {
				yIterationBlock.value(new StBlockClosure2<Void, Integer, Double>() {
					public Void value(final Integer mediumY, final Double solidY) {
						xIterationBlock.value(new StBlockClosure2<Void, Integer, Double>() {
							public Void value(final Integer mediumX, final Double solidX) {
								if (aJunSolid_.containsPoint_(new Jun3dPoint(solidX, solidY, solidZ))) {
									this_.medium().atX_y_z_put_(mediumX, mediumY, mediumZ, paint_);
								}
								return null;
							}
						});
						return null;
					}
				});
				return null;
			}
		});
	}

	/**
	 * Answer the receiver's medium.
	 * 
	 * @return jp.co.sra.jun.threeDImage.basic.Jun3dImage
	 * @category accessing
	 */
	public Jun3dImage medium() {
		return medium;
	}

	/**
	 * Answer the offset point.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category accessing
	 */
	public Jun3dPoint offset() {
		return offset;
	}

	/**
	 * Set the offset point.
	 * 
	 * @param aJun3dPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category accessing
	 */
	public void offset_(Jun3dPoint aJun3dPoint) {
		offset = aJun3dPoint;
	}

	/**
	 * Answer the paint color.
	 * 
	 * @return java.awt.Color
	 * @category accessing
	 */
	public Color paint() {
		return this.paint;
	}

	/**
	 * Set the paint color.
	 * 
	 * @param aColor java.awt.Color
	 * @category accessing
	 */
	public void paint(Color aColor) {
		this.paint = aColor;
	}

	/**
	 * Answer the scale size.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category accessing
	 */
	public Jun3dPoint scale() {
		return this.scale;
	}

	/**
	 * Set the scale size.
	 * 
	 * @param aJun3dPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category accessing
	 */
	public void scale_(Jun3dPoint aJun3dPoint) {
		this.scale = aJun3dPoint;
	}

	/**
	 * Set the receiver's medium.
	 * 
	 * @param aJun3dImage jp.co.sra.jun.threeDImage.basic.Jun3dImage
	 * @category private
	 */
	protected void setMedium_(Jun3dImage aJun3dImage) {
		medium = aJun3dImage;
		clippingBounds = this.defaultClippingBounds();
		paint = this.defaultPaint();
		offset = this.defaultOffset();
		scale = this.defaultScale();
	}

}
