package jp.co.sra.jun.topology.abstracts;

import jp.co.sra.smalltalk.SmalltalkException;
import jp.co.sra.smalltalk.StValued;

import jp.co.sra.jun.topology.elements.JunBody;

/**
 * JunAbstractOperator class
 * 
 *  @author    ASTI Shanghai
 *  @created   UNKNOWN
 *  @updated   2008/05/22 (by nisinaka)
 *  @version   758 (with StPL9.7) based on JunXXX for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunAbstractOperator.java,v 9.9 2010/03/09 09:17:52 nisinaka Exp $
 */
public abstract class JunAbstractOperator extends JunTopology implements StValued<JunBody> {

	/** A JunBody on which the receiver operate. */
	protected JunBody body;

	/**
	 * Create a new instance of <code>JunAbstractOperator</code> and initialize it.
	 * 
	 * @category Instance creation
	 */
	public JunAbstractOperator() {
		super();
	}

	/**
	 * Create a new instance of JunAbstractOperator and initialize it with a JunBody.
	 * 
	 * @param aBody jp.co.sra.jun.topology.elements.JunBody
	 * @category Instance creation
	 */
	public JunAbstractOperator(JunBody aBody) {
		this();
		this.body_(aBody);
	}

	/**
	 * Initialize the receiver when created.
	 * 
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#initialize()
	 * @category initialize-release
	 */
	protected void initialize() {
		super.initialize();
		body = null;
	}

	/**
	 * Answer the body.
	 * 
	 * @return jp.co.sra.jun.topology.elements.JunBody
	 * @category accessing
	 */
	public final JunBody body() {
		return body;
	}

	/**
	 * Set a body.
	 * 
	 * @param aBody jp.co.sra.jun.topology.elements.JunBody
	 * @category accessing
	 */
	public final void body_(JunBody aBody) {
		body = aBody;
	}

	/**
	 * Do the operation.     - In Smalltalk, the method name is "do".
	 * 
	 * @category executing
	 */
	public void doOperation() {
		this.preExecute();
		this.execute();
		this.postExecute();
	}

	/**
	 * Execute the operation.
	 * 
	 * @category executing
	 */
	public abstract void execute();

	/**
	 * Answer the topological element of the receiver.
	 * 
	 * @param aTopologicalElementOrProxy jp.co.sra.jun.topology.abstracts.JunTopologicalElementOrProxy
	 * @return jp.co.sra.jun.topology.abstracts.JunTopologicalElement
	 * @category accessing
	 */
	public JunTopologicalElement getTopologicalElement_(JunTopologicalElementOrProxy aTopologicalElementOrProxy) {
		if (aTopologicalElementOrProxy == null) {
			return null;
		}
		if (aTopologicalElementOrProxy.isTopologicalProxy()) {
			return ((JunTopologicalElementProxy) aTopologicalElementOrProxy).value_(this.body);
		} else {
			return (JunTopologicalElement) aTopologicalElementOrProxy;
		}
	}

	/**
	 * Answer the topological element proxy of the receiver.
	 * 
	 * @param aTopologicalElementOrProxy jp.co.sra.jun.topology.abstracts.JunTopologicalElementOrProxy
	 * @return jp.co.sra.jun.topology.abstracts.JunTopologicalElementProxy
	 * @category accessing
	 */
	public JunTopologicalElementProxy getTopologicalElementProxy_(JunTopologicalElementOrProxy aTopologicalElementOrProxy) {
		return this.getTopologicalElementProxy_advise_(aTopologicalElementOrProxy, null);
	}

	/**
	 * Answer the topological element proxy of the receiver with advise.
	 * 
	 * @param aTopologicalElementOrProxy jp.co.sra.jun.topology.abstracts.JunTopologicalElementOrProxy
	 * @param aTopologicalElementProxyOrNull jp.co.sra.jun.topology.abstracts.JunTopologicalElementProxy
	 * @return jp.co.sra.jun.topology.abstracts.JunTopologicalElementProxy
	 * @category accessing
	 */
	public JunTopologicalElementProxy getTopologicalElementProxy_advise_(JunTopologicalElementOrProxy aTopologicalElementOrProxy, JunTopologicalElementProxy aTopologicalElementProxyOrNull) {
		if (aTopologicalElementOrProxy == null) {
			return null;
		}
		return aTopologicalElementOrProxy.asProxyIn_advise_(this.body(), aTopologicalElementProxyOrNull);
	}

	/**
	 * Answer the inverse operation of the receiver.
	 * 
	 * @return jp.co.sra.jun.topology.abstracts.JunAbstractOperator
	 * @category inverting
	 */
	public abstract JunAbstractOperator inverse();

	/**
	 * Answer <code>true</code> if the receiver is a kind of Euler Operator, otherwise <code>false</code>.
	 * 
	 * @return boolean
	 * @category testing
	 */
	public boolean isEulerOperator() {
		return false;
	}

	/**
	 * Answer <code>true</code> if the receiver is a kind of Geometric Operator, otherwise <code>false</code>.
	 * 
	 * @return boolean
	 * @category testing
	 */
	public boolean isGeometricOperator() {
		return false;
	}

	/**
	 * Answer <code>true</code> if the receiver is a kind of Global Operator, otherwise <code>false</code>.
	 * 
	 * @return boolean
	 * @category testing
	 */
	public boolean isGlobalOperator() {
		return false;
	}

	/**
	 * Answer <code>true</code> if the receiver can be inverted, otherwise <code>false</code>.
	 * 
	 * @return boolean
	 * @category testing
	 */
	public abstract boolean isInversable();

	/**
	 * Answer <code>true</code> if the post-condition is valid, otherwise <code>false</code>
	 * 
	 * @return boolean
	 * @category executing
	 */
	public boolean postcondition() {
		return true;
	}

	/**
	 * Do something after the execution.
	 * 
	 * @category executing
	 * @throws jp.co.sra.smalltalk.SmalltalkException
	 */
	public void postExecute() {
		if (!postcondition()) {
			throw SmalltalkException.Error("Invalid operation");
		}
	}

	/**
	 * Answer <code>true</code> if the pre-condition is valid, otherwise <code>false</code>.
	 * 
	 * @return boolean
	 * @category executing
	 */
	public boolean precondition() {
		return true;
	}

	/**
	 * Do something before the execution.
	 * 
	 * @category executing
	 * @throws jp.co.sra.smalltalk.SmalltalkException
	 */
	public void preExecute() {
		if (!precondition()) {
			throw SmalltalkException.Error("Invalid operation");
		}
	}

	/**
	 * Undo the operation.     - In Smalltalk, the method name is "undo".
	 * 
	 * @category executing
	 */
	public void undoOperation() {
		this.inverse().doOperation();
	}

	/**
	 * Answer the value of the operator.
	 * 
	 * @return java.lang.Object (jp.co.sra.jun.topology.elements.JunBody)
	 * @see jp.co.sra.smalltalk.StValued#value()
	 * @category executing
	 */
	public JunBody value() {
		this.doOperation();
		return this.body();
	}

}
