package jp.co.sra.jun.topology.globaloperators;

import jp.co.sra.smalltalk.StBlockClosure1;

import jp.co.sra.jun.geometry.basic.Jun3dPoint;
import jp.co.sra.jun.geometry.basic.JunAngle;
import jp.co.sra.jun.geometry.curves.Jun3dLine;
import jp.co.sra.jun.geometry.surfaces.Jun3dPolygon;
import jp.co.sra.jun.topology.abstracts.JunAbstractOperator;
import jp.co.sra.jun.topology.abstracts.JunGlobalOperator;
import jp.co.sra.jun.topology.elements.JunBody;
import jp.co.sra.jun.topology.elements.JunEdge;
import jp.co.sra.jun.topology.elements.JunLoop;
import jp.co.sra.jun.topology.elements.JunVertex;
import jp.co.sra.jun.topology.support.JunGeometryTransformer;

/**
 * JunPlaneSCALE class
 * 
 *  @author    ASTI Shanghai
 *  @created   UNKNOWN
 *  @updated   2008/05/26 (by nisinaka)
 *  @version   758 (with StPL9.7) based on JunXXX for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunPlaneSCALE.java,v 9.9 2010/03/09 09:17:55 nisinaka Exp $
 */
public class JunPlaneSCALE extends JunGlobalOperator {

	protected Jun3dPoint point;
	protected Jun3dPoint vector;
	protected double factor;

	/**
	 * Create a new instance of <code>JunPlaneSCALE</code> and initialize it.
	 * 
	 * @param aJunBody jp.co.sra.jun.topology.elements.JunBody
	 * @param aJun3dPoint1 jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @param aJun3dPoint2 jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @param aNumber double
	 * @category Instance creation
	 */
	public JunPlaneSCALE(JunBody aJunBody, Jun3dPoint aJun3dPoint1, Jun3dPoint aJun3dPoint2, double aNumber) {
		super();
		this.body_(aJunBody);
		this.point_(aJun3dPoint1);
		this.vector_(aJun3dPoint2);
		this.factor_(aNumber);
	}

	/**
	 * Create a new instance of <code>JunPlaneSCALE</code> and initialize it.
	 * 
	 * @param aJunBody jp.co.sra.jun.topology.elements.JunBody
	 * @param aJun3dPoint1 jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @param aJun3dPoint2 jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @param aNumber double
	 * @return jp.co.sra.jun.topology.globaloperators.JunPlaneSCALE
	 * @deprecated since Jun700, using JunPlaneSCALE(JunBody, Jun3dPoint, Jun3dPoint, double)
	 * @category Instance creation
	 */
	public static final JunPlaneSCALE Body_point_vector_factor_(JunBody aJunBody, Jun3dPoint aJun3dPoint1, Jun3dPoint aJun3dPoint2, double aNumber) {
		return new JunPlaneSCALE(aJunBody, aJun3dPoint1, aJun3dPoint2, aNumber);
	}

	/**
	 * Initialize the receiver when created.
	 * 
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#initialize()
	 * @category initialize-release
	 */
	protected void initialize() {
		super.initialize();
		point = null;
		vector = null;
		factor = Double.NaN;
	}

	/**
	 * Execute the receiver.
	 * 
	 * @see jp.co.sra.jun.topology.abstracts.JunAbstractOperator#execute()
	 * @category executing
	 */
	public void execute() {
		final JunPlaneSCALE self = this;
		final Jun3dLine scaleAxis = this.scaleAxis();
		StBlockClosure1<Jun3dPoint, Jun3dPoint> scaleBlock = new StBlockClosure1<Jun3dPoint, Jun3dPoint>() {
			public Jun3dPoint value(Jun3dPoint p) {
				Jun3dPoint pVector = p.minus_(self.point());
				double pDistance = pVector.length();
				if (pDistance < pVector.ACCURACY) {
					return p;
				} else {
					JunAngle angle = scaleAxis.angleWithLine_(new Jun3dLine(self.point(), p));
					double qLength = angle.cos() * pDistance;
					Jun3dPoint qVector = self.unitVector().multipliedBy_(qLength);
					if (qVector.innerProduct_(pVector) < 0) {
						qVector = qVector.negated();
					}
					Jun3dPoint rVector = pVector.minus_(qVector);
					return qVector.multipliedBy_(self.factor()).plus_(rVector).plus_(self.point());
				}
			}
		};
		final JunGeometryTransformer transformer = new JunGeometryTransformer(scaleBlock);

		this.body().vertexesDo_(new StBlockClosure1<Void, JunVertex>() {
			public Void value(JunVertex vertex) {
				if (vertex.point() != null) {
					vertex.point_(vertex.point().transform_(transformer));
				}
				return null;
			}
		});
		this.body().edgesDo_(new StBlockClosure1<Void, JunEdge>() {
			public Void value(JunEdge edge) {
				Jun3dLine curve = (Jun3dLine) edge.basicCurve();
				if (curve != null) {
					edge.setCurve_((curve.transform_(transformer)));
				}
				return null;
			}
		});
		this.body().loopsDo_(new StBlockClosure1<Void, JunLoop>() {
			public Void value(JunLoop loop) {
				Jun3dPolygon surface = (Jun3dPolygon) loop.basicSurface();
				if (surface != null) {
					loop.setSurface_(surface.transform_(transformer));
				}
				return null;
			}
		});
	}

	/**
	 * Answer the factor.
	 * 
	 * @return double
	 * @category accessing
	 */
	public final double factor() {
		return factor;
	}

	/**
	 * Set the factor.
	 * 
	 * @param aNumber double
	 * @category accessing
	 */
	public final void factor_(double aNumber) {
		factor = aNumber;
	}

	/**
	 * Answer the inverse operation of the receiver.
	 * 
	 * @return jp.co.sra.jun.topology.abstracts.JunAbstractOperator
	 * @see jp.co.sra.jun.topology.abstracts.JunAbstractOperator#inverse()
	 * @category inverting
	 */
	public JunAbstractOperator inverse() {
		return new JunPlaneSCALE(this.body(), this.point(), this.vector(), 1.0d / this.factor());
	}

	/**
	 * Answer the point.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category accessing
	 */
	public final Jun3dPoint point() {
		return point;
	}

	/**
	 * Set the point.
	 * 
	 * @param aJun3dPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category accessing
	 */
	public final void point_(Jun3dPoint aJun3dPoint) {
		point = aJun3dPoint;
	}

	/**
	 * Answer the receiver's scale axis.
	 * 
	 * @return jp.co.sra.jun.geometry.curves.Jun3dLine
	 * @category accessing
	 */
	public Jun3dLine scaleAxis() {
		return new Jun3dLine(this.point(), this.point().plus_(this.unitVector()));
	}

	/**
	 * Answer the unit vector.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category accessing
	 */
	public Jun3dPoint unitVector() {
		return this.vector().unitVector();
	}

	/**
	 * Answer the vector.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category accessing
	 */
	public final Jun3dPoint vector() {
		return vector;
	}

	/**
	 * Set the vector.
	 * 
	 * @param aJun3dPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category accessing
	 */
	public final void vector_(Jun3dPoint aJun3dPoint) {
		vector = aJun3dPoint;
	}

}
