package jp.co.sra.jun.topology.globaloperators;

import jp.co.sra.smalltalk.StBlockClosure1;

import jp.co.sra.jun.geometry.basic.Jun3dPoint;
import jp.co.sra.jun.geometry.curves.Jun3dLine;
import jp.co.sra.jun.geometry.surfaces.Jun3dPolygon;
import jp.co.sra.jun.topology.abstracts.JunAbstractOperator;
import jp.co.sra.jun.topology.abstracts.JunGlobalOperator;
import jp.co.sra.jun.topology.elements.JunBody;
import jp.co.sra.jun.topology.elements.JunEdge;
import jp.co.sra.jun.topology.elements.JunLoop;
import jp.co.sra.jun.topology.elements.JunVertex;
import jp.co.sra.jun.topology.support.JunGeometryTransformer;

/**
 * JunPointSCALE class
 * 
 *  @author    ASTI Shanghai
 *  @created   UNKNOWN
 *  @updated   2008/05/26 (by nisinaka)
 *  @version   758 (with StPL9.7) based on JunXXX for Smalltalk
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunPointSCALE.java,v 9.9 2010/03/09 09:17:55 nisinaka Exp $
 */
public class JunPointSCALE extends JunGlobalOperator {

	protected Jun3dPoint point;
	protected double factor;

	/**
	 * Create a new instance of <code>JunPointSCALE</code> and initialize it.
	 * 
	 * @param aJunBody jp.co.sra.jun.topology.elements.JunBody
	 * @param aNumber double
	 * @category Instance creation
	 */
	public JunPointSCALE(JunBody aJunBody, double aNumber) {
		this(aJunBody, Jun3dPoint.Zero(), aNumber);
	}

	/**
	 * Create a new instance of <code>JunPointSCALE</code> and initialize it.
	 * 
	 * @param aJunBody jp.co.sra.jun.topology.elements.JunBody
	 * @param aJun3dPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @param aNumber double
	 * @category Instance creation
	 */
	public JunPointSCALE(JunBody aJunBody, Jun3dPoint aJun3dPoint, double aNumber) {
		super();
		this.body_(aJunBody);
		this.point_(aJun3dPoint);
		this.factor_(aNumber);
	}

	/**
	 * Create a new instance of JunPointSCALE and initialize it.
	 * 
	 * @param aJunBody jp.co.sra.jun.topology.elements.JunBody
	 * @param aNumber double
	 * @return jp.co.sra.jun.topology.globaloperators.JunPointSCALE
	 * @deprecated since Jun700, using JunPointSCALE(JunBody, double)
	 * @category Instance creation
	 */
	public static final JunPointSCALE Body_factor_(JunBody aJunBody, double aNumber) {
		return new JunPointSCALE(aJunBody, aNumber);
	}

	/**
	 * Create a new instance of JunPointSCALE and initialize it.
	 * 
	 * @param aJunBody jp.co.sra.jun.topology.elements.JunBody
	 * @param aJun3dPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @param aNumber double
	 * @return jp.co.sra.jun.topology.globaloperators.JunPointSCALE
	 * @deprecated since Jun700, using JunPointSCALE(JunBody, Jun3dPoint, double)
	 * @category Instance creation
	 */
	public static final JunPointSCALE Body_point_factor_(JunBody aJunBody, Jun3dPoint aJun3dPoint, double aNumber) {
		return new JunPointSCALE(aJunBody, aJun3dPoint, aNumber);
	}

	/**
	 * Initialize the receiver when created.
	 * 
	 * @see jp.co.sra.jun.system.framework.JunAbstractObject#initialize()
	 * @category initialize-release
	 */
	protected void initialize() {
		super.initialize();
		point = null;
		factor = Double.NaN;
	}

	/**
	 * Execute the receiver.
	 * 
	 * @see jp.co.sra.jun.topology.abstracts.JunAbstractOperator#execute()
	 * @category executing
	 */
	public void execute() {
		final JunPointSCALE self = this;
		StBlockClosure1<Jun3dPoint, Jun3dPoint> scaleBlock = new StBlockClosure1<Jun3dPoint, Jun3dPoint>() {
			public Jun3dPoint value(Jun3dPoint p) {
				Jun3dPoint relativePosition = (p.minus_(self.point()));
				return relativePosition.multipliedBy_(factor).plus_(self.point());
			}
		};
		final JunGeometryTransformer transformer = new JunGeometryTransformer(scaleBlock);

		this.body().vertexesDo_(new StBlockClosure1<Void, JunVertex>() {
			public Void value(JunVertex vertex) {
				if (vertex.point() != null) {
					vertex.point_(vertex.point().transform_(transformer));
				}
				return null;
			}
		});
		this.body().edgesDo_(new StBlockClosure1<Void, JunEdge>() {
			public Void value(JunEdge edge) {
				Jun3dLine curve = (Jun3dLine) edge.basicCurve();
				if (curve != null) {
					edge.setCurve_(curve.transform_(transformer));
				}
				return null;
			}
		});
		this.body().loopsDo_(new StBlockClosure1<Void, JunLoop>() {
			public Void value(JunLoop loop) {
				Jun3dPolygon surface = (Jun3dPolygon) loop.basicSurface();
				if (surface != null) {
					loop.setSurface_(surface.transform_(transformer));
				}
				return null;
			}
		});
	}

	/**
	 * Answer the factor.
	 * 
	 * @return double
	 * @category accessing
	 */
	public final double factor() {
		return factor;
	}

	/**
	 * Set the factor.
	 * 
	 * @param aNumber double
	 * @category accessing
	 */
	public final void factor_(double aNumber) {
		factor = aNumber;
	}

	/**
	 * Answer the inverse operation of the receiver.
	 * 
	 * @return jp.co.sra.jun.topology.abstracts.JunAbstractOperator
	 * @see jp.co.sra.jun.topology.abstracts.JunAbstractOperator#inverse()
	 * @category inverting
	 */
	public JunAbstractOperator inverse() {
		return new JunPointSCALE(this.body(), this.point(), 1.0d / this.factor());
	}

	/**
	 * Answer the point.
	 * 
	 * @return jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category accessing
	 */
	public final Jun3dPoint point() {
		return point;
	}

	/**
	 * Set the point.
	 * 
	 * @param aJun3dPoint jp.co.sra.jun.geometry.basic.Jun3dPoint
	 * @category accessing
	 */
	public final void point_(Jun3dPoint aJun3dPoint) {
		point = aJun3dPoint;
	}

}
