package jp.co.sra.smalltalk;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

/**
 * StSortedCollection class
 * 
 *  @author    nisinaka
 *  @created   1998/12/02 (by nisinaka)
 *  @updated   1999/08/05 (by nisinaka)
 *  @updated   2008/05/01 (by nisinaka)
 *  @version   9.7
 *  @copyright 1999-2010 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2010 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: StSortedCollection.java,v 9.11 2010/03/09 09:18:07 nisinaka Exp $
 */
public class StSortedCollection<E> extends StObject {

	/** The block closure for comparing two elements. */
	protected StBlockClosure2<Boolean, E, E> sortBlock = null;

	/** The collection which holds the elements in order. */
	protected ArrayList<E> _elements;

	/**
	 * Create a new instance of StSortedCollection and initialize it.
	 *
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @category Instance creation
	 */
	public StSortedCollection(StBlockClosure2<Boolean, E, E> aBlock) {
		this();
		this.sortBlock_(aBlock);
	}

	/**
	 * Create a new instance of StSortedCollection and initialize it.
	 *
	 * @param aBlock jp.co.sra.smalltalk.StBlockCosure
	 * @param anArray java.lang.Object[]
	 * @category Instance creation
	 */
	public StSortedCollection(StBlockClosure2<Boolean, E, E> aBlock, E[] anArray) {
		this(aBlock);

		for (int i = 0; i < anArray.length; i++) {
			this.add_(anArray[i]);
		}
	}

	/**
	 * Create a new instance of StSortedCollection with a sort block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @param aCollection java.util.Collection
	 * @category Instance creation
	 */
	public StSortedCollection(StBlockClosure2<Boolean, E, E> aBlock, Collection<E> aCollection) {
		this(aBlock);

		Iterator<E> i = aCollection.iterator();
		while (i.hasNext()) {
			this.add_(i.next());
		}
	}

	/**
	 * Create a new instance of StSortedCollection and initialize it.
	 *
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @category Instance creation
	 * @deprecated since Jun701, use Generics.
	 */
	public StSortedCollection(StBlockClosure<?, ?, ?, ?> aBlock) {
		this();
		this.sortBlock_(aBlock);
	}

	/**
	 * Create a new instance of StSortedCollection and initialize it.
	 *
	 * @param aBlock jp.co.sra.smalltalk.StBlockCosure
	 * @param anArray java.lang.Object[]
	 * @category Instance creation
	 * @deprecated since Jun701, use Generics.
	 */
	public StSortedCollection(StBlockClosure<?, ?, ?, ?> aBlock, E[] anArray) {
		this(aBlock);

		for (int i = 0; i < anArray.length; i++) {
			this.add_(anArray[i]);
		}
	}

	/**
	 * Create a new instance of StSortedCollection with a sort block.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @param aCollection java.util.Collection
	 * @category Instance creation
	 * @deprecated since Jun701, use Generics.
	 */
	public StSortedCollection(StBlockClosure<?, ?, ?, ?> aBlock, Collection<E> aCollection) {
		this(aBlock);

		Iterator<E> i = aCollection.iterator();
		while (i.hasNext()) {
			this.add_(i.next());
		}
	}

	/**
	 * Create a new instance of StSortedCollection and initialize it.
	 * Should not be created without a sorted block.
	 *
	 * @category Instance creation
	 */
	private StSortedCollection() {
		_elements = new ArrayList<E>();
	}

	/**
	 * Set the sort block of the receiver.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure2
	 * @category accessing
	 */
	public void sortBlock_(StBlockClosure2<Boolean, E, E> aBlock) {
		sortBlock = aBlock;
	}

	/**
	 * Set the sort block of the receiver.
	 * 
	 * @param aBlock jp.co.sra.smalltalk.StBlockClosure
	 * @category accessing
	 * @deprecated since Jun701, use Generics.
	 */
	public void sortBlock_(final StBlockClosure<?, ?, ?, ?> aBlock) {
		sortBlock = new StBlockClosure2<Boolean, E, E>() {
			public Boolean value(E e1, E e2) {
				return (Boolean) aBlock.value_value_(e1, e2);
			}
		};
	}

	/**
	 * Answer the element at the specified index.
	 * 
	 * @param index int
	 * @return java.lang.Object
	 * @category accessing
	 */
	public E at_(int index) {
		return _elements.get(index);
	}

	/**
	 * Answer the first element of the collection.
	 * 
	 * @return java.lang.Object
	 * @category accessing
	 */
	public E first() {
		return this.at_(0);
	}

	/**
	 * Answer the last element of the collection.
	 * 
	 * @return java.lang.Object
	 * @category accessing
	 */
	public E last() {
		return this.at_(this.size() - 1);
	}

	/**
	 * Answer my size.
	 * 
	 * @return int
	 * @category accessing
	 */
	public int size() {
		return _elements.size();
	}

	/**
	 * Add the new object as one of the receiver's element.
	 * 
	 * @param newObject java.lang.Object
	 * @return java.lang.Object
	 * @category adding
	 */
	public E add_(E newObject) {
		if (_elements.isEmpty()) {
			_elements.add(newObject);
		} else {
			int nextIndex = this.indexForInserting_(newObject);
			_elements.add(nextIndex, newObject);
		}
		return newObject;
	}

	/**
	 * Answer the receiver's elements as array.
	 * 
	 * @return java.lang.Object[]
	 * @category converting
	 */
	public Object[] _asArray() {
		return _elements.toArray();
	}

	/**
	 * Answer the receiver's elements as a collection.
	 * 
	 * @return java.util.Collection
	 * @category converting
	 */
	public Collection<E> _asCollection() {
		return new ArrayList<E>(_elements);
	}

	/**
	 * Compare the two elements.
	 * 
	 * @param e1 java.lang.Object
	 * @param e2 java.lang.Object
	 * @return boolean
	 * @category private
	 */
	protected boolean compareElement_to_(E e1, E e2) {
		if (sortBlock == null) {
			throw new SmalltalkException("No sort block.");
		}
		return sortBlock.value(e1, e2);
	}

	/**
	 * Answer the index for inserting the new object.
	 * 
	 * @param newObject java.lang.Object
	 * @return int
	 * @category private
	 */
	protected int indexForInserting_(E newObject) {
		int low = 0;
		int high = this.size() - 1;
		int index = (high + low) / 2;
		while (low <= high) {
			if (this.compareElement_to_(this.at_(index), newObject)) {
				low = index + 1;
			} else {
				high = index - 1;
			}
			index = (high + low) / 2;
		}
		return low;
	}

}
